/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"

bool is_playable_level = false;
bool is_not_demo = true;

void draw_select_level_dialog(void);
void draw_about_dialog(void);

byte arrowhead_left_image_data[];
byte arrowhead_right_image_data[];
image_type* arrowhead_left_image;
image_type* arrowhead_right_image;

// custom logic - full file name to load
char full_filename[POP_MAX_PATH] = "";

void load_arrowhead_images(void) {
    // Make a dummy palette for decode_image().
    dat_pal_type dat_pal;
    memset(&dat_pal, 0, sizeof(dat_pal));
    dat_pal.vga[1].r = dat_pal.vga[1].g = dat_pal.vga[1].b = 0x3F; // white
    if (arrowhead_left_image == NULL) {
        arrowhead_left_image = decode_image((image_data_type*) arrowhead_left_image_data, &dat_pal);
    }
    if (arrowhead_right_image == NULL) {
        arrowhead_right_image = decode_image((image_data_type*) arrowhead_right_image_data, &dat_pal);
    }
}

#define MAX_MENU_ITEM_LENGTH 32

typedef struct pause_menu_item_type pause_menu_item_type;
struct pause_menu_item_type {
    int id;
    pause_menu_item_type* previous;
    pause_menu_item_type* next;
    void* required;
    char text[MAX_MENU_ITEM_LENGTH];
};

enum pause_menu_item_ids {
    PAUSE_MENU_RESUME,
    PAUSE_MENU_ABOUT,
    PAUSE_MENU_DIFFICULTY,
    PAUSE_MENU_WINDOW_MODE,
    PAUSE_MENU_SOUND_TOGGLE,
    PAUSE_MENU_MUSIC_TOGGLE,
    PAUSE_MENU_SAVE_GAME,
    PAUSE_MENU_LOAD_GAME,
    PAUSE_MENU_LOAD_LEVEL,
    PAUSE_MENU_RESTART_GAME,
    PAUSE_MENU_RESTART_LEVEL,
    PAUSE_MENU_QUIT_GAME
};

pause_menu_item_type pause_menu_items[] = {
    {.id = PAUSE_MENU_RESUME,        .text = "RESUME (ESC)"},
    {.id = PAUSE_MENU_ABOUT,         .text = "ABOUT THE GAME..."},
    {.id = PAUSE_MENU_DIFFICULTY,    .text = "DIFFICULTY", .required = &is_playable_level },
    {.id = PAUSE_MENU_WINDOW_MODE,   .text = "MODE: WINDOW (ALT-ENTER)" },
#if defined __APPLE__ || __MACH__
    {.id = PAUSE_MENU_SOUND_TOGGLE,  .text = "SOUND ON (CMD-S)" },
    {.id = PAUSE_MENU_MUSIC_TOGGLE,  .text = "AMBIENT MUSIC ON (CMD-M)", .required = &show_music_menu },
#else
    {.id = PAUSE_MENU_SOUND_TOGGLE,  .text = "SOUND ON (CTRL-S)" },
    {.id = PAUSE_MENU_MUSIC_TOGGLE,  .text = "AMBIENT MUSIC ON (CTRL-M)", .required = &show_music_menu },
#endif
    {.id = PAUSE_MENU_SAVE_GAME,     .text = "SAVE GAME (F6)", .required = &is_playable_level },
    {.id = PAUSE_MENU_LOAD_GAME,     .text = "LOAD GAME (F9)", .required = &is_not_demo },
    {.id = PAUSE_MENU_LOAD_LEVEL,    .text = "LOAD ANOTHER LEVEL...", .required = &is_playable_level },
#if defined __APPLE__ || __MACH__
    {.id = PAUSE_MENU_RESTART_LEVEL, .text = "RESTART LEVEL (CMD-A)" },
    {.id = PAUSE_MENU_RESTART_GAME,  .text = "RESTART GAME (CMD-R)", .required = &is_not_demo },
    {.id = PAUSE_MENU_QUIT_GAME,     .text = "QUIT GAME (CMD-Q)"},
#else
    {.id = PAUSE_MENU_RESTART_LEVEL, .text = "RESTART LEVEL (CTRL-A)" },
    {.id = PAUSE_MENU_RESTART_GAME,  .text = "RESTART GAME (CTRL-R)", .required = &is_not_demo },
    {.id = PAUSE_MENU_QUIT_GAME,     .text = "QUIT GAME (CTRL-Q)"},
#endif
};

int hovering_pause_menu_item = PAUSE_MENU_RESUME;
pause_menu_item_type* next_pause_menu_item;
pause_menu_item_type* previous_pause_menu_item;
byte pause_menu_alpha;
int current_dialog_box;
const char* current_dialog_text;
char level_text[8];
word menu_current_level;
bool need_close_menu;

enum menu_dialog_ids {
    DIALOG_NONE,
    DIALOG_RESTORE_DEFAULT_SETTINGS,
    DIALOG_CONFIRM_QUIT,
    DIALOG_SELECT_LEVEL,
};

int scroll_position = 0;
int menu_control_y;
int menu_control_x;
int menu_control_back;

char* get_level_text(void) {
    switch (menu_current_level) {
        case 7:
            snprintf(level_text, sizeof(level_text), "6B");
            break;
        case 14:
            snprintf(level_text, sizeof(level_text), "12B");
            break;
        case 8:
        case 9:
        case 10:
        case 11:
        case 12:
        case 13:
            snprintf(level_text, sizeof(level_text), "%d", menu_current_level - 1);
            break;
        case 15:
            snprintf(level_text, sizeof(level_text), "%d", menu_current_level - 2);
            break;
        default:
            snprintf(level_text, sizeof(level_text), "%d", menu_current_level);
            break;
    }
    return level_text;
}

void init_pause_menu_items(pause_menu_item_type* first_item, int item_count) {
    if (item_count > 0) {
        for (int i = 0; i < item_count; ++i) {
            pause_menu_item_type* item = first_item + i;
            item->previous = (first_item + MAX(0, i - 1));
            item->next = (first_item + MIN(item_count - 1, i + 1));
        }
        pause_menu_item_type* last_item = first_item + (item_count - 1);
        first_item->previous = last_item;
        last_item->next = first_item;
    }
}

void init_menu(void) {
    load_arrowhead_images();

    init_pause_menu_items(pause_menu_items, COUNT(pause_menu_items));
}

bool is_mouse_over_rect(rect_type* rect) {
    return (mouse_x >= rect->left && mouse_x < rect->right && mouse_y >= rect->top && mouse_y < rect->bottom);
}

// Maps the cursor position into a coordinate between (0,0) and (320,200) and sets mouse_x, mouse_y and mouse_moved.
void read_mouse_state(void) {
    float scale_x, scale_y;
    SDL_RenderGetScale(renderer_, &scale_x, &scale_y);
    int logical_width, logical_height;
    SDL_RenderGetLogicalSize(renderer_, &logical_width, &logical_height);
    int logical_scale_x = logical_width / 320; // These may be higher than 1, if 4:3 aspect ratio scaling is enabled.
    int logical_scale_y = logical_height / 200;
    scale_x *= logical_scale_x;
    scale_y *= logical_scale_y;
    if (!(scale_x > 0 && scale_y > 0 && logical_scale_x > 0 && logical_scale_y > 0)) return;
    SDL_Rect viewport;
    SDL_RenderGetViewport(renderer_, &viewport); // Get the width/height of the 'black bars' around the rendering area.
    viewport.x /= logical_scale_x;
    viewport.y /= logical_scale_y;
    int last_mouse_x = mouse_x;
    int last_mouse_y = mouse_y;
    SDL_GetMouseState(&mouse_x, &mouse_y);
    mouse_x = (int) ((float)mouse_x / scale_x - viewport.x + 0.5f);
    mouse_y = (int) ((float)mouse_y / scale_y - viewport.y + 0.5f);
    mouse_moved = (last_mouse_x != mouse_x || last_mouse_y != mouse_y);
}

rect_type explanation_rect = {170, 20, 200, 300};
int highlighted_setting_id = PAUSE_MENU_RESUME;
int controlled_area = 0; // Whether the focus is on the left (0) or right (1) part of the screen in the Settings menu.
int next_setting_id = 0; // For navigating up/down.
int previous_setting_id = 0;
bool at_scroll_up_boundary; // When navigating up using keyboard/controller, whether we also need to scroll up
bool at_scroll_down_boundary; // When navigating down using keyboard/controller, whether we also need to scroll down

void play_menu_sound(int sound_id) {
    set_sound_pos(drawn_room, 4, 1);
    play_sound(sound_id);
    play_next_sound();
}

void reset_paused_menu(void) {
    controlled_area = 0;
    hovering_pause_menu_item = PAUSE_MENU_RESUME;
}

// custom logic - load auto-saved level
char* load_level_from_autosave(void) {
    snprintf(full_filename, sizeof(full_filename), "%s/LEVEL%s.SAV", auto_save_folder, get_level_text());
    if (!file_exists(full_filename)) {
        return "";
    }
    return full_filename;
}

// custom logic - reload load level 1
void load_level_1(void) {
    need_close_menu = true;
    stop_sounds();
    rem_min = settings->start_minutes_left;
    rem_tick = settings->start_ticks_left;
    hitp_beg_lev = hitp_curr = MIN(settings->start_hitp, settings->max_hitp_allowed);
    total_saves = 0;
    if (current_level == 1) {
        is_restart_level = 1;
    } else {
        quick_loading = 1;
        show_black_screen(6, 0);
        quick_loading = 0;
        next_level = 1;
    }
    is_shadow_flame = false;
    reset_shadow_palette();
}

const char* about_message = "\nUnofficial Prince of Persia v1.5\nBased on the game by Jordan Mechner\n\nCreated by Dmitry S.\n"
                            "Using modified SDLPoP by David Nagy\n\n"
                            "See DOCS/README.TXT for more information\n"
                            "This game requires moves absent in Prince 1";

void pause_menu_clicked(pause_menu_item_type* item) {
    play_menu_sound(sound_22_loose_shake_3);
    switch(item->id) {
        default:
            break;
        case PAUSE_MENU_RESUME:
            need_close_menu = true;
            break;
        case PAUSE_MENU_ABOUT:
            draw_about_dialog();
            break;
        case PAUSE_MENU_DIFFICULTY:
            if (cheats_enabled) {
                stop_sounds();
                play_menu_sound(sound_8_bumped);
                break;
            }
            switch (max_saves_per_level) {
                case 0:
                    max_saves_per_level = 3;
                    break;
                case 1:
                    max_saves_per_level = 0;
                    break;
                default:
                    max_saves_per_level = 1;
                    break;
            }
            config_save();
            break;
        case PAUSE_MENU_WINDOW_MODE:
            toggle_fullscreen();
            need_close_menu = true;
            break;
        case PAUSE_MENU_SOUND_TOGGLE:
            last_key_scancode = SDL_SCANCODE_S | WITH_CTRL;
            break;
        case PAUSE_MENU_MUSIC_TOGGLE:
            last_key_scancode = SDL_SCANCODE_M | WITH_CTRL;
            break;
        case PAUSE_MENU_SAVE_GAME:
            if (Kid.alive < 0) need_quick_save = 1;
            need_close_menu = true;
            break;
        case PAUSE_MENU_LOAD_GAME:
            need_quick_load = 1;
            need_close_menu = true;
            stop_sounds();
            break;
        // custom logic - load previously auto-saved level
        case PAUSE_MENU_LOAD_LEVEL:
            draw_select_level_dialog();
            bool exit_dialog = (key_states[SDL_SCANCODE_ESCAPE] & keystate_held) || menu_control_back;
            if (exit_dialog) {
                break;
            }
            if (menu_current_level == 1) {
                load_level_1();
                break;
            }
            need_auto_load = load_level_from_autosave();
            if (strlen(need_auto_load) == 0 && !exit_dialog) {
                play_menu_sound(sound_8_bumped);
                show_dialog("Requested level has not been played yet", default_dialog);
                pause_menu_clicked(item);
            } else {
                need_close_menu = true;
                stop_sounds();
            }
            break;
        case PAUSE_MENU_RESTART_GAME:
            last_key_scancode = SDL_SCANCODE_R | WITH_CTRL;
            break;
        case PAUSE_MENU_RESTART_LEVEL:
            last_key_scancode = SDL_SCANCODE_A | WITH_CTRL;
            break;
        case PAUSE_MENU_QUIT_GAME:
            current_dialog_box = DIALOG_CONFIRM_QUIT;
            current_dialog_text = "Quit Prince?";
            break;
    }
    clear_menu_controls(); // prevent "click-through" because the screen changes
}

void draw_pause_menu_item(pause_menu_item_type* item, rect_type* parent, int* y_offset, int inactive_text_color) {
    if (item->required != NULL) {
        if (*(sbyte*)item->required == false) {
            return; // skip this item (disabled)
        }
    }

    rect_type text_rect = *parent;
    text_rect.top += *y_offset;
    int text_color = inactive_text_color;

    rect_type selection_box = text_rect;
    selection_box.bottom = selection_box.top + 8;
    selection_box.top -= 3;

    bool highlighted = (hovering_pause_menu_item == item->id);
    if (have_mouse_input && is_mouse_over_rect(&selection_box)) {
        hovering_pause_menu_item = item->id;
        highlighted = true;
    }

    if (highlighted) {
        previous_pause_menu_item = item->previous;
        next_pause_menu_item = item->next;
        // Skip over disabled items (such as the CHEATS menu in non-cheat mode)
        if (previous_pause_menu_item->required != NULL) {
            while (*(sbyte*)previous_pause_menu_item->required == false) {
                previous_pause_menu_item = previous_pause_menu_item->previous;
                if (previous_pause_menu_item->required == NULL) break;
            }
        }
        if (next_pause_menu_item->required != NULL) {
            while (*(sbyte*)next_pause_menu_item->required == false) {
                next_pause_menu_item = next_pause_menu_item->next;
                if (next_pause_menu_item->required == NULL) break;
            }
        }
        text_color = color_15_brightwhite;
        draw_rect_contours(&selection_box, color_7_lightgray);

        if (mouse_clicked) {
            if (is_mouse_over_rect(&selection_box)) {
                pause_menu_clicked(item);
            }
        } else if (pressed_enter) {
            pause_menu_clicked(item);
        }

    }
    char* text = item->text;
    if (is_sound_on != 0 && strncmp("SOUND ON (CTRL-S)", text, strlen(text)) == 0) {
        text = "SOUND OFF (CTRL-S)";
    }
    if (is_music_on != 0 && strncmp("AMBIENT MUSIC ON (CTRL-M)", text, strlen(text)) == 0) {
        text = "AMBIENT MUSIC OFF (CTRL-M)";
    }
    if (strncmp("DIFFICULTY", text, 10) == 0) {
        if (cheats_enabled) {
            text = "DIFFICULTY: CHEATING";
            if (hovering_pause_menu_item == item->id) {
                text_color = color_7_lightgray;
            }
        } else {
            switch (max_saves_per_level) {
                case 0:
                    text = "DIFFICULTY: HARD (NO QUICK SAVE)";
                    break;
                case 1:
                    text = "DIFFICULTY: MEDIUM (1 SAVE/LEVEL)";
                    break;
                case 3:
                    text = "DIFFICULTY: EASY (3 SAVES/LEVEL)";
                    break;
                default:
                    max_saves_per_level = 3;
                    text = "DIFFICULTY: EASY (3 SAVES/LEVEL)";
                    break;
            }
        }
    }

    if (strncmp("MODE:", text, 5) == 0) {
        if (SDL_GetWindowFlags(window_) & SDL_WINDOW_FULLSCREEN_DESKTOP) {
            text = "MODE: WINDOW (ALT-ENTER)";
        } else {
            text = "MODE: FULL SCREEN (ALT-ENTER)";
        }
    }

    show_text_with_color(&text_rect, text_align_center, text_align_top, text, text_color);
    *y_offset += 13;
}

void draw_pause_menu(void) {
    is_playable_level = current_level > 0 && current_level < 15;
    is_not_demo = !demo_mode;

    pause_menu_alpha = 120;
    draw_rect_with_alpha(&screen_rect, color_0_black, pause_menu_alpha);
    draw_rect_with_alpha(&rect_bottom_text, color_0_black, 0); // Transparent so that the text "GAME PAUSED" is visible.
    rect_type pause_rect_outer = {0, 60, 192, 260};
    rect_type pause_rect_inner;
    shrink2_rect(&pause_rect_inner, &pause_rect_outer, 5, 5);

    if (!have_mouse_input) {
        if (menu_control_y == 1) {
            play_menu_sound(sound_21_loose_shake_2);
            hovering_pause_menu_item = next_pause_menu_item->id;
        } else if (menu_control_y == -1) {
            play_menu_sound(sound_21_loose_shake_2);
            hovering_pause_menu_item = previous_pause_menu_item->id;
        } else if (menu_control_y == -100) {
            play_menu_sound(sound_21_loose_shake_2);
            hovering_pause_menu_item = 0;
        } else if (menu_control_y == 100) {
            play_menu_sound(sound_21_loose_shake_2);
            hovering_pause_menu_item = COUNT(pause_menu_items) - 1;
        }
    }

    int y_offset = 14;
    for (size_t i = 0; i < COUNT(pause_menu_items); ++i) {
        draw_pause_menu_item(&pause_menu_items[i], &pause_rect_inner, &y_offset, color_15_brightwhite);
    }
}

bool were_settings_changed;

void draw_image_with_blending(image_type* image, int xpos, int ypos) {
    SDL_Rect src_rect = {0, 0, image->w, image->h};
    SDL_Rect dest_rect = {xpos, ypos, image->w, image->h};
    SDL_SetColorKey(image, SDL_TRUE, 0);
    if (SDL_BlitSurface(image, &src_rect, current_target_surface, &dest_rect) != 0) {
        sdlperror("SDL_BlitSurface");
        quit(1);
    }
}

enum dialog_button_ids {
    DIALOG_BUTTON_CANCEL,
    DIALOG_BUTTON_OK,
};

void confirmation_dialog_result(int which_dialog, int button) {
    if (button == DIALOG_BUTTON_OK) {
        if (which_dialog == DIALOG_CONFIRM_QUIT) {
            last_key_scancode = SDL_SCANCODE_Q | WITH_CTRL;
            key_test_quit();
        }
    } else {
        play_menu_sound(sound_22_loose_shake_3);
    }
}

rect_type cancel_text_rect = {104, 162, 118, 212};
rect_type cancel_highlight_rect = {103, 162, 116, 212};
rect_type ok_text_rect = {104, 108, 118, 158};
rect_type ok_highlight_rect = {103, 108, 116, 158};

void draw_confirmation_dialog(int which_dialog, const char* text) {
    int highlighted_button = DIALOG_BUTTON_OK;
    int old_highlighted_button = -1;
    for (;;) {
        process_events();
        key_test_paused_menu(key_test_quit());
        process_additional_menu_input();

        if (menu_control_back == 1) {
            confirmation_dialog_result(which_dialog, DIALOG_BUTTON_CANCEL);
            break;
        }

        if (have_mouse_input) {
            if (is_mouse_over_rect(&ok_highlight_rect)) {
                highlighted_button = DIALOG_BUTTON_OK;
            } else if (is_mouse_over_rect(&cancel_highlight_rect)) {
                highlighted_button = DIALOG_BUTTON_CANCEL;
            }
        }

        if (menu_control_x < 0) {
            highlighted_button = DIALOG_BUTTON_OK;
        } else if (menu_control_x > 0) {
            highlighted_button = DIALOG_BUTTON_CANCEL;
        } else if (mouse_clicked || pressed_enter) {
            confirmation_dialog_result(which_dialog, highlighted_button);
            break;
        }

        if (highlighted_button != old_highlighted_button) {
            old_highlighted_button = highlighted_button;
            // Need to redraw the dialog box.
            uint32_t clear_color = SDL_MapRGBA(current_target_surface->format, 0, 0, 0, 128);
            SDL_FillRect(overlay_surface, NULL, clear_color);
            dialog_method_2_frame(copyprot_dialog);
            rect_type rect;
            shrink2_rect(&rect, &copyprot_dialog->text_rect, 2, 1);
            rect.bottom -= 14;
            show_text_with_color(&rect, text_align_center, text_align_center, text, color_15_brightwhite);
            clear_kbd_buf();

            rect_type* highlight_rect;
            int ok_text_color, cancel_text_color;
            if (highlighted_button == DIALOG_BUTTON_OK) {
                highlight_rect = &ok_highlight_rect;
                ok_text_color = color_15_brightwhite;
                cancel_text_color = color_7_lightgray;
            } else {
                highlight_rect = &cancel_highlight_rect;
                ok_text_color = color_7_lightgray;
                cancel_text_color = color_15_brightwhite;
            }
            draw_rect(highlight_rect, color_8_darkgray);
            show_text_with_color(&ok_text_rect, text_align_center, text_align_center, "OK", ok_text_color);
            show_text_with_color(&cancel_text_rect, text_align_center, text_align_center, "Cancel", cancel_text_color);
            update_screen();
        }

        SDL_Delay(1); // Prevent 100% cpu usage.

    }
    current_dialog_box = 0;
    clear_menu_controls();
}

int get_current_level(void) {
    return (current_level < 6 || (current_level == 6 && falling_exit_checkpoint < 4)) ? current_level : current_level + 1;
}

int get_prev_level(void) {
    int level = MAX(1, menu_current_level - 1);
    if (level == get_current_level()) {
        if (level > 1) {
            return level - 1;
        } else {
            return menu_current_level;
        }
    }
    return level;
}

int get_next_level(void) {
    int level = MIN(15, menu_current_level + 1);
    int curr_level = get_current_level();
    if (level == curr_level) {
        if (level < 15) {
            return level + 1;
        } else {
            return menu_current_level;
        }
    }
    return level;
}

void draw_about_dialog(void) {
    clear_menu_controls();
    uint32_t clear_color = SDL_MapRGBA(current_target_surface->format, 0, 0, 0, 128);
    SDL_FillRect(overlay_surface, NULL, clear_color);
    show_dialog(about_message, small_text_dialog);
}

void draw_select_level_dialog(void) {
    int prev_level_number = -1;
    clear_menu_controls();
    for (;;) {
        process_events();
        key_test_paused_menu(key_test_quit());
        process_additional_menu_input();

        if (menu_control_back == 1) {
            clear_menu_controls();
            menu_control_back = 1;
            play_menu_sound(sound_22_loose_shake_3);
            return;
        }

        if (mouse_clicked) {
            if ((mouse_x >= 136 && mouse_x <= 150) &&
                    (mouse_y >= 84 && mouse_y <= 96)) {
                mouse_clicked = 0;
                menu_current_level = get_prev_level();
                play_menu_sound(sound_21_loose_shake_2);
            } else if ((mouse_x >= 170 && mouse_x <= 184) &&
                       (mouse_y >= 84 && mouse_y <= 96)) {
                mouse_clicked = 0;
                menu_current_level = get_next_level();
                play_menu_sound(sound_21_loose_shake_2);
            }
        } else if (menu_control_x < 0) {
            menu_current_level = get_prev_level();
            play_menu_sound(sound_21_loose_shake_2);
        } else if (menu_control_x > 0) {
            menu_current_level = get_next_level();
            play_menu_sound(sound_21_loose_shake_2);
        }

        if (mouse_clicked || pressed_enter) {
            break;
        }

        if (menu_current_level != prev_level_number) {
            prev_level_number = menu_current_level;
            // Need to redraw the dialog box.
            uint32_t clear_color = SDL_MapRGBA(current_target_surface->format, 0, 0, 0, 128);
            SDL_FillRect(overlay_surface, NULL, clear_color);
            dialog_method_2_frame(copyprot_dialog);
            rect_type rect;
            shrink2_rect(&rect, &copyprot_dialog->text_rect, 2, 1);
            rect.top -= 20;
            rect.bottom -= 14;
            show_text_with_color(&rect, text_align_center, text_align_center, "SELECT LEVEL NUMBER", color_15_brightwhite);
            clear_kbd_buf();
            rect_type input_rect = {84, 64, 98, 256};
            get_level_text();
            font_type* saved_font = textstate.ptr_font;
            textstate.ptr_font = &hc_font;
            show_text_with_color(&input_rect, text_align_center, text_align_center, level_text, color_15_brightwhite);
            textstate.ptr_font = saved_font;
            draw_image_with_blending(arrowhead_right_image, 175, input_rect.top + 3);
            draw_image_with_blending(arrowhead_left_image, 145 - 3, input_rect.top + 3);

            rect_type instruction_rect = {107, 64, 121, 256};
            show_text_with_color(&instruction_rect, text_align_center, text_align_center, "Press ENTER to load the level", color_7_lightgray);

            update_screen();
        }

        SDL_Delay(1); // Prevent 100% cpu usage.

    }
    clear_menu_controls();
}

int need_full_menu_redraw_count;

void draw_menu(void) {
    menu_current_level = current_level == 1 ? 2 : get_current_level() - 1;

    escape_key_suppressed = key_states[SDL_SCANCODE_ESCAPE] & keystate_held;
    surface_type* saved_target_surface = current_target_surface;
    current_target_surface = overlay_surface;

    need_close_menu = false;
    while (!need_close_menu) {
        clear_menu_controls();
        process_events();
        if (process_key(true) != 0) {
            break; // Menu was forcefully closed, for example by pressing Ctrl+A.
        }
        process_additional_menu_input();

        if (current_dialog_box != DIALOG_NONE) {
            if (current_dialog_box == DIALOG_SELECT_LEVEL) {
                draw_select_level_dialog();
            } else {
                draw_confirmation_dialog(current_dialog_box, current_dialog_text);
            }
            current_dialog_box = DIALOG_NONE;
            clear_menu_controls();
        }

        if (is_menu_shown == 1) {
            is_menu_shown = -1; // reset the menu if the menu is drawn for the first time
            need_full_menu_redraw_count = 2;
            reset_paused_menu();
        }

        if (menu_control_back == 1) {
            break; // Close the menu.
        }

        if (have_mouse_input || have_keyboard_or_controller_input) {
            // The menu is updated+drawn within the same routine, so redrawing may be necessary after the first time.
            // TODO: Maybe in the future fully separate updating from drawing?
            need_full_menu_redraw_count = 2;
        } else {
            if (need_full_menu_redraw_count == 0) {
                SDL_Delay(1);
                continue; // Don't redraw if there is no input to process (save CPU cycles).
            }
        }

        font_type* saved_font = textstate.ptr_font;
        textstate.ptr_font = &hc_small_font;
        draw_pause_menu();
        textstate.ptr_font = saved_font;
        if (!need_close_menu) {
            update_screen();
        }

        --need_full_menu_redraw_count;
    }

    current_target_surface = saved_target_surface;
}

void clear_menu_controls(void) {
    pressed_enter = 0;
    mouse_moved = 0;
    mouse_clicked = 0;
    mouse_button_clicked_right = 0;
    have_mouse_input = 0;
    have_keyboard_or_controller_input = 0;
    menu_control_x = 0;
    menu_control_y = 0;
    menu_control_back = 0;
    menu_control_scroll_y = 0;
}

void process_additional_menu_input(void) {
    read_mouse_state();
    have_keyboard_or_controller_input = (menu_control_x || menu_control_y || menu_control_back || pressed_enter);
    have_mouse_input = (mouse_moved || mouse_clicked || mouse_button_clicked_right || menu_control_scroll_y);

    dword flags = SDL_GetWindowFlags(window_);
    if (flags & SDL_WINDOW_FULLSCREEN_DESKTOP) {
        if (have_mouse_input) {
            SDL_ShowCursor(SDL_ENABLE);
        } else if (have_keyboard_or_controller_input) {
            SDL_ShowCursor(SDL_DISABLE);
        }
    } else {
        SDL_ShowCursor(SDL_ENABLE);
    }
}

bool joy_ABXY_buttons_released;
bool joy_xy_released;
Uint64 joy_xy_timeout_counter;

int key_test_paused_menu(int key) {
    menu_control_x = 0;
    menu_control_y = 0;
    menu_control_back = 0;

    if (mouse_button_clicked_right) {
        menu_control_back = 1; // Can use RMB to close menus.
    }

    if (is_joyst_mode) {
        int y_threshold = 14000;
        int x_threshold = 26000; // Less sensitive, to prevent accidentally changing a setting.
        int joy_x = 0;
        int joy_y = 0;
        if (joy_button_states[joyinput_dpad_left] & keystate_held) {
            joy_x = -1;
        } else if (joy_button_states[joyinput_dpad_right] & keystate_held) {
            joy_x = 1;
        }
        if (joy_button_states[joyinput_dpad_up] & keystate_held) {
            joy_y = -1;
        } else if (joy_button_states[joyinput_dpad_down] & keystate_held) {
            joy_y = 1;
        }
        if (joy_axis[SDL_CONTROLLER_AXIS_LEFTY] < -y_threshold) {
            joy_y = -1;
        } else if (joy_axis[SDL_CONTROLLER_AXIS_LEFTY] > y_threshold) {
            joy_y = 1;
        } else if (joy_axis[SDL_CONTROLLER_AXIS_LEFTX] < -x_threshold) {
            joy_x = -1;
        } else if (joy_axis[SDL_CONTROLLER_AXIS_LEFTX] > x_threshold) {
            joy_x = 1;
        }

        float needed_timeout_s = 0.1f; // Delay for hold-down repeated input.
        if (joy_x == 0 && joy_y == 0) {
            joy_xy_released = true;
            joy_xy_timeout_counter = 0;
        } else {
            if (joy_xy_released) {
                needed_timeout_s = 0.3f; // The delay is longer for the first repetition.
                joy_xy_released = false;
            }
            Uint64 current_counter = SDL_GetPerformanceCounter();
            if (current_counter > joy_xy_timeout_counter) {
                menu_control_x = joy_x;
                menu_control_y = joy_y;
                joy_xy_timeout_counter = current_counter + (Uint64)((float)SDL_GetPerformanceFrequency() * needed_timeout_s);
                return 0; // cancel other input.
            }
        }

        if (!(joy_button_states[joyinput_a] & keystate_held) &&
                !(joy_button_states[joyinput_y] & keystate_held) &&
                !(joy_button_states[joyinput_b] & keystate_held)) {
            joy_ABXY_buttons_released = true;
        } else if (joy_ABXY_buttons_released) {
            joy_ABXY_buttons_released = false;
            if (joy_button_states[joyinput_a] & keystate_held) {
                key = SDL_SCANCODE_RETURN;
                joy_button_states[joyinput_a] = 0; // Prevent 'down' input being passed to the controls if the game is unpaused.
            } else if (joy_button_states[joyinput_b] & keystate_held) {
                key = SDL_SCANCODE_ESCAPE;
            }
        }
    }

    switch(key) {
        default:
            if (key & WITH_CTRL) {
                need_close_menu = true;
                return key; // Allow Ctrl+R, etc.
            } else {
                break;
            }
        case SDL_SCANCODE_HOME:
        case SDL_SCANCODE_KP_7:
            menu_control_y = -100;
            break;
        case SDL_SCANCODE_END:
        case SDL_SCANCODE_KP_1:
            menu_control_y = 100;
            break;
        case SDL_SCANCODE_UP:
        case SDL_SCANCODE_KP_8:
            menu_control_y = -1;
            break;
        case SDL_SCANCODE_DOWN:
        case SDL_SCANCODE_KP_2:
            menu_control_y = 1;
            break;
        case SDL_SCANCODE_RIGHT:
        case SDL_SCANCODE_KP_6:
            menu_control_x = 1;
            break;
        case SDL_SCANCODE_LEFT:
        case SDL_SCANCODE_KP_4:
            menu_control_x = -1;
            break;
        case SDL_SCANCODE_RETURN:
        case SDL_SCANCODE_KP_ENTER:
        case SDL_SCANCODE_SPACE:
        case SDL_SCANCODE_RETURN | WITH_SHIFT:
        case SDL_SCANCODE_KP_ENTER | WITH_SHIFT:
        case SDL_SCANCODE_SPACE | WITH_SHIFT:
            pressed_enter = 1;
            break;
        case SDL_SCANCODE_ESCAPE:
            menu_control_back = 1;
            break;
        case SDL_SCANCODE_F6:
        case SDL_SCANCODE_F6 | WITH_SHIFT:
            if (Kid.alive < 0) need_quick_save = 1;
            need_close_menu = true;
            break;
        case SDL_SCANCODE_F9:
        case SDL_SCANCODE_F9 | WITH_SHIFT:
            need_quick_load = 1;
            need_close_menu = true;
            break;
    }
    return 0;
}

void menu_was_closed(void) {
    is_paused = 0;
    is_menu_shown = 0;
    escape_key_suppressed = key_states[SDL_SCANCODE_ESCAPE] & keystate_held;
    if (were_settings_changed) {
        //save_ingame_settings();
        were_settings_changed = false;
    }
    // In fullscreen mode, hide the mouse cursor (because it is only needed in the menu).
    dword flags = SDL_GetWindowFlags(window_);
    if (flags & SDL_WINDOW_FULLSCREEN_DESKTOP) {
        SDL_ShowCursor(SDL_DISABLE);
    } else {
        SDL_ShowCursor(SDL_ENABLE);
    }
    // custom logic - restart ambient music after the menu was closed
    if (music_paused) {
        Mix_ResumeMusic();
    }
    if (Mix_PlayingMusic() == 0) {
        // digital sound can suspend ambient music
        if (!music_paused && music_suspended_timer > 0) {
            music_suspended_timer = 0;
        }
        start_ambient_music(true);
    }
    music_paused = false;
}

byte arrowhead_left_image_data[] = {
    IMAGE_DATA(5, 3, 1),
    BINARY_8( _, _, 1, _, _, _, _, _ ),
    BINARY_8( _, 1, 1, _, _, _, _, _ ),
    BINARY_8( 1, 1, 1, _, _, _, _, _ ),
    BINARY_8( _, 1, 1, _, _, _, _, _ ),
    BINARY_8( _, _, 1, _, _, _, _, _ ),
};

byte arrowhead_right_image_data[] = {
    IMAGE_DATA(5, 3, 1),
    BINARY_8( 1, _, _, _, _, _, _, _ ),
    BINARY_8( 1, 1, _, _, _, _, _, _ ),
    BINARY_8( 1, 1, 1, _, _, _, _, _ ),
    BINARY_8( 1, 1, _, _, _, _, _, _ ),
    BINARY_8( 1, _, _, _, _, _, _, _ ),
};
