/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"
#include "zip.h"

#define MAX_MUSIC_ID 9
#define MAX_SOUND_ID 81

// Background music player implementation - inspired by SDLPoP JP

const char* enhanced_sound_file = "DATA/MT32SND.DAT";
struct zip_t *enhanced_sound_dat = NULL;

bool heartbeat_music_found = false;
bool music_found = false;

bool music_initialized = false;

const int ambient_music_volume = 45;
const int level_music_gain[15] = { 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 55, 2, 2, 2, 1 };

Mix_Music* ogg_sound_tracks[MAX_SOUND_ID + 1] = { NULL };
Mix_Music* ogg_music_tracks[MAX_MUSIC_ID + 1] = { NULL };

int last_sound_id = -1;

enum musicids {
    music_0_heartbeat_sound = 0,
    music_1_level_music = 1,
    music_2_level_music = 2,
    music_3_level_music = 3,
    music_4_level_music = 4,
    music_5_level_music = 5,
    music_6_level_music = 6,
    music_7_fight_music = 7,
    music_8_fight_music = 8,
    music_9_fight_music = 9
};

bool enhanced_sound_file_exists(void) {
    struct stat st = {0};
    return stat(locate_file(enhanced_sound_file), &st) != -1;
}

void load_enhanced_sounds(void) {
    enhanced_sound_dat = zip_open(enhanced_sound_file, 0, 'r');
    if (enhanced_sound_dat == NULL) {
        printf("Enhanced sound file '%s' was not found. Enhanced sound and ambient music are disabled.\n", enhanced_sound_file);
        enable_positional_audio = false;
    }
}

void unload_enhanced_sounds(void) {
    if (enhanced_sound_dat != NULL) {
        zip_close(enhanced_sound_dat);
    }
}

void* get_enhanced_sound_data(const char* entry, size_t* length) {
    if (zip_entry_open(enhanced_sound_dat, entry) < 0) {
        zip_entry_close(enhanced_sound_dat);
        return NULL;
    }
    void* buf;
    zip_entry_read(enhanced_sound_dat, &buf, length);
    zip_entry_close(enhanced_sound_dat);
    return buf;
}

void load_music(void) {
    int num;
    if (!enable_positional_audio || music_initialized) {
        return;
    }
    load_enhanced_sounds();
    {
        // scope
        char file[POP_MAX_PATH];
        snprintf(file, sizeof(file), "SOUNDHB.OGG");
        size_t length;
        void* buffer = get_enhanced_sound_data(file, &length);
        if (buffer != NULL) {
            SDL_RWops* src = SDL_RWFromMem(buffer, length);
            ogg_music_tracks[0] = Mix_LoadMUSType_RW(src, MUS_OGG, SDL_TRUE);
            heartbeat_music_found = true;
        }
    }
    for (num = 1; num <= MAX_MUSIC_ID; num++) {
        char file[POP_MAX_PATH];
        snprintf(file, sizeof(file), "MUSIC%d.OGG", num);
        size_t length;
        void* buffer = get_enhanced_sound_data(file, &length);
        if (buffer != NULL) {
            SDL_RWops* src = SDL_RWFromMem(buffer, length);
            ogg_music_tracks[num] = Mix_LoadMUSType_RW(src, MUS_OGG, SDL_TRUE);
            if (!music_found) {
                music_found = true;
            }
        }
    }
    for (num = 0; num <= MAX_SOUND_ID; num++) {
        if (!is_foreground_sound(num)) {
            continue;
        }
        char file[POP_MAX_PATH];
        snprintf(file, sizeof(file), "SOUND%d.OGG", num);
        size_t length;
        void* buffer = get_enhanced_sound_data(file, &length);
        if (buffer != NULL) {
            SDL_RWops* src = SDL_RWFromMem(buffer, length);
            ogg_sound_tracks[num] = Mix_LoadMUSType_RW(src, MUS_OGG, SDL_TRUE);
        }
    }
    feather_fall_track_exists = ogg_sound_tracks[sound_39_low_weight] != NULL;
    unload_enhanced_sounds();
    music_initialized = true;
}

void free_music(void) {
    int num;
    stop_all_sounds();
    Mix_HaltMusic();
    for (num = 0; num <= MAX_MUSIC_ID; num++) {
        Mix_Music* ogg_music_track = ogg_music_tracks[num];
        if (ogg_music_track != NULL) {
            Mix_FreeMusic(ogg_music_track);
        }
    }
    for (num = 0; num <= MAX_SOUND_ID; num++) {
        Mix_Music* ogg_sound_track = ogg_sound_tracks[num];
        if (ogg_sound_track != NULL) {
            Mix_FreeMusic(ogg_sound_track);
        }
    }
    Mix_Quit();
}

void ogg_sound_callback(void) {
    curr_ogg_sound_id = -1;
    Mix_HookMusicFinished(NULL);
    ogg_fore_sound_playing = false;
    music_suspended = 0;
}

bool is_interruptable_sound(int sound_id) { // does not play over other sounds
    switch (sound_id) {
        case sound_39_low_weight:
            return true;
        default:
            return false;
    }
}

bool is_foreground_sound(int sound_id) { // plays without stopping other sounds
    if (sound_id > MAX_FORE_SOUND_ID) { // additional sounds
        return true;
    }
    switch (sound_id) {
        case sound_25_presentation:
        case sound_29_meet_Jaffar:
        case sound_30_big_potion:
        case sound_32_mouse_music:
        case sound_33_small_potion:
        case sound_37_victory:
        case sound_39_low_weight:
        case sound_43_victory_Jaffar:
            return true;
        default:
            return false;
    }
}

bool is_exclusive_sound(int sound_id) { // background sounds that stop ambient music
    switch (sound_id) {
        case sound_24_death_regular:
        case sound_28_death_in_fight:
        case sound_36_out_of_time:
        case sound_41_end_level_music:
            return true;
        case sound_0_fell_to_death:
        case sound_46_chomped:
        case sound_48_spiked:
            return hitp_curr == 0;
        default:
            return false;
    }
}

int play_ogg_fore_sound(int sound_id, double sound_pos, bool trigger_callback) {
    if (!is_positional_audio) {
        return -2;
    }
    if (ogg_sound_tracks[sound_id] == NULL) {
        return -1;
    }
    if (ogg_fore_sound_playing) {
        if (last_sound_id == sound_id) {
            return 0;
        }
        if (is_interruptable_sound(sound_id)) {
            return 0;
        }
    }
    Mix_HaltMusic();
    Mix_VolumeMusic(MIX_MAX_VOLUME);
    if (trigger_callback) {
        ogg_fore_sound_playing = true;
        music_suspended = 1;
        Mix_HookMusicFinished(ogg_sound_callback);
    }
    last_sound_id = sound_id;
    curr_ogg_sound_id = sound_id;
    if (sound_pos != 0) {
        return Mix_FadeInMusicPos(ogg_sound_tracks[sound_id], 0, 0, sound_pos);
    }
    return Mix_PlayMusic(ogg_sound_tracks[sound_id], 0);
}

// determine whether to play fight music
bool is_fight_music(void) {
    // do not play fight music with one hit point
    if (hitp_curr < 2 && !is_kid_shadow) {
        return false;
    }

    // falling skeleton on top row of level 3 is not a fight
    if (current_level == 3 && Kid.room == 14) {
        return false;
    }

    // drunk purple weirdo on level 6A is not a fight
    // prince is concussed on level 6B
    if (current_level == 6) {
        return false;
    }

    // guards on level 10 do not fight after the level door has been opened
    if (current_level == 10 && leveldoor_open >= DEAD_KID_LEVEL_10_TIMER) {
        return false;
    }

    // only skeletons, Jaffar and his vision can see shadow
    if (is_kid_shadow &&
            Guard.charid != charid_4_skeleton &&
            current_level != 13 &&
            current_level != 14) {
        return false;
    }

    if (is_kid_mouse()) {
        return false;
    }

    if (Guard.charid == charid_1_shadow || Guard.charid == charid_24_mouse) {
        return false;
    }

    if (music_mode == 1 &&
            Guard.alive < 0 &&
            Guard.direction != dir_56_none &&
            (Guard.action == actions_3_in_midair || Guard.action == actions_4_in_freefall)) {
        return true;
    }

    if (music_mode == 1 && falling_guard.room != 0 && falling_guard.direction != dir_56_none) {
        return true;
    }

    if (is_dynamic_guard_fight > 0 &&
            dynamic_guard_count[Kid.room - 1] < dynamic_guard.frame + 1 &&
            Guard.direction != dir_56_none &&
            Guard.alive >= 0 &&
            Guard.alive < DROPDEAD_FRAMES &&
            level.guards_tile[Kid.room - 1] < MAX_TILES_PER_ROOM) {
        return true;
    }

    if (Guard.alive >= 0 || Guard.direction == dir_56_none) {
        return can_inactive_guard_see_kid == 2;
    }

    if (guard_dropping_down > 0 &&
            !tile_is_floor(get_tile(Guard.room, Guard.curr_col, Guard.curr_row)) &&
            curr_tile2 != tiles_20_wall) {
        return true;
    }

    if (can_guard_see_kid == 1 &&
            Guard.sword == sword_2_drawn &&
            Kid.sword == sword_2_drawn &&
            Kid.direction != Guard.direction) {
        return true;
    }

    if (music_mode == 0 &&
            can_guard_see_kid == 2 &&
            ((Kid.frame >= frame_45_turn && Kid.frame <= frame_52_turn) ||
             (Kid.frame >= frame_141_climbing_7 && Kid.frame <= frame_149_climbing_15))) {
        return false;
    }

    if (can_guard_see_kid == 2 && Kid.direction != Guard.direction && Guard.sword == sword_2_drawn) {
        return true;
    }

    if (can_guard_see_kid != 1 &&
            guard_kid_dist() < (Kid.sword == sword_2_drawn ? 43 : 26)) {
        return true;
    }

    if (music_mode == 1 &&
            can_guard_see_kid &&
            (Kid.sword == sword_2_drawn ||
             (Kid.frame >= frame_7_run && Kid.frame <= frame_14_run) ||
             (Kid.frame >= frame_53_runturn && Kid.frame <= frame_65_runturn))) {
        return true;
    }

    if (dynamic_guard_timer > 0 && (can_guard_see_kid == 2 || droppedout)) {
        return true;
    }

    if (Guard.charid == charid_4_skeleton &&
            can_guard_see_kid != 2 &&
            droppedout == 0) {
        return false;
    }

    return Guard.sword == sword_2_drawn &&
           guard_skill != GATE_GUARDING_SKILL &&
           dynamic_guard_timer == 0 &&
           (can_guard_see_kid != 1 ||
            (!is_guarding_gate && Kid.direction == Guard.direction && Kid.frame >= frame_16_standing_jump_1 && Kid.frame <= frame_44_running_jump_5) ||
            (!is_guarding_gate && Kid.direction == Guard.direction && Kid.frame >= frame_102_start_fall_1 && Kid.frame <= frame_108_fall_land_2));
}

bool set_track(int num) {
    if (track != NULL && num == track_number) {
        return false;
    }
    track_number = num;
    track = ogg_music_tracks[num];
    return true;
}

void play_track(int num, bool fadeIn) {
    bool is_track_set = set_track(num);
    if (music_suspended == 0 && music_suspended_timer == 0) {
        if (!is_track_set && !ogg_fore_sound_playing && Mix_PlayingMusic()) {
            return;
        }
        if (track == NULL) {
            return;
        }
        if (fadeIn) {
            Mix_FadeInMusic(track, -1, 2000);
        } else {
            Mix_PlayMusic(track, -1);
        }
    } else {
        Mix_HaltMusic();
    }
}

bool is_music_found(void) {
    return music_found;
}

bool is_heartbeat_music(void) {
    if (!heartbeat_music_found) {
        return false;
    }
    if (is_kid_shadow) {
        return false;
    }
    if (current_level == 6 && falling_exit_checkpoint >= 4 && hitp_curr > 0 && leveldoor_open == 0) {
        return true;
    }
    return hitp_curr == 1;
}

void start_ambient_music(bool fadeIn) {
    if (feather_fall_timer || ogg_fore_sound_playing || level_restart_timer > 0 || end_of_level_music) {
        return;
    }
    if (!is_positional_audio) {
        Mix_HaltMusic();
        return;
    }
    if (is_heartbeat_music()) {
        play_track(music_0_heartbeat_sound, fadeIn);
        return;
    }
    if (!is_music_on || music_mode < 0) {
        Mix_HaltMusic();
        return;
    }
    if (current_level == 0 || current_level >= 15) {
        return;
    }
    int music_gain = 0;
    bool fight_music = is_fight_music();
    if (!fight_music) {
        music_gain = level_music_gain[current_level];
    }
    Mix_VolumeMusic(ambient_music_volume + music_gain);
    if (fight_music) {
        if (current_level == 10) {
            if (ogg_music_tracks[music_9_fight_music] != NULL) {
                Mix_VolumeMusic(MIX_MAX_VOLUME);
                play_track(music_9_fight_music, fadeIn);
            } else {
                music_mode = -2;
            }
        } else if (current_level == 14) {
            if (ogg_music_tracks[music_7_fight_music] != NULL) {
                Mix_VolumeMusic(ambient_music_volume + 2);
                play_track(music_7_fight_music, fadeIn);
            } else {
                music_mode = -2;
            }
        } else if (ogg_music_tracks[music_8_fight_music] != NULL) {
            play_track(music_8_fight_music, fadeIn);
        } else {
            music_mode = -2;
        }
        return;
    }
    switch (current_level) {
        case 4:
        case 5:
            if (ogg_music_tracks[music_2_level_music] != NULL) {
                play_track(music_2_level_music, fadeIn);
            } else {
                music_mode = -2;
            }
            break;
        case 6:
            if (ogg_music_tracks[music_5_level_music] != NULL) {
                if (falling_exit_checkpoint < 4) {
                    play_track(music_5_level_music, fadeIn);
                } else if (leveldoor_open != 0) {
                    play_track(music_4_level_music, fadeIn);
                }
            } else {
                music_mode = -2;
            }
            break;
        case 10:
            if (ogg_music_tracks[music_6_level_music] != NULL) {
                play_track(music_6_level_music, fadeIn);
            } else {
                music_mode = -2;
            }
            break;
        case 11:
        case 12:
        case 13:
            if (ogg_music_tracks[music_3_level_music] != NULL) {
                play_track(music_3_level_music, fadeIn);
            } else {
                music_mode = -2;
            }
            break;
        case 14:
            if (ogg_music_tracks[music_4_level_music] != NULL) {
                play_track(music_4_level_music, fadeIn);
            } else {
                music_mode = -2;
            }
            break;
        default:
            if (ogg_music_tracks[music_1_level_music] != NULL) {
                play_track(music_1_level_music, fadeIn);
            } else {
                music_mode = -2;
            }
            break;
    }
}

void stop_ambient_music(void) {
    if (!is_sound_on || ogg_fore_sound_playing || feather_fall_timer > 0) {
        return;
    }
    Mix_HaltMusic();
}
