/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"
#include <ctype.h>

// outbound dynamic guard coordinates
#define OUTBOUND_LEFT_X 41
#define OUTBOUND_RIGHT_X 215

// .ini file parser adapted from https://gist.github.com/OrangeTide/947070
int ini_load(const char *filename, void (*report)(const char *name, const char *value)) {
    char name[64];
    char value[256];
    char section[128] = "";
    char *s;
    FILE *f;
    int cnt;

    f = fopen(filename, "r");
    if (!f) {
        return -1;
    }

    while (!feof(f)) {
        if (fscanf(f, "[%127[^]#\n]]\n", section) == 1) {
        } else if ((cnt = fscanf(f, " %63[^=#\n] = %255[^#\n]", name, value))) {
            if (cnt == 1) {
                *value = 0;
            }
            for (s = name + strlen(name) - 1; s > name && isspace(*s); s--) {
                *s = 0;
            }
            for (s = value + strlen(value) - 1; s > value && isspace(*s); s--) {
                *s = 0;
            }
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
            if (strcasecmp("renderer", name) != 0 || strcasecmp("Windows", section) == 0)
#elif defined __linux__ || unix || __unix || __unix__
            if (strcasecmp("renderer", name) != 0 || strcasecmp("Linux", section) == 0)
#elif defined __APPLE__ || __MACH__
            if (strcasecmp("renderer", name) != 0 || strcasecmp("MacOS", section) == 0)
#endif
                report(name, value);
        }
        if (fscanf(f, " #%*[^\n]")) {}
        if (fscanf(f, " \n")) {}
    }

    fclose(f);
    return 0;
}

game_settings_type global_settings = {
    .start_minutes_left = 75,
    .start_ticks_left = 719, // 59.92 sec (60 resets time)
    .start_hitp = 3,
    .max_hitp_allowed = 12,
    .vga_palette = VGA_PALETTE_DEFAULT,
    .level_6_seq_potion_column = 5,
    .level_6_seq_potion_row = 1,
    .level_6_seq_sword_left_x = 186,
    .level_6_seq_sword_top_y = 81,
    .mirror_level = 6,
    .mirror_room = 1,
    .mirror_column = 5,
    .mirror_row = 0,
    .falling_exit_level = 6,
    .falling_exit_room = 22,
    .falling_entry_level = 7,
    .falling_entry_room = 17,
    .mouse_level = 8,
    .mouse_room = 16,
    .mouse_start_x = 200,
    .mouse_2_level = 12,
    .mouse_2_room = 2,
    .mouse_2_room_left_first_empty_col = 6,
    .mouse_2_potion_column = 8,
    .mouse_2_potion_row = 0,
    .mouse_2_start_x = 170,
    .mouse_2_crouch_delay = 32,
    .loose_tiles_level = 13,
    .loose_tiles_room_1 = 23,
    .loose_tiles_room_2 = 16,
    .loose_tiles_first_tile = 22,
    .loose_tiles_last_tile = 29,
    .loose_floor_delay = 11,
    .loose_shake_count = 3, // number of times a loose tile shakes before falling
    .jaffar_victory_flash_time = 18,
    .win_level = 14,
    .win_room = 5,
    .flame_level = 14,
    .flaming_sword_room = 9,
    .flaming_sword_tilepos = 7,
    .falling_bridge_delay = 68,
    .inverse_level = 13, // level where the gates, potions and level doors are inversed
    .moving_wall_delay = 8,
    .moving_wall_level = 13,
    .moving_wall_room = 7,
    .moving_wall_row = 1,
    .moving_wall_col_left = 3,
    .moving_wall_col_right = 8,
    .prison_level = 15,
    .fat_guard_level = 8,
    .fat_guard_skill = 9,

    .default_game_speed = 5, // smaller is faster
    .one_hp_game_speed = 7,
    .fast_game_speed = 3,
    .slow_game_speed = 6,

    // data:02B2
    // 0 = dungeon, 1 = palace/temple
    .tbl_level_type = {0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0},
    // 0 = first type of dungeon/palace, 1 = second type of dungeon/palace, 2 = temple
    .tbl_level_subtype = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 1, 1, 1, 0},
    // Prince of Persia 1.3+ functionality
    .tbl_level_color = {0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 4, 0},
    // data:03D4
    // 0 = guard, 1 = fat, 2 = skeleton, 3 = Jaffar, 4 = kid (experimental), 5 = Jaffar's vision, 6 = purple weirdo, 7 = bird
    .tbl_guard_type = {0, 0, 0, 2, 0, 0, 6, 0, 0, 0, 7, 0, 2, 5, 3, 0},
    // data:0EDA
    .tbl_guard_hp = {3, 3, 3, 3, 3, 4, 5, 4, 3, 5, 5, 5, 3, 5, 6, 3},
    .tbl_cutscenes_by_index = {0, 1, 2, 3, 4, 5, 0, 7, 8, 0, 4, 11, 12, 13, 9, 15},
    .tbl_entry_pose = {0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0},
    .tbl_seamless_exit = {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 23, -1, -1, -1},

    // empty potion color animation (levels 0-15)
    // non-standard potions are implemented using empty potions
    .tbl_empty_potion_color = {
        -1, // no color
            -1,
            -1,
            color_15_brightwhite,
            -1,
            color_11_brightcyan,
            color_13_brightmagenta,
            color_14_brightyellow,
            color_14_brightyellow,
            color_14_brightyellow,
            color_15_brightwhite,
            -1,
            color_14_brightyellow,
            color_12_brightred,
            color_7_lightgray,
            -1
        },

    // empty potion size (levels 0-15)
    .tbl_empty_potion_size = {
        0, // small,
        0,
        0,
        1, // large
        0,
        1,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        1,
        0,
        0
    },

    .tbl_guard_dat = {
        "GUARD.DAT",
        "FAT.DAT",
        "SKEL1.DAT",
        "VIZIER.DAT",
        "KID.DAT", // experimental (not used)
        "VIZVISN.DAT",
        "PURPLE.DAT",
        "BIRD.DAT"
    },
    // custom level graphics files
    .tbl_level_dat = {
        "DUNGEON1.DAT", // demo level
        "DUNGEON1.DAT", // level 1
        "DUNGEON1.DAT", // level 2
        "DUNGEON1.DAT", // level 3
        "PALACE1.DAT",  // level 4
        "PALACE1.DAT",  // level 5
        "PALACE1.DAT",  // level 6 and 6B
        "DUNGEON1.DAT", // level 7
        "DUNGEON1.DAT", // level 8
        "DUNGEON1.DAT", // level 9
        "TEMPLE.DAT",   // level 10
        "DUNGEON2.DAT", // level 11
        "DUNGEON2.DAT", // level 12
        "DUNGEON2.DAT", // level 13 (12A)
        "PALACE2.DAT",  // level 14 (13)
        "DUNGEON1.DAT", // level 15 (prison)
    },
    // wda flag for palace levels
    .tbl_level_wda = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0},

    // demo skills
    .demo_guard_skill = 3,
    .demo_kid_skill = 6,

    // guard skills (skill 8 is a gate keeping guard with stats like skill 3)

    // probability of striking when prince is in range (out of 255)
    .strikeprob      = { 61, 100,  61,  61,  61,  40, 100, 220,  61,  48, 220, 150},
    // probability of striking after a block (out of 255)
    .restrikeprob    = {  0,   0,   0,   5,   5, 175,  16,   8,   5, 255, 255, 150},
    // probability of blocking a strike (out of 255)
    .blockprob       = {  0, 150, 150, 200, 200, 255, 200, 250, 200, 255, 255, 255},
    // probability of implied blocking (right after a block) (out of 255)
    .impblockprob    = {  0,  61,  61, 100, 100, 145, 100, 250, 100, 145, 255, 175},
    // probability of advancing into the range (out of 255)
    .advprob         = {255, 200, 200, 200, 255, 255, 200,   0, 200, 255,  24,  48},
    // number of time ticks being idle after getting hurt
    .refractimer     = { 16,  16,  16,  16,   8,   8,   8,   8,  16,   8,   0,   0},
    // number of time ticks being idle after getting punched without a hitpoint taken
    .refractimer2    = { 10,   7,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0},
    // number of additional hit points (tbl_guard_hp defines the number of hitpoints per level)
    .extrastrength   = {  0,   0,   0,   0,   1,   1,   0,   0,   0,   2,   0,   0},

    // Jaffar's parry probability (out of 255) that provides a significantly improved defense
    .jaffarparryprob = 200,

    // shadow initialization
    // {frame, x coord, y coord, direction, column (not needed), row, action, fall_x, fall_y}
    .init_shad_6b  = {frame_15_stand,   173, 118,  dir_0_right, 0, 1, actions_0_stand,    0, 0},
    .init_shad_7   = {frame_15_stand,   129, 81,   dir_0_right, 0, 0, actions_0_stand,    0, 0},
    .init_shad_8   = {frame_109_crouch, 173, 55,   dir_0_right, 0, 0, actions_1_run_jump, 0, 0},
    .init_shad_9   = {frame_15_stand,   180, 118,  1,           0, 1, actions_0_stand,    0, 0}, // direction 1 makes shadow moonwalk
    .init_shad_9_2 = {frame_15_stand,   80,  137,  dir_0_right, 0, 1, actions_0_stand,    0, 0},
    .init_shad_10  = {frame_15_stand,   228, 179,  dir_FF_left, 0, 2, actions_0_stand,    0, 0},

    // auto moves
    .demo_moves = {
        {0x00, 0},
        {0x01, 1},
        {0x10, 0},
        {0x16, 2},
        {0x1A, 0},
        {0x26, 6},
        {0x3C, 2},
        {0x40, 0},
        {0x41, 1},
        {0x47, 5},
        {0x53, 0},
        {0x54, 1},
        {0x62, 0},
        {0x6A, 2},
        {0x6E, 0},
        {0x88, 2},
        {0x8C, 0},
        {0x8F, 9},
        {0x92, 0},
        {0x96, 5},
        {0xA3, 6},
        {0xAE, 3},
        {0x9E, 0},
        {0xB2, 6},
        {0xCC, 1},
        {0xFA, 0},
        {0xFF, -1},
    },
    .demo_moves_2 = {
        {0x00, 0},
        {0x01, 1},
        {0x2E, 2},
        {0x2F, 6},
        {0x49, 0},
        {0x4C, 2},
        {0x52, 0},
        {0x59, 9},
        {0x5D, 0},
        {0x5E, 1},
        {0x86, 5},
        {0x93, 6},
        {0xA0, 0},
        {0xA3, 4},
        {0xBD, 0},
        {0xC6, 2},
        {0xC9, 1},
        {0xD6, 8},
        {0xDF, 0},
        {0xF2, 1},
        {0x115, 0},
        {0x11B, 3},
        {0x120, -1},
    },
    .level_6_intro_seq = {
        {0x02, -3}, // pick up item
        {0x05, 50},
        {0x0F, -2}, // play draw sword sound
        {0x1D, -1}, // gate closed sound
        {0x1E, 13},
        {0x29, 1},
    },
    .level_6_drink_seq = {
        {0x21, 5},
        {0x34, 43},
        {0x3E, 20}
    },
    .shad_level_6b = {
        {0x00, 0},
        {0x01, 2},
        {0x01, 1},
        {0x16, 1},
        {0x1E, 1},
        {0x26, 1},
        {0x2E, 1},
        {0x59, -2},
    },
    .shad_level_7 = {
        {0x00, 0},
        {0x01, 1},
        {0x1A, 5},
        {0x20, 6},
        {0x40, 1},
        {0x59, -2},
    },
    .shad_level_8 = {
        {0x00, 4},
        {0x22, 1},
        {0x3D, 7},
        {0x59, -2},
    },
    .shad_level_9 = {
        {0x00, 0},
        {0x12, 1},
        {0x1C, 6},
        {0x31, 0},
        {0x43, 2},
        {0x44, 1},
        {0x56, 6},
        {0x5A, 3},
        {0x6F, 1},
        {0x87, -2},
    },
    .shad_level_9_2 = {
        {0x00, 0},
        {0x22, 1},
        {0x3D, 7},
        {0x59, -2},
    },

    .dead_guards = {
        [0] = {0}, // level 0 (demo)
        [1] = {
            // level 1
            16,                       // room #
            dir_FF_left,              // direction
            117,                      // x-coordinate
            0,                        // row
            0,                        // current frame (assigned during the game)
            guard_light_blue,         // optional guard color
            frame_185_dead,           // initial frame
            0,                        // unused
            0,                        // unused
            0xFF                      // 0xFF = alternate dead frame
        },
        [2 ... 15] = {0} // levels 2-15
    },

    // inactive guards - 1 per level
    .inactive_guards = {
        [0 ... 3] = {0}, // levels 0 (demo) - 3
        [4] = {
            // level 4
            24,                       // room #
            dir_FF_left,              // direction
            168,                      // x-coordinate
            1,                        // row
            0,                        // current frame (assigned during the game)
            guard_purple,             // optional guard color (0 = match active room guard)
            frame_166_stand_inactive, // initial frame
            0,                        // unused
            0,                        // unused
            0                         // 0xFF = alternate dead frame
        },
        [5 ... 6] = {0}, // levels 5 - 6
        [7] = {
            // level 7
            22,
            dir_FF_left,
            152,
            2,
            0,
            guard_dark_orange,
            frame_166_stand_inactive,
            0,
            0,
            0
        },
        [8] = {0}, // level 8
        [9] = {
            // level 9
            17,
            dir_0_right,
            78,
            1,
            0,
            guard_light_orange,
            frame_166_stand_inactive,
            0,
            0,
            0
        },
        [10] = {
            // level 10
            5,
            dir_0_right,
            138,
            2,
            0,
            0,
            frame_166_stand_inactive,
            0,
            0,
            0
        },
        [11] = {
            // level 11
            18,
            dir_0_right,
            138,
            2,
            0,
            guard_light_orange,
            frame_166_stand_inactive,
            0,
            0,
            0
        },
        [12] = {
            // level 12
            23,
            dir_0_right,
            114,
            1,
            0,
            0,
            frame_166_stand_inactive,
            0,
            0,
            0
        },
        [13 ... 15] = {0}, // levels 13-15
    },

    .dynamic_guards = {
        [0] = { // demo level
            [0 ... 2] = {0}, // array index = (room # minus 1), 24 rooms per level
            [3] = {
                4,                 // room # (should be equal to "array index + 1")
                dir_0_right,       // direction
                OUTBOUND_LEFT_X,   // x (outbound_left_x or outbound_right_x)
                0,                 // row
                0,                 // additional # of dynamic guards (0 = one dynamic guard)
                guard_dark_red,    // guards' color
                3,                 // skill (defaults to "settings->demo_guard_skill" on level 0 in demo mode)
                0,                 // unused
                0,                 // unused
                0                  // 0xFF = alternate dead frame
            },
            [4 ... 23] = {0},
        },
        [1] = { // level 1
            [0 ... 23] = {0},
        },
        [2] = { // level 2
            [0 ... 2] = {0},
            [3] = {
                4,
                dir_0_right,
                OUTBOUND_LEFT_X,
                1,
                1,
                guard_dark_orange,
                1,
                0,
                0,
                0
            },
            [4 ... 13] = {0},
            [14] = {
                15,
                dir_0_right,
                OUTBOUND_LEFT_X,
                1,
                0,
                guard_dark_red,
                1,
                0,
                0,
                0
            },
            [15 ... 23] = {0},
        },
        [3] = { // level 3
            [0 ... 23] = {0},
        },
        [4] = { // level 4
            [0 ... 20] = {0},
            [21] = {
                22,
                dir_FF_left,
                OUTBOUND_RIGHT_X,
                0,
                0,
                guard_purple,
                3,
                0,
                0,
                0
            },
            [22 ... 23] = {0},
        },
        [5] = { // level 5
            [0 ... 10] = {0},
            [11] = {
                12,
                dir_0_right,
                OUTBOUND_LEFT_X,
                1,
                0,
                guard_yellow,
                4,
                0,
                0,
                0
            },
            [12 ... 23] = {0},
        },
        [6] = { // level 6
            [0 ... 23] = {0},
        },
        [7] = { // level 7
            [0 ... 23] = {0},
        },
        [8] = { // level 8
            [0 ... 22] = {0},
            [23] = {
                24,
                dir_FF_left,
                OUTBOUND_RIGHT_X,
                2,
                0,
                guard_light_blue,
                4,
                0,
                0,
                0
            },
        },
        [9] = { // level 9
            [0 ... 11] = {0},
            [12] = {
                13,
                dir_FF_left,
                OUTBOUND_RIGHT_X,
                2,
                0,
                guard_green,
                4,
                0,
                0,
                0
            },
            [13 ... 23] = {0},
        },
        [10] = { // level 10
            [0 ... 12] = {0},
            [13] = {
                14,
                dir_FF_left,
                OUTBOUND_RIGHT_X,
                0,
                0,
                0,
                4,
                0,
                0,
                0
            },
            [14 ... 23] = {0},
        },
        [11] = { // level 11
            [0 ... 13] = {0},
            [14] = {
                15,
                dir_FF_left,
                OUTBOUND_RIGHT_X,
                0,
                0,
                guard_light_blue,
                4,
                0,
                0,
                0
            },
            [15 ... 23] = {0},
        },
        [12] = { // level 12
            [0 ... 23] = {0},
        },
        [13] = { // level 13
            [0 ... 23] = {0},
        },
        [14] = { // level 14
            [0 ... 23] = {0},
        },
        [15] = { // level 15
            [0 ... 23] = {0},
        },
    },

    .torch_level_colors = {
        0,  // level 0 (demo)  - default orange
        0,  // level 1
        0,  // level 2
        0,  // level 3
        0,  // level 4
        0,  // level 5
        51, // levels 6/6B     - purple
        0,  // level 7
        0,  // level 8
        0,  // level 9
        11, // level 10        - argon blue
        0,  // level 11
        0,  // level 12
        0,  // level 13 (12B)
        11, // level 14 (13)   - argon blue
        0,  // level 15 (14)
    },
    .torch_level_colors_dimmed = {
        37,  // level 0 (demo)  - dim orange
        37,  // level 1
        37,  // level 2
        37,  // level 3
        37,  // level 4
        37,  // level 5
        17,  // levels 6/6B     - dim purple
        37,  // level 7
        37,  // level 8
        37,  // level 9
        6,   // level 10        - dim blue
        37,  // level 11
        37,  // level 12
        37,  // level 13 (12B)
        6,   // level 14 (13)   - dim blue
        37,  // level 15 (14)
    },
    .dark_levels = { 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0 },
    .level_names = {
        "DEMO LEVEL",
        "LEVEL 1: THE GREAT ESCAPE",
        "LEVEL 2: MORE GUARDS",
        "LEVEL 3: I SEE DEAD PEOPLE",
        "LEVEL 4: THE PALACE",
        "LEVEL 5: SPILLED POISON",
        "LEVEL 6: OOH! SPARKLY",
        "LEVEL 6B: THE HANGOVER",
        "LEVEL 7: WHERE AM I?",
        "LEVEL 8: IT'S A TRAP",
        "LEVEL 9: ORIENTATION MATTERS",
        "LEVEL 10: TEMPLE AND FLAME",
        "LEVEL 11: THAT PESKY GUARD",
        "LEVEL 12: THE PALACE TOWER",
        "LEVEL 12B: JAFFAR IS WATCHING",
        "LEVEL 13: ENTER THE DRAGON"
    },
    .sound_prio_table = {
        0x0F, // sound_0_fell_to_death
        0x1E, // sound_1_falling
        0x02, // sound_2_tile_crashing
        0x32, // sound_3_button_pressed
        0x37, // sound_4_gate_closing
        0x33, // sound_5_gate_opening
        0x09, // sound_6_gate_closing_fast
        0x20, // sound_7_gate_stop
        0x29, // sound_8_bumped
        0x50, // sound_9_grab
        0x11, // sound_10_sword_vs_sword
        0x12, // sound_11_sword_moving
        0x0C, // sound_12_guard_hurt
        0x0B, // sound_13_kid_hurt
        0x28, // sound_14_leveldoor_closing
        0x31, // sound_15_leveldoor_sliding
        0x02, // sound_16_fat_fart
        0x36, // sound_17_soft_land
        0x7D, // sound_18_drink
        0x31, // sound_19_draw_sword
        0x91, // sound_20_loose_shake_1
        0x96, // sound_21_loose_shake_2
        0x9B, // sound_22_loose_shake_3
        0xA0, // sound_23_footstep
        0x01, // sound_24_death_regular
        0x01, // sound_25_presentation
        0x01, // sound_26_embrace
        0x01, // sound_27_cutscene_2_4_6_12
        0x01, // sound_28_death_in_fight
        0x13, // sound_29_meet_Jaffar
        0x01, // sound_30_big_potion
        0x02, // sound_31_flame_hurts_Jaffar
        0x01, // sound_32_mouse_music
        0x01, // sound_33_small_potion
        0x0C, // sound_34_skeleton_hurt
        0x01, // sound_35_cutscene_8_9
        0x01, // sound_36_out_of_time
        0x01, // sound_37_victory
        0x01, // sound_38_blink
        0x00, // sound_39_low_weight
        0x01, // sound_40_cutscene_12_short_time
        0x01, // sound_41_end_level_music
        0x1E, // sound_42_guard_falling
        0x01, // sound_43_victory_Jaffar
        0x21, // sound_44_skel_alive
        0x01, // sound_45_jump_through_mirror
        0x09, // sound_46_chomped
        0x10, // sound_47_chomper
        0x0A, // sound_48_spiked
        0x16, // sound_49_spikes_out
        0x01, // sound_50_story_2_princess
        0x20, // sound_51_princess_door_opening
        0x01, // sound_52_story_4_Jaffar_leaves
        0x01, // sound_53_story_3_Jaffar_comes
        0x01, // sound_54_intro_music
        0x01, // sound_55_story_1_absence
        0x01, // sound_56_ending_music
        0x02, // sound_57_sword_striked
        0x02, // sound_58_say_whew
        0x03, // sound_59_guard_snoring
        0x17, // sound_60_spikes_in
        0x0B, // sound_61_shadow_separation
        0x00
    },
    .tbl_quotes = {
        " \"(****/****) Incredibly realistic. . . The\n"
        " adventurer character actually looks human as he\n"
        " runs, jumps, climbs, and hangs from ledges.\"\n"
        "\n"
        "                                    Computer Entertainer"
        "\n"
        "\n"
        "\n"
        "\n"
        " \"A tremendous achievement. . . Mechner has crafted\n"
        " the smoothest animation ever seen in a game of this\n"
        " type.\n"
        "\n"
        " \"PRINCE OF PERSIA is the STAR WARS of its field.\"\n"
        "\n"
        "                                   Computer Gaming World",
        " \"An unmitigated delight. . . comes as close to\n"
        " (perfection) as any arcade game has come in a long,\n"
        " long time. . . what makes this game so wonderful (am\n"
        " I gushing?) is that the little onscreen character\n"
        " does not move like a little onscreen character -- he\n"
        " moves like a person.\"\n"
        "\n"
        "                                                  Nibble",
        " Brilliant mod! Amazing mod! After I can compile on\n"
        " Linux, I'll upload all the videos of it. It's just too\n"
        " brilliant!\n"
        "\n"
        "                                    atrueprincefanfrom18",
        NULL
    },
    .aspect_ratio = false,
    .native_ratio = false,
    .run_in_window = true,
    .positional_audio = true,
    .controller_threshold = 8000,
    .controller_rumble = true,
    .enable_vsync = false,
};

static void ini_callback(const char *name, const char *value) {
    if (strcmp("renderer", name) == 0) {
        if (strcmp("auto", value) != 0) {
            strcpy(global_settings.renderer, value);
        }
        return;
    }
    if (strcmp("enable_vsync", name) == 0) {
        if (strcmp("yes", value) == 0) {
            global_settings.enable_vsync = true;
        }
        return;
    }
    if (strcmp("full_screen", name) == 0) {
        if (strcmp("yes", value) == 0) {
            global_settings.run_in_window = false;
        }
        return;
    }
    if (strcmp("aspect_ratio", name) == 0) {
        if (strcmp("original", value) == 0) {
            global_settings.aspect_ratio = true;
        } else if (strcmp("native", value) == 0) {
            global_settings.native_ratio = true;
        }
        return;
    }
    if (strcmp("enhanced_audio", name) == 0) {
        if (strcmp("no", value) == 0) {
            global_settings.positional_audio = false;
        }
        return;
    }
    if (strcmp("controller_threshold", name) == 0) {
        int threshold = atoi(value);
        if (threshold > 0) {
            global_settings.controller_threshold = threshold;
        }
        return;
    }
    if (strcmp("controller_rumble", name) == 0) {
        if (strcmp("no", value) == 0) {
            global_settings.controller_rumble = false;
        }
        return;
    }
}

void load_global_options(void) {
    const char* ini_file = "PRINCE.INI";
    strcpy(global_settings.renderer, "");
    if (ini_load(locate_file(ini_file), ini_callback) != 0) {
        printf("File %s was not found. Using the default settings.\n", ini_file);
    }
    settings = &global_settings;
    enable_positional_audio = settings->positional_audio && enhanced_sound_file_exists();
    enable_lighting = enable_pause_menu = 1;
}
