/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"
#include "proto.h"

void draw_extras(void);

// custom logic - rename folders and files
char screenshot_filename[POP_MAX_PATH] = "SHOT.PNG";
int screenshot_index = 0;

// Use incrementing numbers and a separate folder, like DOSBox.
void make_screenshot_filename(void) {
    create_folder_if_missing(save_folder);
    create_folder_if_missing(screenshot_folder);
    // Find the first unused filename:
    for (;;) {
        snprintf(screenshot_filename, sizeof(screenshot_filename), "%s/SHOT%03d.PNG", screenshot_folder, screenshot_index);
        if (!file_exists(screenshot_filename)) {
            return;
        }
        screenshot_index++;
    }
}

#define EVENT_OFFSET 1 // Add this number to displayed event numbers. Use 1 for Apoplexy compatibility.

void show_result(int result, const char* what) {
    // custom logic - play sound on successful save
    if (result == 0) {
        if (is_positional_audio && !ogg_fore_sound_playing) {
            play_sound(sound_65_camera_shutter);
        }
        flash_color = color_15_brightwhite;
        flash_time = 1;
        return;
    }
    display_text("COULD NOT SAVE %s", what);
    text_time_total = 24;
    text_time_remaining = 24;
}

// Save a screenshot.
void save_screenshot(bool want_extras) {
    make_screenshot_filename();
    // custom logic - draw extras when Ctrl-F12 is pressed (in cheats mode only)
    if (cheats_enabled && want_extras) {
        draw_extras();
    }
    int result = IMG_SavePNG(get_final_surface(), screenshot_filename);
    show_result(result, "SCREENSHOT");
}

// Switch to the given room and draw it.
void switch_to_room(int room, char_type old_guard, word old_guard_hp) {
    drawn_room = room;

    if (room == PLACEHOLDER_ROOM) {
        for (int tilepos = 0; tilepos < MAX_TILES_PER_ROOM; tilepos++) {
            load_alter_mod(tilepos);
        }
    } else {
        load_room_links();
    }

    if (settings->tbl_level_type[current_level]) {
        gen_palace_wall_colors();
    }

    if (room != PLACEHOLDER_ROOM) {
        // for potion bubbles
        for (int tilepos = 0; tilepos < MAX_TILES_PER_ROOM; tilepos++) {
            int tile_type = curr_room_tiles[tilepos] & 0x1F;
            if (tile_type == tiles_10_potion) {
                int modifier = curr_room_modif[tilepos];
                int potion_type = modifier >> 3;
                if (potion_type < potion_7_empty_small && (modifier & 7) == 0) {
                    curr_room_modif[tilepos] += prandom(5) + 1;
                }
            }
        }

        char_type2 old_inctv_grd = inactive_guard[room - 1];

        // enter guards
        leave_guard();
        enter_guard();

        // restore shadow/the mouse
        if (old_guard.charid > 0) {
            if (Guard.alive >= 0 || Guard.direction == dir_56_none) {
                Guard = old_guard;
                guardhp_curr = old_guard_hp;
            }
            inactive_guard[room - 1] = old_inctv_grd;
        }
    }

    redraw_screen(1);
}

bool event_used[256] = { false };
bool has_trigger_potion = false;

// delta vectors for room links
const int dx[4] = {-1, +1,  0,  0};
const int dy[4] = { 0,  0, -1, +1};

int xpos[NUMBER_OF_ROOMS + 1] = {0};
int ypos[NUMBER_OF_ROOMS + 1] = {0};

// custom logic - shows guard stats
void draw_guard_stats(rect_type rect, int skill, int hitp) {
    char guard_info[30];
    if (hitp < 0) {
        if (skill < 0) {
            snprintf(guard_info, sizeof(guard_info), "sk* hp*");
        } else {
            snprintf(guard_info, sizeof(guard_info), "sk%d hp*", skill);
        }
    } else {
        snprintf(guard_info, sizeof(guard_info), "sk%d hp%d", skill, hitp);
    }
    show_text_with_color(&rect, text_align_center, text_align_top, guard_info, color_10_brightgreen);
}

// Show annotations for non-visible things, like: room bounds, room numbers, door events, loose floors, potion types, special events, ...
// (this will make the function even more like a cheat)
void draw_extras(void) {
    // ambiguous tiles
    // The editor branch has something similar...
    for (int tilepos = 0; tilepos < MAX_TILES_PER_ROOM; tilepos++) {
        int tile_type = curr_room_tiles[tilepos] & 0x1F;
        int modifier = curr_room_modif[tilepos];
        int row = tilepos / 10;
        int col = tilepos % 10;
        int y = row * 63 + 3;
        int x = col * 32;

        // special floors
        rect_type floor_rect = {y + 60 - 3, x, y + 63 - 3, x + 32};

        // loose floors
        if (tile_type == tiles_11_loose) {
            int color = color_15_brightwhite;
            if (get_tile_ctrl_bits(drawn_room, tilepos, 0) == 1) {
                color = color_7_lightgray; // brittle floor
            } else if (get_tile_ctrl_bits(drawn_room, tilepos, 1) == 1) {
                color = color_12_brightred; // totally loose floor
            }
            show_text_with_color(&floor_rect, text_align_center, text_align_top, "~~~~", color);
        }

        // buttons
        if (tile_type == tiles_15_opener) {
            show_text_with_color(&floor_rect, text_align_center, text_align_top, "^^^^", color_10_brightgreen);
        }
        if (tile_type == tiles_6_closer) {
            floor_rect.top -= 2;
            show_text_with_color(&floor_rect, text_align_center, text_align_top, "xxxx", color_9_brightblue); // Only the top half is visible, looks like an inverted "^" or a tiny "v".
        }

        // potion types
        if (tile_type == tiles_10_potion) {
            char* empty_potion_type;
            byte empty_potion_color;
            empty_potion_color = settings->tbl_empty_potion_color[current_level];
            if (current_level == settings->mouse_2_level && drawn_room == settings->mouse_2_room) {
                empty_potion_color = color_15_brightwhite;
            }
            switch (empty_potion_color) {
                case color_7_lightgray:
                    empty_potion_type = "mouse";
                    break;
                case color_11_brightcyan:
                    empty_potion_type = "antidote";
                    break;
                case color_12_brightred: {
                    if (current_level == settings->inverse_level) {
                        empty_potion_type = "---";
                        empty_potion_color = color_9_brightblue;
                    } else {
                        empty_potion_type = "x";
                    }
                    break;
                }
                case color_13_brightmagenta:
                    empty_potion_type = "crazy";
                    break;
                case color_14_brightyellow:
                    empty_potion_type = "light";
                    break;
                case color_15_brightwhite:
                    empty_potion_type = "change";
                    break;
                default:
                    empty_potion_type = "x";
                    empty_potion_color = color_8_darkgray;
                    break;
            }
            struct pot_type {
                int color;
                const char* text;
            } pot_types[7] = {
                {empty_potion_color, empty_potion_type}, // empty
                {color_12_brightred, "+1"}, // heal
                {color_12_brightred, "+++"}, // life
                {color_10_brightgreen, "slow\nfall"}, // slow fall
                {color_10_brightgreen, "flip"}, // upside down
                {color_9_brightblue, "-1"}, // hurt
                {color_9_brightblue, "trig"}, // open
            };
            int potion_type = modifier >> 3;
            // custom logic - handle spilled potions
            if (potion_type >= potion_9_spilled_left_small && potion_type <= potion_12_spilled_right_large) {
                potion_type = modifier & 0x07;
            }
            if (potion_type >= potion_0_custom && potion_type < potion_7_empty_small) {
                rect_type pot_rect = { y + 26, x, y + 46, x + 60 };
                show_text_with_color(&pot_rect, text_align_center, text_align_top, pot_types[potion_type].text, pot_types[potion_type].color);
            }
        }

        // triggered door events
        if (tile_type == tiles_6_closer || tile_type == tiles_15_opener
                || (has_trigger_potion && drawn_room == 8 && tilepos == 0) // triggered when player drinks an open potion
           ) {
            // custom logic - do not show event on a tile that does not open anything on level 13
            if (current_level == 14 && drawn_room == 19 && tilepos == 2) {
                continue;
            }

            int first_event = modifier;
            int last_event = modifier;
            while (last_event < 256 && get_doorlink_next(last_event)) last_event++;
            char events[256 * 4] = ""; // More than enough space to list all the numbers from 0 to 255.
            int events_pos = 0;
            for (int event = first_event; event <= last_event && (size_t) events_pos < sizeof(events); event++) {
                int len = snprintf(events + events_pos, sizeof(events) - events_pos, "%d ", event + EVENT_OFFSET);
                if (len < 0) break; // snprintf might return -1 if the buffer is too small.
                events_pos += len;
            }
            --events_pos;
            if (events_pos > 0 && (size_t) events_pos < sizeof(events)) events[events_pos] = '\0'; // trim trailing space
            rect_type buttonmod_rect = {y, x, y + 60 - 3, x + 32};
            show_text_with_color(&buttonmod_rect, text_align_center, text_align_bottom, events, color_14_brightyellow);
        }

        // TODO: Add an option to merge events pointing to the same tile?

        rect_type events_rect = {y, x, y + 63 - 3, x + 32 - 7};

        // custom logic - level 12B exit door closer event
        if (current_level == 13 && drawn_room == 12 && tile_type == tiles_16_level_door_left) {
            show_text_with_color(&events_rect, text_align_center, text_align_bottom, "1", color_14_brightyellow);
            continue;
        }

        // door events that point here
        char events[256 * 4] = "";
        int events_pos = 0;
        for (int event = 0; event < 256 && (size_t) events_pos < sizeof(events); event++) {
            if (event_used[event] && get_doorlink_room(event) == drawn_room && get_doorlink_tile(event) == tilepos) {
                int len = snprintf(events + events_pos, sizeof(events) - events_pos, "%d ", event + EVENT_OFFSET);
                if (len < 0) break;
                events_pos += len;
            }
        }
        --events_pos;
        if (events_pos > 0 && (size_t) events_pos < sizeof(events)) events[events_pos] = '\0'; // trim trailing space
        if (events[0] != '\0') {
            show_text_with_color(&events_rect, text_align_center, text_align_bottom, events, color_14_brightyellow);
        }

        // special events
        char* special_event = NULL;
        // not marked: level 1 falling entry
        if (current_level == 1 && drawn_room == 5 && tilepos == 2) {
            special_event = "start\ntrig\n\n\n"; // triggered at start
        }
        // "teleport" on level 1
        if (current_level == 1 && drawn_room == 10 && tilepos == 22) {
            special_event = "\n\n\nteleport";
        }
        if (current_level == 1 && drawn_room == 18 && tilepos == 2) {
            special_event = "teleport\ndest";
        }
        if (tile_type == tiles_15_opener && get_tile_ctrl_bits(drawn_room, tilepos, 1) != 0) {
            special_event = "sticky\n\n\n\n"; // gets stuck after being pressed
        }
        if (current_level == 3) {
            if (
                (drawn_room == 1 && tilepos == 15 && tile_type == tiles_21_skeleton) ||
                (drawn_room == 14 && tilepos == 2 && tile_type == tiles_21_skeleton) ||
                (drawn_room == 6 && tilepos == get_tilepos(6, 2) && tile_type == tiles_21_skeleton) ||
                (drawn_room == 17 && tilepos == get_tilepos(4, 2) && tile_type == tiles_21_skeleton)
            ) {
                special_event = "skel wake"; // skeleton wakes
            }
        }
        if (current_level == 12 && drawn_room == 24 && tilepos == 17 && tile_type == tiles_21_skeleton) {
            special_event = "skel wake"; // skeleton wakes
        }
        if (current_level == 3 && drawn_room == 3 && tilepos == 14) {
            special_event = "skel cont"; // skeleton continue
        }
        if (current_level == 6 && drawn_room == 1 && tilepos == 15) {
            special_event = "mirror"; // mirror appears
        }
        if (current_level == 6 && drawn_room == 22 && tilepos == 22) {
            special_event = "exit\ndown"; // exit by falling
        }
        if (current_level == 10 && drawn_room == 8 && tilepos == 22) {
            special_event = "door\nkeeper"; // responsible for the level door
        }
        if (current_level == 10 && drawn_room == 20 && tilepos == 24 && !shadow_initialized) {
            special_event = "death trap"; // level door with a lot of guards
        }
        if (current_level == 14 && drawn_room == 11 && tilepos == 13) {
            special_event = "prison"; // prison exist
        }
        if (current_level == 4 && drawn_room == 20 && tilepos == 15 && tile_type == tiles_20_wall) {
            special_event = "jump\nthru"; // jump through wall
        }
        if (current_level == 14 && drawn_room == 11 && tilepos == 11) {
            special_event = "walk\nthru\n<-"; // walk through doortop with floor
        }
        if (current_level == 2 && drawn_room == 24 && tilepos == 9) {
            special_event = "fake*"; // fake wall
        }
        if (current_level == 14 && drawn_room == 19 && tilepos == 9) {
            special_event = "fake*"; // fake wall
        }
        // not marked: level 12 shadow
        if (current_level == 12 &&
                tile_type != tiles_1_floor &&
                falling_bridge_timer < 0xFF &&
                ((drawn_room == 2 && tilepos < 8) ||
                 (drawn_room == 13 && tilepos >= 6 && tilepos <= 9))) {
            special_event = "floor"; // floors appear
        }
        if (current_level == 12 && drawn_room == 13 && (tilepos == 0 || tilepos == 10) && !is_kid_shadow) {
            special_event = "exit"; // exit by entering the room on the left
        }
        if (current_level == 13 && drawn_room == level.roomlinks[23 - 1].up && (tilepos >= 22 && tilepos < 27)) {
            special_event = "fall"; // falling loose floors
        }
        if (current_level == 13 && drawn_room == level.roomlinks[16 - 1].up && tilepos == 29) {
            special_event = "fall"; // falling loose floors
        }
        if (current_level == 13 && drawn_room == 7 && tilepos >= 3 && tilepos <= 8 && (tile_type == tiles_1_floor || tile_type == tiles_19_torch)) {
            special_event = "sliding\nwall";
        }
        if (current_level == 13 && (drawn_room == 2 || drawn_room == 13) && tilepos % 10 == 9) {
            special_event = "prev\n->"; // previous level
        }
        if (current_level == 14 && drawn_room == 23 && tilepos == 10) {
            special_event = "end\n<-"; // end of the game
        }
        if (has_trigger_potion && drawn_room == 8 && tilepos == 0) {
            special_event = "blue\ntrig"; // triggered when player drinks an open potion
        }
        if (special_event) {
            rect_type event_rect = {y, x - 10, y + 63, x + 32 + 10};
            if (strcmp("skel wake", special_event) == 0 && (drawn_room == 1 || drawn_room == 6)) { // wake up after the level door is opened
                show_text_with_color(&event_rect, text_align_center, text_align_center, special_event, color_6_brown);
            } else if (strcmp("skel cont", special_event) == 0) {
                show_text_with_color(&event_rect, text_align_center, text_align_center, special_event, color_6_brown);
            } else if (strcmp("spike", special_event) == 0 || strcmp("prison", special_event) == 0) {
                show_text_with_color(&event_rect, text_align_center, text_align_center, special_event, color_12_brightred);
            } else {
                show_text_with_color(&event_rect, text_align_center, text_align_center, special_event, color_14_brightyellow);
            }
        }

        // start pos
        if (level.start_room == drawn_room && level.start_pos == tilepos) {
            byte start_dir = level.start_dir;
            if (current_level == 1 || current_level == 13) start_dir ^= 0xFF; // falling/running entry
            char* start_text = (start_dir == dir_0_right) ? "start\n->" : "start\n<-";
            rect_type start_rect = {y, x - 10, y + 63, x + 32 + 10};
            show_text_with_color(&start_rect, text_align_center, text_align_center, start_text, color_14_brightyellow);
        }

        // guard info
        word room_minus_1 = drawn_room - 1;
        if (Guard.room == drawn_room &&
                Guard.alive < 0 &&
                Guard.direction != dir_56_none &&
                tilepos == Guard.curr_row * 10 + MAX(MIN(Guard.curr_col, 0), 9)) {
            loadshad();
            load_frame_to_obj();
            int screen_x = calc_screen_x_coord(obj_x);
            // Put it above the guard's head.
            if (Guard.direction == dir_0_right) {
                screen_x -= 15;
            } else {
                screen_x += 15;
            }
            if (screen_x > 290) {
                screen_x = 290;
            } else if (screen_x < 30) {
                screen_x = 30;
            }
            if (settings->tbl_guard_type[current_level] == guard_type_Jaffar_vision && row == 2) y -= 126;
            rect_type event_rect = {y + 2, screen_x - 16 - 10, y + 63, screen_x + 16 + 15};
            draw_guard_stats(
                event_rect,
                Guard.charid != charid_1_shadow && Guard.charid != charid_24_mouse ? guard_skill : -1,
                Guard.charid == charid_2_guard ? guardhp_max : -1);
        } else if (inactive_guard[room_minus_1].room == drawn_room &&
                   inactive_guard[room_minus_1].direction != dir_56_none &&
                   tilepos == inactive_guard[room_minus_1].row * 10 + get_tile_div_mod_m7(inactive_guard[room_minus_1].x)) {
            int screen_x = calc_screen_x_coord(inactive_obj_x);
            // Put it above the guard's head.
            if (inactive_guard[room_minus_1].direction == dir_0_right) {
                screen_x -= 15;
            } else {
                screen_x += 15;
            }
            if (screen_x > 290) {
                screen_x = 290;
            } else if (screen_x < 30) {
                screen_x = 30;
            }
            rect_type event_rect = { y + 2, screen_x - 16 - 10, y + 63, screen_x + 16 + 15};
            draw_guard_stats(
                event_rect,
                level.guards_skill[room_minus_1],
                settings->tbl_guard_type[current_level] != guard_type_skeleton ? (settings->tbl_guard_hp[current_level] - 1) : -1
            );
        }
        if (dynamic_guard.room == drawn_room) {
            char guard_info[20];
            rect_type event_rect = {y + 50, x - 5, y + 113, x + 75};
            if (dynamic_guard.direction == dir_0_right && tilepos == dynamic_guard.row * 10) {
                if (dynamic_guard.frame == 0) {
                    strncpy(guard_info, ">> +1 guard", sizeof(guard_info));
                } else {
                    snprintf(guard_info, sizeof(guard_info), ">> +%d guards", dynamic_guard.frame + 1);
                }
                show_text_with_color(&event_rect, text_align_center, text_align_top, guard_info, color_10_brightgreen);
            } else if (dynamic_guard.direction == dir_FF_left && tilepos == dynamic_guard.row * 10 + 8) {
                if (dynamic_guard.frame == 0) {
                    strncpy(guard_info, "+1 guard <<", sizeof(guard_info));
                } else {
                    snprintf(guard_info, sizeof(guard_info), "+%d guards <<", dynamic_guard.frame + 1);
                }
                show_text_with_color(&event_rect, text_align_center, text_align_top, guard_info, color_10_brightgreen);
            }
        }
    }

    // room number
    char room_num[6];
    snprintf(room_num, sizeof(room_num), "%d", drawn_room);
    rect_type text_rect = {5, 2, 16, 22};
    method_5_rect(&text_rect, 0, color_8_darkgray);
    show_text_with_color(&text_rect, text_align_center, text_align_center, room_num, color_15_brightwhite);

    // grid lines
    rect_type vline = {0, 0, 192, 1};
    method_5_rect(&vline, 0, color_15_brightwhite);
    rect_type hline = {3, 0, 4, 320};
    method_5_rect(&hline, 0, color_15_brightwhite);
}

// Save a "screenshot" of the whole level.
void save_level_screenshot(bool want_extras) {
    // TODO: Disable in the intro or if a cutscene is active?

    // Restrict this to cheat mode. After all, it's like using H/J/U/N or opening the level in an editor.
    if (!cheats_enabled) return;

    // custom logic - do not allow screenshots in the first room of level 1 while guard is falling
    if (current_level == 1 && drawn_room == 1 && Guard.direction != dir_56_none && Guard.action > actions_1_run_jump) {
        return;
    }

    // custom logic - do not save the map for levels 6A
    if (current_level == 6 && falling_exit_checkpoint < 4) {
        display_text_bottom("LEVEL 6 DOES NOT HAVE A MAP");
        text_time_total = 24;
        text_time_remaining = 24;
        return;
    }

    word prev_ud_timer = upside_down_timer;
    upside_down_timer = 0;

    // First, figure out where to put each room.
    // We don't stop on broken room links, because the resulting map might still be usable.

    int processed[NUMBER_OF_ROOMS + 1] = {0};
    for (int room = 1; room <= NUMBER_OF_ROOMS; room++) {
        xpos[room] = 0;
        ypos[room] = 0;
    }
    xpos[drawn_room] = 0;
    ypos[drawn_room] = 0;
    int queue_start = 0;
    int queue_end = 1;
    // custom logic - we always start from the level start room to be consistent
    int queue[NUMBER_OF_ROOMS] = {level.start_room};
    // Assemble a map based on room links.
    while (queue_start < queue_end) {
        int room = queue[queue_start++];
        // custom logic - only scan level 6B rooms
        if (current_level == 6 &&
                room != 12 &&
                room != 17 &&
                room != 19 &&
                room != 20 &&
                room != 21 &&
                room != 22) {
            continue;
        }
        byte* roomlinks = (byte*)(&level.roomlinks[room - 1]);
        for (int direction = 0; direction < 4; direction++) {
            // custom logic - do not scan rooms above # 18 on level 1
            if (current_level == 1 && room == 18 && direction == room_dir_up) {
                continue;
            }
            // custom logic - do not scan above room 12 on level 6B
            if (current_level == 6 && room == 12 && direction == room_dir_up) {
                continue;
            }
            // custom logic - remove the end of the game room
            if (current_level == 14 && room == 23 && !debug_cheats_enabled) {
                continue;
            }
            int other_room = roomlinks[direction];
            if (other_room >= 1 && other_room <= NUMBER_OF_ROOMS && !processed[other_room]) {
                int other_x = xpos[room] + dx[direction];
                int other_y = ypos[room] + dy[direction];
                xpos[other_room] = other_x;
                ypos[other_room] = other_y;
                processed[other_room] = 1;
                queue[queue_end++] = other_room;
            }
        }
    }

    // Find the bounds of the level.
    // The starting room is mapped to x=0,y=0, so 0 is a good initial value for max and min.
    int min_x = 0, max_x = 0, min_y = 0, max_y = 0;
    for (int room = 1; room <= NUMBER_OF_ROOMS; room++) {
        if (xpos[room] < min_x) min_x = xpos[room];
        if (xpos[room] > max_x) max_x = xpos[room];
        if (ypos[room] < min_y) min_y = ypos[room];
        if (ypos[room] > max_y) max_y = ypos[room];
    }

    // Position for rooms that would clash with other rooms: Below the normally mapped rooms.
    int clash_y = max_y + 1;
    int clash_x = min_x;

#define MAX_MAP_SIZE NUMBER_OF_ROOMS
    int map[MAX_MAP_SIZE][MAX_MAP_SIZE] = {{0}};
    for (int room = 1; room <= NUMBER_OF_ROOMS; room++) {
        if (processed[room]) {
            int x, y;
again:
            y = ypos[room] - min_y;
            x = xpos[room] - min_x;
            if (x >= 0 && y >= 0 && x < MAX_MAP_SIZE && y < MAX_MAP_SIZE) {
                if (map[y][x]) {
                    // Try to find some other place for this room:
                    // Put this room to the bottom of the map.
                    xpos[room] = clash_x;
                    ypos[room] = clash_y;
                    clash_x++;
                    if (xpos[room] > max_x) max_x = xpos[room];
                    if (ypos[room] > max_y) max_y = ypos[room];
                    goto again; // Force bounds check, just to be sure.
                }
                map[y][x] = room;
            }
        }
    }

    int map_width = max_x - min_x + 1;
    int map_height = max_y - min_y + 1;
    int map_center = (map_width - 1) / 2;

    // Now we have the arrangement, let's make the picture!

    int image_width = map_width * 320;
    int image_height = map_height * 189 + 3 + 8;

    // custom logic - remove the rooms on the left of the exit room on level 12 when kid is not shadow
    if (current_level == 12 && !is_kid_shadow) {
        image_width -= 320;
    }

    SDL_Surface* map_surface = SDL_CreateRGBSurface(0, image_width, image_height, 32, 0, 0, 0, 0);
    if (map_surface == NULL) {
        sdlperror("SDL_CreateRGBSurface (map_surface)");
        return;
    }

    // TODO: Background color for places where there is no room?

    // TODO: Add an option for displaying all unreachable rooms?

    has_trigger_potion = false;

    // Is there a trigger potion on the level?
    for (int room = 1; room <= NUMBER_OF_ROOMS; room++) {
        if (processed[room]) {
            get_room_address(room);
            for (int tilepos = 0; tilepos < MAX_TILES_PER_ROOM; tilepos++) {
                int tile_type = curr_room_tiles[tilepos] & 0x1F;
                if (tile_type == tiles_10_potion && curr_room_modif[tilepos] >> 3 == 6) {
                    has_trigger_potion = true;
                }
            }
        }
    }

    memset(event_used, 0, sizeof(event_used));

    // Find out which door events are used:
    for (int room = 1; room <= NUMBER_OF_ROOMS; room++) {
        if (processed[room]) {
            get_room_address(room);
            for (int tilepos = 0; tilepos < MAX_TILES_PER_ROOM; tilepos++) {
                int tile_type = curr_room_tiles[tilepos] & 0x1F;
                if (tile_type == tiles_6_closer || tile_type == tiles_15_opener
                        || (has_trigger_potion && room == 8 && tilepos == 0) // triggered when player drinks an open potion
                   ) {
                    int modifier = curr_room_modif[tilepos];
                    for (int index = modifier; index < 256; index++) {
                        event_used[index] = true;
                        if (!get_doorlink_next(index)) break;
                    }
                }
            }
        }
    }

    int old_room = drawn_room;
    char_type old_guard = { 0 };
    word old_guard_hp = 0;
    if (Guard.charid == charid_1_shadow || Guard.charid == charid_24_mouse) {
        old_guard = Guard;
        old_guard_hp = guardhp_curr;
    }
    for (int y = 0; y < map_height; y++) {
        for (int x = 0; x < map_width; x++) {
            int room = map[y][x];
            int x_center = map_center;
            // custom logic - remove repeating rooms from the map
            if (current_level == 6) {
                if (room == 4 ||
                        room == 5 ||
                        room == 13 ||
                        room == 14 ||
                        room == 16 ||
                        room == 18) {
                    continue;
                }
            }
            SDL_Rect dest_rect;
            dest_rect.x = x * 320;
            dest_rect.y = y * 189;
            // custom logic - remove the rooms on the left of the exit room on level 12 when kid is not shadow
            if (current_level == 12 && !is_kid_shadow) {
                if (x == 0) {
                    continue;
                } else {
                    dest_rect.x -= 320;
                    ++x_center;
                }
            }
            if (room) {
                placeholder_id = 0;
                switch_to_room(room, old_guard, old_guard_hp);

                if (want_extras) draw_extras();

                // moved from above
                draw_guard_hp(0, MAX_GUARD_HP); // otherwise guard HPs still stay on screen if some guards have extra HP

                // custom logic - clear kid's hitpoints
                clear_kid_hp(hitp_max);

                // custom logic - erase bottom text
                erase_bottom_text(1);

                // show level # and name on the bottom of the map
                if (y == map_height - 1 && x == x_center) {
                    show_level();
                }

                SDL_BlitSurface(onscreen_surface_, NULL, map_surface, &dest_rect);
            } else {
                placeholder_id = x << 8;
                placeholder_id |= y << 4;
                get_room_address(PLACEHOLDER_ROOM);

                if (x > 0 && map[y][x - 1]) {
                    room_L = map[y][x - 1];
                } else {
                    room_L = 0;
                }
                if (x < map_width - 1 && map[y][x + 1]) {
                    room_R = map[y][x + 1];
                } else {
                    room_R = 0;
                }
                if (y > 0) {
                    room_A = map[y - 1][x];
                    if (x > 0 && map[y - 1][x - 1]) {
                        room_AL = map[y - 1][x - 1];
                    }
                    if (x < map_width - 1 && map[y - 1][x + 1]) {
                        room_AR = map[y - 1][x + 1];
                    }
                } else {
                    room_A = 0;
                }
                if (y < map_height - 1) {
                    room_B = map[y + 1][x];
                    if (x > 0 && map[y + 1][x - 1]) {
                        room_BL = map[y + 1][x - 1];
                    }
                    if (x < map_width - 1 && map[y + 1][x + 1]) {
                        room_BR = map[y + 1][x + 1];
                    }
                } else {
                    room_B = 0;
                }

                switch_to_room(PLACEHOLDER_ROOM, old_guard, old_guard_hp);

                if (want_extras) {
                    rect_type vline = {0, 0, 192, 1};
                    method_5_rect(&vline, 0, color_15_brightwhite);
                    rect_type hline = {3, 0, 4, 320};
                    method_5_rect(&hline, 0, color_15_brightwhite);
                }

                draw_guard_hp(0, MAX_GUARD_HP); // otherwise guard HPs still stay on screen if some guards have extra HP
                clear_kid_hp(hitp_max);
                erase_bottom_text(1);
                if (y == map_height - 1 && x == x_center) {
                    show_level();
                }
                SDL_BlitSurface(onscreen_surface_, NULL, map_surface, &dest_rect);
                for (int tilepos = 0; tilepos < MAX_TILES_PER_ROOM; tilepos++) {
                    curr_room_modif[tilepos] = 0;
                }
            }
        }
    }
    placeholder_id = 0;
    switch_to_room(old_room, old_guard, old_guard_hp);

    make_screenshot_filename();
    int result = IMG_SavePNG(map_surface, screenshot_filename);
    show_result(result, "LEVEL MAP");

    SDL_FreeSurface(map_surface);

    // custom logic - reset upside down mode
    if (prev_ud_timer > 0) {
        upside_down_timer = prev_ud_timer;
        need_full_redraw = 1;
    }

    // custom logic - redraw kid's and guard's lifes
    if (Guard.room != Kid.room) {
        draw_guard_hp(0, MAX_GUARD_HP);
    }
    draw_kid_hp(hitp_curr, hitp_max);
}
