/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"
#include "ctype.h"
#include <setjmp.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846264f
#endif

// data:461E
dat_type *dathandle;

// data:4C08
word need_redraw_because_flipped;

// custom logic - displays slot type number for cheats
byte slot_type_number = 0;

// seg000:0000
void pop_main(void) {
    if (check_param("--help") || check_param("-h") || check_param("?")) {
        display_help();
        exit(0);
    }

    find_exe_dir();
    if (chdir(exe_dir) != 0) {
        // do nothing
    }

    load_global_options();

    // custom logic - load sound and music settings
    config_load();

    if (check_param("nomusic")) is_music_on = 0;

    if (check_param("mute")) is_sound_on = 0;
    turn_sound_on_off((is_sound_on != 0) * 15); // Turn off sound/music if those options were set.

    // custom logic - set ambient music mixer settings
    set_music();

    // Fix bug: with start_in_blind_mode enabled, moving objects are not displayed until blind mode is toggled off+on??
    need_drects = 1;

    char sprintf_temp[100];
    int i;

    dathandle = open_dat("PRINCE.DAT");

    set_gr_mode();

    init_timer(60);

    set_hc_pal();

#if defined WIN32 || _WIN32 || WIN64 || _WIN64
    // custom logic - turns off num lock in Windows because Shift is not detected when Num Lock is on
    set_numlock_state(false);
#endif

    cheats_enabled = check_param("megahit") != NULL;
    debug_cheats_enabled = check_param("saycheese") != NULL && cheats_enabled;

    if (cheats_enabled) {
        if (check_param("6b")) {
            start_level = 6;
            falling_exit_checkpoint = 4;
        } else if (check_param("12b")) {
            start_level = 13;
        } else {
            for (i = 14; i >= 0; --i) {
                snprintf(sprintf_temp, sizeof(sprintf_temp), "%d", i);
                if (check_param(sprintf_temp)) {
                    // custom logic - follow level numbers from the game
                    switch (i) {
                        case 14:
                        case 13:
                            start_level = i + 1;
                            break;
                        case 0:
                            play_demo_level = 1;
                        // fallthrough!
                        default:
                            start_level = i;
                            break;
                    }
                    break;
                }
            }
        }
        // custom logic - kid has sword on levels other than demo, 1, 6B or 14 (prison)
        if (start_level > 1 &&
                (start_level != 6 || falling_exit_checkpoint != 4) &&
                start_level != settings->prison_level) {
            have_sword = 1;
        }
    }

    // custom logic - allows setting a random seed for debugging
    if (debug_cheats_enabled) {
        const char* param_value = check_param("seed:");
        if (param_value != NULL) {
            custom_seed = atoi(param_value + 5);
            random_seed = custom_seed;
            seed_was_init = 1;
        }
    }

    demo_mode = check_param("demo") != NULL;

    if (demo_mode && start_level < 0) {
        start_level = 0;
    }

    current_target_surface = onscreen_surface_;
    show_loading();
    set_joy_mode();

    init_dialogs();
    init_menu();
    init_game_main();
}

byte* level_var_palettes;

// seg000:024F
void init_game_main(void) {
    doorlink_loc = level.doorlinks1;
    doorlink_map = level.doorlinks2;
    prandom(1);
    // Guard palettes
    guard_palettes = (byte*) load_from_opendats_alloc(10, "bin", NULL, NULL);
    // (blood, hurt flash) #E00030 = red
    set_pal(12, 0x38, 0x00, 0x0C);
    // Level color variations (1.3)
    level_var_palettes = load_from_opendats_alloc(20, "bin", NULL, NULL);
    // custom logic - loading fat guard as a secondary guard in a separate chtab
    load_chtab_from_file(id_chtab_10_fat_guard, 750, "FAT.DAT", 1 << 8);
    // PRINCE.DAT: sword
    chtab_addrs[id_chtab_0_sword] = load_sprites_from_file(700, 1 << 2);
    // PRINCE.DAT: flame, sword on floor, potion
    chtab_addrs[id_chtab_1_flameswordpotion] = load_sprites_from_file(150, 1 << 3);
    // custom logic - Jaffar's vision has a sword without a partial hand
    chtab_addrs[id_chtab_11_sword_no_hand] = load_sprites_from_file(1700, 1 << 2);
    // custom logic - skeleton sword
    chtab_addrs[id_chtab_12_skel_sword] = load_sprites_from_file(2700, 1 << 2);
    // custom logic - alternate skeleton sword
    chtab_addrs[id_chtab_13_alt_skel_sword] = load_sprites_from_file(3700, 1 << 2);
    // custom logic - flame's sword
    chtab_addrs[id_chtab_14_flaming_sword] = load_sprites_from_file(700, 1 << 2);
    // custom logic - PoP2 style sword on the floor
    chtab_addrs[id_chtab_15_pop2_sword] = load_sprites_from_file(1150, 1 << 3);
    close_dat(dathandle);
    init_custom_graphics(); // custom logic - initializes custom graphics
    init_lighting(); // must be called after init_custom_graphics()
    load_more_palettes(); // custom logic - additional graphics palettes
    load_music(); // custom logic - loads ambient music and enhanced sound effects
    load_all_sounds(); // standard game sounds

    hof_read();
    start_game();
}

// seg000:249D
void show_quotes(word which_quote) {
    word key;
    draw_rect(&screen_rect, color_0_black);
    show_text(&screen_rect, text_align_left, text_align_center, settings->tbl_quotes[which_quote]);
    start_timer(timer_0, 0x384);
    while (!has_timer_stopped(timer_0) && (key = key_test_quit()) == 0) {
        if (key) break;
        if (key_states[SDL_SCANCODE_LSHIFT] & keystate_held || key_states[SDL_SCANCODE_RSHIFT] & keystate_held) {
            key_states[SDL_SCANCODE_LSHIFT] = key_states[SDL_SCANCODE_RSHIFT] = 0;
            break;
        }
        idle();
    }
}

// data:02C2
word first_start = 1;
// data:4C38
jmp_buf setjmp_buf;
// seg000:0358
void start_game(void) {
    int quote;
    Mix_HaltMusic();
    // Prevent filling of stack.
    // start_game is called from many places to restart the game, for example:
    // process_key, play_frame, draw_game_frame, play_level, control_kid, end_sequence, expired
    if (first_start) {
        first_start = 0;
        setjmp(setjmp_buf);
    } else {
        draw_rect(&screen_rect, color_0_black);
        clear_screen_and_sounds();
        if (quotes_needed) {
            quotes_needed = 0;
            for (quote = 0; quote < str_arr_size(settings->tbl_quotes); ++quote) {
                show_quotes(quote);
            }
            if (demo_mode) {
                quit(0);
            }
        }
        longjmp(setjmp_buf, -1);
    }
    release_title_images(); // added
    free_optsnd_chtab(); // added
    if (start_level < 0) {
        show_title();
    } else {
        init_game(start_level);
    }
}

// All these functions return true on success, false otherwise.

FILE* quick_fp;

int process_save(void* data, size_t data_size) {
    return fwrite(data, data_size, 1, quick_fp) == 1;
}

int process_load(void* data, size_t data_size) {
    return fread(data, data_size, 1, quick_fp) == 1;
}

typedef int process_func_type(void* data, size_t data_size);

int quick_process(process_func_type process_func) {
    int ok = 1;
#define process(x) ok = ok && process_func(&(x), sizeof(x))
    // level
    word old_current_level = current_level;
    process(current_level);
    // custom logic - in debug cheat mode Shift-F9 does not save or load level information
    // but recreates it on top of the existing level
    if (debug_cheats_enabled &&
            process_func == process_load &&
            (key_states[SDL_SCANCODE_LSHIFT] & keystate_held || key_states[SDL_SCANCODE_RSHIFT] & keystate_held)) {
        if (current_level != old_current_level) {
            current_level = old_current_level;
            return 0;
        }
        fseek(quick_fp, sizeof(level), SEEK_CUR);
    } else {
        process(level);
    }
    process(next_level);
    process(checkpoint);
    process(upside_down_timer);
    process(drawn_room);
    process(mobs_count);
    process(mobs);
    process(trobs_count);
    process(trobs);
    process(leveldoor_open);
    // kid
    process(Kid);
    process(hitp_curr);
    process(hitp_max);
    process(hitp_beg_lev);
    process(grab_timer);
    process(holding_sword);
    process(united_with_shadow);
    process(have_sword);
    process(kid_sword_strike);
    process(pickup_obj_type);
    process(offguard);
    // guard
    process(Guard);
    process(Char);
    process(Opp);
    process(guardhp_curr);
    process(guardhp_max);
    process(demo_index);
    process(demo_time);
    process(guard_notice_timer);
    process(guard_skill);
    process(shadow_initialized);
    process(guard_refrac);
    process(justblocked);
    process(droppedout);
    // collision
    process(curr_row_coll_room);
    process(curr_row_coll_flags);
    process(below_row_coll_room);
    process(below_row_coll_flags);
    process(above_row_coll_room);
    process(above_row_coll_flags);
    process(prev_collision_row);
    // flash
    process(flash_color);
    process(flash_time);
    // sounds
    process(need_level1_music);
    process(is_screaming);
    process(feather_fall_timer);
    process(last_loose_sound);
    // random
    process(random_seed);
    // remaining time
    process(rem_min);
    process(rem_tick);
    // saved controls
    process(control_x);
    process(control_y);
    process(control_shift);
    process(control_forward);
    process(control_backward);
    process(control_up);
    process(control_down);
    process(control_shift2);
    process(ctrl1_forward);
    process(ctrl1_backward);
    process(ctrl1_up);
    process(ctrl1_down);
    process(ctrl1_shift2);
    // custom logic - additional values
    process(jaffar_presented);
    process(jaffar_vision_presented);
    process(took_antidote);
    process(level_7_tile_crashed);
    process(is_shadow_flame);
    process(initial_shadow_potion_counter);
    process(shadow_potion_counter);
    process(potion_room);
    process(potion_pos);
    process(potion_timer);
    process(potion_size);
    process(falling_exit_checkpoint);
    process(falling_exit_final_direction);
    process(level_2_right_door);
    process(level_restart_timer);
    process(skeleton_timer);
    process(level_6_sequence_timer);
    process(enable_lighting);
    process(keep_sword_drawn);
    process(is_kid_shadow);
    process(Dead);
    process(dead_kid_timer);
    process(dead_guard1);
    process(dead_guard2);
    process(outbound_dead_guard1);
    process(outbound_dead_guard2);
    process(loose_tile_counter);
    process(is_shadow_text_shown);
    process(is_dynamic_guard_fight);
    process(outbound_dead_guard3);
    process(level_rem_min);
    process(level_rem_tick);
    process(guard_dropping_down);
    process(delayed_sound_timer);
    process(delayed_sound);
    process(music_suspended);
    process(music_suspended_timer);
    process(dynamic_guard);
    process(dynamic_guard_count);
    process(dynamic_guard_timer);
    process(inactive_guard);
    process(inactive_guard_timer);
    process(dead_guard3);
    process(jaffar_delay_timer);
    process(blink_timer);
    process(guard_knocked_tile);
    process(lighting_timer);
    process(skeleton_appeared);
    process(hard_jump_fall);
    process(falling_bridge_timer);
    process(inactive_guard_skill);
    process(inactive_guard_fall_timer);
    process(inactive_seq_ptr);
    process(super_jump_fall);
    process(super_jump_timer);
    process(super_jump_room);
    process(super_jump_col);
    process(super_jump_row);
    process(jumped_through_mirror);
    process(is_guard_notice);
    process(is_inactive_guard_notice);
    process(can_guard_see_kid);
    process(can_inactive_guard_see_kid);
    process(total_saves);
    process(saves_per_level);
    process(kid_cheat_pal_offset);
    process(charid_hurt_by_sword);
    process(skeleton_shake_timer);
    process(should_turn_around);
    process(should_inactive_turn_around);
    process(seamless);
    process(moving_wall_timer);
    process(moving_wall_direction);
    process(mouse_potion_timer);
    process(falling_guard);
    process(separate_from_shadow_counter);
    process(separate_from_shadow_counter_prev);
    process(separate_from_shadow_reset_timer);
    process(justturned_engarde);
    process(dead_kid_tile);
    process(skeleton_death_timer);
    process(guard_punch_count);
    process(music_mode);
    process(had_sword);
#undef process
    return ok;
}

const char* config_file = "PRINCE.CFG";

const char* quick_file = "PRINCE.SAV";
const char quick_version[] = "V1.50r3 ";
char quick_control[] = "........";

const char* get_config_path(char* custom_path_buffer, size_t max_len) {
    snprintf(custom_path_buffer, max_len, "%s/%s", save_folder, config_file);
    return custom_path_buffer;
}

const char* get_quick_path(char* custom_path_buffer, size_t max_len) {
    snprintf(custom_path_buffer, max_len, "%s/%s", save_folder, quick_file);
    return custom_path_buffer;
}

// custom logic - get auto-save file path
const char* get_auto_path(char* custom_path_buffer, size_t max_len, int level, bool is_b_level) {
    if (is_b_level) {
        snprintf(custom_path_buffer, max_len, "%s/LEVEL%dB.SAV", auto_save_folder, level);
    } else {
        snprintf(custom_path_buffer, max_len, "%s/LEVEL%d.SAV", auto_save_folder, level);
    }
    return custom_path_buffer;
}

// custom logic - print load/save setting error message
void config_failed(const char* operation, const char* setting) {
    printf("Could not %s %s enabled option", operation, setting);
}

// custom logic - load sound and music settings
void config_load(void) {
    char custom_config_path[POP_MAX_PATH];
    create_folder_if_missing(save_folder);
    const char* path = get_config_path(custom_config_path, sizeof(custom_config_path));
    FILE* config_fp = fopen(path, "rb");
    if (config_fp == NULL) return;
    if (fread(&is_sound_on, 1, 2, config_fp) != 2) {
        config_failed("load", "sound");
    }
    if (fread(&is_music_on, 1, 2, config_fp) != 2) {
        config_failed("load", "music");
    }
    if (fread(&max_saves_per_level, 1, 2, config_fp) != 2) {
        max_saves_per_level = 3;
    }
    fclose(config_fp);
}

// custom logic - save sound and music settings
void config_save(void) {
    char custom_config_path[POP_MAX_PATH];
    create_folder_if_missing(save_folder);
    const char* path = get_config_path(custom_config_path, sizeof(custom_config_path));
    FILE* config_fp = fopen(path, "wb");
    if (config_fp == NULL) return;
    if (fwrite(&is_sound_on, 1, 2, config_fp) != 2) {
        config_failed("save", "sound");
    }
    if (fwrite(&is_music_on, 1, 2, config_fp) != 2) {
        config_failed("save", "music");
    }
    if (fwrite(&max_saves_per_level, 1, 2, config_fp) != 2) {
        config_failed("save", "max saves");
    }
    fclose(config_fp);
}

int quick_save(bool increase_saves) {
    int ok = 0;
    char custom_quick_path[POP_MAX_PATH];
    create_folder_if_missing(save_folder);
    const char* path = get_quick_path(custom_quick_path, sizeof(custom_quick_path));
    quick_fp = fopen(path, "wb");
    if (quick_fp != NULL) {
        // custom logic - increase saves per game and level
        if (increase_saves) {
            ++saves_per_level;
            ++total_saves;
            if (total_saves > 999) total_saves = 999;
        }
        process_save((void*) quick_version, COUNT(quick_version));
        ok = quick_process(process_save);
        fclose(quick_fp);
        quick_fp = NULL;
    }
    return ok;
}

// custom logic - auto save levels 2-13
int auto_save(int level, bool is_b_level) {
    int ok = 0;
    char custom_auto_path[POP_MAX_PATH];
    create_folder_if_missing(save_folder);
    create_folder_if_missing(auto_save_folder);
    const char* path = get_auto_path(custom_auto_path, sizeof(custom_auto_path), level, is_b_level);
    quick_fp = fopen(path, "wb");
    if (quick_fp != NULL) {
        process_save((void*) quick_version, COUNT(quick_version));
        ok = quick_process(process_save);
        fclose(quick_fp);
        quick_fp = NULL;
    }
    if (!ok) {
        return 0;
    }
    return quick_save(false);
}

// custom logic - auto save levels 2-13
void auto_save_level(int level) {
    if (skip_auto_save) {
        skip_auto_save = 0;
        return;
    }
    // do not autosave dead kid
    if (Kid.alive >= 0) {
        return;
    }
    if (level < 2 || level > 14) {
        return;
    }
    // entering level 12 from level 13
    if (level == 12 && (Char.room == 2 || Char.room == 21)) {
        return;
    }
    // meeting the princess condition
    if (level == settings->win_level && leveldoor_open >= 2) {
        return;
    }
    switch (level) {
        case 6:
            auto_save(6, falling_exit_checkpoint > 0);
            break;
        case 13:
            auto_save(12, true);
            break;
        case 14:
            auto_save(13, false);
            break;
        default:
            auto_save(level, false);
            break;
    }
};

void restore_room_after_quick_load(word prev_level) {
    skip_auto_save = 1;

    // custom logic - reset need music flag after load
    need_level1_music = 0;

    // custom logic - reset blink timer
    if (blink_timer > 11) {
        blink_timer = 0;
    }

    // custom logic - start on level 6B
    if (current_level == 6 && falling_exit_checkpoint == 3) {
        // stop gate sounds
        get_tile(1, 9, 1);
        curr_room_modif[curr_tilepos] = 0xFF;
        ++falling_exit_checkpoint;
        music_suspended_timer = 12;
        next_room = drawn_room = 17;
        load_room_links();
        anim_tile_modif();
    }

    // custom logic - reset music state
    ogg_fore_sound_playing = false;
    if (music_suspended == 1 && (delayed_sound_timer == 0 || delayed_sound.sound_id < 0)) {
        music_suspended = 0;
    }
    if (prev_level == 0 || prev_level == 15) {
        show_music_menu = enable_positional_audio &&
                          check_param("nomusic") == NULL &&
                          is_music_found() ? 1 : 0;
        start_ambient_music(true);
    }

    int level = next_level;
    reset_level_unused_fields(false);
    load_lev_spr(current_level);
    // custom logic - level 6 has a different color palette until kid kills the guard
    if (current_level == 6 && !is_level_6_sequence() && (falling_exit_checkpoint < 4 || leveldoor_open != 1)) {
        load_lev6_palette();
    }
    // custom logic - reset kid palette
    if (cheats_enabled) {
        kid_cheat_pal_offset = -1;
    } else if (kid_cheat_pal_offset != -1) {
        kid_cheat_pal_offset = 0;
    }
    free_chtab(chtab_addrs[id_chtab_2_kid]);
    chtab_addrs[id_chtab_2_kid] = NULL;
    load_kid_sprite();
    if (is_kid_shadow && is_shadow_flame) {
        load_flame_sprite();
    }
    next_level = level;
    // reset last guard color
    last_guard_color = 0;
    different_room = 1;
    next_room = drawn_room;
    load_room_links();
    resurrect_time = 0;
    draw_game_frame(); // for falling

    hitp_delta = guardhp_delta = 1; // force HP redraw
    // Don't draw guard HP if a previously viewed room (with the H,J,U,N keys) had a guard but the current room doesn't have one.
    if (Guard.room != drawn_room) {
        draw_guard_hp(0, MAX_GUARD_HP);
    }

    draw_hp();
    loadkid_and_opp();

    // Get rid of "press button" message if kid was dead before quickload.
    text_time_total = text_time_remaining = 0;
    exit_room_timer = 0;
    find_all_rooms_from(drawn_room);

    // custom logic - falling from level 6B to level 7 needs a transition
    if (current_level == 6 && next_level == 7) {
        show_black_screen(1, 0);
    }
}

int quick_load_with_path(const char* path) {
    int ok = 0;
    quick_fp = fopen(path, "rb");
    if (quick_fp != NULL) {
        // check quicksave version is compatible
        process_load(quick_control, COUNT(quick_control));
        if (strcmp(quick_control, quick_version) != 0) {
            fclose(quick_fp);
            quick_fp = NULL;
            return 0;
        }

        stop_all_sounds();
        stop_ambient_music();

        quick_loading = 1;
        show_black_screen(6, 0);
        quick_loading = 0;

        word prev_level = current_level;

        ok = quick_process(process_load);
        fclose(quick_fp);
        quick_fp = NULL;

        restore_room_after_quick_load(prev_level);
        update_screen();
    }
    return ok;
}

int quick_load(void) {
    char custom_quick_path[POP_MAX_PATH];
    const char* path = get_quick_path(custom_quick_path, sizeof(custom_quick_path));
    return quick_load_with_path(path);
}

void check_quick_op(void) {
    // custom logic - no quick operations in demo mode
    if (demo_mode) {
        return;
    }
    // custom logic - do not quicksave on level 6 starting sequence
    if (is_level_6_sequence() && need_quick_save && !need_auto_save) {
        need_quick_save = 0;
        return;
    }
    // custom logic - no quick save on the demo level or level 14
    byte max_saves;
    if (need_quick_save && current_level > 0 && current_level < 15) {
        max_saves = cheats_enabled ? 255 : max_saves_per_level;
        // cannot quick save in the middle of Jaffar fight
        if (current_level == settings->flame_level && can_guard_see_kid && guardhp_curr > 0) {
            goto loc_save_not_allowed;
        }
        if (max_saves == 0) {
loc_save_not_allowed:
            display_text_bottom("NOT ALLOWED");
        } else if (saves_per_level >= max_saves) {
            display_text_bottom("SAVE LIMIT PER LEVEL REACHED");
        } else if (hitp_curr > 0 && level_restart_timer == 0 && quick_save(true)) {
            display_text("GAME SAVED (%d/%d)", saves_per_level, max_saves);
        } else {
            display_text_bottom("GAME NOT SAVED");
        }
        need_quick_save = 0;
        text_time_total = 24;
        text_time_remaining = 24;
    }
    if (need_quick_load) {
        bool quick_load_from_title;
        quick_load_from_title = need_quick_load >= 2;
        need_quick_load = 0;
        // custom logic - if quick save was not done on this level, restart it
        if (saves_per_level == 0 &&
                !quick_load_from_title &&
                ((!debug_cheats_enabled && !cheats_enabled) ||
                 (!(key_states[SDL_SCANCODE_LSHIFT] & keystate_held) && !(key_states[SDL_SCANCODE_RSHIFT] & keystate_held))) &&
                current_level > 0 &&
                current_level < 15) {
            Kid.alive = -1; // resets the remaining time
            last_key_scancode = SDL_SCANCODE_A | WITH_CTRL;
            return;
        }
        if (quick_load()) {
            display_text_bottom("GAME LOADED");
        } else {
            display_text_bottom("NO SAVED GAME");
        }
        text_time_total = 24;
        text_time_remaining = 24;
    } else if (strlen(need_auto_load) > 0) {
        skip_auto_save = 1;
        if (quick_load_with_path(need_auto_load)) {
            show_level();
        } else {
            display_text_bottom("LOADING FAILED");
        }
        need_auto_load = "";
        text_time_total = 24;
        text_time_remaining = 24;
    } else if (need_auto_save) {
        need_auto_save = 0;
        auto_save_level(current_level);
    }
}

// custom logic - shadow gets united with kid
bool unite_shadow_with_kid(void) {
    if (is_kid_shadow &&
            Kid.frame == frame_15_stand &&
            Dead.frame == frame_185_dead &&
            Dead.direction != dir_56_none &&
            Dead.room == Kid.room &&
            Dead.row == Kid.curr_row &&
            shad_kid_dist() <= 10) {
        // body died from a fall or a loose tile
        if (Dead.color == 0) {
            hitp_delta = - hitp_curr;
            return true;
        }
        Kid.fall_y = 0;
        Kid.x = Dead.x;
        Kid.direction = Dead.direction;
        control_forward = control_backward = control_up = control_down = release_arrows();
        seqtbl_offset_kid_char(seq_71_dying);
        play_seq();
        dead_kid_timer = 12;
        // on level 13 open left most gate after Jaffar dies
        if (current_level == settings->flame_level && leveldoor_open == 2) {
            ++leveldoor_open;
            get_tile(8, 0, 0);
            trigger_button(0, 0, -1);
        }
        return true;
    }
    return false;
}

// custom logic - can kid separate from shadow
sbyte can_kid_separate_from_shadow(void) {
    byte old_x;
    // kid cannot separate from shadow before he unites with it
    // or on the prison level
    if ((current_level == 10 && shadow_initialized == 0) ||
            current_level < 10 ||
            current_level > 14) {
        return 0;
    }
    if (is_kid_shadow ||
            Char.alive >= 0 ||
            (Char.frame != frame_15_stand && Char.frame != frame_45_turn) ||
            united_with_shadow > 3) {
        return 0;
    }
    old_x = 0;
    // loading tiles around kid
    loadkid();
loc_determine_col:
    determine_col();
    // kid is separating behind a gate in the room on the left
    if (Char.x <= 57 && Char.room == drawn_room) {
        goto_other_room(0);
        savekid();
        draw_guard_hp(0, MAX_GUARD_HP);
        next_room = Char.room;
    }
    short tile_at = get_tile_at_char();
    byte tile_at_stuck = get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 1;
    short tile_behind = get_tile_behind_char();
    byte tile_behind_stuck = get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 1;
    // kid cannot turn into shadow near dangerous tiles
    if (!tile_is_floor(tile_at) ||
            (!tile_is_floor(tile_behind) && tile_behind != tiles_20_wall) ||
            (tile_at == tiles_11_loose && !tile_at_stuck) ||
            (tile_behind == tiles_11_loose && !tile_behind_stuck) ||
            tile_at == tiles_2_spike ||
            tile_behind == tiles_2_spike ||
            tile_at == tiles_18_chomper ||
            tile_behind == tiles_18_chomper) {
        if (old_x > 0) {
            Char.x = old_x;
        }
        return 0;
    }
    // bump kid back near walls, doortops and closed gates
    if (separate_from_shadow_counter == (SHADOW_SEPARATION_COUNT * 2) &&
            ((Char.direction == dir_0_right &&
              (wall_type2(get_tile_at_char()) == wall_type_1_right || wall_type2(curr_tile2) == wall_type_4_wall)) ||
             (Char.direction == dir_FF_left &&
              (wall_type2(get_tile_infrontof_char()) == wall_type_1_right || wall_type2(curr_tile2) == wall_type_4_wall)))
       ) {
        short distance = distance_to_edge_weight();
        if (distance <= 7) {
            old_x = Char.x;
            Char.x = char_dx_forward(distance - 9);
            goto loc_determine_col;
        }
    }
    return 1;
}

// custom logic - turn kid into shadow, if separation is possible
void separate_kid_from_shadow(void) {
    // check whether it is safe to separate
    sbyte can_separate = can_kid_separate_from_shadow();
    if (can_separate != 1) {
        return;
    }
    sbyte char_x = Char.x;
    // bump kid back if he is close to the edge
    if (distance_to_edge_weight() < 4) {
        Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
        if (Char.direction == dir_0_right) {
            Char.x += 7;
        }
    }
    // bump to the middle of the tile
    if (Char.x < 75 && (Char.direction == dir_0_right || Char.x < 60)) {
        Char.x = x_bump[/*0 +*/ ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
        if (Char.direction == dir_0_right) {
            Char.x += 7;
        }
    } else if (Char.x > 182 && (Char.direction == dir_FF_left || Char.x > 193)) {
        Char.x = x_bump[9 + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
        if (Char.direction == dir_0_right) {
            Char.x += 7;
        }
    }
    if (char_x != Char.x) {
        determine_col();
        get_tile_at_char();
        if (curr_tile2 == tiles_20_wall) {
            in_wall();
        }
        if (!tile_is_floor(curr_tile2) ||
                (curr_tile2 == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0) ||
                curr_tile2 == tiles_2_spike ||
                curr_tile2 == tiles_18_chomper) {
            Char.x = char_x;
            determine_col();
            united_with_shadow = 2;
            return;
        }
    }
    savekid();
    // turn kid into shadow
    text_time_remaining = 1;
    have_sword = is_shadow_flame;
    control_forward = control_backward = control_up = control_down = release_arrows();
    initial_shadow_potion_counter = shadow_potion_counter = 0;
    music_suspended_timer = dead_kid_timer = united_with_shadow = 12;
}

// custom logic - is kid going up the stairs?
bool going_up_stairs(void) {
    return Kid.curr_seq >= seqtbl_offset_get_seq(seq_70_go_up_on_level_door) &&
           Kid.curr_seq < seqtbl_offset_get_seq(seq_95_Jaffar_stand_PV1); // Jaffar stand intro is the next sequence in the table
}

// seg000:04CD
int process_key(bool from_paused) {
    char sprintf_temp[80];
    bool is_demo, is_lev_6_seq;
    int key;
    const char* answer_text = NULL;
    word need_show_text;
    need_show_text = 0;
    is_demo = current_level == 0 && !play_demo_level;
    is_lev_6_seq = is_level_6_sequence();
    key = key_test_quit();
    if (is_paused && is_menu_shown) {
        key = key_test_paused_menu(key);
        if (key == 0) return 0;
    }
    if (is_ending_sequence) return 0;
    // custom logic - do not react to most keys while kid is dead on level 10
    if (((key != 0 &&
            key != SDL_SCANCODE_F9 &&
            key != (SDL_SCANCODE_A | WITH_CTRL) &&
            key != (SDL_SCANCODE_R | WITH_CTRL)) || control_shift) &&
            current_level == 10 &&
            Kid.alive >= 0 &&
            Kid.room == 8 &&
            holding_sword &&
            ((leveldoor_open > 0 && leveldoor_open < DEAD_KID_LEVEL_10_TIMER) ||
             (shadow_initialized > 0 && shadow_initialized < 18))) {
        return 0;
    }
    if (start_level < 0) {
        if (((key || control_shift) && !control_alt) || (is_joyst_mode && need_quick_load)) {
            if (key == SDL_SCANCODE_F9 || need_quick_load) need_quick_load = 2; // 2 = from title screen
            if (key == SDL_SCANCODE_F1) { // custom logic - start tutorial from title screen
                play_demo_level = 0;
                start_level = 0;
            } else if (key == (SDL_SCANCODE_S | WITH_CTRL)) { // custom logic - sound on/off
                turn_sound_on_off((!is_sound_on) * 15);
                config_save();
                return 0;
            } else {
                start_level = 1;
            }
            draw_rect(&screen_rect, color_0_black);
            if (is_global_fading) {
                fade_palette_buffer->proc_restore_free(fade_palette_buffer);
                is_global_fading = 0;
            }
            start_game();
        }
    }
    bool shift_or_enter = control_shift || key == SDL_SCANCODE_RETURN || key == SDL_SCANCODE_KP_ENTER;
    if (shift_or_enter && !is_demo) { // demo level cannot be restarted with a key
        // If the Kid died, enter or shift will restart the level.
        if (rem_min != 0 && Kid.alive > 6) {
            key = SDL_SCANCODE_A | WITH_CTRL; // ctrl-a
            // custom logic - shift can skip end-of-level
        } else if (end_of_level_music == 1) {
            key_states[SDL_SCANCODE_LSHIFT] = key_states[SDL_SCANCODE_RSHIFT] = 0;
            end_of_level_music = 0;
            stop_all_sounds();
        }
    }
    if (key == 0) return 0;
    if (is_keyboard_mode) clear_kbd_buf();

    switch (key) {
        case SDL_SCANCODE_ESCAPE: // esc
        case SDL_SCANCODE_ESCAPE | WITH_SHIFT: // allow pause while grabbing
            if (is_cutscene || is_ending_sequence || going_up_stairs()) {
                break;
            }
            enable_pause_menu = 1;
            is_paused = 1;
            is_menu_shown = 1;
            break;
        // custom logic - Ctrl-P pauses the game without the menu
        case SDL_SCANCODE_P | WITH_CTRL: // ctrl-p
            if (from_paused || is_cutscene || is_ending_sequence || going_up_stairs() || is_lev_6_seq) {
                break;
            }
            if (is_menu_shown == 0) {
                enable_pause_menu = 0;
                is_paused = 1;
            }
            break;
        case SDL_SCANCODE_SPACE: // space
            is_show_time = 1;
            break;
        case SDL_SCANCODE_A | WITH_CTRL: // ctrl-a
            if (is_cutscene || is_ending_sequence || going_up_stairs()) {
                break;
            }
            // custom logic - do not restart the level while falling into level 6B
            if (current_level == 6 && falling_exit_checkpoint > 0 && falling_exit_checkpoint < 4) {
                break;
            }
            // custom logic - reset time and saves when restarting level
            if (Kid.alive < 0) {
                if (level_rem_min >= rem_min) {
                    rem_min = level_rem_min;
                    rem_tick = level_rem_tick;
                }
                saves_per_level = 0;
            }
            stop_all_sounds();
            Mix_HaltMusic();
            is_restart_level = 1;
            show_black_screen(6, 0);
            break;
        case SDL_SCANCODE_G | WITH_CTRL: // ctrl-g
            answer_text = "PRESS F6 TO SAVE GAME";
            need_show_text = 1;
            break;
        case SDL_SCANCODE_J | WITH_CTRL: // ctrl-j
            if (is_demo) {
                break;
            }
            if (set_joy_mode()) {
                answer_text = "JOYSTICK FOUND";
            } else {
                answer_text = "JOYSTICK NOT FOUND";
            }
            need_show_text = 1;
            break;
        case SDL_SCANCODE_R | WITH_CTRL: // ctrl-r
            if (is_cutscene || is_ending_sequence || demo_mode) {
                break;
            }
            start_level = -1;
            if (is_menu_shown) menu_was_closed(); // Do necessary cleanup.
            start_game();
            break;
        case SDL_SCANCODE_S | WITH_CTRL: // ctrl-s
            turn_sound_on_off((!is_sound_on) * 15);
            config_save();
            answer_text = "SOUND OFF";
            if (is_sound_on) {
                answer_text = "SOUND ON";
            }
            need_show_text = 1;
            break;
        // custom logic - turn music on/off
        case SDL_SCANCODE_M | WITH_CTRL: // ctrl-m
            if (enable_positional_audio && show_music_menu) {
                if (is_music_on) {
                    is_music_on = 0;
                    answer_text = "AMBIENT MUSIC OFF";
                    start_ambient_music(true);
                } else {
                    is_music_on = 1;
                    music_mode = is_fight_music() ? 1 : 0;
                    answer_text = "AMBIENT MUSIC ON";
                    music_suspended = 0;
                    if (ogg_fore_sound_playing) {
                        delayed_music_start = 1;
                    } else {
                        start_ambient_music(true);
                    }
                }
                if (music_mode == -2) {
                    answer_text = "MUSIC FILE NOT FOUND";
                }
                config_save();
                need_show_text = 1;
            }
            break;
        case SDL_SCANCODE_V | WITH_CTRL: // ctrl-v
            answer_text = PRINCE_VERSION;
            need_show_text = 1;
            break;
        case SDL_SCANCODE_V | WITH_SHIFT: { // shift-v
            SDL_version version;
            SDL_GetVersion(&version);
            snprintf(sprintf_temp, sizeof(sprintf_temp),
                     "SDL %u.%u.%u  %s",
                     version.major, version.minor, version.patch, renderer_name);
            answer_text = sprintf_temp;
            need_show_text = 1;
            break;
        }
        case SDL_SCANCODE_O | WITH_CTRL: { // ctrl-o
            char cpu_type[5];
            switch (sizeof(void*) * 8) {
                case 64:
                    snprintf(cpu_type, sizeof(cpu_type), "X64");
                    break;
                default:
                    snprintf(cpu_type, sizeof(cpu_type), "X86");
                    break;
            }
            snprintf(sprintf_temp, sizeof(sprintf_temp), "OS: %s  PROC: %s", SDL_GetPlatform(), cpu_type);
            to_ucase(sprintf_temp);
            answer_text = sprintf_temp;
            need_show_text = 1;
            break;
        }
        case SDL_SCANCODE_F6:
        case SDL_SCANCODE_F6 | WITH_SHIFT:
            if (is_cutscene || is_ending_sequence || going_up_stairs()) {
                break;
            }
            if (Kid.alive < 0) need_quick_save = 1;
            break;
        case SDL_SCANCODE_F9:
        case SDL_SCANCODE_F9 | WITH_SHIFT:
            if (is_cutscene || is_ending_sequence || going_up_stairs()) {
                break;
            }
            // custom logic - do not auto-save on Shift-L
            skip_auto_save = 1;
            need_quick_load = 1;
            break;
    }
    if (cheats_enabled && !is_cutscene && !is_ending_sequence && !going_up_stairs() && !is_demo) {
        switch (key) {
            case SDL_SCANCODE_L | WITH_SHIFT: // shift-l
                // custom logic - do not switch levels during auto-save
                if (need_auto_save) {
                    break;
                }
                // custom logic - do not auto-save on Shift-L
                skip_auto_save = 1;
                key_states[SDL_SCANCODE_LSHIFT] = 0;
                key_states[SDL_SCANCODE_RSHIFT] = 0;
                if (current_level >= (debug_cheats_enabled ? 15 : 14)) {
                    rem_min = settings->start_minutes_left;
                    rem_tick = settings->start_ticks_left;
                    hitp_beg_lev = hitp_curr = hitp_max = MIN(settings->start_hitp, settings->max_hitp_allowed);
                    total_saves = 0;
                    next_level = 1;
                    need_level1_music = 1;
                    // custom logic - reset shadow palette
                    is_shadow_flame = false;
                    reset_shadow_palette();
                } else {
                    // custom logic - level 6B entry
                    if (current_level == 6 && falling_exit_checkpoint < 4) {
                        next_level = current_level;
                        falling_exit_checkpoint = 4;
                        is_restart_level = 1;
                    } else {
                        // custom logic - assume kid picked up the sword on level 1
                        if (current_level == 1) have_sword = 1;
                        next_level = current_level + 1;
                    }
                }
                stop_all_sounds();
                stop_ambient_music();
                break;
            case SDL_SCANCODE_C: // c
            case SDL_SCANCODE_C | WITH_SHIFT: // shift-c
                // custom logic - this cheat does not work on level 6A
                if (!debug_cheats_enabled && current_level == 6 && falling_exit_checkpoint < 4) {
                    break;
                }
                if ((key & WITH_SHIFT) == WITH_SHIFT) {
                    snprintf(sprintf_temp, sizeof(sprintf_temp), "AL%d AR%d BL%d BR%d", room_AL, room_AR, room_BL, room_BR);
                } else {
                    snprintf(sprintf_temp, sizeof(sprintf_temp), "S%d L%d R%d A%d B%d", drawn_room, room_L, room_R, room_A, room_B);
                }
                answer_text = sprintf_temp;
                need_show_text = 1;
                break;
            case SDL_SCANCODE_D: // d
                if (debug_cheats_enabled) {
                    // custom logic - display kid debug info with D
                    snprintf(sprintf_temp,
                             sizeof(sprintf_temp),
                             "KF%d KP%d KX%d KY%d KS%d",
                             Kid.frame,
                             get_tilepos(Kid.curr_col, Kid.curr_row),
                             Kid.x,
                             Kid.y,
                             keep_sword_drawn);
                    answer_text = sprintf_temp;
                    need_show_text = 1;
                }
                break;
            case SDL_SCANCODE_D | WITH_CTRL: // ctrl-d
                if (debug_cheats_enabled) {
                    // custom logic - display guard debug info with Ctrl-D
                    snprintf(sprintf_temp,
                             sizeof(sprintf_temp),
                             "GS%d GF%d GP%d GX%d GY%d",
                             guard_skill,
                             Guard.frame,
                             get_tilepos(Guard.curr_col, Guard.curr_row),
                             Guard.x,
                             Guard.y);
                    answer_text = sprintf_temp;
                    need_show_text = 1;
                }
                break;
            case SDL_SCANCODE_D | WITH_SHIFT: // shift-d
                // custom logic - display dead guard debug info with Shift-D
                if (debug_cheats_enabled) {
                    int guard_pos1, guard_pos2, guard_pos3, guard_pos4, guard_pos5;
                    char_type2 slot1, slot2, slot3;
                    char* slot_type;
                    if (slot_type_number == 1) {
                        slot_type = "OGS";
                        slot1 = outbound_dead_guard1[drawn_room - 1];
                        slot2 = outbound_dead_guard2[drawn_room - 1];
                        slot3 = outbound_dead_guard3[drawn_room - 1];
                    } else if (slot_type_number == 2) {
                        slot_type = "IGS";
                        slot1 = inactive_guard[drawn_room - 1];
                        slot2 = (struct char_type2) {
                            0
                        };
                        slot3 = (struct char_type2) {
                            0
                        };
                    } else {
                        slot_type = "DGS";
                        slot1 = dead_guard1[drawn_room - 1];
                        slot2 = dead_guard2[drawn_room - 1];
                        slot3 = dead_guard3[drawn_room - 1];
                    }
                    snprintf(sprintf_temp,
                             sizeof(sprintf_temp),
                             "GA%d GP%d RGP%d %s%d:%d:%d",
                             Guard.room == drawn_room && Guard.direction != dir_56_none ? Guard.alive : 255,
                             Guard.room == drawn_room &&
                             (guard_pos1 = get_tilepos(Guard.curr_col, Guard.curr_row)) < MAX_TILES_PER_ROOM &&
                             guard_pos1 >= 0 ?
                             (Guard.direction != dir_56_none ? guard_pos1 : -2) :
                             -1,
                             level.guards_tile[drawn_room - 1] < MAX_TILES_PER_ROOM &&
                             (guard_pos2 = get_tilepos(MAX(get_tile_div_mod_m7(level.guards_x[drawn_room - 1]), 0), level.guards_tile[drawn_room - 1] / 10)) < MAX_TILES_PER_ROOM &&
                             guard_pos2 >= 0 ? guard_pos2 : -1,
                             slot_type,
                             slot1.room == drawn_room &&
                             (guard_pos3 = get_tilepos(MAX(get_tile_div_mod_m7(slot1.x), 0), slot1.row)) < MAX_TILES_PER_ROOM &&
                             guard_pos3 >= 0 ?
                             (slot1.direction != dir_56_none ? guard_pos3 : -2) :
                             -1,
                             slot2.room == drawn_room &&
                             (guard_pos4 = get_tilepos(MAX(get_tile_div_mod_m7(slot2.x), 0), slot2.row)) < MAX_TILES_PER_ROOM &&
                             guard_pos4 >= 0 ?
                             (slot2.direction != dir_56_none ? guard_pos4 : -2) :
                             -1,
                             slot3.room == drawn_room &&
                             (guard_pos5 = get_tilepos(MAX(get_tile_div_mod_m7(slot3.x), 0), slot3.row)) < MAX_TILES_PER_ROOM &&
                             guard_pos5 >= 0 ?
                             (slot3.direction != dir_56_none ? guard_pos5 : -2) :
                             -1
                            );
                    if (slot_type_number >= 2) {
                        slot_type_number = 0;
                    } else {
                        ++slot_type_number;
                    }
                    answer_text = sprintf_temp;
                    need_show_text = 1;
                }
                break;
            case SDL_SCANCODE_MINUS:
            case SDL_SCANCODE_KP_MINUS:		// '-' --> subtract time cheat
                if (rem_min > 1) --rem_min;
                text_time_total = 0;
                text_time_remaining = 0;
                is_show_time = 1;
                break;
            case SDL_SCANCODE_EQUALS | WITH_SHIFT: // '+'
            case SDL_SCANCODE_KP_PLUS:	   // '+' --> add time cheat
                if (rem_min < 999) ++rem_min;
                text_time_total = 0;
                text_time_remaining = 0;
                is_show_time = 1;
                break;
            case SDL_SCANCODE_R: // R --> revive kid cheat
                if (Kid.alive >= 0 && Kid.room != 0) {
                    stop_all_sounds();
                    release_arrows();
                    resurrect_time = 20;
                    Kid.alive = -1;
                    holding_sword = 0;
                    start_chompers();
                    // restore inactive guard if active guard is dead
                    if (Guard.alive >= 0 &&
                            inactive_guard[Kid.room - 1].room == Kid.room &&
                            inactive_guard[Kid.room - 1].direction != dir_56_none) {
                        on_guard_killed();
                    }
                    erase_bottom_text(1);
                    music_suspended = music_suspended_timer = 0;
                    if (music_mode >= 0) {
                        music_mode = is_fight_music() ? 1 : 0;
                    }
                    start_ambient_music(false);
                }
                break;
            case SDL_SCANCODE_K: // K --> kill guard cheat
loc_scancode_k:
                if (Guard.charid != charid_4_skeleton && Guard.charid != charid_24_mouse) {
                    guardhp_delta = -guardhp_curr;
                    // custom logic - guards in a different room should collapse
                    if (Guard.room != Kid.room || Guard.room != drawn_room || current_level == settings->win_level) {
                        Guard.alive = 0;
                        // kill inactive guard as well
                        inactive_guard_dead();
                    }
                }
                break;
            case SDL_SCANCODE_I | WITH_SHIFT: // shift+I --> invert cheat
                // custom logic - do not invert screen during the initial level 6 sequence
                if (is_lev_6_seq) {
                    break;
                }
                if (upside_down_timer == 0) {
                    upside_down_timer = 1080; // 90 seconds
                } else {
                    upside_down_timer = 0;
                }
                toggle_upside();
                break;
            // custom logic - reduce guard's skill to 0
            case SDL_SCANCODE_E | WITH_SHIFT: // shift-E
                if (Guard.alive < 0 && guard_skill > 0) {
                    guard_skill = 0;
                }
                break;
            case SDL_SCANCODE_W | WITH_SHIFT: // shift+W --> feather fall cheat
                // custom logic - do not enable feather fall during the initial level 6 sequence
                if (is_lev_6_seq) {
                    break;
                }
                feather_fall();
                break;
            case SDL_SCANCODE_H: // H --> view room to the left
                // custom logic - this cheat requires "saycheese" parameter
                if (!debug_cheats_enabled) {
                    break;
                }
                if (room_L) {
                    draw_guard_hp(0, MAX_GUARD_HP);
                    next_room = room_L;
                }
                break;
            case SDL_SCANCODE_J: // J --> view room to the right
                // custom logic - this cheat requires "saycheese" parameter
                if (!debug_cheats_enabled) {
                    break;
                }
                if (room_R) {
                    draw_guard_hp(0, MAX_GUARD_HP);
                    next_room = room_R;
                }
                break;
            case SDL_SCANCODE_U: // U --> view room above
                // custom logic - this cheat requires "saycheese" parameter
                if (!debug_cheats_enabled) {
                    break;
                }
                word room_above = upside_down_timer > 0 ? room_B : room_A;
                if (room_above) {
                    draw_guard_hp(0, MAX_GUARD_HP);
                    next_room = room_above;
                }
                break;
            case SDL_SCANCODE_N: // N --> view room below
                // custom logic - this cheat requires "saycheese" parameter
                if (!debug_cheats_enabled) {
                    break;
                }
                word room_below = upside_down_timer > 0 ? room_A : room_B;
                if (room_below) {
                    draw_guard_hp(0, MAX_GUARD_HP);
                    next_room = room_below;
                }
                break;
            case SDL_SCANCODE_B | WITH_CTRL: // ctrl-b --> Go back to the room where the prince is. (Undo H,J,U,N.)
                draw_guard_hp(0, MAX_GUARD_HP);
                next_room = Kid.room;
                break;
            case SDL_SCANCODE_B | WITH_SHIFT: // shift-b
                // custom logic - this cheat requires "saycheese" parameter
                if (!debug_cheats_enabled) {
                    break;
                }
                is_blind_mode = !is_blind_mode;
                if (is_blind_mode) {
                    draw_rect(&rect_top, color_0_black);
                } else {
                    need_full_redraw = 1;
                }
                break;
            case SDL_SCANCODE_L | WITH_CTRL: // ctrl-l
                if (enable_lighting > 0) {
                    enable_lighting = 0;
                } else {
                    enable_lighting = 1 - enable_lighting;
                }
                need_full_redraw = 1;
                break;
            case SDL_SCANCODE_S | WITH_SHIFT: // shift-s
                // custom logic - hitpoint cheats do not work with shadow
                if ((!is_kid_shadow || is_shadow_flame) && resurrect_time == 0 && hitp_curr < hitp_max) {
                    play_sound(sound_33_small_potion); // small potion (cheat)
                    flash_color = color_15_brightwhite;
                    flash_time = 1;
                    hitp_delta = 1;
                }
                break;
            case SDL_SCANCODE_M | WITH_SHIFT: // shift-m
                if (is_cutscene || is_ending_sequence || going_up_stairs() || is_lev_6_seq) {
                    break;
                }
                if (debug_cheats_enabled) {
                    release_arrows();
                    if (Kid.frame == frame_15_stand && !is_kid_shadow && current_level < 15) {
                        if (is_positional_audio) {
                            set_sound_pos(Kid.room, Kid.curr_col, Kid.curr_row);
                            play_sound(sound_70_mouse_squeak);
                        }
                        seqtbl_offset_kid_char(seq_106_mouse);
                        play_kid_seq();
                        flash_color = color_7_lightgray;
                        flash_time = 6;
                    } else if (Kid.frame >= frame_186_mouse_1 && Kid.frame <= frame_188_mouse_stand) {
                        seqtbl_offset_kid_char(seq_2_stand);
                        play_kid_seq();
                    }
                }
                break;
            case SDL_SCANCODE_T | WITH_SHIFT: // shift-t
                // custom logic - do not add lives during the initial level 6 sequence
                if (is_lev_6_seq) {
                    break;
                }
                // custom logic - hitpoint cheats do not work with shadow
                if ((!is_kid_shadow || is_shadow_flame) && resurrect_time == 0 && (hitp_curr != hitp_max || hitp_max < settings->max_hitp_allowed)) {
                    play_sound(sound_30_big_potion); // big potion (cheat)
                    flash_color = color_15_brightwhite;
                    flash_time = 4;
                    add_life();
                }
                break;
            // custom logic - added POP2 like cheat takes away 1 hitpoint from kid
            case SDL_SCANCODE_K | WITH_SHIFT: // shift-k
                // custom logic - do not kill kid during the initial level 6 sequence
                if (is_lev_6_seq) {
                    break;
                }
                // custom logic - hitpoint cheats do not work with shadow
                if (hitp_curr > 0 && (!is_kid_shadow || is_shadow_flame) && resurrect_time == 0) {
                    play_sound(sound_13_kid_hurt); // small blue potion (cheat)
                    hitp_delta = -1;
                }
                break;
            case SDL_SCANCODE_K | WITH_CTRL | WITH_SHIFT: // ctrl-shift-k
                // custom logic - skeletons, mouse and Jaffar's vision are unkillable
                if (guardhp_curr == 1) {
                    goto loc_scancode_k;
                }
                if (guardhp_curr > 0 &&
                        Guard.room == Kid.room &&
                        Guard.charid != charid_4_skeleton &&
                        Guard.charid != charid_24_mouse) {
                    if (Guard.charid == charid_1_shadow) { // hurting shadow hurts kid
                        play_sound(sound_13_kid_hurt);
                        hitp_delta = -1;
                    } else {
                        set_sound_pos(Guard.room, Guard.curr_col, Guard.curr_row);
                        play_sound(sound_12_guard_hurt);
                        guardhp_delta = -1;
                    }
                }
                break;
            // custom logic - turn kid into shadow on levels 10 (after uniting with shadow) and 11-13
            case SDL_SCANCODE_A | WITH_SHIFT: // shift-a
                separate_kid_from_shadow();
                break;
            case SDL_SCANCODE_O | WITH_SHIFT: // shift-o
                // custom logic - do not get sword during the initial level 6 sequence
                if (is_lev_6_seq) {
                    break;
                }
                if (!have_sword && hitp_curr > 0 && Kid.frame == frame_15_stand) {
                    pickup_obj_type = -1;
                    proc_get_object();
                    seqtbl_offset_kid_char(seq_91_get_sword);
                    have_sword = 1;
                }
                break;
            case SDL_SCANCODE_T:
                is_timer_displayed = 1 - is_timer_displayed; // toggle
                break;
            case SDL_SCANCODE_F:
                is_feather_timer_displayed = 1 - is_feather_timer_displayed; // toggle
                break;
            case SDL_SCANCODE_I:
                is_upside_down_timer_displayed = 1 - is_upside_down_timer_displayed; // toggle
                break;
        }
    }
    // custom logic - do not display text in demo mode
    if (need_show_text && !is_demo) {
        display_text_bottom(answer_text);
        text_time_total = 24;
        text_time_remaining = 24;
    }
    return 1;
}

// custom logic - play ambient music
void play_ambient_music(void) {
    if (is_heartbeat_music() &&
            is_positional_audio &&
            !ogg_fore_sound_playing &&
            !feather_fall_timer) {
        if (Kid.alive >= 0 || hitp_curr == 0) {
            stop_ambient_music();
        } else {
            start_ambient_music(true);
        }
    } else if (music_mode >= 0 &&
               is_positional_audio &&
               is_screaming != 1 &&
               !ogg_fore_sound_playing &&
               !feather_fall_timer) {
        bool is_music_playing = Mix_PlayingMusic();
        if (Kid.alive >= 0 || hitp_curr == 0) {
            stop_ambient_music();
        } else if (track != NULL &&
                   music_suspended == 0 &&
                   music_suspended_timer == 0 &&
                   delayed_sound_timer == 0 &&
                   inactive_guard_timer == 0 &&
                   roomleave_result < 0 &&
                   ((Kid.action != actions_3_in_midair &&
                     Kid.action != actions_4_in_freefall &&
                     Kid.action != actions_5_bumped &&
                     (Kid.frame < frame_135_climbing_1 || Kid.frame > frame_149_climbing_15)
                    ) || Kid.frame == frame_109_crouch || !is_music_playing) &&
                   ((Guard.curr_col >= -1 && Guard.curr_col <= 10) || can_guard_see_kid < 2 || !is_music_playing) &&
                   (!check_sound_playing() || !is_exclusive_sound(current_sound))) {
            int mode = is_fight_music() ? 1 : 0;
            if (music_mode != mode) {
                stop_ambient_music();
                music_mode = mode;
                start_ambient_music(mode != 1);
            } else if (!is_music_playing) {
                start_ambient_music(true);
            }
        }
    } else {
        if (is_screaming != 1) {
            stop_ambient_music();
        }
        if (is_sound_on &&
                feather_fall_timer &&
                !check_sound_playing() &&
                Mix_PlayingMusic() == 0) {
            play_sound(sound_39_low_weight);
        }
    }
}

// custom logic - kid says "whew" after climbing up an almost closed gate
// or when the gate is closed on level 1 in front of a guard
void say_whew(void) {
    int gate_trob_index;
    // climbing up an almost closed gate
    if (Char.direction == dir_FF_left &&
            Char.frame == frame_149_climbing_15 &&
            get_tile_at_char() == tiles_4_gate &&
            curr_room_modif[curr_tilepos] > 0 &&
            curr_room_modif[curr_tilepos] <= 48) {
        gate_trob_index = find_trob_at_loc(curr_room, curr_tilepos);
        if (gate_trob_index < 0 || trobs[gate_trob_index].type != 0) {
            return;
        }
        delayed_sound_timer = 3; // wait for the next few frames
        delayed_sound = (struct delayed_sound_type) {
            sound_58_say_whew, 0, 0, 0, 0, 0, 0
        };
        // gate closed quickly infront of the guard on level 1
    } else if (!have_sword &&
               current_level == 1 &&
               trob.room == 15 &&
               trob.tilepos == 7 &&
               Kid.room == trob.room &&
               Kid.curr_row == trob.tilepos / 10 &&
               Guard.alive < 0 &&
               Guard.room == trob.room &&
               Guard.curr_row == Kid.curr_row &&
               Guard.direction == dir_0_right &&
               (Guard.curr_col == 6 || can_guard_see_kid == 1)) {
        delayed_sound_timer = music_suspended_timer = 6; // wait for the next few frames
        delayed_sound = (struct delayed_sound_type) {
            sound_58_say_whew, 0, 0, 0, 0, 0, 0
        };
    }
}

// seg000:08EB
void play_frame(void) {
    do_mobs();
    process_trobs();
    process_custom_events();
    check_skel();
    check_can_guard_see_kid();
    check_can_inactive_guard_see_kid();
    play_ambient_music();
    // if level is restarted, return immediately
    if (play_kid_frame()) return;
    play_guard_frame();
    if (resurrect_time == 0) {
        check_sword_hurting();
        check_sword_hurt();
    }
    check_sword_vs_sword();
    do_delta_hp();
    exit_room();
    check_the_end();
    check_guard_fallout();
    check_falling_guard();
    // custom logic - kid says "whew" after climbing up an almost closed gate
    say_whew();
    if (current_level == 1) {
        // custom logic - slam the top gate on level 1, room 1 when kid jumps
        if (current_level == 1 &&
                get_tile(5, 9, 0) == tiles_4_gate &&
                curr_room_modif[curr_tilepos] > 0 &&
                Kid.room == 1 &&
                !tile_is_floor(get_tile_at_kid(Kid.x)) &&
                (Kid.curr_row == 0 || Kid.curr_row == 1) &&
                (Kid.curr_col == 0 || Kid.curr_col == 1 || Kid.curr_col == 2)) {
            get_tile(5, 2, 0);
            trigger_button(1, tiles_6_closer, -1);
        }
    } else if (current_level == settings->falling_exit_level) {
        // Special event: level 6 falling exit
        if (roomleave_result == -2) {
            Kid.y = -1;
            stop_all_sounds();
            stop_ambient_music();
            ++next_level;
        }
    } else if (settings->tbl_seamless_exit[current_level] >= 0) {
        // Special event: level 12 running exit
        if (Kid.room == settings->tbl_seamless_exit[current_level] &&
                Kid.alive < 0 &&
                !is_kid_shadow) {
            ++next_level;
            // custom logic - restore missing hitpoints on level 12B
            hitp_curr = hitp_max;
            // Sounds must be stopped, because play_level_2() checks next_level only if there are no sounds playing.
            stop_all_sounds();
            seamless = 1;
        }
    }
    show_time();
    // expiring doesn't count on Jaffar/princess level
    if (current_level < settings->win_level && rem_min == 0) {
        expired();
    }
}

// seg000:09B6
void draw_game_frame(void) {
    short ticks_remaining;
    if (need_full_redraw) {
        redraw_screen(0);
        need_full_redraw = 0;
    } else {
        if (different_room) {
            drawn_room = next_room;
            if (settings->tbl_level_type[current_level]) {
                gen_palace_wall_colors();
            }
            // custom logic - restores collapsed guard
            inactive_guard_restore();
            redraw_screen(1);
        } else {
            if (need_redraw_because_flipped) {
                need_redraw_because_flipped = 0;
                redraw_screen(0);
            } else {
                memset(&table_counts, 0, sizeof(table_counts));
                draw_moving();
                draw_tables();
                if (is_blind_mode) {
                    draw_rect(&rect_top, color_0_black);
                }
                if (upside_down_timer > 0) {
                    flip_screen(offscreen_surface);
                }
                while (drects_count--) {
                    copy_screen_rect(&drects[drects_count]);
                }
                if (upside_down_timer > 0) {
                    flip_screen(offscreen_surface);
                }
                drects_count = 0;
            }
        }
    }

    play_next_sound();
    // Note: texts are identified by their total time!
    if (text_time_remaining == 1) {
        // If the text's is about to expire:
        if (text_time_total == RESTART_AFTER_DEATH_DELAY) {
            // press button to continue
            // In this case, restart the game.
            start_level = -1;
            start_game();
        } else {
            // Otherwise, just clear it.
            erase_bottom_text(1);
        }
    } else {
        if (text_time_remaining != 0) {
            --text_time_remaining;
            if (text_time_total == RESTART_AFTER_DEATH_DELAY && text_time_remaining < 72) {
                // press button to continue
                // Blink the message:
                ticks_remaining = text_time_remaining % 12;
                if (ticks_remaining > 3) {
                    erase_bottom_text(0);
                } else {
                    if (ticks_remaining == 3) {
                        display_press_button();
                        current_sound_is_positional = false;
                        play_sound_from_buffer(sound_pointers[sound_38_blink]); // press button blink
                    }
                }
            }
        }
    }
}

// seg000:0B12
void anim_tile_modif(void) {
    word tilepos;
    for (tilepos = 0; tilepos < MAX_TILES_PER_ROOM; ++tilepos) {
        switch (get_curr_tile(tilepos)) {
            case tiles_10_potion:
                start_anim_potion(drawn_room, tilepos);
                break;
            case tiles_19_torch:
            case tiles_30_torch_with_debris:
                start_anim_torch(drawn_room, tilepos);
                break;
            // custom logic - tile 26 is a torch without a floor in dungeons
            case tiles_26_torch_empty:
                if (settings->tbl_level_type[current_level] == DUNGEON_LEVEL) {
                    start_anim_torch(drawn_room, tilepos);
                }
                break;
            case tiles_22_sword:
                start_anim_sword(drawn_room, tilepos);
                break;
            // custom logic - start animating balcony stars
            case tiles_23_balcony_left:
                start_anim_balcony(drawn_room, tilepos);
                break;
        }
    }

    // Animate torches in the rightmost column of the left-side room as well, because they are visible in the current room.
    for (int row = 0; row <= 2; row++) {
        // custom logic - start animating visible balcony stars of the left-side room as well
        if (get_tile(room_L, 8, row) == tiles_23_balcony_left) {
            start_anim_balcony(room_L, row * 10 + 8);
        }
        switch (get_tile(room_L, 9, row)) {
            case tiles_19_torch:
            case tiles_30_torch_with_debris:
                start_anim_torch(room_L, row * 10 + 9);
                break;
            // custom logic - tile 26 is a torch without a floor in dungeons
            case tiles_26_torch_empty:
                if (settings->tbl_level_type[current_level] == DUNGEON_LEVEL) {
                    start_anim_torch(room_L, row * 10 + 9);
                }
                break;
        }
    }
}

// custom logic - returns if the level uses dungeon 2 graphics
bool is_dungeon_2(void) {
    return current_level <= 15 &&
           settings->tbl_level_type[current_level] == DUNGEON_LEVEL &&
           settings->tbl_level_subtype[current_level] == DUNGEON_OR_PALACE_2_LEVEL;
}

// custom logic - returns if the level uses temple graphics
bool is_temple(void) {
    return current_level <= 15 &&
           settings->tbl_level_type[current_level] == PALACE_LEVEL &&
           settings->tbl_level_subtype[current_level] == TEMPLE_LEVEL;
}

// custom logic - load additional palettes
void load_more_palettes(void) {
    byte* flaming_sword_palette;
    dat_type* dathandle;
    dathandle = open_dat("PRINCE15.DAT");
    level_6_env_palette = load_from_opendats_alloc(4001, "bin", NULL, NULL);
    level_6_wall_palette = load_from_opendats_alloc(4002, "bin", NULL, NULL);
    flame_palette = load_from_opendats_alloc(4003, "bin", NULL, NULL);
    flaming_sword_palette = load_from_opendats_alloc(4004, "bin", NULL, NULL);
    close_dat(dathandle);
    set_chtab_palette(chtab_addrs[id_chtab_14_flaming_sword], flaming_sword_palette, 0x10);
}

// custom logic - level 6 has a different color palette
void load_lev6_palette(void) {
    if (level_6_env_palette == NULL || level_6_wall_palette == NULL) {
        printf("Could not load additional level 6 palette");
        return;
    }
    set_pal_arr(0x50, 0x10, (rgb_type*)level_6_env_palette);
    set_pal_arr(0x60, 0x10, (rgb_type*)level_6_wall_palette);
    set_chtab_palette(chtab_addrs[id_chtab_6_environment], level_6_env_palette, 0x10);
    set_chtab_palette(chtab_addrs[id_chtab_7_environmentwall], level_6_wall_palette, 0x10);
}

// seg000:0D20
void load_lev_spr(int level) {
    short guardtype;
    const char* filename;
    current_level = next_level = level;
    draw_rect(&screen_rect, color_0_black);
    free_optsnd_chtab();
    filename = settings->tbl_level_dat[level];
    load_chtab_from_file(id_chtab_6_environment, 200, filename, 1 << 5);
    load_more_opt_graf(filename);
    guardtype = settings->tbl_guard_type[level];
    if (guardtype != -1) {
        // custom logic - dungeon 2 uses a different skeleton file
        if (is_dungeon_2() && strcmp("SKEL1.DAT", settings->tbl_guard_dat[guardtype]) == 0) {
            load_chtab_from_file(id_chtab_5_guard, 750, "SKEL2.DAT", 1 << 8);
        } else {
            load_chtab_from_file(id_chtab_5_guard, 750, settings->tbl_guard_dat[guardtype], 1 << 8);
        }
    }
    load_chtab_from_file(id_chtab_7_environmentwall, 360, filename, 1 << 6);
    // (palace wall pattern) #D49850 = light brown
    set_pal(6, 0x35, 0x26, 0x14);
    // Level colors (1.3)
    if (level_var_palettes != NULL) {
        int level_color = settings->tbl_level_color[level];
        if (level_color != 0) {
            byte *env_pal = level_var_palettes + 0x30 * (level_color - 1);
            set_pal_arr(0x50, 0x10, (rgb_type*)env_pal);
            set_chtab_palette(chtab_addrs[id_chtab_6_environment], env_pal, 0x10);
            // custom logic - temple uses monochrome wall palette that should not change colors
            if (is_temple()) {
                return;
            }
            byte* wall_pal = level_var_palettes +
                             0x30 * (level_color - 1) +
                             0x30 * settings->tbl_level_type[level];
            set_pal_arr(0x60, 0x10, (rgb_type*)wall_pal);
            set_chtab_palette(chtab_addrs[id_chtab_7_environmentwall], wall_pal, 0x10);
            if (level_color == 4) {
                // (palace wall pattern) #A07864 = brown
                set_pal(6, 0x28, 0x1E, 0x19);
            }
        }
    }
}

// seg000:0E6C
void load_level(void) {
    dat_type* dathandle;
    dathandle = open_dat("LEVELS.DAT");
    load_from_opendats_to_area(current_level + 2000, &level, sizeof(level), "bin");
    close_dat(dathandle);

    alter_mods_allrm();
    reset_level_unused_fields(true); // added
}

void reset_level_unused_fields(bool loading_clean_level) {
    // Entirely unused fields in the level format: reset to zero for now.
    // Those fields were used by Jordan Mechner's level editor.
    memset(level.fill_1, 0, sizeof(level.fill_1));
    memset(level.fill_2, 0, sizeof(level.fill_2));
    memset(level.fill_3, 0, sizeof(level.fill_3));

    // level.used_rooms is 25 on some levels. Limit it to the actual number of rooms.
    if (level.used_rooms > 24) level.used_rooms = 24;

    // For these fields, only use the bits that are actually used, and set the rest to zero.
    // Good for repurposing the unused bits in the future.
    int i;
    for (i = 0; i < level.used_rooms; ++i) {
        level.guards_skill[i] &= 0x0F; // 4 bits in use
    }

    // In savestates, additional information may be stored (e.g. remembered guard hp) - should not reset this then!
    if (loading_clean_level) {
        for (i = 0; i < level.used_rooms; ++i) {
            level.guards_color[i] &= 0x0F; // 4 bits in use (other 4 bits repurposed as remembered guard hp)
        }
    }
}

// custom logic - Jaffar's vision reappears in a different spot when
// he gets close to an edge of the room
void check_jaffar_vision_reappearing(void) {
    if (current_level == 13 &&
            Char.alive < 0 &&
            guardhp_curr > 0 &&
            Char.room == 1 &&
            Char.curr_row == 2 &&
            Char.sword == sword_2_drawn) {
        if (Char.direction == dir_FF_left && Char.x > x_bump[9 + ONSCREEN_COLUMN_OFFSET]) {
            Char.direction = dir_0_right;
        } else if (Char.direction == dir_0_right && Char.x < x_bump[3 + ONSCREEN_COLUMN_OFFSET]) {
            Char.direction = dir_FF_left;
        } else {
            return;
        }
        Char.curr_col = 5;
        Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
        make_guard_notice(false);
        stop_sounds();
        play_sound(sound_45_jump_through_mirror);
    }
}

// seg000:0EA8
// returns 1 if level is restarted, 0 otherwise
int play_kid_frame(void) {
    loadkid_and_opp();
    load_fram_det_col();
    play_kid();
    if (is_restart_level) {
        return 1;
    }
    // custom logic - if kid is dead, undo the upside down mode
    if (upside_down_timer > 0 && Char.alive >= 0 && is_dead()) {
        upside_down_timer = 0;
        need_redraw_because_flipped = 1;
    }
    // custom logic - before level 10 kid should not be able to run into shadow.
    // This is possible on level 7.
    if (current_level < 10 &&
            jumped_through_mirror == 0 &&
            Char.action == actions_1_run_jump &&
            Guard.action == actions_1_run_jump &&
            Guard.charid == charid_1_shadow &&
            Char.room == Guard.room &&
            Char.curr_row == Guard.curr_row &&
            ABS(Char.x - Guard.x) <= 7) {
        Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
        Char.fall_y = 0;
        seqtbl_offset_char(seq_47_bump); // bump into opponent
        play_seq();
    }
    if (Char.room != 0 && Char.direction != dir_56_none) {
        play_seq();
        fall_accel();
        fall_speed();
        load_frame_to_obj();
        load_fram_det_col();
        set_char_collision();
        bump_into_opponent();
        bump_into_inactive_opponent();
        check_collisions();
        check_bumped();
        check_gate_push();
        check_action();
        check_press();
        check_spike_near();
        if (resurrect_time <= 0) {
            check_spiked();
            check_chomped_kid();
        }
        check_knock();
    }
    savekid();
    return 0;
}

// seg000:0F48
void play_guard_frame(void) {
    if (Guard.direction != dir_56_none) {
        loadshad_and_opp();
        load_fram_det_col();
        play_guard();
        if (Char.direction != dir_56_none) {
            // custom logic - shadow and mouse should not stop if kid is not in the room
            // until they reach the end of the room
            if (Char.charid == charid_1_shadow || Char.charid == charid_24_mouse) {
                // shadow should not disappear between rooms on level 10
                if (current_level == 10 && Char.charid == charid_1_shadow) {
                    goto loc_play_guard;
                }
                if (Char.x >= 44 && Char.x < 211 && Char.curr_row < 3) {
                    goto loc_play_guard;
                }
                clear_char();
                saveshad();
                // activate inactive guard
                if (inactive_guard[drawn_room - 1].room == drawn_room && inactive_guard[drawn_room - 1].direction != dir_56_none) {
                    enter_guard();
                }
                return;
                // custom logic - skeleton on the bridge on level 12 should continue falling if he is in a different room
            } else if (Char.charid == charid_4_skeleton && current_level == 12 && falling_bridge_timer != 0) {
                goto loc_play_guard;
            }
        }
        if (Char.room == drawn_room && Char.direction != dir_56_none) {
loc_play_guard:
            play_seq();
            check_jaffar_vision_reappearing();
            fall_accel();
            fall_speed();
            load_frame_to_obj();
            load_fram_det_col();
            set_char_collision();
            check_guard_bumped();
            check_action();
            check_press();
            check_spike_near();
            check_spiked();
            check_chomped_guard();
            // custom logic - knock should be checked for kid and guards
            check_knock();
        }
        saveshad();
    }
    // custom logic - checks whether inactive guard is on a button
    check_inactive_press();
    // custom logic - control inactive guard
    autocontrol_inactive_guard();
    play_inactive_seq();
    // custom logic - check for chomped dead or inactive guards
    check_dead_or_inactive_guards_chomped();
}

void dijkstra_visit_room(byte room) {
    if (!(room > 0 && room <= level.used_rooms)) return;
    room_offset_type* current_node = &room_offsets[room - 1];
    current_node->visited = 1;

    byte next_distance = current_node->distance + 1;
    sbyte curr_offset_x = current_node->dx;
    sbyte curr_offset_y = current_node->dy;
    link_type links = level.roomlinks[room - 1];
    room_offset_type* left =  (links.left > 0 && links.left <= 24)   ? &room_offsets[links.left - 1] : NULL;
    room_offset_type* right = (links.right > 0 && links.right <= 24) ? &room_offsets[links.right - 1] : NULL;
    room_offset_type* up =    (links.up > 0 && links.up <= 24)       ? &room_offsets[links.up - 1] : NULL;
    room_offset_type* down =  (links.down > 0 && links.down <= 24)   ? &room_offsets[links.down - 1] : NULL;
    if (left != NULL && next_distance < left->distance) {
        left->distance = next_distance;
        left->dx = curr_offset_x - 1;
        left->dy = curr_offset_y;
    }
    if (right != NULL && next_distance < right->distance) {
        right->distance = next_distance;
        right->dx = curr_offset_x + 1;
        right->dy = curr_offset_y;
    }
    if (up != NULL && next_distance < up->distance) {
        up->distance = next_distance;
        up->dx = curr_offset_x;
        up->dy = curr_offset_y - 1;
    }
    if (down != NULL && next_distance < down->distance) {
        down->distance = next_distance;
        down->dx = curr_offset_x;
        down->dy = curr_offset_y + 1;
    }
}

// Find where all reachable rooms are located relative to the current room, using Dijkstra's algorithm.
// https://en.wikipedia.org/wiki/Dijkstra%27s_algorithm
void find_all_rooms_from(byte starting_room) {
    memset(room_offsets, 0, sizeof(room_offsets));
    for (int i = 0; (size_t) i < COUNT(room_offsets); ++i) {
        room_offsets[i].distance = 255;
    }
    room_offsets[starting_room - 1].distance = 0; // The starting node is at distance=0 by definition.
    room_offsets[starting_room - 1].visited = false;

    int current_node = drawn_room;
    for (int i = 1; i <= level.used_rooms; ++i) {
        dijkstra_visit_room(current_node);
        // Find the unvisited node with the lowest distance to the starting node, and repeat.
        current_node = 0;
        int closest_distance = 255;
        for (int j = 1; j <= level.used_rooms; ++j) {
            room_offset_type* node = &room_offsets[j - 1];
            if (!node->visited && node->distance < closest_distance) {
                closest_distance = node->distance;
                current_node = j;
            }
        }
        if (current_node > 0) continue;
        else break; // All nodes visited.
    }
}

// seg000:0FBD
void check_the_end(void) {
    if (next_room != 0 && next_room != drawn_room) {
        drawn_room = next_room;
        load_room_links();
        if (current_level == settings->win_level &&
                drawn_room == settings->win_room &&
                // custom logic - kid is not shadow
                !is_kid_shadow) {
            // custom logic - reset timers
            feather_fall_timer = 0;
            upside_down_timer = 0;
            // custom logic - stop sounds
            stop_all_sounds();
            // Special event: end of game
            end_sequence();
        }
        different_room = 1;
        loadkid();
        anim_tile_modif();
        start_chompers();
        check_fall_flo();
        check_shadow();
        find_all_rooms_from(drawn_room);
    }
}

// seg000:1009
void check_fall_flo(void) {
    // Special event: falling floors
    if (current_level == settings->loose_tiles_level &&
            (drawn_room == settings->loose_tiles_room_1 || drawn_room == settings->loose_tiles_room_2) &&
            Kid.room == drawn_room) {
        get_room_address(curr_room = room_A);
        for (curr_tilepos = settings->loose_tiles_first_tile;
                curr_tilepos <= settings->loose_tiles_last_tile; ++curr_tilepos) {
            // custom logic - check that the tile is actually a loose tile
            if ((curr_room_tiles[curr_tilepos] & 0x1F) != tiles_11_loose) {
                continue;
            }
            // custom logic - last tile in the second room falls down immediately
            if (drawn_room == settings->loose_tiles_room_2 && curr_tilepos == settings->loose_tiles_last_tile) {
                make_loose_fall(settings->loose_floor_delay);
            } else {
                make_loose_fall(-(prandom(0xFF) & 0x0F));
            }
        }
    }
}

void get_joystick_state(int raw_x, int raw_y, int axis_state[2]) {

#define DEGREES_TO_RADIANS (M_PI/180.0)

    // check if the X/Y position is within the 'dead zone' of the joystick
    int dist_squared = raw_x * raw_x + raw_y * raw_y;
    if (dist_squared < settings->controller_threshold * settings->controller_threshold) {
        axis_state[0] = control_0_released;
        axis_state[1] = control_0_released;
    } else {
        double angle = atan2(raw_y, raw_x); // angle of the joystick: 0 = right, >0 = downward, <0 = upward

        if (fabs(angle) < (60 * DEGREES_TO_RADIANS)) // 120 degree range facing right
            axis_state[0] = control_1_held_right;

        else if (fabs(angle) > (120 * DEGREES_TO_RADIANS)) // 120 degree range facing left
            axis_state[0] = control_m1_held_left;

        else {
            // joystick is neutral horizontally, so the control should be released
            // however: prevent stop running if the Kid was already running / trying to do a running-jump
            // (this tweak makes it a bit easier to do (multiple) running jumps)
            if (!(angle < 0 /*facing upward*/ && Kid.action == actions_1_run_jump)) {
                axis_state[0] = control_0_released;
            }
        }

        if (angle < (-30 * DEGREES_TO_RADIANS) && angle > (-150 * DEGREES_TO_RADIANS)) // 120 degree range facing up
            axis_state[1] = Kid.sword == sword_2_drawn && upside_down_timer > 0 ? control_0_released : control_m1_held_up;

        // down slightly less sensitive than up (prevent annoyance when your thumb slips down a bit by accident)
        // (not sure if this adjustment is really necessary)
        else if (angle > (35 * DEGREES_TO_RADIANS) && angle < (145 * DEGREES_TO_RADIANS)) // 110 degree range facing down
            // custom logic - do not put sword away with rotating stick control
            axis_state[1] = Kid.sword == sword_2_drawn && upside_down_timer == 0 ? control_0_released : control_1_held_down;

        else {
            // joystick is neutral vertically, so the control should be released
            // however: should prevent unintended standing up when attempting to crouch-hop
            if (!((Kid.frame >= frame_108_fall_land_2 && Kid.frame <= frame_112_stand_up_from_crouch_3)
                    && angle > 0 /* facing downward */)) {
                axis_state[1] = control_0_released;
            }
        }
    }
}

// seg000:1051
void read_joyst_control(void) {
    int key_state = keystate_held | keystate_held_new;

    get_joystick_state(joy_axis_max[SDL_CONTROLLER_AXIS_LEFTX], joy_axis_max[SDL_CONTROLLER_AXIS_LEFTY], joy_left_stick_states);
    get_joystick_state(joy_axis_max[SDL_CONTROLLER_AXIS_RIGHTX], joy_axis_max[SDL_CONTROLLER_AXIS_RIGHTY], joy_right_stick_states);

    if (joy_left_stick_states[0] == control_m1_held_left ||
            joy_right_stick_states[0] == control_m1_held_left ||
            joy_button_states[joyinput_dpad_left] & key_state) {
        control_x = control_m1_held_left;
    }

    if (joy_left_stick_states[0] == control_1_held_right ||
            joy_right_stick_states[0] == control_1_held_right ||
            joy_button_states[joyinput_dpad_right] & key_state) {
        control_x = control_1_held_right;
    }

    if (joy_left_stick_states[1] == control_m1_held_up ||
            joy_right_stick_states[1] == control_m1_held_up ||
            joy_button_states[joyinput_dpad_up] & key_state ||
            joy_button_states[joyinput_y] & key_state) {
        control_y = upside_down_timer > 0 ? control_1_held_down : control_m1_held_up;
    }

    if (joy_left_stick_states[1] == control_1_held_down ||
            joy_right_stick_states[1] == control_1_held_down ||
            joy_button_states[joyinput_dpad_down] & key_state ||
            joy_button_states[joyinput_a] & key_state) {
        control_y = upside_down_timer > 0 ? control_m1_held_up : control_1_held_down;
    }

    if (joy_button_states[joyinput_x] & key_state ||
            joy_axis_max[SDL_CONTROLLER_AXIS_TRIGGERLEFT] > settings->controller_threshold ||
            joy_axis_max[SDL_CONTROLLER_AXIS_TRIGGERRIGHT] > settings->controller_threshold) {
        control_shift = control_m1_held;
    }

    // custom logic - joystick B button draws the sword
    if (joy_button_states[joyinput_b] & key_state) {
        control_alt = control_m1_held;
    }

    // custom logic - quick save when left and right shoulder buttons are pressed
    if (joy_button_states[joyinput_left_shoulder] & key_state && joy_button_states[joyinput_right_shoulder] & key_state) {
        joy_button_states[joyinput_left_shoulder] = control_0_released;
        joy_button_states[joyinput_right_shoulder] = control_0_released;
        need_quick_load = 1;
    } else {
        control_shoulder = (joy_button_states[joyinput_left_shoulder] & key_state) ||
                           (joy_button_states[joyinput_right_shoulder] & key_state)
                           ? control_m1_held
                           : control_0_released;
    }
}

// seg000:10EA
void draw_kid_hp(short curr_hp, short max_hp) {
    short drawn_hp_index;
    // custom logic - do not show kid's hitpoints during the level 6 animation
    if (current_level == 6 && falling_exit_checkpoint == 0 && level_6_sequence_timer > 0) {
        return;
    }
    // custom logic - do not redraw hitpoints when shadow has been killed not in a fight
    if (hitp_curr == 0 && is_kid_shadow && (!is_shadow_flame || holding_sword == 0)) {
        return;
    }
    for (drawn_hp_index = curr_hp; drawn_hp_index < max_hp; ++drawn_hp_index) {
        // empty HP
        if (is_kid_shadow) {
            method_6_blit_img_to_scr(get_image(id_chtab_2_kid, 216, 0), drawn_hp_index * 6, 194, blitters_9_black);
        } else {
            method_6_blit_img_to_scr(get_image(id_chtab_2_kid, 217, 0), drawn_hp_index * 6, 194, blitters_0_no_transp);
        }
    }
    for (drawn_hp_index = 0; drawn_hp_index < curr_hp; ++drawn_hp_index) {
        // full HP
        method_6_blit_img_to_scr(get_image(id_chtab_2_kid, 216, 0), drawn_hp_index * 6, 194, blitters_0_no_transp);
    }
}

// custom logic - clears kid hp
void clear_kid_hp(short max_hp) {
    short drawn_hp_index;
    for (drawn_hp_index = 0; drawn_hp_index < max_hp; ++drawn_hp_index) {
        // empty HP
        method_6_blit_img_to_scr(get_image(id_chtab_2_kid, 216, 0), drawn_hp_index * 6, 194, blitters_9_black);
    }
}

// seg000:1159
void draw_guard_hp(short curr_hp, short max_hp) {
    short drawn_hp_index;
    short guard_charid;
    short guard_type;
    short guard_color;
    chtab_type* chtab;
    if (chtab_addrs[id_chtab_5_guard] == NULL) return;
    guard_charid = Guard.charid;
    guard_color = Guard.color;
    // custom logic - do not draw hit points for bowing birdmen
    if (current_level == 10 &&
            leveldoor_open >= DEAD_KID_LEVEL_10_TIMER &&
            !is_kid_shadow) {
        curr_hp = 0;
        max_hp = MAX_GUARD_HP;
    }
    // custom logic - shadow and mouse have no visible hit points
    // but inactive guard hit points should still be visible
    if (guard_charid == charid_1_shadow || guard_charid == charid_24_mouse) {
        max_hp = MAX_GUARD_HP;
        if (inactive_guard[drawn_room - 1].room == drawn_room) {
            curr_hp = (level.guards_color[drawn_room - 1] & 0xF0) >> 4;
            guard_color = level.guards_color[drawn_room - 1] & 0x0F;
            if (curr_hp == 1 && blink_timer % 2 == 0) {
                curr_hp = 0;
            }
        } else {
            curr_hp = 0;
        }
    }
    guard_type = settings->tbl_guard_type[current_level];
    if (guard_charid != charid_4_skeleton &&
            // custom logic - Jaffar's vision has no visible lives
            guard_type != guard_type_Jaffar_vision
       ) {
        chtab = chtab_addrs[id_chtab_5_guard];
        // custom logic - setting the correct palette color for the guard
        if (guard_type == guard_type_guard) {
            if (current_level == settings->fat_guard_level && guard_color == 0) {
                chtab = chtab_addrs[id_chtab_10_fat_guard];
            } else if (guard_color != last_guard_color && guard_color > 0 && guard_color <= GUARD_COLORS) {
                set_chtab_palette(chtab, &guard_palettes[0x30 * guard_color - 0x30], 0x10);
                last_guard_color = guard_color;
            }
        }
        for (drawn_hp_index = curr_hp; drawn_hp_index < max_hp; ++drawn_hp_index) {
            method_6_blit_img_to_scr(chtab->images[0], 314 - drawn_hp_index * 6, 194, blitters_9_black);
        }
        for (drawn_hp_index = 0; drawn_hp_index < curr_hp; ++drawn_hp_index) {
            method_6_blit_img_to_scr(chtab->images[0], 314 - drawn_hp_index * 6, 194, blitters_0_no_transp);
        }
    }
}

// seg000:11EC
void add_life(void) {
    short hpmax = hitp_max;
    ++hpmax;
    if (hpmax > settings->max_hitp_allowed) hpmax = settings->max_hitp_allowed;
    hitp_max = hpmax;
    set_health_life();
}

// seg000:1200
void set_health_life(void) {
    hitp_delta = hitp_max - hitp_curr;
}

// seg000:120B
void draw_hp(void) {
    if (hitp_delta) {
        draw_kid_hp(hitp_curr, hitp_max);
    }
    // custom logic - blink on 1 live or when kid is resurrected
    if (resurrect_time == 1) {
        draw_kid_hp(hitp_curr, hitp_max);
    } else if (resurrect_time > 1) {
        // blinking hitpoint
        if (blink_timer % 2 == 0) {
            draw_kid_hp(hitp_curr, 0);
        } else {
            draw_kid_hp(0, hitp_curr);
        }
    } else if (hitp_curr == 1 && !is_kid_shadow) {
        if (blink_timer % 4 == 0) {
            draw_kid_hp(hitp_curr, 0);
        } else {
            draw_kid_hp(0, hitp_curr);
        }
    }
    if (guardhp_delta) {
        draw_guard_hp(guardhp_curr, guardhp_max);
    }
    // custom logic - guard's current hitpoint blinks after landing 2 punches in a fist fight
    if (!have_sword && Kid.sword == sword_2_drawn) {
        if (guardhp_curr > 0 &&
                guard_punch_count == 2 &&
                Kid.sword == sword_2_drawn &&
                blink_timer % 2 == 0) {
            draw_guard_hp(guardhp_curr - 1, guardhp_max);
            return;
        } else if (guardhp_curr > 1 || Kid.sword == sword_2_drawn) {
            draw_guard_hp(guardhp_curr, guardhp_max);
            return;
        }
    }
    // custom logic - shadow and mouse have no visible hit points
    if (Guard.charid == charid_1_shadow || Guard.charid == charid_24_mouse) {
        draw_guard_hp(0, MAX_GUARD_HP);
        return;
    }
    if (guardhp_curr == 1) {
        if (blink_timer % 2 == 0) {
            draw_guard_hp(1, 0);
        } else {
            draw_guard_hp(0, 1);
        }
    }
}

// seg000:127B
void do_delta_hp(void) {
    word prev_hitp = hitp_curr;
    hitp_curr = MIN(MAX(hitp_curr + hitp_delta, 0), hitp_max);
    guardhp_curr = MIN(MAX(guardhp_curr + guardhp_delta, 0), guardhp_max);
    // custom logic - reset guard punch count after a hitpoint loss
    if (guardhp_delta < 0 && guard_punch_count > 0) {
        guard_punch_count = 0;
    }
    // custom logic - stop music to switch to/from heartbeat
    if (music_mode >= 0 && !ogg_fore_sound_playing && !is_kid_shadow) {
        if ((hitp_curr == 1 && prev_hitp > 1) ||
                (hitp_curr != 1 && prev_hitp == 1)) {
            Mix_HaltMusic();
        }
    }
}

// seg000:12C5
void play_sound(int sound_id) {
    bool is_sound_positional = want_positional_sound;
    want_positional_sound = false;
    // custom logic - do not play new sounds after a level has been restarted
    if (is_restart_level) {
        return;
    }
    // custom logic - do not interrupt out of time sound
    if (current_sound == sound_36_out_of_time && check_sound_playing()) {
        return;
    }
    // custom logic - do not play certain sounds for shadow
    if (Char.charid == charid_1_shadow || (Char.charid == charid_0_kid && is_kid_shadow)) {
        bool is_delayed_sound = delayed_sound.sound_id == sound_id && delayed_sound_timer == 0;
        if (sound_id == sound_0_fell_to_death ||
                sound_id == sound_8_bumped ||
                sound_id == sound_9_grab ||
                sound_id == sound_16_fat_fart ||
                sound_id == sound_18_drink ||
                sound_id == sound_23_footstep ||
                sound_id == sound_30_big_potion ||
                (!is_delayed_sound && sound_id == sound_17_soft_land) ||
                (!is_delayed_sound && sound_id == sound_13_kid_hurt) ||
                (!is_shadow_flame && sound_id == sound_11_sword_moving) ||
                (!is_shadow_flame && sound_id == sound_19_draw_sword) ||
                (!is_shadow_flame && sound_id == sound_34_hit_by_sword)) {
            return;
        }
    }
    // custom logic - sounds that make inactive guard turn
    if (inactive_guard[drawn_room - 1].room == drawn_room) {
        if (sound_id == sound_8_bumped ||
                sound_id == sound_9_grab ||
                sound_id == sound_10_sword_vs_sword ||
                sound_id == sound_12_guard_hurt ||
                sound_id == sound_13_kid_hurt ||
                sound_id == sound_23_footstep ||
                (Char.charid == charid_0_kid && sound_id == sound_0_fell_to_death) ||
                (Char.charid == charid_0_kid && sound_id == sound_17_soft_land)) {
            is_inactive_guard_notice = 1;
        }
    }
    // custom logic - do not play hurt or victory sounds after kid is dead
    if ((sound_id == sound_13_kid_hurt || sound_id == sound_37_victory) && Kid.alive >= 0) {
        return;
    }
    // custom logic - play foreground sounds on a separate channel under specific conditions
    if (is_positional_audio && curr_ogg_sound_id != sound_79_united_with_shadow) {
        // victory and death sounds are different in the temple, flame level or
        // when kid has turned into shadow
        if (is_kid_shadow ||
                (current_level == settings->flame_level && holding_sword) ||
                is_temple()) {
            switch (sound_id) {
                case sound_24_death_regular:
                    sound_id = sound_64_death_temple;
                    break;
                case sound_28_death_in_fight:
                    sound_id = sound_66_death_in_fight_temple;
                    break;
                case sound_37_victory:
                    sound_id = sound_63_temple_victory;
                    break;
            }
        }
        if (is_exclusive_sound(sound_id)) {
            if (ogg_fore_sound_playing) {
                Mix_HaltMusic();
                ogg_fore_sound_playing = false;
            }
            if (feather_fall_timer) {
                Mix_HaltMusic();
                feather_fall_timer = 0;
            }
        }
        if (sound_id == sound_39_low_weight && feather_fall_timer > 0) {
            double ticks_per_sec = get_ticks_per_sec(timer_1);
            double sound_pos = (FEATHER_FALL_IN_SEC * ticks_per_sec - feather_fall_timer) / ticks_per_sec;
            if (play_ogg_fore_sound(sound_id, sound_pos, false) == 0) {
                return;
            }
        } else if (play_ogg_fore_sound(sound_id, 0, true) == 0) {
            return;
        }
    }
    // custom logic - ignore sounds that do not exist for non-positional audio
    if (sound_id > MAX_FORE_SOUND_ID) {
        next_sound = -1;
        next_sound_is_positional = false;
        return;
    }
    if (next_sound < 0 || settings->sound_prio_table[sound_id] <= settings->sound_prio_table[next_sound]) {
        if (NULL == sound_pointers[sound_id]) {
            return;
        }
        if (enable_positional_audio && is_sound_positional) {
            // Don't interrupt the previous sound if this one is further away
            if (!(sound_id == next_sound && want_sound_distance_loss < next_sound_distance_loss)) {
                next_sound_is_positional = is_sound_positional;
                next_sound_dir_x = want_sound_dir_x;
                next_sound_distance_loss = want_sound_distance_loss;
            }
        } else {
            next_sound_is_positional = false;
        }
        next_sound = sound_id;
    }
}

// custom logic - sets positional sound distance from kid with a specified loss multiplier
void set_sound_pos_with_dist_loss(int room, int col, int row, float distance_loss_multiplier) {
    if (!enable_positional_audio) return;
    want_positional_sound = false;
    want_sound_distance_loss = 1.0f;
    want_sound_dir_x = 0.0f;
    if (!(room > 0 && room <= level.used_rooms)) return;
    room_offset_type room_offset = room_offsets[room - 1];
    if (room_offset.visited) {
        want_positional_sound = true;
        // Calculate the position of the sound within the room
        int tile_x = col * 32 + 16;
        int tile_y = 3 + row * 63;
        int sound_x = tile_x - (320 / 2); // offset in pixels from the center of the room
        int sound_y = tile_y - (192 / 2);
        // Place the sound in the other room
        sound_x += room_offset.dx * 320;
        sound_y += room_offset.dy * 189;
        if (upside_down_timer > 0) sound_y *= -1;
        // The listener is placed at some distance along the z-axis (away from the screen)
        float distance_z = 3.0f * 320.0f; // could be tweaked
        float distance_x = (float) ABS(sound_x);
        float distance_y = (float) ABS(sound_y);
        float sound_distance = sqrtf(distance_x * distance_x + distance_y * distance_y + distance_z * distance_z); // Assumes rooms are square...
        want_sound_distance_loss = 1.0f / MAX(1.0f, log2f(sound_distance / distance_z) * distance_loss_multiplier);
        float angle_in_xy_plane = atan2f((float) sound_y, (float) sound_x);
        float angle_with_z_axis = asinf(distance_z / sound_distance);

        want_sound_dir_x = cosf(angle_in_xy_plane);
        want_sound_dir_x *= cosf(angle_with_z_axis);
        want_sound_dir_x *= 5.0f; // amplify the directional effect
    }
}

// custom logic - sets positional sound distance from kid
void set_sound_pos(int room, int col, int row) {
    // Make distant sounds louder when ambient music is playing
    float distance_loss_multiplier = Mix_PlayingMusic() == 0 ? 6.0f : 4.5f;
    set_sound_pos_with_dist_loss(room, col, row, distance_loss_multiplier);
}

// seg000:1304
void play_next_sound(void) {
    // custom logic - ignore sounds that do not exist for non-positional audio
    if (next_sound >= 0 && next_sound <= MAX_FORE_SOUND_ID) {
        if (!check_sound_playing() ||
                (sound_interruptible[current_sound] != 0 && settings->sound_prio_table[next_sound] <= settings->sound_prio_table[current_sound])
           ) {
            if (enable_positional_audio) {
                current_sound_is_positional = next_sound_is_positional;
                current_sound_dir_x = next_sound_dir_x;
                current_sound_distance_loss = next_sound_distance_loss;
            } else {
                current_sound_is_positional = false;
            }
            current_sound = next_sound;
            play_sound_from_buffer(sound_pointers[current_sound]);
            // custom logic - slammed gate should not reset "whew" as the next sound
        } else if (current_sound == sound_6_gate_closing_fast && next_sound == sound_58_say_whew) {
            return;
            // custom logic - fallen tile should not reset "oww" as the next sound
        } else if (current_sound == sound_2_tile_crashing && next_sound == sound_13_kid_hurt) {
            return;
        }
    }
    next_sound = -1;
}

// seg000:1353
void check_sword_vs_sword(void) {
    if (!have_sword) {
        return;
    }
    if (Kid.frame == frame_167_blocked || Guard.frame == frame_167_blocked) {
        play_sound(sound_10_sword_vs_sword); // sword vs. sword
    }
}

// seg000:136A
void load_chtab_from_file(int chtab_id, int resource, const char *filename, int palette_bits) {
    dat_type* dathandle;
    if (chtab_addrs[chtab_id] != NULL) return;
    dathandle = open_dat(filename);
    chtab_addrs[chtab_id] = load_sprites_from_file(resource, palette_bits);
    close_dat(dathandle);
}

// seg000:13BA
void free_all_chtabs_from(int first) {
    word chtab_id;
    free_peels();
    for (chtab_id = first; chtab_id < 10; ++chtab_id) {
        if (chtab_addrs[chtab_id]) {
            free_chtab(chtab_addrs[chtab_id]);
            chtab_addrs[chtab_id] = NULL;
        }
    }
}

// seg009:12EF
void load_one_optgraf(chtab_type* chtab_ptr, dat_pal_type *pal_ptr, int base_id, int min_index, int max_index) {
    short index;
    for (index = min_index; index <= max_index; ++index) {
        image_type* image = load_image(base_id + index + 1, pal_ptr);
        if (image != NULL) chtab_ptr->images[index] = image;
    }
}

byte optgraf_min[] = {0x01, 0x1E, 0x4B, 0x4E, 0x56, 0x65, 0x7F, 0x0A};
byte optgraf_max[] = {0x09, 0x1F, 0x4D, 0x53, 0x5B, 0x7B, 0x8F, 0x11};

// seg000:13FC
void load_more_opt_graf(const char *filename) {
    dat_type* dathandle;
    dat_shpl_type area;
    short graf_index;
    dathandle = NULL;
    for (graf_index = 0; graf_index < 8; ++graf_index) {
        // loading optional resources for all levels
        if (dathandle == NULL) {
            dathandle = open_dat(filename);
            load_from_opendats_to_area(200, &area, sizeof(area), "pal");
            area.palette.row_bits = 0x20;
        }
        load_one_optgraf(chtab_addrs[id_chtab_6_environment], &area.palette, 1200, optgraf_min[graf_index] - 1, optgraf_max[graf_index] - 1);
    }
    if (dathandle != NULL) {
        close_dat(dathandle);
    }
}

// seg000:148D
int do_paused(void) {
    word key;
    key = 0;
    next_room = 0;
    control_shift = control_0_released;
    control_alt = control_0_released;
    control_y = control_0_released;
    control_x = control_0_released;
    // custom logic - do not read controls during demo
    if (current_level > 0 || play_demo_level) {
        if (is_joyst_mode) {
            read_joyst_control();
        } else {
            read_keyb_control();
        }
    }
    key = process_key(false);
    if (is_ending_sequence && is_paused) {
        is_paused = 0; // fix being able to pause the game during the ending sequence
    }
    if (is_paused) {
        // custom logic - do not pause ogg sounds
        if (!ogg_fore_sound_playing) {
            Mix_PauseMusic();
            music_paused = true;
        }
        // custom logic - stop feather fall music
        bool is_sound_playing = check_sound_playing();
        if (feather_fall_timer && is_sound_playing && current_sound == sound_39_low_weight) {
            stop_sounds();
        }
        // custom logic - stop level door sliding sound
        if (is_sound_playing && current_sound == sound_15_leveldoor_sliding) {
            stop_sounds();
        }
        display_text_bottom("GAME PAUSED");
        if (enable_pause_menu || is_menu_shown) {
            draw_menu();
            menu_was_closed();
        } else {
            is_paused = 0;
            // busy waiting?
            do {
                idle();
                delay_ticks(1);
            } while (!process_key(true));
        }
        erase_bottom_text(1);
        // custom logic - restart ambient music after the menu was closed
    } else if (music_paused) {
        Mix_ResumeMusic();
    }
    // As we processed input for current gameplay change all input to reflect their current status
    for (int i = 0; i < SDL_NUM_SCANCODES; i++) {
        key_states[i] &= ~keystate_held_new;
    }
    for (int i = 0; i < JOYINPUT_NUM; i++) {
        joy_button_states[i] &= ~keystate_held_new;
    }
    for (int i = 0; i < JOY_AXIS_NUM; i++) {
        joy_axis_max[i] = joy_axis[i];
    }
    return (key || control_shift) && !control_alt;
}

// seg000:1500
void read_keyb_control(void) {
    byte button_up, button_down, button_left, button_right;
    int key_state = keystate_held | keystate_held_new;
    // custom logic - invert up/down and related keys in upside down mode
    if (upside_down_timer > 0)  {
        button_up = key_states[SDL_SCANCODE_DOWN] & key_state
                    || key_states[SDL_SCANCODE_PAGEDOWN] & key_state
                    || key_states[SDL_SCANCODE_END] & key_state
                    || key_states[SDL_SCANCODE_KP_1] & key_state
                    || key_states[SDL_SCANCODE_KP_2] & key_state
                    || key_states[SDL_SCANCODE_KP_3] & key_state
                    || key_states[SDL_SCANCODE_KP_5] & key_state;
        button_down = key_states[SDL_SCANCODE_CLEAR] & key_state
                      || key_states[SDL_SCANCODE_UP] & key_state
                      || key_states[SDL_SCANCODE_HOME] & key_state
                      || key_states[SDL_SCANCODE_PAGEUP] & key_state
                      || key_states[SDL_SCANCODE_KP_7] & key_state
                      || key_states[SDL_SCANCODE_KP_8] & key_state
                      || key_states[SDL_SCANCODE_KP_9] & key_state;
        // custom logic - playing alpha keys in non-cheat mode
        if (!cheats_enabled) {
            button_up = button_up
                        || key_states[SDL_SCANCODE_X] & key_state
                        || key_states[SDL_SCANCODE_C] & key_state
                        || key_states[SDL_SCANCODE_Z] & key_state
                        || key_states[SDL_SCANCODE_S] & key_state;
            button_down = button_down
                          || key_states[SDL_SCANCODE_W] & key_state
                          || key_states[SDL_SCANCODE_E] & key_state
                          || key_states[SDL_SCANCODE_Q] & key_state;
        }
    } else {
        button_up = key_states[SDL_SCANCODE_UP] & key_state
                    || key_states[SDL_SCANCODE_PAGEUP] & key_state
                    || key_states[SDL_SCANCODE_HOME] & key_state
                    || key_states[SDL_SCANCODE_KP_7] & key_state
                    || key_states[SDL_SCANCODE_KP_8] & key_state
                    || key_states[SDL_SCANCODE_KP_9] & key_state;
        button_down = key_states[SDL_SCANCODE_CLEAR] & key_state
                      || key_states[SDL_SCANCODE_DOWN] & key_state
                      || key_states[SDL_SCANCODE_PAGEDOWN] & key_state
                      || key_states[SDL_SCANCODE_END] & key_state
                      || key_states[SDL_SCANCODE_KP_1] & key_state
                      || key_states[SDL_SCANCODE_KP_2] & key_state
                      || key_states[SDL_SCANCODE_KP_3] & key_state
                      || key_states[SDL_SCANCODE_KP_5] & key_state;
        // custom logic - playing alpha keys in non-cheat mode
        if (!cheats_enabled) {
            button_up = button_up
                        || key_states[SDL_SCANCODE_W] & key_state
                        || key_states[SDL_SCANCODE_E] & key_state
                        || key_states[SDL_SCANCODE_Q] & key_state;
            button_down = button_down
                          || key_states[SDL_SCANCODE_X] & key_state
                          || key_states[SDL_SCANCODE_C] & key_state
                          || key_states[SDL_SCANCODE_Z] & key_state
                          || key_states[SDL_SCANCODE_S] & key_state;
        }
    }

    button_left = key_states[SDL_SCANCODE_LEFT] & key_state
                  || key_states[SDL_SCANCODE_HOME] & key_state
                  || key_states[SDL_SCANCODE_END] & key_state
                  || key_states[SDL_SCANCODE_KP_1] & key_state
                  || key_states[SDL_SCANCODE_KP_4] & key_state
                  || key_states[SDL_SCANCODE_KP_7] & key_state;
    button_right = key_states[SDL_SCANCODE_RIGHT] & key_state
                   || key_states[SDL_SCANCODE_PAGEUP] & key_state
                   || key_states[SDL_SCANCODE_PAGEDOWN] & key_state
                   || key_states[SDL_SCANCODE_KP_3] & key_state
                   || key_states[SDL_SCANCODE_KP_6] & key_state
                   || key_states[SDL_SCANCODE_KP_9] & key_state;
    // custom logic - playing alpha keys in non-cheat mode
    if (!cheats_enabled) {
        button_left = button_left
                      || key_states[SDL_SCANCODE_Q] & key_state
                      || key_states[SDL_SCANCODE_A] & key_state
                      || key_states[SDL_SCANCODE_Z] & key_state;
        button_right = button_right
                       || key_states[SDL_SCANCODE_E] & key_state
                       || key_states[SDL_SCANCODE_D] & key_state
                       || key_states[SDL_SCANCODE_C] & key_state;
    }

    if (button_up) {
        control_y = control_m1_held_up;
    } else if (button_down) {
        control_y = control_1_held_down;
    }
    if (button_left) {
        control_x = control_m1_held_left;
    } else if (button_right) {
        control_x = control_1_held_right;
    }

    control_shift = -(key_states[SDL_SCANCODE_LSHIFT] & key_state ||
                      key_states[SDL_SCANCODE_RSHIFT] & key_state);

    // custom logic - control_alt is used for manual sword control
    if ((key_states[SDL_SCANCODE_LALT] & key_state ||
            key_states[SDL_SCANCODE_RALT] & key_state) &&
            control_shift >= control_0_released) {
        control_alt = control_m1_held;
    } else if (key_states[SDL_SCANCODE_INSERT] & key_state ||
               key_states[SDL_SCANCODE_KP_0] & key_state) {
        control_alt = control_m2_set;
    }

    if (cheats_enabled && debug_cheats_enabled) {
        if (key_states[SDL_SCANCODE_LSHIFT] & key_state ||
                key_states[SDL_SCANCODE_RSHIFT] & key_state) {
            if (Guard.room == drawn_room && Guard.direction != dir_56_none) {
                if (key_states[SDL_SCANCODE_RIGHTBRACKET] & key_state) ++Guard.x;
                else if (key_states[SDL_SCANCODE_LEFTBRACKET] & key_state) --Guard.x;
            }
        } else {
            if (key_states[SDL_SCANCODE_RIGHTBRACKET] & key_state) {
                ++Char.x;
                Char.fall_x = Char.direction == dir_FF_left ? -1 : 1;
            } else if (key_states[SDL_SCANCODE_LEFTBRACKET] & key_state) {
                --Char.x;
                Char.fall_x = Char.direction == dir_0_right ? -1 : 1;
            }
        }
    }
}

// seg000:156D
void copy_screen_rect(const rect_type *source_rect_ptr) {
    const rect_type* target_rect_ptr;
    rect_type target_rect;
    if (upside_down_timer > 0) {
        target_rect_ptr = &target_rect;
        /**target_rect_ptr*/target_rect = *source_rect_ptr;
        /*target_rect_ptr->*/target_rect.top = 192 - source_rect_ptr->bottom;
        /*target_rect_ptr->*/target_rect.bottom = 192 - source_rect_ptr->top;
    } else {
        target_rect_ptr = source_rect_ptr;
    }
    method_1_blit_rect(onscreen_surface_, offscreen_surface, target_rect_ptr, target_rect_ptr, 0);
    update_lighting(target_rect_ptr);
}

// seg000:15E9
// custom logic - control upside down mode with timer
void toggle_upside(void) {
    need_redraw_because_flipped = 1;
    stop_all_sounds();
    play_sound(is_positional_audio ? sound_69_upside_down : sound_45_jump_through_mirror);
}

// seg000:15F8
void feather_fall(void) {
    alternate_flash = false;
    flash_color = color_2_green;
    flash_time = 2;
    feather_fall_timer = FEATHER_FALL_IN_SEC * get_ticks_per_sec(timer_1);
    music_suspended_timer = feather_fall_timer;
    stop_sounds();
    stop_ambient_music();
    play_sound(sound_39_low_weight); // low weight
}

// seg000:172C
void gen_palace_wall_colors(void) {
    dword old_randseed;
    word prev_color;
    short row;
    short subrow;
    word color_base;
    short column;
    word color;

    old_randseed = random_seed;
    random_seed = drawn_room;

    // custom logic - level 6 has specific room colors
    if (current_level == 6) {
        switch (random_seed) {
            // rooms 3 and 7 should change colors as kid falls
            case 3:
            case 7:
                random_seed += falling_exit_checkpoint;
                break;
            // rooms 4 and 13 should look like room 16
            case 4:
            case 13:
                random_seed = 16;
                break;
            // rooms 14 and 18 should look like room 5
            case 14:
            case 18:
                random_seed = 5;
                break;
        }
    }

    prandom(1); // discard
    for (row = 0; row < 3; row++) {
        for (subrow = 0; subrow < 4; subrow++) {
            if (subrow % 2) {
                color_base = 0x61; // 0x61..0x64 in subrow 1 and 3
            } else {
                color_base = 0x66; // 0x66..0x69 in subrow 0 and 2
            }
            prev_color = -1;
            for (column = 0; column <= 10; ++column) {
                do {
                    color = color_base + prandom(3);
                } while (color == prev_color);
                palace_wall_colors[44 * row + 11 * subrow + column] = color;
                //palace_wall_colors[row][subrow][column] = color;
                prev_color = color;
            }
        }
    }
    random_seed = old_randseed;
}

// data:042E
const rect_type rect_titles = {106, 24, 195, 296};

// seg000:17E6
void show_title(void) {
    dont_reset_time = 0;
    if (offscreen_surface) free_surface(offscreen_surface); // missing in original
    offscreen_surface = make_offscreen_buffer(&screen_rect);
    load_title_images(1);
    current_target_surface = offscreen_surface;
    idle(); // modified
    do_paused();

    draw_palace_image();
    fade_in_2(0x1000);
    method_1_blit_rect(onscreen_surface_, offscreen_surface, &screen_rect, &screen_rect, blitters_0_no_transp);
    current_sound_is_positional = false;
    current_sound = sound_54_intro_music; // added
    play_sound_from_buffer(sound_pointers[sound_54_intro_music]); // main theme
    start_timer(timer_0, 0x82);
    draw_full_image(TITLE_PRESENTS);
    draw_full_image(TITLE_TUTORIAL);
    do_wait(timer_0);

    start_timer(timer_0, 0xCD);
    method_1_blit_rect(onscreen_surface_, offscreen_surface, &rect_titles, &rect_titles, blitters_0_no_transp);
    draw_palace_image();
    draw_full_image(TITLE_TUTORIAL);
    do_wait(timer_0);

    start_timer(timer_0, 0x41);
    method_1_blit_rect(onscreen_surface_, offscreen_surface, &rect_titles, &rect_titles, blitters_0_no_transp);
    draw_palace_image();
    draw_full_image(TITLE_GAME);
    draw_full_image(TITLE_TUTORIAL);
    do_wait(timer_0);

    start_timer(timer_0, 0x10E);
    method_1_blit_rect(onscreen_surface_, offscreen_surface, &rect_titles, &rect_titles, blitters_0_no_transp);
    draw_palace_image();
    draw_full_image(TITLE_TUTORIAL);
    do_wait(timer_0);

    start_timer(timer_0, 0xEB);
    method_1_blit_rect(onscreen_surface_, offscreen_surface, &rect_titles, &rect_titles, blitters_0_no_transp);
    draw_palace_image();
    draw_full_image(TITLE_POP);
    draw_full_image(TITLE_HACKED_BY);
    draw_full_image(TITLE_TUTORIAL);
    do_wait(timer_0);

    method_1_blit_rect(onscreen_surface_, offscreen_surface, &screen_rect, &screen_rect, blitters_0_no_transp);
    draw_full_image(STORY_FRAME);
    draw_full_image(STORY_ABSENCE);
    current_target_surface = onscreen_surface_;
    while (check_sound_playing()) {
        idle();
        do_paused();
        delay_ticks(1);
    }
    play_sound_from_buffer(sound_pointers[sound_55_story_1_absence]); // story 1: In the absence
    transition_ltr();
    pop_wait(timer_0, is_positional_audio ? 0x34A : 0x258);
    fade_out_2(0x800);
    release_title_images();

    load_intro(0, &pv_scene);

    load_title_images(1);
    current_target_surface = offscreen_surface;
    draw_full_image(STORY_FRAME);
    draw_full_image(STORY_MARRY);
    fade_in_2(0x800);
    draw_palace_image();
    draw_full_image(TITLE_POP);
    draw_full_image(TITLE_HACKED_BY);
    while (check_sound_playing()) {
        idle();
        do_paused();
        delay_ticks(1);
    }
    transition_ltr();
    pop_wait(timer_0, 0x78);
    draw_full_image(STORY_FRAME);
    draw_full_image(STORY_CREDITS);
    transition_ltr();
    pop_wait(timer_0, 0x168);
    if (hof_count) {
        draw_full_image(STORY_FRAME);
        draw_full_image(HOF_POP);
        show_hof();
        transition_ltr();
        pop_wait(timer_0, 0xF0);
    }
    current_target_surface = onscreen_surface_;
    while (check_sound_playing()) {
        idle();
        do_paused();
        delay_ticks(1);
    }
    fade_out_2(0x1800);
    free_surface(offscreen_surface);
    offscreen_surface = NULL; // added
    release_title_images();
    play_demo_level = 0;
    start_level = 1;
    init_game(start_level);
}

Uint64 last_transition_counter;

// seg000:1BB3
void transition_ltr(void) {
    short position;
    rect_type rect;
    rect.top = 0;
    rect.bottom = 200;
    rect.left = 0;
    rect.right = 2;
    // Estimated transition fps based on the speed of the transition on an Apple IIe.
    // See: https://www.youtube.com/watch?v=7m7j2VuWhQ0
    int transition_fps = 120;
    Uint64 counters_per_frame = perf_frequency / transition_fps;
    last_transition_counter = SDL_GetPerformanceCounter();
    int overshoot = 0;
    for (position = 0; position < 320; position += 2) {
        method_1_blit_rect(onscreen_surface_, offscreen_surface, &rect, &rect, 0);
        rect.left += 2;
        rect.right += 2;
        if (overshoot > 0 && overshoot < 10) {
            --overshoot;
            continue; // On slow systems (e.g. Raspberry Pi), allow the animation to catch up, before refreshing the screen.
        }
        idle(); // modified
        do_paused();
        // Add an appropriate delay until the next frame, so that the animation isn't instantaneous on fast CPUs.
        for (;;) {
            Uint64 current_counter = SDL_GetPerformanceCounter();
            int frametimes_elapsed = (int)((current_counter / counters_per_frame) - (last_transition_counter / counters_per_frame));
            if (frametimes_elapsed > 0) {
                overshoot = frametimes_elapsed - 1;
                last_transition_counter = current_counter;
                break; // Proceed to the next frame.
            } else {
                SDL_Delay(1);
            }
        }

    }
}

// seg000:1C0F
void release_title_images(void) {
    if (chtab_title50) {
        free_chtab(chtab_title50);
        chtab_title50 = NULL;
    }
    if (chtab_title40) {
        free_chtab(chtab_title40);
        chtab_title40 = NULL;
    }
}

// seg000:1C3A
void draw_full_image(enum full_image_id id) {
    image_type* decoded_image;
    int xpos, ypos, blit;
    if (id >= MAX_FULL_IMAGES) return;
    if (NULL == *full_image[id].chtab) return;
    decoded_image = (*full_image[id].chtab)->images[full_image[id].id];
    blit = full_image[id].blitter;
    xpos = full_image[id].xpos;
    ypos = full_image[id].ypos;
    if (blit != blitters_0_no_transp && blit != blitters_10h_transp) {
        method_3_blit_mono(decoded_image, xpos, ypos, blitters_0_no_transp, blit);
    } else {
        method_6_blit_img_to_scr(decoded_image, xpos, ypos, blit);
    }
}

// seg000:1D2C
void load_kid_sprite(void) {
    load_chtab_from_file(id_chtab_2_kid, 400, "KID.DAT", 1 << 7);
}

// custom logic - turns shadow into flame
void load_flame_sprite(void) {
    set_chtab_palette(chtab_addrs[id_chtab_2_kid], flame_palette, 0x10);
}

// seg000:1F02
void clear_screen_and_sounds(void) {
    stop_all_sounds();
    current_target_surface = onscreen_surface_;
    is_cutscene = 0;
    is_ending_sequence = false; // added
    free_all_chtabs_from(2);
    current_level = -1;
}

void load_all_sounds(void) {
    dat_type* digi1_dat = NULL;
    dat_type* midi_dat = NULL;
    short current;
    digi1_dat = open_dat("DIGISND.DAT");
    midi_dat = open_dat("MIDISND.DAT");
    // custom logic - use OGG files for positional audio
    if (settings->positional_audio) {
        load_enhanced_sounds();
        load_sound_names();
    }
    for (current = 0; current <= MAX_FORE_SOUND_ID; ++current) {
        if (sound_pointers[current] != NULL) {
            continue;
        }
        sound_pointers[current] = load_sound(current);
    }
    if (settings->positional_audio) {
        unload_enhanced_sounds();
    }
    if (midi_dat) close_dat(midi_dat);
    if (digi1_dat) close_dat(digi1_dat);
}

// seg000:22BB
void free_optsnd_chtab(void) {
    free_all_chtabs_from(id_chtab_3_princessinstory);
}

// seg000:22C8
void load_title_images(int bgcolor) {
    dat_type* dathandle;
    dathandle = open_dat("TITLE.DAT");
    chtab_title40 = load_sprites_from_file(40, 1 << 11);
    chtab_title50 = load_sprites_from_file(50, 1 << 12);
    close_dat(dathandle);
    // background of text frame
    SDL_Color color;
    if (bgcolor) {
        // RGB(4,0,18h) = #100060 = dark blue
        set_pal((find_first_pal_row(1 << 11) << 4) + 14, 0x04, 0x00, 0x18);
        color.r = 0x10;
        color.g = 0x00;
        color.b = 0x60;
        color.a = 0xFF;
    } else {
        // RGB(20h,0,0) = #800000 = dark red
        set_pal((find_first_pal_row(1 << 11) << 4) + 14, 0x20, 0x00, 0x00);
        color.r = 0x80;
        color.g = 0x00;
        color.b = 0x00;
        color.a = 0xFF;
    }
    if (NULL != chtab_title40) {
        SDL_SetPaletteColors(chtab_title40->images[0]->format->palette, &color, 14, 1);
    }
}

// custom logic - check for prison level, set timer and play out-of-time music
void check_prison_level(void) {
    if (current_level != settings->prison_level) {
        return;
    }
    is_show_time = 0;
    if (is_sound_on) {
        play_sound(sound_36_out_of_time);
        level_restart_timer = 168;
    } else {
        level_restart_timer = 90;
    }
    display_text_bottom("YOU HAVE BEEN CAPTURED");
    text_time_total = 48;
    text_time_remaining = 48;
}

// seg000:2489
void show_loading(void) {
    struct stat st = {0};
    char custom_quick_path[POP_MAX_PATH];
    byte wait_delay = 72;
    int i = 0;
    char help_str[100];
    char* moves_str = "This game uses special moves absent from the original game.\nStart the tutorial or read the \"DOCS/README.TXT\" file.";
    char* demo_str = "Press F1 to start the tutorial.";
    int key;
    strcpy(help_str, demo_str);
    // custom logic - use PoP2 yellowish text
    show_text_with_color(&screen_rect, text_align_center, text_align_center, "Loading. . .", color_20_yellowish_text);
    // custom logic - increase the delay and allow F9 to load the last saved game
    if (start_level == -1 && stat(locate_file(get_quick_path(custom_quick_path, sizeof(custom_quick_path))), &st) != -1) {
        strcat(help_str, "\nPress F9 to load the last saved game.");
    }
    if (start_level == -1) {
        textstate.ptr_font = &hc_small_font;
        show_text_with_color(&screen_rect, text_align_center, text_align_top, moves_str, color_15_brightwhite);
        textstate.ptr_font = &hc_font;
        show_text_with_color(&screen_rect, text_align_center, text_align_bottom, help_str, color_15_brightwhite);
    } else {
        wait_delay = 3;
    }
    update_screen();
    do {
        key = key_test_quit();
        if (key == SDL_SCANCODE_F1) {
            play_demo_level = 0;
            start_level = 0;
            break;
        } else if (key == SDL_SCANCODE_F9) {
            need_quick_load = 2; // 2 = from title screen
            start_level = 1;
            break;
        } else if (key == SDL_SCANCODE_SPACE) {
            break;
        }
        if (key_states[SDL_SCANCODE_LSHIFT] & keystate_held || key_states[SDL_SCANCODE_RSHIFT] & keystate_held) {
            key_states[SDL_SCANCODE_LSHIFT] = key_states[SDL_SCANCODE_RSHIFT] = control_0_released;
            break;
        }
        if (key_states[SDL_SCANCODE_RETURN] & keystate_held || key_states[SDL_SCANCODE_KP_ENTER] & keystate_held) {
            key_states[SDL_SCANCODE_RETURN] = key_states[SDL_SCANCODE_KP_ENTER] = control_0_released;
            break;
        }
        if (joy_button_states[joyinput_x] & keystate_held) {
            joy_button_states[joyinput_x] = control_0_released;
            break;
        }
        if (joy_button_states[joyinput_left_shoulder] & keystate_held && joy_button_states[joyinput_right_shoulder] & keystate_held) {
            joy_button_states[joyinput_left_shoulder] = control_0_released;
            joy_button_states[joyinput_right_shoulder] = control_0_released;
            need_quick_load = 2;
            start_level = 1;
            break;
        }
        idle();
        delay_ticks(6);
    } while (i++ < wait_delay);
}
