/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"
#include <string.h>

// data:3D1A
sbyte distance_mirror;

// seg003:0000
void init_game(int level) {
    if (offscreen_surface) {
        free_surface(offscreen_surface); // missing in original
        offscreen_surface = NULL;
    }
    offscreen_surface = make_offscreen_buffer(&rect_top);
    // custom logic - sets kid palette offset in cheats mode
    kid_cheat_pal_offset = cheats_enabled ? -1 : 0;
    load_kid_sprite();
    text_time_remaining = 0;
    text_time_total = 0;
    is_show_time = 0;
    seamless = 0;
    enable_lighting = 0;
    // custom logic - reset total saves
    total_saves = 0;
    // custom logic - falling checkpoint on level 6
    if (level != 6) {
        falling_exit_checkpoint = 0;
    }
    // custom logic - kid's default direction at the end of level 6 is to the left
    if (level != 7 || check_param("7")) {
        falling_exit_final_direction = dir_FF_left;
    }
    // custom logic - alternating flash
    alternate_flash = false;
    // custom logic - quick loading flag
    quick_loading = 0;
    // custom logic - reset timers
    feather_fall_timer = 0;
    upside_down_timer = 0;
    // custom logic - disable/enable ambient music
    track = NULL;
    track_number = 0;
    show_music_menu = enable_positional_audio &&
                      level > 0 &&
                      level < 15 &&
                      check_param("nomusic") == NULL &&
                      is_music_found() ? 1 : 0;
    music_mode = enable_positional_audio ? 0 : -1;
    music_paused = false;
    // custom logic - init ogg sounds
    ogg_fore_sound_playing = false;
    resurrect_time = 0;
    if (!dont_reset_time) {
        rem_min = settings->start_minutes_left;   // 60
        rem_tick = settings->start_ticks_left;    // 719
        hitp_beg_lev = settings->start_hitp;      // 3
    }
    need_level1_music = (level == 1);
    play_level(level);
}

// custom logic - initialize loose tile counter
void init_loose_tile_counter(void) {
    byte room, tile;
    for (room = 0; room < level.used_rooms; ++room) {
        for (tile = 0; tile < MAX_TILES_PER_ROOM; ++tile) {
            loose_tile_counter[room][tile] = 0;
        }
    }
}

// seg003:005C
void play_level(int level_number) {
    cutscene_ptr_type cutscene_func;
    for (;;) {
        end_of_level_music = 0;
        // custom logic - make certain levels dark
        if (level_number > 0 &&
                settings->tbl_seamless_exit[level_number - 1] == -1 && // previous level does not have a seamless exit
                checkpoint == 0) { // we are not returning from level 12B to level 12
            enable_lighting = settings->dark_levels[level_number] != 0 ? 1 : 0;
        }
        if (level_number != current_level && // not a level restart
                // custom logic - do not play cut scene when entering level 12 from level 12B
                checkpoint == 0) {
            // custom - reset music mode if music file was not found for the previous level
            if (music_mode == -2) {
                music_mode = 0;
            }
            // custom logic - reset saves per level
            if (level_number != 15 && current_level != 15) {
                saves_per_level = 0;
            }
            if (level_number < 0 || level_number > 15) {
                printf("Tried to load cutscene for level %d, not in 0..15\n", level_number);
                quit(1);
            }
            cutscene_func = tbl_cutscenes[settings->tbl_cutscenes_by_index[level_number]];
            // custom logic - always show the princess waiting 2 cutscene when the time is less than 5 min
            if (level_number != 15 && cutscene_func != NULL && cutscene_func != cutscene_8 && rem_min < 5) {
                cutscene_func = cutscene_short_time;
            }
            if (current_level != 15 && cutscene_func != NULL) {
                flash_time = 0;
                load_intro(level_number > 2, cutscene_func);
            }
            // custom logic - stores whether prince had sword on the previous level
            had_sword = have_sword;
        }
        // custom logic - kid cannot be shadow at the beginning of the level
        is_kid_shadow = 0;
        load_lev_spr(level_number);
        load_level();
        // custom logic - shadow turned into flame
        is_shadow_flame = false;
        // custom logic - reset shadow/flame palette on level 13
        reset_shadow_palette();
        // custom logic - start music after a brief delay
        music_suspended = 0;
        if (level_number == 1) { // starting the game
            stop_ambient_music();
            music_suspended_timer = 60;
        } else if (level_number == 6 || level_number == 7) { // falling exit
            stop_ambient_music();
            music_suspended_timer = 120;
        } else if (seamless) {
            music_suspended_timer = 0;
        } else { // regular exit
            stop_ambient_music();
            music_suspended_timer = 12;
        }
        // setting the music track - it will not start playing due to the suspended timer
        if (!seamless) {
            start_ambient_music(true);
        }
        pos_guards();
        clear_coll_rooms();
        clear_saved_ctrl();
        // custom logic - initialize loose tile counter
        init_loose_tile_counter();
        drawn_room = 0;
        mobs_count = 0;
        trobs_count = 0;
        next_sound = -1;
        holding_sword = 0;
        grab_timer = 0;
        can_guard_see_kid = 0;
        united_with_shadow = 0;
        flash_time = 0;
        leveldoor_open = 0;
        // custom logic - shadow can only jump out of the mirror once per game
        if (level_number <= settings->mirror_level) {
            jumped_through_mirror = 0;
        }
        // custom logic - save remaining time in the level
        level_rem_min = rem_min;
        level_rem_tick = rem_tick;
        // custom logic - tracking whether the kid met Jaffar
        jaffar_presented = 0;
        // custom logic - Jaffar's timer is a separate variable
        jaffar_delay_timer = 0;
        // custom logic - tracking whether the kid met Jaffar's vision
        jaffar_vision_presented = false;
        // custom logic - tracking antidote on level 5
        took_antidote = 0;
        // custom logic - tracking shadow crashing the floor
        level_7_tile_crashed = false;
        // custom logic - timer that kills kid on level 12B when there is nowhere to run
        level_13_nowhere_to_run_timer = -1;
        // custom logic - tracks whether shadow appeared on this level
        shadow_initialized = 0;
        // custom logic - an ability to restart the level on a timer
        level_restart_timer = 0;
        // custom logic - set blink timer to 0
        blink_timer = 0;
        // custom logic - delayed sound
        delayed_sound_timer = 0;
        delayed_sound = (struct delayed_sound_type) {
            -1, 0, 0, 0, 0, 0, 0
            };
        // custom logic - do not reset number of potions shadow drank between levels 7 and 10
        if (level_number >= 7 && level_number <= 10) {
            shadow_potion_counter = initial_shadow_potion_counter;
            // custom logic - do not reset shadow potion counter on level 6B
        } else if (level_number != 6 || falling_exit_checkpoint == 0) {
            initial_shadow_potion_counter = 0;
            shadow_potion_counter = 0;
        }
        // custom logic - reset timer for the last skeleton on level 3
        skeleton_timer = 0;
        // custom logic - dead kid timer
        dead_kid_timer = 0;
        // custom logic - dead kid tile, allows restoring buttons and spikes when uniting with shadow or resurrecting
        dead_kid_tile = (struct stuck_button_type) {
            0, 0, 0, 0
        };
        // custom logic - inactive guard timer
        inactive_guard_timer = 0;
        // custom logic - inactive guard notice
        is_inactive_guard_notice = 0;
        // custom logic - dynamic guard timer
        dynamic_guard_timer = 0;
        // custom logic - dynamic guard fight indicator
        is_dynamic_guard_fight = 0;
        // custom logic - should shadow separation text be displayed
        is_shadow_text_shown = 0;
        // custom logic - control separating kid from shadow
        separate_from_shadow_counter = 0;
        separate_from_shadow_counter_prev = 0;
        separate_from_shadow_reset_timer = 0;
        // custom logic - guard is following kid into the room below
        guard_dropping_down = 0;
        // custom logic - can inactive guard see kid?
        can_inactive_guard_see_kid = 0;
        // custom logic - an indicator that a guard knocked down a tile
        guard_knocked_tile = 0;
        // custom logic - tracks whether a skeleton has appeared on the level
        skeleton_appeared = 1 - enable_positional_audio;
        // custom logic - times skeletons that shake
        skeleton_shake_timer = 0;
        // custom logic - hard jump fall
        hard_jump_fall = 0;
        // custom logic - level 6 initial sequence timer
        level_6_sequence_timer = 0;
        // custom logic - tracking last drank potion
        potion_room = 0;
        potion_pos = -1;
        potion_timer = 0;
        potion_size = 0;
        // custom logic - blink when flame is hit with sword
        lighting_timer = 0;
        // custom logic - super jump fall indicators
        super_jump_timer = 0;
        super_jump_fall = 0;
        // custom logic - inactive guard fall timer
        inactive_guard_fall_timer = 0;
        // custom logic - inactive guard skill reset
        inactive_guard_skill = 0xFF;
        // custom logic - moving wall timer reset
        moving_wall_timer = 0;
        // custom logic - falling bridge timer reset
        falling_bridge_timer = 0;
        // custom logic - mouse potion animation timer reset
        mouse_potion_timer = 0;
        // custom logic - reset last guard color
        last_guard_color = 0;
        // custom logic - skeleton death timer
        skeleton_death_timer = 0;
        // custom logic - guard punch count
        guard_punch_count = 0;
        demo_index = 0;
        demo_time = 0;
        guardhp_curr = 0;
        hitp_delta = 0;
        Guard.charid = charid_2_guard;
        Guard.direction = dir_56_none;
        do_startpos();
        checkpoint = 0;
        // custom logic - no sword on levels tutorial, 1, 6 and 15
        if (level_number <= 1 || level_number == 6 || level_number == settings->prison_level) {
            have_sword = 0;
            // it is impossible to get to level 7 without a sword
        } else if (level_number >= 7) {
            have_sword = 1;
        }
        find_start_level_door();
        // busy waiting?
        while (check_sound_playing() && !do_paused()) idle();
        stop_sounds();
        draw_level_first();
        // custom logic - check for prison level
        check_prison_level();
        // custom logic - reset Shift state
        const Uint8* state = SDL_GetKeyboardState(NULL);
        if (state[SDL_SCANCODE_LSHIFT]) key_states[SDL_SCANCODE_LSHIFT] |= keystate_held | keystate_held_new;
        if (state[SDL_SCANCODE_RSHIFT]) key_states[SDL_SCANCODE_RSHIFT] |= keystate_held | keystate_held_new;
        // custom logic - demo level instructions
        if (level_number == 0 && play_demo_level == 1) {
            ++play_demo_level;
            display_press_button();
            text_time_total = text_time_remaining = 10000;
            const char* game_instructions =
                "Now try to pass the demo level yourself!\n\n"
                "Press ESC for the game menu options to start a real level or exit the game.";
            show_dialog(game_instructions, instructions_dialog);
            erase_bottom_text(1);
        }
        level_number = play_level_2();
        free_peels();
    }
}

// custom logic - change spikes to debris on level 12 when entered from level 12
void load_level_12_from_13(void) {
    for (int tilecol = 6; tilecol <= 9; ++tilecol) {
        if (get_tile(11, tilecol, 2) == tiles_2_spike) {
            curr_room_tiles[curr_tilepos] = tiles_14_debris;
            curr_room_modif[curr_tilepos] = 0;
        }
    }
    for (int tilecol = 0; tilecol <= 7; ++tilecol) {
        if (get_tile(21, tilecol, 2) == tiles_2_spike) {
            curr_room_tiles[curr_tilepos] = tiles_14_debris;
            curr_room_modif[curr_tilepos] = 0;
        }
    }
}

// seg003:01A3
void do_startpos(void) {
    word x, prev_x, prev_y, delta_x;
    delta_x = 0;
    // custom logic - start level 3 in a different place if prince
    // had entered the left door on level 2
    if (current_level == 3 && level_2_right_door == 0) {
        level.start_dir = dir_FF_left;
        level.start_room = 4;
        level.start_pos = 25;
    }
    // custom logic - configuring starting position for level 6B
    else if (current_level == 6 && falling_exit_checkpoint >= 4) {
        level.start_dir = dir_0_right;
        level.start_room = 12;
        level.start_pos = 22;
        load_lev6_palette();
        if (hitp_curr > 0 || Char.fall_y == 0) {
            find_all_rooms_from(level.start_room);
            stop_sounds();
            play_sound(sound_1_falling);
        }
    }
    // custom logic - configuring starting position for level 7
    else if (current_level == 7) {
        level.start_dir = falling_exit_final_direction;
        if (level.start_dir == dir_0_right) {
            delta_x = -1;
        }
    }
    next_room = Char.room = level.start_room;
    prev_x = Char.x;
    prev_y = Char.y;
    x = level.start_pos;
    Char.curr_col = x % 10 + (delta_x != 0 ? 1 : 0);
    Char.curr_row = x / 10;
    Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET + delta_x;
    // Start in the opposite direction (and turn into the correct one).
    Char.direction = ~ level.start_dir;
    if (seamless == 0) {
        if (current_level != 0 && current_level != settings->prison_level) {
            x = hitp_beg_lev;
        } else {
            // HP on demo or prison levels
            x = settings->start_hitp;
        }
        hitp_max = hitp_curr = x;
    }
    if (settings->tbl_entry_pose[current_level] == 1) {
        // Special event: press button + falling entry
        get_tile(5, 2, 0);
        trigger_button(0, 0, -1);
        seqtbl_offset_char(seq_7_fall); // fall
    } else if (settings->tbl_entry_pose[current_level] == 2) {
        // Special event: running entry
        seqtbl_offset_char(seq_84_run); // run
        // custom logic - reset the fall on level 6B
    } else if (current_level == 6 && falling_exit_checkpoint >= 4) {
        seqtbl_offset_char(seq_2_stand); // reset the fall
        // custom logic - entering level 12 from level 13
    } else if (current_level == 12 && checkpoint) {
        load_level_12_from_13();
        next_room = Char.room = (checkpoint == 1 ? 2 : 21);
        Char.direction = dir_0_right;
        Char.x = prev_x;
        determine_col();
        Char.y = prev_y;
        Char.curr_row = y_to_row_mod4(Char.y);
        if (checkpoint == 1) {
            seqtbl_offset_char(seq_18_fall_after_standing_jump);
        } else {
            if (Char.action != actions_4_in_freefall) {
                if (Char.alive >= 0) {
                    seqtbl_offset_char(seq_22_crushed);
                } else {
                    seqtbl_offset_char(seq_2_stand);
                }
            } else {
                seqtbl_offset_char(seq_19_fall);
            }
        }
    } else if (current_level == settings->prison_level) { // custom logic - kid is sitting in a cell
        seqtbl_offset_char(seq_125_sit_in_cell);
    } else {
        seqtbl_offset_char(seq_5_turn); // turn
        // custom logic - do not count this turn as a part of separating from shadow
        separate_from_shadow_counter_prev = 1;
    }
    if (current_level == 1) {
        // custom logic - add a dropped large empty potion next to the drunk guard
        if (get_tile(16, 4, 0) == tiles_10_potion) {
            curr_room_modif[curr_tilepos] = 0x60;
        }
        // custom logic - the gate in the escaped cell should be partially opened
        if (get_tile(24, 9, 0) == tiles_4_gate) {
            curr_room_modif[curr_tilepos] = 85;
        }
        // custom logic - configuring level 6
    } else if (current_level == 6 && falling_exit_checkpoint == 0) {
        // set starting sequence timer
        if (Char.room == level.start_room &&
                get_tile(level.start_room, settings->level_6_seq_potion_column, settings->level_6_seq_potion_row) == tiles_10_potion) {
            upside_down_timer = 0;
            level_6_sequence_timer = 60;
        }
        // purple weirdo's potion is empty
        if (get_tile(6, 6, 1) == tiles_10_potion) {
            curr_room_modif[curr_tilepos] = 0x40;
        }
        // custom logic - randomize blue potions on level 8, room 19
    } else if (current_level == 8) {
        if (custom_seed != 0) {
            random_seed = custom_seed;
            seed_was_init = 1;
        }
        int door_opener_potion_tile = prandom(4);
        if (get_tile(19, 5, 0) == tiles_10_potion) {
            curr_room_modif[curr_tilepos] = (door_opener_potion_tile == 0 ? 0x30 : 0x28);
        }
        if (get_tile(19, 3, 1) == tiles_10_potion) {
            curr_room_modif[curr_tilepos] = (door_opener_potion_tile == 1 ? 0x30 : 0x28);
        }
        if (get_tile(19, 5, 1) == tiles_10_potion) {
            curr_room_modif[curr_tilepos] = (door_opener_potion_tile == 2 ? 0x30 : 0x28);
        }
        if (get_tile(19, 7, 1) == tiles_10_potion) {
            curr_room_modif[curr_tilepos] = (door_opener_potion_tile == 3 ? 0x30 : 0x28);
        }
        if (get_tile(19, 8, 2) == tiles_10_potion) {
            curr_room_modif[curr_tilepos] = (door_opener_potion_tile == 4 ? 0x30 : 0x28);
        }
        // custom logic - the level 9 potion gate opener tile is stuck
    } else if (current_level == 9) {
        get_tile(23, 8, 2);
        curr_room_tiles[curr_tilepos] = tiles_5_stuck;
        // custom logic - the level 10 door opener tile is stuck
    } else if (current_level == 10) {
        get_tile(5, 2, 2);
        curr_room_tiles[curr_tilepos] = tiles_5_stuck;
        // custom logic - the level door on level 12B should appear to be open
    } else if (current_level == settings->inverse_level) {
        if (get_tile(12, 3, 1) == tiles_16_level_door_left) {
            curr_room_modif[curr_tilepos] = 42;
        }
    }

    set_start_pos();
}

// seg003:028A
void set_start_pos(void) {
    word prev_falling_exit_checkpoint;
    prev_falling_exit_checkpoint = falling_exit_checkpoint;
    // custom logic - if kid falls into level 6B not alive, kill him and restart level 6
    if (current_level == 6 &&
            falling_exit_checkpoint >= 4 &&
            Char.alive == 0 &&
            !check_param("6b")
       ) {
        falling_exit_checkpoint = 0;
        hitp_curr = 0;
        shadow_potion_counter = 0;
        Char.fall_y = 0;
        Char.fall_x = 0;
    } else {
        Char.alive = -1;
    }
    Char.charid = charid_0_kid;
    Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
    knock = 0;
    // custom logic - entering level 12 from level 13
    if (current_level != 12 || !checkpoint) {
        is_screaming = 0;
        feather_fall_timer = 0;
        Char.fall_y = 0;
        Char.fall_x = 0;
    }
    offguard = 0;
    Char.sword = sword_0_sheathed;
    droppedout = 0;
    play_seq();
    // custom logic - do not show room 12 when the prince is falling into level 6B
    if (current_level == 6 &&
            Char.room == 12 &&
            prev_falling_exit_checkpoint >= 4) {
        goto_other_room(3);
    } else if (current_level == /*7*/ settings->falling_entry_level && Char.room == /*17*/ settings->falling_entry_room) {
        // Special event: level 7 falling entry
        // level 7, room 17: show room below
        goto_other_room(3);
    }
    savekid();
}

// seg003:02E6
void find_start_level_door(void) {
    short tilepos;
    get_room_address(Kid.room);
    for (tilepos = 0; tilepos < MAX_TILES_PER_ROOM; ++tilepos) {
        if ((curr_room_tiles[tilepos] & 0x1F) == tiles_16_level_door_left) {
            start_level_door(Kid.room, tilepos);
        }
    }
}

// seg003:0326
void draw_level_first(void) {
    next_room = Kid.room;
    check_the_end();
    if (settings->tbl_level_type[current_level]) {
        gen_palace_wall_colors();
    }
    draw_rect(&screen_rect, color_0_black);
    show_level();
    redraw_screen(0);
    draw_kid_hp(hitp_curr, hitp_max);
    check_quick_op();
    // Busy waiting!
    start_timer(timer_1, 5);
    do_simple_wait(1);
}

// seg003:037B
void redraw_screen(int drawing_different_room) {
    //remove_flash();
    if (drawing_different_room) {
        draw_rect(&rect_top, color_0_black);
    }

    different_room = 0;

    if (is_blind_mode) {
        draw_rect(&rect_top, color_0_black);
    } else {
        // custom logic - reset last guard color
        last_guard_color = 0;
        need_drects = 0;
        redraw_room();
        redraw_lighting();
        if (is_keyboard_mode) {
            clear_kbd_buf();
        }
        is_blind_mode = 1;
        draw_tables();
        if (is_keyboard_mode) {
            clear_kbd_buf();
        }
        is_blind_mode = 0;
        memset(table_counts, 0, sizeof(table_counts));
        draw_moving();
        draw_tables();
        if (is_keyboard_mode) {
            clear_kbd_buf();
        }
        need_drects = 1;
        if (upside_down_timer > 0) {
            flip_screen(offscreen_surface);
        }
        copy_screen_rect(&rect_top);
        if (upside_down_timer > 0) {
            flip_screen(offscreen_surface);
        }
        if (is_keyboard_mode) {
            clear_kbd_buf();
        }
    }
    exit_room_timer = 2;
}

// custom logic - level 6 feels faster (prince is drunk) in rooms
// before the mirror and slower in rooms after the mirror.
void set_timer_length_custom(void) {
    // custom logic - set default speed game when walking up stairs
    if (blink_timer == 16) {
        set_timer_length(timer_1, settings->default_game_speed);
        return;
    }
    // custom logic - kid is faster on level 6A except the mirror room
    if (current_level == 6 &&
            !is_level_6_sequence() &&
            Kid.room != 1 &&
            Kid.room != 3 &&
            Kid.room != 7 &&
            hitp_curr > 1) {
        if (falling_exit_checkpoint < 4) {
            // kid is fast on level 6
            // speed when fighting (smaller is faster)
            if (Kid.sword == sword_2_drawn) {
                // speed when fighting (smaller is faster)
                set_timer_length(timer_1, settings->fast_game_speed + 1);
            } else {
                // speed when not fighting (smaller is faster)
                set_timer_length(timer_1, settings->fast_game_speed);
            }
        } else {
            if (leveldoor_open != 0) {
                goto loc_default_game_speed;
            }
            // kid is slower on level 6B
            if (Kid.sword == sword_2_drawn) {
                // speed when fighting (smaller is faster)
                set_timer_length(timer_1, settings->slow_game_speed + 1);
            } else {
                // speed when not fighting (smaller is faster)
                set_timer_length(timer_1, settings->slow_game_speed);
            }
        }
    } else if (hitp_curr == 1 && !is_kid_shadow) {
        // custom logic - kid is sluggish with a single hit point
        if (Kid.sword == sword_2_drawn) {
            // speed when fighting (smaller is faster)
            set_timer_length(timer_1, settings->one_hp_game_speed + 1);
        } else {
            // speed when not fighting (smaller is faster)
            set_timer_length(timer_1, settings->one_hp_game_speed);
        }
    } else {
loc_default_game_speed:
        if (Kid.sword == sword_2_drawn) {
            // speed when fighting (smaller is faster)
            set_timer_length(timer_1, settings->default_game_speed + 1);
        } else {
            // speed when not fighting (smaller is faster)
            set_timer_length(timer_1, settings->default_game_speed);
        }
    }
}

// seg003:04F8
// Returns a level number:
// - The current level if it was restarted.
// - The next level if the level was completed.
int play_level_2(void) {
    reset_timer(timer_1);
    while (1) { // main loop
        check_quick_op();
        // custom logic - level 6 feels faster (prince is drunk) in rooms
        // before the mirror and slower in rooms after the mirror.
        set_timer_length_custom();
        guardhp_delta = 0;
        hitp_delta = 0;
        timers();
        play_frame();
        if (is_restart_level) {
            is_restart_level = 0;
            // custom logic - restart the previous level if the prison level was restarted
            return current_level != settings->prison_level ? current_level : settings->prison_level - 1;
        } else {
            if (next_level == current_level || check_sound_playing() || ogg_fore_sound_playing) {
                draw_game_frame();
                // custom logic - reset auto save skip flag
                if (skip_auto_save && next_level == current_level) {
                    skip_auto_save = 0;
                }
                flash_if_hurt();
                remove_flash_if_hurt();
                do_simple_wait(timer_1);
            } else {
                stop_all_sounds();
                // custom logic - do not reset initial level hitpoints on the prison level
                if (next_level != settings->prison_level) {
                    hitp_beg_lev = hitp_max;
                }
                falling_exit_checkpoint = 0;
                // custom logic - setting the initial number of potions shadow drank on the next level
                initial_shadow_potion_counter = shadow_potion_counter;
                // custom logic - kid's default direction at the end of level 6 is to the left
                if (next_level != 7) {
                    falling_exit_final_direction = dir_FF_left;
                }
                // custom logic - auto save level
                // level 7 needs to be saved immediately due to a falling start
                if (next_level == 7) {
                    auto_save_level(next_level);
                } else {
                    need_auto_save = 1;
                }
                return next_level;
            }
        }
    }
}

// seg003:0576
void redraw_at_char(void) {
    short x_top_row;
    short tile_col;
    short tile_row;
    short x_col_left;
    short x_col_right;
    if (Char.sword >= sword_2_drawn ||
            (Char.charid > charid_0_kid && Char.action == actions_3_in_midair)) { // guard in the air can leave artifacts in temple
        // If char is holding sword, it makes redraw-area bigger.
        if (Char.direction >= dir_0_right) {
            if (++char_col_right > 9) char_col_right = 9;
        } else {
            if (--char_col_left < 0) char_col_left = 0;
        }
    }
    if (Char.charid == charid_0_kid) {
        x_top_row = MIN(char_top_row, prev_char_top_row);
        x_col_right = MAX(char_col_right, prev_char_col_right);
        x_col_left = MIN(char_col_left, prev_char_col_left);
    } else {
        x_top_row = char_top_row;
        x_col_right = char_col_right;
        x_col_left = char_col_left;
    }
    // custom logic - fix for gates overlapping kid when climbing/jumping up
    if (Char.charid == charid_0_kid &&
            Char.direction == dir_FF_left &&
            obj_y < y_land[Char.curr_row + 1] &&
            x_col_left > Char.curr_col &&
            (curr_room_tiles[get_tilepos(Char.curr_col, char_bottom_row)] & 0x1F) == tiles_4_gate) {
        x_col_left = Char.curr_col;
        x_col_right = Char.curr_col;
    }
    for (tile_row = x_top_row; tile_row <= char_bottom_row; ++tile_row) {
        for (tile_col = x_col_left; tile_col <= x_col_right; ++tile_col) {
            set_redraw_fore(get_tilepos(tile_col, tile_row), 1);
        }
    }
    if (Char.charid == charid_0_kid) {
        prev_char_top_row = char_top_row;
        prev_char_col_right = char_col_right;
        prev_char_col_left = char_col_left;
    }
}

// custom logic - redraws all tiles except chompers and torches without a floor
// when falling/hanging next to the tile
void set_redraw_at_char2(short tilepos, byte frames) {
    int tile;
    if (curr_room > 0 && tilepos >= 0 && tilepos < MAX_TILES_PER_ROOM) {
        tile = curr_room_tiles[tilepos] & 0x1F;
    } else {
        tile = tiles_20_wall; // out of bounds
    }
    // double rendering a chomper can cause artifacts
    if (tile == tiles_18_chomper) {
        return;
    }
    // kid gets redrawn incorrectly when falling or hanging off the floor tile
    // at column 8 when there is a doortop in the position 9
    if (tile == tiles_12_doortop && Char.charid == charid_0_kid && tilepos % 10 == 9) {
        return;
    }
    // do not redraw empty tiles with torches that only exist on dungeon levels
    if (tile == tiles_26_torch_empty && settings->tbl_level_type[current_level] == DUNGEON_LEVEL) {
        return;
    }
    // on level 13 only redraw big pillar floors when falling left next to other pillars
    if (current_level == settings->win_level &&
            tile == tiles_8_bigpillar_bottom &&
            Char.direction == dir_FF_left &&
            Char.action == actions_3_in_midair) {
        redraw_frames_floor_overlay[tilepos] = frames;
        return;
    }
    set_redraw2(tilepos, frames);
}

// seg003:0645
void redraw_at_char2(void) {
    short char_action;
    short char_frame;
    void (* redraw_func)(short, byte);
    char_action = Char.action;
    char_frame = Char.frame;
    redraw_func = &set_redraw_at_char2; // custom logic - do not redraw chompers
    // frames 78..80: grab
    if (char_frame < frame_78_jumphang || char_frame >= frame_80_jumphang) {
        // frames 135..149: climb up
        if (char_frame >= frame_137_climbing_3 && char_frame < frame_145_climbing_11) {
            redraw_func = &set_redraw_floor_overlay;
        } else {
            // frames 102..106: fall
            if (char_action != actions_2_hang_climb && char_action != actions_3_in_midair &&
                    char_action != actions_4_in_freefall && char_action != actions_6_hang_straight &&
                    (char_action != actions_5_bumped || char_frame < frame_102_start_fall_1 || char_frame > frame_106_fall)) {
                return;
            }
        }
    }
    for (tile_col = char_col_right; tile_col >= char_col_left; --tile_col) {
        if (char_action != actions_2_hang_climb) {
            redraw_func(get_tilepos(tile_col, char_bottom_row), 1);
        }
        if (char_top_row != char_bottom_row) {
            redraw_func(get_tilepos(tile_col, char_top_row), 1);
        }
    }
}

// seg003:0706
void check_knock(void) {
    byte charid;
    sbyte row;
    // custom logic - knock should be checked for kid and guards as well
    if (knock) {
        charid = (knock & 0xF0) >> 4;
        row = (knock & 0x0F) - 2;
        if (Char.charid == charid) {
            do_knock(Char.room, row, Char.curr_col);
            knock = 0;
        }
    }
}

// custom logic - shakes skeletons that can only rise when the level door is open
void do_shake_skel(byte tilepos, byte col, byte row, byte op) {
    if (resurrect_time > 0) {
        return;
    }
    curr_room_tiles[tilepos] = (op == 0 ? tiles_21_skeleton : tiles_1_floor);
    redraw_height = 24;
    set_redraw_full(tilepos, 1);
    set_wipe(tilepos, 1);
    set_redraw_full(tilepos + 1, 1);
    set_wipe(tilepos + 1, 1);
    if (op) {
        Char.room = drawn_room;
        Char.curr_row = row;
        Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
        Char.curr_col = col;
        Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
        Char.direction = Kid.x <= Char.x ? dir_FF_left : dir_0_right;
        seqtbl_offset_char(seq_88_skel_wake_up);
        play_seq();
        guard_skill = 4;
        Char.alive = 0;
        guardhp_max = guardhp_curr = 3;
        Char.fall_x = Char.fall_y = 0;
        Char.sword = sword_0_sheathed;
        Char.charid = charid_4_skeleton;
        saveshad();
    } else {
        loadshad();
        clear_char();
        guardhp_max = guardhp_curr = 0;
        saveshad();
    }
}

// custom logic - inactive guard follows kid when active guard gets killed
void check_kid_droppedout(void) {
    if (droppedout == 1 || is_kid_shadow || Guard.alive >= 0 || Kid.curr_row != Guard.curr_row + 1) {
        return;
    }
    short distance;
    char_type old_char = Char;
    char_type old_opp = Opp;
    loadshad_and_opp();
    distance = char_opp_dist();
    if (distance >= 0) {
        loadkid_and_opp();
        if (Char.curr_row != Opp.curr_row + 1 ||
                (Char.curr_col < Opp.curr_col && Opp.direction != dir_FF_left) ||
                (Char.curr_col > Opp.curr_col && Opp.direction != dir_0_right)) {
            droppedout = 0;
        } else {
            droppedout = 1;
            check_can_guard_follow_kid();
        }
    }
    Char = old_char;
    Opp = old_opp;
}

// seg003:0735
void timers(void) {
    if (united_with_shadow > 0) {
        --united_with_shadow;
        if (united_with_shadow == 0) {
            --united_with_shadow;
        }
    }
    // custom logic - guard notice timer triggers the is_guard_notice event after the delay
    if (guard_notice_timer > 0) {
        --guard_notice_timer;
        if (guard_notice_timer == 0) {
            make_guard_notice(false);
        }
    }
    // custom logic - Jaffar initial delay timer
    if (jaffar_delay_timer > 0) {
        --jaffar_delay_timer;
    }
    if (resurrect_time > 0) {
        --resurrect_time;
        // custom logic - guards cannot hurt kid; hitpoints are not flashing
    } else if (resurrect_time < 0) {
        ++resurrect_time;
    }
    // custom logic - suspended music timer
    if (music_suspended_timer > 0) {
        --music_suspended_timer;
    }
    // custom logic - feather fall is now timer driven
    if (feather_fall_timer > 0 && !is_paused) {
        --feather_fall_timer;
        if (feather_fall_timer == 0) {
            if (music_suspended_timer > 1) {
                music_suspended_timer = 1;
            }
            stop_ambient_music();
            if (check_sound_playing() && current_sound == sound_39_low_weight) {
                stop_sounds();
            }
        }
    }
    // custom logic - level restart timer
    if (level_restart_timer > 0) {
        if (current_level == 10 && level_restart_timer == DEAD_KID_LEVEL_10_TIMER_P12) {
            need_full_redraw = 1;
        }
        --level_restart_timer;
        if (current_level == 10 && Kid.room == 20) {
            if (level_restart_timer == 0) {
                stop_all_sounds();
                Kid.y = y_land[3 + ONSCREEN_ROW_OFFSET];
                Kid.curr_seq = seqtbl_offset_get_seq(seq_22_crushed);
                holding_sword = 1;
            } else if (level_restart_timer % 2 == 0 && hitp_curr > 0) {
                hitp_delta = -1;
                if (!is_positional_audio) {
                    play_sound(sound_13_kid_hurt);
                }
            }
        } else {
            if (level_restart_timer == 0) {
                if (current_level == settings->prison_level) {
                    play_level(settings->prison_level - 1);
                } else {
                    is_restart_level = 1;
                }
            }
        }
    }
    // custom logic - blink timer
    if (blink_timer < 12) {
        ++blink_timer;
        if (blink_timer == 12) {
            blink_timer = 0;
        }
        if (feather_fall_timer && blink_timer == 6) {
            if (alternate_flash) {
                flash_color = color_2_green;
                flash_time = 2;
            }
            alternate_flash = !alternate_flash;
        } else if (hitp_curr == 1 && blink_timer == 6 && !is_kid_shadow) {
            if (alternate_flash) {
                flash_color = color_12_brightred;
                flash_time = 2;
            }
            alternate_flash = !alternate_flash;
        }
    }
    // custom logic - control upside down mode with timer
    if (upside_down_timer > 0 && current_level == next_level) {
        --upside_down_timer;
        if (upside_down_timer == 0) {
            toggle_upside();
        }
    }
    // custom logic - delayed sound timer
    if (delayed_sound_timer > 0) {
        --delayed_sound_timer;
        if (delayed_sound_timer == 0) {
            if (delayed_sound.sound_id == -2) {
                delayed_sound = (struct delayed_sound_type) {
                    -1, 0, 0, 0, 0, 0, 0
                    };
                stop_sounds();
            } else if (current_sound != sound_36_out_of_time && delayed_sound.sound_id >= 0) {
                // victory sound does not interrupt other sounds
                if (delayed_sound.sound_id == sound_37_victory && (check_sound_playing() || ogg_fore_sound_playing)) {
                    if (current_sound != sound_37_victory &&
                            (curr_ogg_sound_id != sound_37_victory && curr_ogg_sound_id != sound_63_temple_victory)) {
                        delayed_sound_timer = 1;
                    }
                    goto skip_delayed_sound_timer;
                }
                // dead kid cannot say whew
                if (delayed_sound.sound_id == sound_58_say_whew && (Kid.alive >= 0 || hitp_curr == 0)) {
                    goto skip_delayed_sound_timer;
                }
                // play sound
                if (is_positional_audio && Kid.alive < 0 && delayed_sound.repeat == 0) {
                    stop_sounds();
                }
                set_sound_pos(delayed_sound.room, delayed_sound.col, delayed_sound.row);
                play_sound(delayed_sound.sound_id);
                // special cases
                if (((Guard.direction == dir_56_none && delayed_sound.play_victory_sound == 1) ||
                        delayed_sound.play_victory_sound == 3)) { // inactive guard died
                    delayed_sound_timer = 6;
                    delayed_sound = (struct delayed_sound_type) {
                        sound_37_victory, 0, 0, 0, 0, 0, 0
                    };
                } else if (delayed_sound.play_victory_sound == 2) { // skeleton on level 3 fell down
                    // skeleton makes a loose tile fall down
                    get_tile(delayed_sound.room, 2, 1);
                    make_loose_fall(1);
                } else if (current_level == 3 &&
                           delayed_sound.room == 16 &&
                           delayed_sound.sound_id == sound_47_chomper &&
                           delayed_sound.repeat > 0) { // started chompers on level 3, room 16
                    if (delayed_sound.row == 0) {
                        delayed_sound_timer = 2;
                        delayed_sound = (struct delayed_sound_type) {
                            sound_47_chomper, 16, 3, 2, 0, 0, 1
                        };
                    } else if (delayed_sound.col == 3) {
                        delayed_sound_timer = 1;
                        delayed_sound = (struct delayed_sound_type) {
                            sound_47_chomper, 16, 4, 2, 0, 0, 1
                        };
                    } else if (delayed_sound.col == 4) {
                        delayed_sound_timer = 1;
                        delayed_sound = (struct delayed_sound_type) {
                            sound_47_chomper, 16, 5, 2, 0, 0, 0
                        };
                    }
                    // created loose floors on level 10, room 16
                } else if (current_level == 10 && Kid.room == 17 && delayed_sound.repeat > 0) {
                    if (delayed_sound.sound_id == sound_20_loose_shake_1) {
                        delayed_sound_timer = 2;
                        delayed_sound = (struct delayed_sound_type) {
                            sound_21_loose_shake_2, 16, 0, 2, 0, 0, delayed_sound.repeat
                        };
                    } else if (delayed_sound.sound_id == sound_21_loose_shake_2) {
                        delayed_sound_timer = 2;
                        delayed_sound = (struct delayed_sound_type) {
                            sound_22_loose_shake_3, 16, 0, 2, 0, 0, delayed_sound.repeat
                        };
                    } else if (delayed_sound.sound_id == sound_22_loose_shake_3 && delayed_sound.repeat > 1) {
                        delayed_sound_timer = 2;
                        delayed_sound = (struct delayed_sound_type) {
                            sound_20_loose_shake_1, 16, 0, 2, 0, 0, delayed_sound.repeat - 1
                        };
                    }
                    // blink sound can be repeated once per second
                } else if (delayed_sound.sound_id == sound_38_blink && delayed_sound.repeat > 0) {
                    delayed_sound_timer = 12;
                    delayed_sound = (struct delayed_sound_type) {
                        sound_38_blink, 0, 0, 0, 0, 0, delayed_sound.repeat
                    };
                }
                // press a button after the sound starts
                if (delayed_sound.button_type == tiles_6_closer || delayed_sound.button_type == tiles_15_opener) {
                    get_tile(delayed_sound.room, delayed_sound.col, delayed_sound.row);
                    if (curr_tile2 == tiles_6_closer || curr_tile2 == tiles_15_opener) {
                        trigger_button(1, delayed_sound.button_type, curr_room_modif[curr_tilepos]);
                    }
                }
            }
            if (delayed_sound_timer == 0) {
                delayed_sound = (struct delayed_sound_type) {
                    -1, 0, 0, 0, 0, 0, 0
                    };
            }
        }
    }
skip_delayed_sound_timer:
    // custom logic - dead kid timer
    if (dead_kid_timer > 0) {
        --dead_kid_timer;
        if (dead_kid_timer == 11) {
            if (current_level != 6) {
                if (is_kid_shadow && Dead.room == Kid.room && Dead.direction != dir_56_none) {
                    Kid.direction = Dead.direction;
                    Kid.x = Dead.x;
                    Kid.x += dir_front[Kid.direction + 1] * 3;
                }
                seqtbl_offset_kid_char(seq_71_dying);
                play_seq();
            }
        } else if (dead_kid_timer == 7 && is_kid_shadow && is_shadow_flame) {
            load_flame_sprite();
        } else if (dead_kid_timer == 0) {
            if (current_level == 6 && Kid.room == 17) {
                if (hitp_curr > 0) {
                    Kid.direction = dir_0_right;
                    Kid.x += 7;
                    Kid.y = y_land[Kid.curr_row + ONSCREEN_ROW_OFFSET];
                    seqtbl_offset_kid_char(seq_49_stand_up_from_crouch);
                    play_seq();
                }
                goto skip_dead_kid_timer;
            }
            if (Kid.frame != frame_185_dead) {
                goto skip_dead_kid_timer;
            }
            if (is_kid_shadow) {
                Dead.direction = dir_56_none;
                // cannot unite with flame
                if (is_shadow_flame) {
                    stop_all_sounds();
                    play_sound(sound_1_falling);
                    take_hp(ALL_HP);
                    flash_color = color_12_brightred;
                    flash_time = 12;
                    goto skip_dead_kid_timer;
                }
                have_sword = 1;
                if (shadow_potion_counter > 0 && shadow_potion_counter <= 10) {
                    for (; shadow_potion_counter > 0; shadow_potion_counter--) {
                        add_life();
                    }
                }
                united_with_shadow = resurrect_time = 20;
            } else {
                Dead.room = Kid.room;
                Dead.direction = Kid.direction;
                Dead.x = Kid.x;
                Dead.row = Kid.curr_row;
                Dead.color = 0;
                Dead.color |= hitp_curr << 4;
                get_tile(Dead.room, get_tile_div_mod_m7(Dead.x), Dead.row);
                if (curr_tile2 == tiles_6_closer || curr_tile2 == tiles_15_opener) {
                    died_on_button(charid_1_shadow);
                }
                can_guard_see_kid = 0;
                play_sound(is_positional_audio ? sound_78_separated_from_shadow : sound_45_jump_through_mirror);
                loadkid();
            }
            is_kid_shadow = !is_kid_shadow;
            if (hitp_curr == 1) {
                start_ambient_music(false);
            }
            reset_shadow_palette();
            clear_kid_hp(hitp_max);
            draw_kid_hp(hitp_curr, hitp_max);
            if (Guard.alive >= 0 &&
                    Kid.room == drawn_room &&
                    !is_kid_shadow &&
                    dynamic_guard.room == drawn_room &&
                    (dynamic_guard_count[drawn_room - 1] == 0 ||
                     (dynamic_guard_count[drawn_room - 1] == 1 && dynamic_guard.frame > 0))) {
                leave_guard();
                enter_dynamic_guard();
            }
            if (!is_kid_shadow) {
                restore_dead_kid_tile();
            }
            make_guard_notice(false);
            seqtbl_offset_kid_char(seq_142_rise);
            play_seq();
        }
    } else if (current_level == settings->flame_level && dead_kid_timer < 0) {
        ++dead_kid_timer;
        if (dead_kid_timer == 0) {
            have_sword = 0;
            is_shadow_flame = false;
            reset_shadow_palette();
            clear_kid_hp(hitp_max);
            if (Kid.alive < 0) {
                hitp_curr = hitp_max = MIN(settings->start_hitp, settings->max_hitp_allowed);
            }
            draw_kid_hp(hitp_curr, hitp_max);
        }
    }
skip_dead_kid_timer:
    // custom logic - inactive guard timer
    if (inactive_guard_timer > 0) {
        --inactive_guard_timer;
        loadshad();
        if (is_dead() && inactive_guard_timer > 1) {
            inactive_guard_timer = 2;
        }
        if (inactive_guard_timer == 2) {
            leave_guard();
            draw_guard_hp(0, MAX_GUARD_HP);
            enter_guard();
            if (guard_overlap && can_inactive_guard_see_kid == 2) {
                Guard.y -= 2;
            }
            make_guard_notice(false);
            check_kid_droppedout();
        } else if (inactive_guard_timer == 0 &&
                   Guard.alive < 0 &&
                   (Guard.sword != sword_2_drawn || is_guarding_gate)) {
            if (!(current_level == settings->mouse_level &&
                    Guard.room == settings->mouse_room &&
                    leveldoor_open > 0)) { // not a mouse room
                if (delayed_sound_timer > 0) {
                    delayed_sound.play_victory_sound = 3;
                } else {
                    delayed_sound_timer = 1;
                    delayed_sound = (struct delayed_sound_type) {
                        sound_37_victory, 0, 0, 0, 0, 0, 0
                    };
                }
            }
        }
    }
    // custom logic - inactive guard fall timer
    if (inactive_guard_fall_timer > 0) {
        --inactive_guard_fall_timer;
        bool is_mirror_room = current_level == settings->mirror_level && drawn_room == settings->mirror_room && jumped_through_mirror == 1;
        // This timer is used in the mirror room to freeze the fall of prince
        // to give the player time to see what has happened.
        if (Guard.frame == (is_mirror_room ? frame_8_run : frame_166_stand_inactive)) {
            inactive_guard_fall_timer = 0;
        }
        if (inactive_guard_fall_timer == 0) {
            if (is_mirror_room) {
                delay_ticks(100); // pause for 10 seconds
            } else if (Guard.charid != charid_1_shadow) {
                // enter a dynamic guard if guard fell into a row below
                inactivate_active_guard();
                enter_dynamic_guard();
            }
        }
    }
    // custom logic - incrementing negative guard notice that acts as a timer
    if (is_inactive_guard_notice < 0) {
        ++is_inactive_guard_notice;
    }
    // custom logic - dynamic guard timer
    if (dynamic_guard_timer > 0) {
        --dynamic_guard_timer;
        check_kid_droppedout();
    }
    // custom logic - resets separating from shadow counter when there is a delay of more
    // than 7 ticks between side arrow key presses
    if (separate_from_shadow_counter > 0 && separate_from_shadow_reset_timer < 7) {
        ++separate_from_shadow_reset_timer;
        if (separate_from_shadow_reset_timer >= 7) {
            separate_from_shadow_reset_timer = 0;
            if (separate_from_shadow_counter > 0) {
                if (separate_from_shadow_counter_prev == separate_from_shadow_counter) {
                    separate_from_shadow_counter_prev = separate_from_shadow_counter = 0;
                } else {
                    separate_from_shadow_counter_prev = separate_from_shadow_counter;
                }
            }
        }
    }
    // custom logic - blink when flame is hit with sword
    if (lighting_timer > 0) {
        --lighting_timer;
        if (lighting_timer == 3) {
            enable_lighting = 1;
            need_full_redraw = 1;
        } else if (lighting_timer == 0) {
            enable_lighting = 0;
            need_full_redraw = 1;
        }
    }
    // Special event: mouse
    // custom logic - there is no delay for the mouse
    if (current_level == settings->mouse_level &&
            Kid.room == settings->mouse_room &&
            Kid.curr_row == 0 &&
            leveldoor_open == 2 &&
            get_tile(16, 3, 0) == tiles_4_gate &&
            curr_room_modif[curr_tilepos] >= 40 &&
            curr_room_modif[curr_tilepos] != 0xFF) {
        ++leveldoor_open;
        do_mouse(settings->mouse_start_x);
    }
    // custom logic - mouse starts running after drinking large white potion on level 12
    if (current_level == settings->mouse_2_level &&
            Kid.room == settings->mouse_2_room &&
            !is_kid_shadow &&
            leveldoor_open > 0 &&
            leveldoor_open < settings->mouse_2_crouch_delay &&
            get_tile(settings->mouse_2_room, settings->mouse_2_potion_column, settings->mouse_2_potion_row) == tiles_10_potion) {
        ++leveldoor_open;
        if (leveldoor_open == 8) {
            play_sound(sound_32_mouse_music);
            do_mouse(settings->mouse_2_start_x);
        }
    }
    // custom logic - timer for the last skeleton on level 3
    if (current_level == 3 &&
            skeleton_timer > 0 &&
            skeleton_timer < 61 &&
            drawn_room == 6) {
        ++skeleton_timer;
    }
    // custom logic - decreasing guard dropping down timer
    if (guard_dropping_down > 1) {
        --guard_dropping_down;
    }
    // custom logic - super high jump
    if (super_jump_timer > 0) {
        --super_jump_timer;
        Kid.curr_row = y_to_row_mod4(Kid.y);
        if (super_jump_timer > 0 &&
                feather_fall_timer == 0 &&
                Kid.frame >= frame_77_jumphang && Kid.frame <= frame_79_jumphang &&
                y_land[Kid.curr_row + ONSCREEN_ROW_OFFSET] - Kid.y > 32) {
            super_jump_timer = 0;
            super_jump_fall = 1;
            Kid.fall_x = 0;
            Kid.fall_y = 0;
            seqtbl_offset_kid_char(seq_7_fall);
            play_seq();
        } else if (super_jump_timer == 0 && Kid.frame == frame_79_jumphang) {
            if (get_tile(super_jump_room, super_jump_col, super_jump_row) == tiles_11_loose &&
                    get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0) {
                make_loose_fall(1); // knocks the true loose tile above
                do_knock(super_jump_room, super_jump_row, super_jump_col); // shakes the rest of loose tiles
            } else if (curr_tile2 == tiles_20_wall || tile_is_floor(curr_tile2)) {
                if (super_jump_row < 2) {
                    Kid.curr_row = super_jump_row + 1;
                    Kid.y = y_land[super_jump_row + 2] + 10;
                }
                do_knock(super_jump_room, super_jump_row, super_jump_col); // shakes loose tiles in the row
            } else if (!tile_is_floor(curr_tile2)) {
                if (super_jump_row == 2) {
                    Kid.room = level.roomlinks[Kid.room - 1].up;
                }
                if (Kid.room != 0) { // there is a room above
                    // positions kid for grabbing
                    Kid.curr_row = super_jump_row + 1;
                    Kid.y = y_land[super_jump_row + 2] - 10;
                    Kid.fall_x = 0;
                    Kid.fall_y = 0;
                    super_jump_fall = 1;
                    hard_jump_fall = 0;
                    // gives kid an ability to grab the above front tile
                    seqtbl_offset_kid_char(seq_19_fall);
                    play_seq();
                }
            }
        }
    }
    // custom logic - moving wall timer
    if (current_level == settings->moving_wall_level && moving_wall_timer < settings->moving_wall_delay) {
        ++moving_wall_timer;
        if (moving_wall_timer == settings->moving_wall_delay) {
            moving_wall_timer = 0;
        }
        if (moving_wall_timer == 1) {
            animate_moving_wall();
        }
    }
    // custom logic - falling bridge timer on level 12
    if (current_level == settings->mouse_2_level &&
            leveldoor_open > 8 &&
            falling_bridge_timer > 0 &&
            falling_bridge_timer < 0xFF &&
            Kid.alive < 0) {
        --falling_bridge_timer;
        if (falling_bridge_timer == 3) {
            play_sound(is_positional_audio ? sound_74_bridge_collapse_music : sound_25_presentation);
        }
        if (falling_bridge_timer == 0) {
            falling_bridge_timer = 0xFF;
            int bridge_col;
            for (bridge_col = 6; bridge_col < 10; bridge_col++) {
                get_tile(13, bridge_col, 0);
                curr_room_tiles[curr_tilepos] = tiles_11_loose;
                curr_room_modif[curr_tilepos] = 0;
                make_loose_fall(bridge_col);
                redraw_height = 24;
                set_wipe(curr_tilepos, 1);
                set_redraw_full(curr_tilepos, 1);
                ++curr_tilepos;
                set_wipe(curr_tilepos, 1);
                set_redraw_full(curr_tilepos, 1);
            }
            for (bridge_col = 0; bridge_col < 8; bridge_col++) {
                get_tile(2, bridge_col, 0);
                curr_room_tiles[curr_tilepos] = tiles_11_loose;
                curr_room_modif[curr_tilepos] = 0;
                make_loose_fall(10 - bridge_col);
                redraw_height = 24;
                set_wipe(curr_tilepos, 1);
                set_redraw_full(curr_tilepos, 1);
                ++curr_tilepos;
                set_wipe(curr_tilepos, 1);
                set_redraw_full(curr_tilepos, 1);
            }
        }
        // custom logic - delay main skeleton fall on level 3
    } else if (current_level == 3 && falling_bridge_timer > 0) {
        --falling_bridge_timer;
    }
    if (skeleton_death_timer > 0) {
        --skeleton_death_timer;
        if (skeleton_death_timer == 0) {
            loadshad();
            short char_col = Char.curr_col;
            bool moved_forward = false;
            while (char_col >= -1 && char_col <= 10) {
                if (get_tile(Char.room, char_col, Char.curr_row) != tiles_1_floor &&
                        curr_tile2 != tiles_6_closer &&
                        curr_tile2 != tiles_15_opener) {
                    if (curr_tile2 == tiles_20_wall ||
                            get_tile(Char.room, char_col + dir_behind[Char.direction + 1], Char.curr_row) == tiles_20_wall) {
                        char_col += dir_front[Char.direction + 1]; // backed up behind the floor tile
                        moved_forward = true;
                    } else {
                        if (moved_forward) break; // avoids infinite moving back and forth
                        char_col += dir_behind[Char.direction + 1]; // moving back until there is a floor tile
                    }
                    continue;
                }
                curr_room_tiles[curr_tilepos] = tiles_21_skeleton;
                set_redraw_full(curr_tilepos, 1);
                set_redraw_full(curr_tilepos + 1, 1);
                break;
            }
            level.guards_tile[Char.room - 1] = 0xFF;
            clear_char();
            saveshad();
            if (Kid.alive < 0) {
                set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
                play_sound(sound_44_skel_alive);
            }
        }
    }
}

// custom logic - process custom events
void process_custom_events(void) {
    // start ambient music after an OGG foresound has finished playing
    if (delayed_music_start == 1 && !ogg_fore_sound_playing) {
        delayed_music_start = 0;
        start_ambient_music(true);
    }
    // show message how to open main menu
    if (need_level1_music > 1 && need_level1_music < 72) {
        ++need_level1_music;
        if (need_level1_music == 72 && Kid.alive < 0) {
            need_level1_music = 0;
            display_text("PRESS ESC FOR MAIN MENU");
        }
    }
    // replace potion with an empty one
    if (potion_room > 0 && potion_pos < MAX_TILES_PER_ROOM) {
        if (++potion_timer > 23) { // length of potion drinking/putting down sequence in frames
            if (get_tile(potion_room, potion_pos % 10, potion_pos / 10) == tiles_1_floor) {
                curr_room_tiles[potion_pos] = tiles_10_potion;
                word potion_type = potion_size >> 3;
                potion_size &= 0x07;
                if (potion_timer < 100) {
                    switch (potion_size) {
                        case 1:
                            curr_room_modif[potion_pos] = 0x40; // large empty potion
                            break;
                        case 2:
                            curr_room_modif[potion_pos] = 0x30; // opener potion
                            break;
                        default:
                            if (hitp_curr == 0 && potion_type == potion_5_hurt) { // died from poison
                                curr_room_modif[potion_pos] = Kid.direction == dir_0_right ? 0x48 : 0x50; // small dropped potion
                                curr_room_modif[potion_pos] |= potion_type;
                                break;
                            }
                            curr_room_modif[potion_pos] = 0x38; // small empty potion
                            break;
                    }
                } else {
                    switch (potion_size) {
                        case 1:
                            curr_room_modif[potion_pos] = Kid.direction == dir_0_right ? 0x58 : 0x60; // large dropped potion
                            if (potion_type > potion_0_custom) {
                                curr_room_modif[potion_pos] |= potion_type;
                            }
                            break;
                        case 2:
                            curr_room_modif[potion_pos] = 0x30; // opener potion
                            break;
                        default:
                            curr_room_modif[potion_pos] = Kid.direction == dir_0_right ? 0x48 : 0x50; // small dropped potion
                            if (potion_type > potion_0_custom) {
                                curr_room_modif[potion_pos] |= potion_type;
                            }
                            break;
                    }
                }
                redraw_height = 35;
                set_wipe(potion_pos, 1);
                set_redraw_full(potion_pos, 1);
                start_anim_potion(potion_room, potion_pos);
                do_knock(potion_room, potion_pos / 10, potion_pos % 10);
                if (potion_room == drawn_room && !check_sound_playing() && delayed_sound_timer == 0 && feather_fall_timer == 0) {
                    delayed_sound_timer = 1;
                    delayed_sound = (struct delayed_sound_type) {
                        sound_17_soft_land, potion_room, potion_pos % 10, potion_pos / 10, 0, 0, 0
                    };
                }
            }
            if (current_level != 6 || Kid.room != level.start_room) {
                potion_timer = 0;
                potion_room = 0;
                potion_pos = -1;
                potion_size = 0;
            }
        }
        // visual effects after drinking a potion in start room of level 6
        if (current_level == 6 && Kid.room == level.start_room && potion_timer > 20) {
            if (potion_timer == 28) {
                music_suspended_timer = 124;
                if (had_sword && get_tile(Kid.room, settings->level_6_seq_potion_column - 1, Kid.curr_row) == tiles_11_loose) {
                    make_loose_fall(1);
                }
            } else if (potion_timer == 32 && had_sword) {
                if (get_tile(Kid.room, settings->level_6_seq_potion_column - 1, Kid.curr_row + 1) == tiles_1_floor) {
                    curr_room_tiles[curr_tilepos] = tiles_22_sword;
                    set_wipe(curr_tilepos, 1);
                    set_redraw_full(curr_tilepos, 1);
                    set_sound_pos(Kid.room, tile_col, tile_row);
                    stop_sounds();
                    play_sound(sound_10_sword_vs_sword);
                }
            } else if (potion_timer == 70) {
                play_sound(is_positional_audio ? sound_69_upside_down : sound_45_jump_through_mirror);
                // switch to the crazy palette
                load_lev6_palette();
                need_full_redraw = 1;
                flash_color = color_13_brightmagenta;
                flash_time = 16;
                level_6_sequence_timer = 0;
            } else if (potion_timer >= 80) {
                display_text_bottom(had_sword ? "WHOA ... MY SWORD!" : "WHOA ... GROOVY!");
                text_time_remaining = 48;
                text_time_total = 48;
                is_show_time = 1;
                potion_room = 0;
                potion_pos = -1;
                potion_timer = 0;
                potion_size = 0;
                draw_kid_hp(hitp_curr, hitp_max);
            }
        }
    }
    // replace skeleton behind the gate with a skeleton tile
    if (current_level == 3 &&
            drawn_room == 17 &&
            get_tile(17, 8, 2) != tiles_21_skeleton &&
            get_tile(17, 6, 2) == tiles_4_gate &&
            curr_room_modif[curr_tilepos] <= GUARD_HEIGHT_GATE_MODIFIER + 4 &&
            can_guard_see_kid &&
            Guard.direction == dir_FF_left &&
            Guard.room == 17 &&
            Guard.curr_row == 2 &&
            (Guard.frame == frame_158_stand_with_sword ||
             Guard.frame == frame_170_stand_with_sword ||
             Guard.frame == frame_171_stand_with_sword) &&
            get_tile_div_mod_m7(Guard.x) >= 6) {
        loadshad();
        if (guard_refrac == 0 && get_tile_div_mod_m7(Guard.x) == 6) {
            guard_refrac = 6;
            Char.x = char_dx_forward(-6);
            determine_col();
            seqtbl_offset_char(seq_57_back_with_sword);
            saveshad();
            play_seq();
        } else {
            Char.x = x_bump[8 + ONSCREEN_COLUMN_OFFSET];
            inactivate_skel(dir_FF_left);
            saveshad();
            play_seq();
            short tilepos = 21;
            // change the opener tile to a stuck tile
            curr_room_tiles[tilepos] = tiles_5_stuck;
            curr_room_modif[tilepos] = 0;
            redraw_height = 0x20;
            set_redraw_full(tilepos, 1);
            set_wipe(tilepos, 1);
            set_redraw_full(tilepos + 1, 1);
            set_wipe(tilepos + 1, 1);
            delayed_sound_timer = 7;
            delayed_sound = (struct delayed_sound_type) {
                sound_37_victory, 0, 0, 0, 0, 0, 0
            };
        }
    }
    // shakes skeletons that can only rise when the level door is open
    if (current_level == 3 &&
            !is_kid_mouse() &&
            Kid.room == 1 &&
            Kid.room == drawn_room &&
            Kid.curr_row == 1 &&
            leveldoor_open == 0) {
        if (skeleton_shake_timer == 0 && ABS(Kid.curr_col - 5) <= 1) {
            skeleton_shake_timer = SKELETON_SHAKE_TIME;
            set_sound_pos(1, 5, 1);
            play_sound(sound_44_skel_alive);
        } else if (skeleton_shake_timer > 0) {
            --skeleton_shake_timer;
            get_tile(1, 5, 1);
            if (skeleton_shake_timer == SKELETON_SHAKE_TIME - 1) {
                do_shake_skel(curr_tilepos, 5, 1, 1);
            } else if (skeleton_shake_timer == SKELETON_SHAKE_TIME - 3) {
                do_shake_skel(curr_tilepos, 5, 1, 0);
            }
        }
    } else if (current_level == 3 &&
               !is_kid_mouse() &&
               Kid.room == 6 &&
               Kid.room == drawn_room &&
               Kid.curr_row == 2 &&
               leveldoor_open == 0) {
        if (skeleton_shake_timer == 0 && ABS(Kid.curr_col - 6) <= 1) {
            skeleton_shake_timer = SKELETON_SHAKE_TIME;
            set_sound_pos(6, 6, 2);
            play_sound(sound_44_skel_alive);
        } else if (skeleton_shake_timer > 0) {
            --skeleton_shake_timer;
            get_tile(6, 6, 2);
            if (skeleton_shake_timer == SKELETON_SHAKE_TIME - 1) {
                do_shake_skel(curr_tilepos, 6, 2, 1);
            } else if (skeleton_shake_timer == SKELETON_SHAKE_TIME - 3) {
                do_shake_skel(curr_tilepos, 6, 2, 0);
            }
        }
    }
    // huge spilled poision potion kills kid without an antidote
    if (Kid.alive < 0 &&
            hitp_curr > 0 &&
            !is_kid_shadow &&
            get_tile_at_kid(Kid.x) == tiles_23_balcony_left &&
            curr_room_modif[curr_tilepos] == 1 /* huge posion potion */) {
        if (took_antidote == 0 || took_antidote == 1) {
            resurrect_time = 0;
            display_text_bottom("FIND THE ANTIDOTE");
            text_time_total = 48;
            text_time_remaining = 48;
            loadkid();
            take_hp(ALL_HP);
            seqtbl_offset_char(seq_26_crouch_while_running);
            savekid();
            stop_sounds();
            play_sound(sound_1_falling);
        }
    }
    // the kid must die to open level door
    if (current_level == 10 &&
            Kid.alive >= 0 &&
            Kid.room == 8 &&
            level_restart_timer == 0 &&
            holding_sword &&
            leveldoor_open < DEAD_KID_LEVEL_10_TIMER) {
        if (get_tile(8, 8, 2) != tiles_14_debris) {
            if (level_restart_timer == 0) {
                need_full_redraw = 1;
            }
            level_restart_timer = 1000;
        } else {
            ++leveldoor_open;
            ++text_time_remaining;
            if (leveldoor_open == DEAD_KID_LEVEL_10_TIMER) {
                enter_shadow_level_10();
            } else if (leveldoor_open == DEAD_KID_LEVEL_10_TIMER - 4) {
                get_tile(5, 2, 2);
                curr_room_tiles[curr_tilepos] = tiles_15_opener;
                trigger_button(1, tiles_15_opener, -1);
            } else if (leveldoor_open == DEAD_KID_LEVEL_10_TIMER - 28) {
                get_tile(5, 7, 2);
                trigger_button(1, tiles_15_opener, -1);
            } else if (leveldoor_open == DEAD_KID_LEVEL_10_TIMER - 40) {
                set_sound_pos(1, 4, 2);
                play_sound(sound_14_leveldoor_closing);
            }
        }
    }
    if (current_level == 13) {
        // open left bottom gate after the level door is opened on level 12B
        if (Kid.room == 1 && leveldoor_open == 1) {
            ++leveldoor_open;
            get_tile(20, 0, 0);
            trigger_button(0, tiles_14_debris, -1);
        }
        // kill kid after passing gate when the level door is closed on level 12B
        if (level_13_nowhere_to_run_timer >= 0 &&
                !is_kid_shadow &&
                Kid.room == 1 &&
                ((Kid.curr_row == 0 && Kid.x >= 92) ||
                 Kid.alive >= 0)) {
            if (++level_13_nowhere_to_run_timer == 1 && Kid.alive < 0 && hitp_curr > 0) {
                if (Kid.frame == frame_40_running_jump_1 && Kid.y < 55) {
                    Kid.y = 60;
                } else if (Kid.frame == frame_41_running_jump_2 && Kid.y < 55) {
                    Kid.y = 56;
                }
                loadkid();
                take_hp(ALL_HP);
                seqtbl_offset_char(seq_26_crouch_while_running);
                stop_sounds();
                play_sound(is_sound_on ? sound_36_out_of_time : sound_24_death_regular);
            } else if (level_13_nowhere_to_run_timer > 36) {
                level_13_nowhere_to_run_timer = -1;
                Kid.direction = dir_56_none;
                Kid.y = y_land[3 + ONSCREEN_ROW_OFFSET];
                int tile_col;
                if (get_tile(1, 4, 0) != tiles_21_skeleton) {
                    curr_room_tiles[curr_tilepos] = tiles_21_skeleton;
                    redraw_height = 35;
                    set_wipe(curr_tilepos, 1);
                    set_redraw_full(curr_tilepos, 1);
                    set_wipe(curr_tilepos + 1, 1);
                    set_redraw_full(curr_tilepos + 1, 1);
                    if (Kid.curr_row == 0) {
                        for (tile_col = 1; tile_col <= 5; tile_col++) {
                            get_tile(1, tile_col, 1);
                            curr_room_tiles[curr_tilepos] = tiles_10_potion;
                            redraw_height = 63;
                            set_wipe(curr_tilepos, 1);
                            set_redraw_full(curr_tilepos, 1);
                            set_redraw_anim_curr();
                            set_wipe(curr_tilepos + 1, 1);
                            set_redraw_full(curr_tilepos + 1, 1);
                            start_anim_potion(1, curr_tilepos);
                        }
                    }
                }
            }
        }
    }
    // on level 2 make the fake wall transparent for 1 tick out 12 when kid is standing near
    if (current_level == 2 &&
            ((drawn_room == 20 && Kid.room == 20) ||
             (drawn_room == 24 && Kid.room == 24))) {
        byte* room19_tiles = &level.fg[(24 - 1) * 30];
        if ((Kid.curr_col == 0 && Kid.curr_row == 0) &&
                Kid.room == 20 &&
                blink_timer == 6) {
            if ((room19_tiles[9] & 0x1F) != tiles_1_floor) {
                room19_tiles[9] = tiles_1_floor;
                need_full_redraw = 1;
            }
        } else if ((room19_tiles[9] & 0x1F) != tiles_20_wall) {
            room19_tiles[9] = tiles_20_wall;
            need_full_redraw = 1;
        }
    }
    // on level 14 make the fake wall transparent for 1 tick out 12 when kid is standing near
    if (current_level == 14 &&
            ((drawn_room == 14 && Kid.room == 14) ||
             (drawn_room == 19 && Kid.room == 19))) {
        byte* room19_tiles = &level.fg[(19 - 1) * 30];
        if (((Kid.curr_col == 0 && Kid.curr_row == 1) ||
                (Kid.curr_col == 1 && Kid.curr_row == 0)) &&
                Kid.room == 14 &&
                blink_timer == 6) {
            if ((room19_tiles[9] & 0x1F) != tiles_1_floor) {
                room19_tiles[9] = tiles_1_floor;
                need_full_redraw = 1;
            }
        } else if ((room19_tiles[9] & 0x1F) != tiles_20_wall) {
            room19_tiles[9] = tiles_20_wall;
            need_full_redraw = 1;
        }
    }
    // on level 13 room 3 some floor tiles disappear the first time prince enters the room
    if (current_level == 14 &&
            Kid.room == 3 &&
            (!is_kid_shadow || !is_shadow_flame) &&
            !is_kid_mouse() &&
            jaffar_presented < 6) {
        if (++jaffar_presented == 6) {
            play_sound(is_positional_audio ? sound_78_separated_from_shadow : sound_25_presentation);
            redraw_height = 63;
            for (int row = 1; row <= 2; row++) {
                for (int col = 5; col <= 7; col++) {
                    get_tile(3, col, row);
                    curr_room_tiles[curr_tilepos] = (row == 2 && col == 5);
                    curr_room_modif[curr_tilepos] = (row != 2 || col != 5);
                    set_wipe(curr_tilepos, 1);
                    set_redraw_full(curr_tilepos, 1);
                }
                set_wipe(curr_tilepos + 1, 1);
                set_redraw_full(curr_tilepos + 1, 1);
            }
        }
    }
}

// seg003:0798
void check_mirror(void) {
    word clip_top;
    if (jumped_through_mirror < 0) {
        jump_through_mirror();
    } else if (jumped_through_mirror == 0) {
        if (get_tile_at_char() == tiles_13_mirror) {
            loadkid();
            load_frame();
            check_mirror_image();
            if (distance_mirror >= 0 && Char.room == drawn_room && Char.sword == sword_0_sheathed) {
                load_frame_to_obj();
                reset_obj_clip();
                clip_top = y_clip[Char.curr_row + 1];
                if (clip_top < obj_y) {
                    obj_clip_top = clip_top;
                    obj_clip_left = (Char.curr_col << 5) + 9;
                    add_objtable(objtype_4_mirror_image);
                }
            }
        }
    }
}

// seg003:080A
void jump_through_mirror(void) {
    // custom logic - on level 6, room 1 change tiles after the jump so kid would fall down
    if (current_level == settings->mirror_level && Char.room == 1) {
        int tile_row, tile_col;
        for (tile_row = 2; tile_row >= 1; tile_row--)
            for (tile_col = 2; tile_col <= 6; tile_col++) {
                get_tile(1, tile_col, tile_row);
                if (tile_row == 2 && tile_col < 5) {
                    curr_room_tiles[curr_tilepos] = tiles_0_empty;
                    curr_room_modif[curr_tilepos] = 1;
                } else if (tile_row == 1 && tile_col == 5) {
                    curr_room_tiles[curr_tilepos] = tiles_3_pillar;
                }
                redraw_height = 63;
                set_wipe(curr_tilepos, 1);
                set_redraw_full(curr_tilepos, 1);
            }
        get_tile(1, 6, 1);
        curr_room_tiles[curr_tilepos] = tiles_15_opener;
        set_wipe(curr_tilepos, 1);
        set_redraw_full(curr_tilepos, 1);
        set_wipe(++curr_tilepos, 1);
        set_redraw_full(--curr_tilepos, 1);
        trigger_button(0, tiles_15_opener, -1);
    }
    erase_bottom_text(0);
    if (jumped_through_mirror == -1) {
        loadkid();
        load_frame();
        check_mirror_image();
        redraw_height = 63;
        set_wipe(10, 1);
        set_redraw_full(10, 1);
        set_wipe(11, 1);
        set_redraw_full(11, 1);
        draw_custom_graphics_backtable();
        inactive_guard_fall_timer = 12;
        Char.charid = charid_1_shadow;
        Char.x = x_bump[5 + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
        play_sound(sound_45_jump_through_mirror);
        saveshad();
        guardhp_max = guardhp_curr = 1;
        // custom logic - take away 1 hit point if kid has more than 3 (max)
        if (current_level == 6 && Char.room == 1) {
            clear_kid_hp(hitp_max);
            if (hitp_max > 3) {
                draw_kid_hp(MIN(hitp_curr, hitp_max - 1), hitp_max - 1);
            } else {
                draw_kid_hp(hitp_curr, hitp_max);
            }
        } else {
            --hitp_curr;
            draw_kid_hp(hitp_curr, hitp_max);
            draw_guard_hp(guardhp_curr, guardhp_max);
        }
    } else {
        play_sound(sound_13_kid_hurt);
        --hitp_curr;
        draw_kid_hp(hitp_curr, hitp_max);
    }
    jumped_through_mirror = 1;
    // custom logic - flash blue when jumping through mirror
    flash_color = color_12_brightred;
    flash_time = 10;
}

// seg003:085B
void check_mirror_image(void) {
    short distance;
    short xpos;
    xpos = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_10_PX_OFFSET;
    distance = distance_to_edge_weight();
    if (Char.direction >= dir_0_right) {
        distance = (~distance) + 14;
    }
    distance_mirror = distance - 2;
    Char.x = (xpos << 1) - Char.x;
    Char.direction = ~Char.direction;
}

// seg003:08AA
void bump_into_opponent(void) {
    // This is called from play_kid_frame, so Char=Kid, Opp=Guard
    short distance;
    // custom logic - bump kid out of the wall
    if (Char.action == actions_5_bumped &&
            get_tile_at_char() == tiles_20_wall &&
            Opp.alive < 0 &&
            Char.direction != Opp.direction &&
            Char.curr_row == Opp.curr_row &&
            ABS(Char.curr_col - Opp.curr_col) <= 1) {
        Char.x = char_dx_forward(distance_to_edge_weight() + 7);
        load_fram_det_col();
        Char.y = y_land[Char.curr_row + 1];
        return;
    }
    if (can_guard_see_kid >= 2 &&
            Char.sword == sword_0_sheathed && // Kid must not be in fighting pose
            Opp.sword != sword_0_sheathed && // but Guard must
            Opp.action <= actions_1_run_jump &&
            Char.direction != Opp.direction // must be facing toward each other
       ) {
        distance = char_opp_dist();
        if (ABS(distance) <= 15) {
            if (Char.fall_y >= 22) {
                return; // don't bump
            }
            // custom logic - do not bump guard in midair
            if (Char.action == actions_3_in_midair ||
                    (Char.action == actions_4_in_freefall && y_land[Char.curr_row + ONSCREEN_ROW_OFFSET] > Char.y)) {
                can_guard_see_kid = 1;
                return;
            }
            // custom logic - don't bump guards in rooms above or below to avoid a room switching bug
            if (Char.curr_row < 0 || Char.curr_row > 2) {
                return;
            }
            // custom logic - do not bump when hanging or falling
            if (Char.frame >= frame_81_hangdrop_1 && Char.frame < frame_106_fall) {
                return;
            }
            // custom logic - makes it easier to jump over purple weirdo on level 6
            if (current_level == 6 &&
                    Char.room == 6 &&
                    Char.frame >= frame_40_running_jump_1 &&
                    Char.frame <= frame_44_running_jump_5) {
                return;
            }
            // custom logic - prevents kid from jumping over guards
            if ((Char.direction == dir_0_right && Char.x > Opp.x) ||
                    (Char.direction == dir_FF_left && Char.x < Opp.x)) {
                Char.x = Opp.x;
            }
            Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
            Char.fall_y = 0;
            // custom logic - reset hard jump fall flag
            hard_jump_fall = 0;
            seqtbl_offset_char(seq_47_bump); // bump into opponent
            play_seq();
        }
    }
}

// seg003:0913
void pos_guards(void) {
    short guard_tile;
    short room_minus_1;
    word crushed_seq;
    char_type2 dead_level_guard;
    char_type2 inactive_level_guard;
    short guard_color;
    // custom logic - initialize the dead kid slot
    Dead.direction = dir_56_none;
    Dead.room = 0;
    Dead.x = 0;
    Dead.row = 0;
    Dead.frame = Dead.frame2 = 0;
    Dead.color = 0;
    // custom logic - reset outbound falling guards
    falling_guard = (struct falling_guard_type) {
        0
    };
    // custom logic - add a guard to the prison level
    if (current_level == settings->prison_level) {
        level.guards_tile[0] = 7;
        level.guards_dir[0] = dir_FF_left;
        level.guards_color[0] = 9; // prison guard palette
        level.guards_skill[0] = 1;
    }
    for (room_minus_1 = 0; room_minus_1 < level.used_rooms; ++room_minus_1) {
        // custom logic - reset opener tiles pressed by guards
        level.opener_col[room_minus_1] = 0;
        level.opener_row[room_minus_1] = 0;
        // custom logic - reset dead guard slots
        dead_guard1[room_minus_1] = (struct char_type2) {
            0
        };
        dead_guard2[room_minus_1] = (struct char_type2) {
            0
        };
        dead_guard3[room_minus_1] = (struct char_type2) {
            0
        };
        outbound_dead_guard1[room_minus_1] = (struct char_type2) {
            0
        };
        outbound_dead_guard2[room_minus_1] = (struct char_type2) {
            0
        };
        outbound_dead_guard3[room_minus_1] = (struct char_type2) {
            0
        };
        // custom logic - reset dynamic guard counts
        dynamic_guard_count[room_minus_1] = 0;
        // custom logic - reset inactive guard coordinates
        inactive_guard[room_minus_1] = (struct char_type2) {
            0
        };
        // position guards
        guard_tile = level.guards_tile[room_minus_1];
        if (guard_tile < MAX_TILES_PER_ROOM) {
            level.guards_x[room_minus_1] = x_bump[guard_tile % 10 + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
            // custom logic - poisoned guard on level 5
            if (current_level == 5 && room_minus_1 + 1 == 15) {
                crushed_seq = seqtbl_offset_get_seq(seq_22_crushed);
                level.guards_seq_lo[room_minus_1] = crushed_seq;
                level.guards_seq_hi[room_minus_1] = crushed_seq >> 8;
                level.guards_dir[room_minus_1] = dir_0_right;
            } else {
                level.guards_seq_hi[room_minus_1] = 0;
            }
        }
    }
    // custom logic - position dead level guard
    dead_level_guard = settings->dead_guards[current_level];
    if (dead_level_guard.room > 0 && dead_level_guard.room <= level.used_rooms) {
        dead_guard3[dead_level_guard.room - 1] = dead_level_guard;
    }
    // custom logic - position inactive level guard
    inactive_level_guard = settings->inactive_guards[current_level];
    if (inactive_level_guard.room > 0 && inactive_level_guard.room <= level.used_rooms) {
        if (settings->tbl_guard_type[current_level] == guard_type_guard) {
            if (inactive_level_guard.color > 0 && inactive_level_guard.color <= GUARD_COLORS) {
                guard_color = inactive_level_guard.color;
            } else {
                guard_color = level.guards_color[inactive_level_guard.room - 1] & 0x0F;
            }
        } else {
            guard_color = 0;
        }
        guard_color |= ((settings->tbl_guard_hp[current_level] - 1) << 4);
        inactive_level_guard.color = guard_color;
        inactive_guard[inactive_level_guard.room - 1] = inactive_level_guard;
    }
    // custom logic - add additional characters on level 6
    if (current_level == 6) {
        // add princess
        if (falling_exit_checkpoint < 4) {
            dead_guard3[0].room = 1;
            dead_guard3[0].direction = dir_FF_left;
            dead_guard3[0].x = x_bump[/*0 +*/ ONSCREEN_COLUMN_OFFSET] + CHAR_10_PX_OFFSET;
            dead_guard3[0].row = 1;
            dead_guard3[0].frame2 = 190; // princess frame
        }
        // dead fat guard
        dead_guard3[16].room = 17;
        dead_guard3[16].direction = dir_0_right;
        dead_guard3[16].x = x_bump[3 + ONSCREEN_COLUMN_OFFSET] + 3;
        dead_guard3[16].row = 2;
        dead_guard3[16].frame2 = 191; // dead fat guard frame
    }
    // custom logic - add skeleton in room 21 when entering level 12 from level 12B
    if (current_level == 12 && level.guards_tile[20] >= 30 && checkpoint) {
        level.guards_tile[20] = get_tilepos(0, 2);
        level.guards_x[20] = x_bump[5 + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
        level.guards_dir[20] = dir_FF_left;
        level.guards_seq_hi[20] = 0;
    }
}

// seg003:0959
void check_can_guard_see_kid(void) {
    /*
    Possible results in can_guard_see_kid:
    0: Guard can't see Kid
    1: Guard can see Kid, but won't come
    2: Guard can see Kid, and will come
    */
    short kid_frame;
    short left_pos;
    short temp;
    short right_pos;
    short guard_type;
    kid_frame = Kid.frame;
    guard_type = settings->tbl_guard_type[current_level];
    if (Guard.charid != charid_1_shadow &&
            Guard.charid != charid_24_mouse &&
            // custom logic - guards do not react to shadow/mouse except for skeletons, Jaffar and his vision
            ((!is_kid_shadow && !is_kid_mouse()) ||
             (Guard.charid == charid_4_skeleton && is_kid_shadow) ||
             guard_type == guard_type_Jaffar ||
             guard_type == guard_type_Jaffar_vision) &&
            kid_frame != 0 && (kid_frame < frame_219_exit_stairs_3 || kid_frame > frame_228_exit_stairs_12) &&
            Guard.direction != dir_56_none && Kid.alive < 0 && Guard.alive < 0 && Kid.room == Guard.room && Kid.curr_row == Guard.curr_row
       ) {
        can_guard_see_kid = 2;
        is_guarding_gate = 0;
        // custom logic - set the dynamic guard fight flag
        if (!is_dynamic_guard_fight &&
                dynamic_guard.room == drawn_room &&
                Kid.curr_row >= dynamic_guard.row &&
                Kid.curr_row - dynamic_guard.row < 2) {
            is_dynamic_guard_fight = 1;
        }
        left_pos = x_bump[Kid.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
        right_pos = x_bump[Guard.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
        // When the kid is hanging on the right side of a doortop, Kid.curr_col points at the doortop tile and a guard on the left side will see the prince.
        // This fixes that.
        if (Kid.direction == dir_FF_left && Kid.action == actions_6_hang_straight) {
            left_pos += 14;
        }
        // custom logic - a running turn in the air or a start of a standing jump does not count as an empty tile
        if (!tile_is_floor(get_tile_at_kid(left_pos)) &&
                ((Kid.frame >= frame_18_standing_jump_3 && Kid.frame <= frame_26_standing_jump_11) ||
                 (Kid.frame >= frame_56_runturn && Kid.frame <= frame_65_runturn))) {
            if (char_opp_dist() < 29) { // within striking range
                if (left_pos < right_pos) {
                    left_pos += 14;
                } else {
                    left_pos -= 14;
                }
            }
        }
        if (left_pos > right_pos) {
            temp = left_pos;
            left_pos = right_pos;
            right_pos = temp;
        }
        // A chomper is on the left side of a tile, so it doesn't count.
        if (get_tile_at_kid(left_pos) == tiles_18_chomper) {
            left_pos += 14;
        }
        // custom logic - wall does not count when kid is hanging straight on the left side
        if (Kid.direction == dir_0_right && Kid.action == actions_6_hang_straight) {
            right_pos -= 14;
        }
        // A gate or a doortop is on the right side of a tile, so it doesn't count.
        if (get_tile_at_kid(right_pos) == tiles_4_gate ||
                curr_tile2 == tiles_7_doortop_with_floor ||
                curr_tile2 == tiles_12_doortop) {
            right_pos -= 14;
        }
        if (right_pos >= left_pos) {
            while (left_pos <= right_pos) {
                // Can't see through these tiles.
                if (get_tile_at_kid(left_pos) == tiles_20_wall ||
                        curr_tile2 == tiles_7_doortop_with_floor ||
                        curr_tile2 == tiles_12_doortop
                   ) {
                    can_guard_see_kid = 0;
                    return;
                }
                // custom logic - a bug that can scan tiles in other rooms can switch the inactive guard
                // with the active guard unnecessarily
                if (curr_room != drawn_room && inactive_guard[drawn_room - 1].room == drawn_room) {
                    check_can_inactive_guard_see_kid();
                    get_tile_at_kid(left_pos);
                    if (can_inactive_guard_see_kid == 2) {
                        left_pos += 14;
                        continue;
                    }
                }
                // Can see through these, but won't go through them.
                if (curr_tile2 == tiles_11_loose ||
                        curr_tile2 == tiles_18_chomper ||
                        (curr_tile2 == tiles_4_gate &&
                         current_level != settings->inverse_level &&
                         curr_room_modif[curr_tilepos] < GUARD_HEIGHT_GATE_MODIFIER) ||
                        !tile_is_floor(curr_tile2)
                   ) {
                    // custom logic - guard can try to strike kid during a super high jump
                    if (tile_col == Kid.curr_col &&
                            tile_is_empty(curr_tile2) &&
                            (super_jump_timer != 0 || super_jump_fall != 0) &&
                            guard_kid_dist() < 29) {
                        left_pos += 14;
                        continue;
                    }
                    // custom logic - guard should move closer to the chomper when he sees kid
                    if (curr_tile2 == tiles_18_chomper && ABS(Guard.x - left_pos) > 21) {
                        can_guard_see_kid = 3; // guard moves forward in engarde mode but kid does not draw his sword
                        left_pos += 14;
                        continue;
                    }
                    // custom logic - continue if kid is standing on a regular loose tile
                    if (tile_col == Kid.curr_col &&
                            curr_tile2 == tiles_11_loose &&
                            get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0 && // not stuck loose (for level 6 guard to work)
                            get_tile_ctrl_bits(curr_room, curr_tilepos, 1) == 0) { // and not totally loose
                        left_pos += 14;
                        continue;
                    }
                    // custom logic - Guard can see kid hanging in front of him over an empty tile
                    if (ABS(Kid.curr_col - Guard.curr_col) == 1 &&
                            Kid.action == actions_2_hang_climb &&
                            !tile_is_floor(curr_tile2)) {
                        return;
                    }
                    // custom logic - Jaffar's vision can walk on air
                    if (settings->tbl_guard_type[current_level] == guard_type_Jaffar_vision && tile_is_empty(curr_tile2)) {
                        return;
                    }
                    can_guard_see_kid = 1;
                    // custom logic - a guard guarding gate does not become inactive at a distance
                    is_guarding_gate = (guard_skill == GATE_GUARDING_SKILL);
                    return;
                }
                left_pos += 14;
            }
        }
    } else {
        can_guard_see_kid = 0;
        is_guarding_gate = 0;
    }
    // custom logic - bug fix for prince not drawing the sword after it was put away
    // even in he goes out of range and comes back
    if (can_guard_see_kid == 0 && offguard != 0) {
        offguard = 0;
    }
}

// custom logic - check if inactive guard can see kid
void check_can_inactive_guard_see_kid(void) {
    short kid_frame;
    short left_pos;
    short temp;
    short right_pos;
    word room_minus_1;
    sbyte inactive_guard_col;
    kid_frame = Kid.frame;
    room_minus_1 = drawn_room - 1;
    if (inactive_guard[room_minus_1].room != drawn_room ||
            inactive_guard[room_minus_1].row != Kid.curr_row ||
            inactive_guard[room_minus_1].direction == dir_56_none ||
            (inactive_guard[room_minus_1].frame == 0 && inactive_guard[room_minus_1].frame2 == 0)) {
        can_inactive_guard_see_kid = 0;
        return;
    }
    if (is_kid_mouse()) {
        can_inactive_guard_see_kid = 0;
        return;
    }
    if (Kid.alive < 0 &&
            kid_frame != 0 &&
            Kid.room == drawn_room &&
            (kid_frame < frame_219_exit_stairs_3 || kid_frame > frame_228_exit_stairs_12)) {
        can_inactive_guard_see_kid = 2;
        inactive_guard_col = get_tile_div_mod_m7(inactive_guard[room_minus_1].x);
        left_pos = x_bump[Kid.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
        right_pos = x_bump[inactive_guard_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
        if (left_pos > right_pos) {
            temp = left_pos;
            left_pos = right_pos;
            right_pos = temp;
        }
        // A chomper is on the left side of a tile, so it doesn't count.
        if (get_tile_at_kid(left_pos) == tiles_18_chomper) {
            left_pos += 14;
        }
        // custom logic - wall does not count when kid is hanging straight on the left side
        if (Kid.direction == dir_0_right && Kid.action == actions_6_hang_straight) {
            right_pos -= 14;
        }
        // A gate or a doortop is on the right side of a tile, so it doesn't count.
        if (get_tile_at_kid(right_pos) == tiles_4_gate ||
                curr_tile2 == tiles_7_doortop_with_floor ||
                curr_tile2 == tiles_12_doortop) {
            right_pos -= 14;
        }
        if (right_pos >= left_pos) {
            while (left_pos <= right_pos) {
                // Can't see through these tiles.
                if (get_tile_at_kid(left_pos) == tiles_20_wall ||
                        curr_tile2 == tiles_7_doortop_with_floor ||
                        curr_tile2 == tiles_12_doortop
                   ) {
                    can_inactive_guard_see_kid = 0;
                    return;
                }
                // Can see through these, but won't go through them.
                if (curr_tile2 == tiles_11_loose ||
                        curr_tile2 == tiles_18_chomper ||
                        (curr_tile2 == tiles_4_gate && curr_room_modif[curr_tilepos] < GUARD_HEIGHT_GATE_MODIFIER) ||
                        !tile_is_floor(curr_tile2)
                   ) {
                    can_inactive_guard_see_kid = 1;
                }
                left_pos += 14;
            }
        }
    } else {
        can_inactive_guard_see_kid = 0;
    }
}

// custom logic - checks whether the guard was following kid in the previous room before he dropped down
bool check_kid_dropped_into_next_room(void) {
    if (Guard.alive >= 0 || Guard.room == 0 || Guard.action > actions_1_run_jump) {
        return false;
    }
    if (Guard.room != level.roomlinks[Kid.room - 1].left && Guard.room != level.roomlinks[Kid.room - 1].right) {
        return false;
    }
    if (!droppedout || Kid.curr_row - Guard.curr_row != 1) {
        return false;
    }
    return (Guard.frame >= frame_163_fighting && Guard.frame <= frame_165_walk_with_sword) ||
           Guard.frame == frame_158_stand_with_sword;
}

// seg003:0A99
byte get_tile_at_kid(int xpos) {
    return get_tile(Kid.room, get_tile_div_mod_m7(xpos), Kid.curr_row);
}

// seg003:0ABA
void do_mouse(int start_x) {
    loadkid();
    // custom logic - inactivate guard in the room
    inactivate_guard(Char.room - 1);
    Char.charid = charid_24_mouse;
    Char.x = start_x;
    Char.curr_row = 0;
    Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
    Char.alive = -1;
    Char.direction = dir_FF_left;
    Char.fall_x = 0;
    Char.fall_y = 0;
    guardhp_curr = 1;
    seqtbl_offset_char(seq_105_mouse_forward); // mouse forward
    play_seq();
    saveshad();
}

// custom logic - has kid turn into mouse?
bool is_kid_mouse(void) {
    return Kid.frame >= frame_186_mouse_1 && Kid.frame <= frame_188_mouse_stand;
}

// seg003:0AFC
int flash_if_hurt(void) {
    if (flash_time != 0) {
        do_flash(flash_color);
        return 1;
    } else if (hitp_delta < 0) {
        if (is_joyst_mode && settings->controller_rumble) {
            if (sdl_haptic != NULL) {
                SDL_HapticRumblePlay(sdl_haptic, 1.0, 100); // rumble at full strength for 100 milliseconds
#if SDL_VERSION_ATLEAST(2,0,9)
            } else if (sdl_controller_ != NULL) {
                SDL_GameControllerRumble(sdl_controller_, 0xFFFF, 0xFFFF, 100);
            } else {
                SDL_JoystickRumble(sdl_joystick_, 0xFFFF, 0xFFFF, 100);
#endif
            }
        }
        // custom logic - do not flash when flame is hurt
        if (!is_shadow_flame) {
            do_flash(color_12_brightred);
        }
        return 1;
    }
    return 0; // not flashed
}

// seg003:0B1A
void remove_flash_if_hurt(void) {
    if (flash_time != 0) {
        --flash_time;
    } else {
        if (hitp_delta >= 0) return;
    }
    remove_flash();
}
