/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"

// custom logic - get sequence word value by its index
word seqtbl_offset_get_seq(short seq_index) {
    return seqtbl_offsets[seq_index];
}

// seg005:000A
void seqtbl_offset_char(short seq_index) {
    Char.curr_seq = seqtbl_offsets[seq_index];
}

// seg005:001D
void seqtbl_offset_opp(int seq_index) {
    Opp.curr_seq = seqtbl_offsets[seq_index];
}

// custom logic - ensure kid does not land/stop at the very edge of a tile in front of an abyss
void check_on_edge(void) {
    bool is_abyss_behind;
    is_abyss_behind = (!tile_is_floor(get_tile_behind_char()) && curr_tile2 != tiles_20_wall) ||
                      (curr_tile2 == tiles_11_loose && curr_room_modif[curr_tilepos] != 0);
    if (is_abyss_behind && (Char.action <= actions_1_run_jump || Char.action == actions_4_in_freefall) && distance_to_edge_weight() == 13) {
        Char.x = char_dx_forward(1);
    }
}

// custom logic - separated screaming from falling logic
void do_scream(void) {
    // custom logic - kid screams differently in specific rooms when a bottomless fall happens
    bool bottomless_room = (current_level == 6 && Char.room == 10) || (current_level == 10 && Char.room == 24);
    // custom logic - only kid screams here
    if (Char.charid == charid_0_kid &&
            !is_kid_shadow &&
            hitp_curr > 0 &&
            (Char.alive < 0 || holding_sword) &&
            Char.fall_y >= (bottomless_room && Char.curr_row > 2 ? 24 : 31) &&
            is_screaming == 0 &&
            feather_fall_timer == 0 &&
            // do not scream during falls on level 6 except room 10 or when kid turns into shadow
            (current_level != 6 || Char.room == 10)) {
        if (is_positional_audio && bottomless_room) {
            play_sound(sound_75_falling_far);
        } else {
            play_sound(sound_1_falling);
            if (!ogg_fore_sound_playing && Mix_PlayingMusic()) {
                Mix_FadeOutMusic(400);
            }
        }
        is_screaming = 1;
    }
}

// checks if kid/guard is falling next to a wall or a doortop
bool falling_near_wall(sbyte direction) {
    return (curr_tile2 == tiles_20_wall && get_tile_ctrl_bits(curr_room, curr_tilepos, 1) == 0) ||
           ((curr_tile2 == tiles_12_doortop || curr_tile2 == tiles_7_doortop_with_floor) &&
            direction == dir_FF_left);
}

// seg005:0030
void do_fall(void) {
    do_scream();
    if (y_land[Char.curr_row + ONSCREEN_ROW_OFFSET] > Char.y) {
        check_grab();
        // Fix for the kid falling through walls
        determine_col();
        if (feather_fall_timer > 0) { // weightless
            get_tile(Char.room, Char.curr_col, y_to_row_mod4(char_top_y + 3));
            if (!falling_near_wall(Char.direction)) {
                get_tile_at_char();
            }
        } else {
            get_tile_at_char();
        }
        if (falling_near_wall(Char.direction)) {
            int delta_x = distance_to_edge_weight();
            // When falling into a wall or doortop after turning or running, delta_x is likely to be either 8, 11 or 12
            // Calling in_wall() here produces the desired fix (no glitching through wall), but only when delta_x >= 10
            // The code below "emulates" in_wall() to always get the same effect as if delta_x == 10
            // (which distance actually looks/behaves best is a matter of preference)
            if (delta_x >= 8) {
                delta_x = -5 + delta_x - 10;
                Char.x = (byte) char_dx_forward(delta_x);
                Char.fall_x = 0; // not in in_wall(), but we do need to cancel horizontal movement
            }
        }
    } else {
        determine_col();
        if (get_tile_at_char() == tiles_20_wall) {
            in_wall();
        } else if (curr_tile2 == tiles_12_doortop && Char.direction == dir_FF_left) {
            if (distance_to_edge_weight() >= 8) { // only intervene if the kid is actually IN FRONT of the tapestry
                in_wall();
            }
        }
        if (tile_is_floor(curr_tile2)) {
            // custom logic - sheathe the sword during a short fall
            if (Char.charid == charid_0_kid && Char.sword == sword_2_drawn && can_guard_see_kid != 2) {
                Char.sword = sword_0_sheathed;
            }
            land();
        } else {
            inc_curr_row();
        }
    }
}

// custom logic - gets the tile in front of falling guard
int get_tile_infrontof_falling_guard(void) {
    return get_tile(falling_guard.room, get_tile_div_mod_m7(falling_guard.x) + dir_front[falling_guard.direction + 1], y_to_row_mod4(falling_guard.y));
}

// custom logic - gets the tile behind falling guard
int get_tile_behind_falling_guard(void) {
    return get_tile(falling_guard.room, get_tile_div_mod_m7(falling_guard.x) + dir_behind[falling_guard.direction + 1], y_to_row_mod4(falling_guard.y));
}

// custom logic - fallen guard lands
void fallen_guard_land(bool is_skeleton) {
    word room_minus_1;
    int sound_id;
    int delta_x;
    sbyte guard_dir;
    int guard_col;
    int guard_x;
    bool follow_kid;
    bool skip_victory_music;
    char_type2 *slot;
    room_minus_1 = falling_guard.room - 1;
    sound_id = -1;
    follow_kid = false;
    skip_victory_music = false;
    if (is_screaming) {
        is_screaming = 0;
        if (Kid.alive < 0) {
            if (ogg_fore_sound_playing && curr_ogg_sound_id == sound_81_fat_guard_falling) {
                stop_all_sounds();
            } else {
                stop_sounds();
            }
        }
    }
    // hard fall - guard loses a hitpoint
    if (!is_skeleton && falling_guard.fall_y >= 22 && falling_guard.fall_y < 33 && falling_guard.hitp > 0) {
        --falling_guard.hitp;
    }
    guard_col = tile_col;
    // landed outside the room
    if (curr_room != falling_guard.room)	{
        if (curr_room == level.roomlinks[room_minus_1].right) {
            guard_col += 10;
        } else if (curr_room == level.roomlinks[room_minus_1].left) {
            guard_col -= 10;
        }
    }
    // skeletons in the second part of level 12 stay alive after a fall
    if (is_skeleton && current_level == settings->mouse_2_level && leveldoor_open > 0) {
        goto loc_alive_guard_fall;
    }
    if (falling_guard.hitp <= 0 ||
            falling_guard.fall_y >= 33 ||
            curr_tile2 == tiles_2_spike ||
            curr_tile2 == tiles_18_chomper) {
        if (curr_tile2 == tiles_15_opener || curr_tile2 == tiles_6_closer) {
            died_on_button(charid_2_guard);
        }
        slot = get_dead_guard_slot(room_minus_1);
        if (slot != NULL) {
            // guard died
            slot->room = falling_guard.room;
            slot->direction = falling_guard.direction;
            slot->x = falling_guard.x;
            slot->row = tile_row;
            slot->color = falling_guard.color;
            // how did guard die?
            if (curr_tile2 == tiles_2_spike) {
                // a stuck spike
                if (curr_room_modif[curr_tilepos] == 0xFF) {
                    if (falling_guard.fall_y < 22) {
                        if (falling_guard.hitp > 0) {
                            slot->room = 0;
                            goto loc_alive_guard_fall;
                        } else {
                            goto loc_dead_guard_fall;
                        }
                    } else if (is_temple()) { // a stuck vertical spike
                        goto loc_dead_guard_fall;
                    }
                }
                sound_id = sound_48_spiked;
                // make spikes harmless for kid
                curr_room_modif[curr_tilepos] = 0xFF;
                slot->frame = slot->frame2 = frame_177_spiked;
                // position guard (bird has a hanging spiked frame with a different dx)
                delta_x = settings->tbl_guard_type[current_level] == guard_type_bird ? 3 - falling_guard.direction : 8;
                if (falling_guard.direction < dir_0_right) {
                    delta_x = -delta_x;
                }
                slot->x = x_bump[guard_col + ONSCREEN_COLUMN_OFFSET] + CHAR_10_PX_OFFSET + delta_x;
            } else {
loc_dead_guard_fall:
                // skeleton does not have a dead frame
                if (is_skeleton) {
                    slot->direction = dir_56_none;
                    sound_id = sound_44_skel_alive;
                    switch (curr_tile2) {
                        case tiles_1_floor:
                        case tiles_5_stuck:
                        case tiles_6_closer:
                        case tiles_10_potion:
                        case tiles_14_debris:
                        case tiles_15_opener:
                        case tiles_22_sword:
                            curr_room_tiles[curr_tilepos] = tiles_21_skeleton;
                            curr_room_modif[curr_tilepos] = 0;
                            if (falling_guard.room == drawn_room) {
                                redraw_height = 24;
                                set_redraw_full(curr_tilepos, 1);
                                set_wipe(curr_tilepos, 1);
                                set_redraw_full(curr_tilepos + 1, 1);
                                set_wipe(curr_tilepos + 1, 1);
                            }
                            // custom logic - do not play victory music for skeleton that falls down
                            if (current_level == 3 && (curr_room == 2 || curr_room == 13)) {
                                skip_victory_music = true;
                            }
                            break;
                    }
                } else {
                    if (feather_fall_timer == 0) {
                        sound_id = sound_0_fell_to_death;
                    }
                    slot->frame = slot->frame2 = frame_185_dead;
                    slot->x = x_bump[guard_col + ONSCREEN_COLUMN_OFFSET] + CHAR_10_PX_OFFSET;
                    if (falling_guard.direction == dir_0_right) {
                        slot->x += 4;
                    }
                }
            }
            slot->x += falling_guard.direction < dir_0_right ? -2 : 2; // dead guard offset
            if (slot == &dead_guard2[room_minus_1] &&
                    dead_guard1[room_minus_1].room == slot->room &&
                    dead_guard1[room_minus_1].direction == slot->direction &&
                    dead_guard1[room_minus_1].row == slot->row &&
                    slot->frame != frame_178_chomped) {
                if (dead_guard1[room_minus_1].x == slot->x) {
                    slot->x += slot->direction == dir_0_right ? 3 : -3;
                }
                if (slot->frame == frame_185_dead &&
                        dead_guard1[room_minus_1].frame == slot->frame &&
                        ABS(dead_guard1[room_minus_1].x - dead_guard2[room_minus_1].x) < 42) {
                    slot->alt_frame = 0xFF;
                }
            } else if ((slot != &dead_guard1[room_minus_1] && slot != &dead_guard2[room_minus_1]) &&
                       (dead_guard1[room_minus_1].room == slot->room || dead_guard2[room_minus_1].room == slot->room) &&
                       (dead_guard1[room_minus_1].direction == slot->direction || dead_guard2[room_minus_1].direction == slot->direction) &&
                       (dead_guard1[room_minus_1].row == slot->row  || dead_guard2[room_minus_1].row == slot->row ) &&
                       (dead_guard1[room_minus_1].x == slot->x || dead_guard2[room_minus_1].x == slot->x) &&
                       slot->frame != frame_178_chomped) {
                slot->x += slot->direction == dir_0_right ? 5 : -5;
            } else if ((slot != &dead_guard1[room_minus_1] && slot != &dead_guard3[room_minus_1]) &&
                       (dead_guard1[room_minus_1].room == slot->room || dead_guard3[room_minus_1].room == slot->room) &&
                       (dead_guard1[room_minus_1].direction == slot->direction || dead_guard3[room_minus_1].direction == slot->direction) &&
                       (dead_guard1[room_minus_1].row == slot->row  || dead_guard3[room_minus_1].row == slot->row ) &&
                       (dead_guard1[room_minus_1].x == slot->x || dead_guard3[room_minus_1].x == slot->x) &&
                       slot->frame != frame_178_chomped) {
                slot->x += slot->direction == dir_0_right ? -3 : 3;
            }
            leave_outbound_guard(room_minus_1, slot->frame, slot->direction, slot->x, slot->row, slot->alt_frame, slot->color);
            do_knock(slot->room, slot->row, guard_col);
            if (!skip_victory_music &&
                    (Guard.alive >= 0 || Guard.direction == dir_56_none || can_guard_see_kid == 1) &&
                    (inactive_guard[Kid.room - 1].room != Kid.room ||
                     inactive_guard[Kid.room - 1].direction == dir_56_none ||
                     can_inactive_guard_see_kid < 2)) {
                stop_ambient_music();
                delayed_sound_timer = music_suspended_timer = 2;
                delayed_sound = (struct delayed_sound_type) {
                    sound_37_victory, 0, 0, 0, 0, 0, 0
                };
            }
        }
    } else {
loc_alive_guard_fall:
        // guard is alive
        if (feather_fall_timer == 0 && falling_guard.fall_y >= 11) {
            sound_id = sound_17_soft_land;
        }
        // guard is hurt
        if (falling_guard.fall_y >= 22 && !is_skeleton) {
            delayed_sound_timer = 1;
            delayed_sound = (struct delayed_sound_type) {
                sound_12_guard_hurt, falling_guard.room, guard_col, tile_row, 0, 0, 0
            };
            // last skeleton on level 12 fell down; play victory sound
        } else if (current_level == 12 &&
                   is_skeleton &&
                   Kid.room == 13 &&
                   Kid.alive < 0 &&
                   Kid.action != actions_3_in_midair &&
                   Kid.action != actions_4_in_freefall) {
            delayed_sound_timer = 2;
            delayed_sound = (struct delayed_sound_type) {
                sound_37_victory, 0, 0, 0, 0, 0, 0
            };
        }
        if (curr_tile2 == tiles_15_opener || curr_tile2 == tiles_6_closer) {
            trigger_button(1, curr_tile2, curr_room_modif[curr_tilepos]);
            if (curr_tile2 == tiles_15_opener) {
                // 0 = default. Stored value is incremented by 1.
                level.opener_col[room_minus_1] = guard_col + 1;
                level.opener_row[room_minus_1] = tile_row + 1;
            }
        }
        if ((roomleave_result == room_dir_left ||
                (roomleave_result == -1 && Kid.curr_col > 5)) &&
                Kid.room == level.roomlinks[room_minus_1].left &&
                !has_active_guard(Kid.room - 1) &&
                Kid.curr_row == tile_row &&
                falling_guard.x <= 130) {
            room_minus_1 = Kid.room - 1;
            guard_x = falling_guard.x > 90 ? 231 : (falling_guard.x + 140); // follow_guard logic
            guard_dir = dir_FF_left;
            follow_kid = true;
        } else if ((roomleave_result == room_dir_right ||
                    (roomleave_result == -1 && Kid.curr_col < 4)) &&
                   Kid.room == level.roomlinks[room_minus_1].right &&
                   !has_active_guard(Kid.room - 1) &&
                   Kid.curr_row == tile_row &&
                   falling_guard.x >= 125) {
            room_minus_1 = Kid.room - 1;
            guard_x = falling_guard.x < 165 ? 25 : (falling_guard.x - 140);  // follow_guard logic
            guard_dir = dir_0_right;
            follow_kid = true;
        } else if (guard_col < 0 && level.roomlinks[room_minus_1].left) {
            room_minus_1 = level.roomlinks[room_minus_1].left - 1;
            guard_x = falling_guard.x + 140;
            guard_dir = falling_guard.direction;
        } else if (guard_col > 9 && level.roomlinks[room_minus_1].right) {
            room_minus_1 = level.roomlinks[room_minus_1].right - 1;
            guard_x = falling_guard.x - 140;
            guard_dir = falling_guard.direction;
        } else {
            guard_x = falling_guard.x;
            guard_dir = falling_guard.direction;
        }
        if (level.guards_tile[room_minus_1] < MAX_TILES_PER_ROOM && level.guards_dir[room_minus_1] != dir_56_none) {
            if (level.guards_seq_hi[room_minus_1] == 0) {
                inactivate_guard(room_minus_1);
            } else {
                move_dead_guard_to_dead_slot(room_minus_1);
            }
        }
        level.guards_x[room_minus_1] = guard_x;
        level.guards_tile[room_minus_1] = get_tilepos(0, tile_row);
        level.guards_dir[room_minus_1] = guard_dir;
        level.guards_skill[room_minus_1] = falling_guard.skill;
        level.guards_color[room_minus_1] = falling_guard.color;
        if (falling_guard.hitp < 16) // can remember 1..15 hp
            level.guards_color[room_minus_1] |= (falling_guard.hitp << 4);
        level.guards_seq_hi[room_minus_1] = 0;
        // face the room where kid is at
        if (falling_guard.direction == dir_FF_left && Kid.room == level.roomlinks[room_minus_1].right) {
            level.guards_x[room_minus_1] += 13;
            level.guards_dir[room_minus_1] = dir_0_right;
        } else if (falling_guard.direction == dir_0_right && Kid.room == level.roomlinks[room_minus_1].left) {
            level.guards_x[room_minus_1] -= 13;
            level.guards_dir[room_minus_1] = dir_FF_left;
        }
    }
    do_knock(room_minus_1 + 1, tile_row, guard_col);
    if (sound_id >= 0 && !is_disappearing_guard() && Kid.alive < 0) {
        stop_sounds();
        set_sound_pos(falling_guard.room, guard_col, tile_row);
        play_sound(sound_id);
    }
    if (drawn_room == room_minus_1 + 1 || follow_kid) {
        enter_guard();
        guard_refrac = 6;
        if (!follow_kid) {
            droppedout = 0;
        }
    } else if (drawn_room != room_minus_1 + 1 &&
               inactive_guard[drawn_room - 1].room == drawn_room &&
               level.guards_seq_hi[drawn_room - 1] == 0) {
        leave_guard();
        draw_guard_hp(0, MAX_GUARD_HP);
        enter_guard();
        make_guard_notice(false);
    }
    if (Guard.alive < 0 && Guard.direction != dir_56_none) {
        char_type old_char = Char;
        loadshad();
        load_fram_det_col();
        if (get_tile_at_char() == tiles_20_wall) {
            if (Char.sword == sword_2_drawn) {
                in_wall_engarde();
            } else {
                in_wall();
            }
            saveshad();
        }
        if (Char.charid != old_char.charid) {
            Char = old_char;
        }
    }
    music_suspended = 0;
    falling_guard = (struct falling_guard_type) {
        0
    };
}

// custom logic - outbound guard is falling
void do_guard_fall(void) {
    word room_below, drawn_room_m1;
    int guard_tile, guard_row;
    bool is_skeleton;
    int delta_x;
    drawn_room_m1 = drawn_room - 1;
    is_skeleton = settings->tbl_guard_type[current_level] == guard_type_skeleton;
    // enter dynamic guards
    if ((Guard.alive >= 0 || Guard.direction == dir_56_none) &&
            dynamic_guard.room == drawn_room &&
            (inactive_guard[drawn_room_m1].room == 0 || !can_inactive_guard_see_kid) &&
            Kid.room == drawn_room &&
            (dynamic_guard_count[drawn_room_m1] == 0 ||
             (dynamic_guard_count[drawn_room_m1] == 1 && dynamic_guard.frame > 0))) {
        if (!is_dynamic_guard_fight) {
            is_dynamic_guard_fight = 2;
        }
        enter_dynamic_guard();
    }
    // only guards scream here
    if (Guard.charid == charid_2_guard &&
            falling_guard.fall_y >= 31 &&
            falling_guard.hitp > 0 &&
            !is_disappearing_guard() &&
            is_screaming == 0) {
        set_sound_pos(falling_guard.room, get_tile_div_mod_m7(falling_guard.x), y_to_row_mod4(falling_guard.y));
        if (is_positional_audio && current_level == settings->fat_guard_level && falling_guard.color == 0) {
            play_sound(sound_81_fat_guard_falling); // fat guard screaming differently
        } else {
            play_sound(sound_42_guard_falling); // guard screaming
        }
        is_screaming = 2;
    }
    if (falling_guard.y >= OUT_OF_BOUNDS_Y) {
        room_below = level.roomlinks[falling_guard.room - 1].down;
        falling_guard.room = room_below;
        if (room_below == 0) {
            on_guard_killed();
            return;
        }
        falling_guard.y -= 189;
    }
    // increase fall speed
    if (feather_fall_timer) {
        ++falling_guard.fall_y;
        if (falling_guard.fall_y > 4) falling_guard.fall_y = 4;
    } else {
        falling_guard.fall_y += 3;
        if (falling_guard.fall_y > 33) falling_guard.fall_y = 33;
    }
loc_guard_fall:
    guard_row = falling_guard.y <= 191 ? y_to_row_mod4(falling_guard.y) : 3;
    short x_dx_weight = 5; // simulates frame 106 (fall)
    if (falling_guard.hitp > 0 && falling_guard.fall_y < 33) {
        x_dx_weight = 9; // simulates frame 166 (stand with sword)
    }
    if (falling_guard.direction == dir_0_right) {
        x_dx_weight = -x_dx_weight;
    }
    guard_tile = get_tile(falling_guard.room, get_tile_div_mod_m7(falling_guard.x + x_dx_weight), guard_row);
    if (guard_row < 0 && (tile_is_floor(guard_tile) || guard_tile == tiles_20_wall)) {
        falling_guard.y += falling_guard.fall_y;
        goto loc_guard_fall;
    }
    // falling guard makes loose tiles fall
    if (guard_tile == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0) {
        guard_knocked_tile = 1;
        make_loose_fall(MAX(settings->loose_floor_delay - 1, 0));
        guard_tile = tiles_0_empty;
        if (falling_guard.fall_y > 11) {
            if (falling_guard.fall_y >= 33) {
                falling_guard.hitp = 0;
            } else if (falling_guard.fall_y >= 22 && falling_guard.hitp > 0) {
                --falling_guard.hitp;
            }
            falling_guard.fall_y = 11;
        }
        set_sound_pos(curr_room, tile_col, tile_row);
        if (falling_guard.fall_y >= 33) {
            play_sound(sound_0_fell_to_death);
        } else {
            play_sound(sound_17_soft_land);
        }
        // in wall
    } else if (guard_tile == tiles_20_wall) {
        delta_x = distance_to_edge(falling_guard.x);
        if (delta_x >= 8 || get_tile_infrontof_falling_guard() == tiles_20_wall) {
            delta_x = 4 - delta_x;
        } else {
            delta_x += 6;
        }
        if (falling_guard.direction < dir_0_right) {
            delta_x = -delta_x;
        }
        falling_guard.x += delta_x;
        guard_tile = get_tile(falling_guard.room, get_tile_div_mod_m7(falling_guard.x), guard_row);
    }
    if (!tile_is_floor(guard_tile) || y_land[guard_row + ONSCREEN_ROW_OFFSET] - falling_guard.y > falling_guard.fall_y) {
        // guard keeps falling
        falling_guard.y += falling_guard.fall_y;
    } else {
        // guard starts falling
        if ((falling_guard.start_fall != 0 || falling_guard.fall_y < 4) && tile_is_floor(guard_tile)) {
            delta_x = distance_to_edge(falling_guard.x + dir_behind[falling_guard.direction + 1] * 5);
            if (!tile_is_floor(get_tile_behind_falling_guard())) {
                falling_guard.x += (14 - delta_x) * dir_behind[falling_guard.direction + 1];
            } else if (!tile_is_floor(get_tile_infrontof_falling_guard())) {
                falling_guard.x += (1 + delta_x) * dir_front[falling_guard.direction + 1];
            }
            falling_guard.start_fall = 0;
            if (falling_guard.fall_y < 4) {
                falling_guard.fall_y = 4;
            }
            goto loc_guard_fall;
        } else {
            // guard lands
            fallen_guard_land(is_skeleton);
        }
    }
    if (falling_guard.start_fall != 0) {
        falling_guard.start_fall = 0;
    }
}

// seg005:0090
void land(void) {
    word seq_id;
    byte landed_on_loose;
    landed_on_loose = 0;
    // custom logic - stop screaming
    if (is_screaming && Char.fall_y >= 31) {
        stop_sounds();
    }
    is_screaming = 0;
    // custom logic - reset super jump fall flag
    super_jump_fall = 0;
    Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
    if (get_tile_at_char() != tiles_2_spike) {
        if (!tile_is_floor(get_tile_infrontof_char()) &&
                distance_to_edge_weight() < 3
           ) {
            Char.x = char_dx_forward(-3);
        } else {
            // A closed gate right in front of the landing spot should not behave like an open floor tile, but like a wall
            // Similar for a tapestry tile (with floor)
            get_tile_infrontof_char();
            if (Char.direction == dir_FF_left && (
                        (curr_tile2 == tiles_4_gate && can_bump_into_gate()) ||
                        (curr_tile2 == tiles_7_doortop_with_floor))
                    && distance_to_edge_weight() < 3
               ) {
                Char.x = char_dx_forward(-3);
            }
        }
        start_chompers();
    } else {
        // fell on spikes
        goto loc_spiked;
    }
    if (Char.alive < 0) {
        // alive
        if ((distance_to_edge_weight() >= 12 &&
                get_tile_behind_char() == tiles_2_spike) ||
                get_tile_at_char() == tiles_2_spike
           ) {
            // fell on spikes
loc_spiked:
            if (is_spike_harmful()) {
                hard_jump_fall = 0;
                spiked();
                return;
            } else if (curr_room_modif[curr_tilepos] == 0 &&
                       Char.charid != charid_4_skeleton && // custom logic - skeletons do not get spiked here
                       (Char.charid != charid_0_kid || !is_kid_shadow)) { // custom logic - kid as shadow cannot get spiked
                // spikes ARE dangerous, just not out yet.
                hard_jump_fall = 0;
                spiked();
                return;
            }
        }
        // custom logic - shadow and skeletons do not die
        if (Char.fall_y < 22 || Char.charid == charid_4_skeleton || (Char.charid == charid_0_kid && is_kid_shadow)) {
            // fell 1 row
            // custom logic - hard land after fall takes away a hit point
            if (hard_jump_fall == 2 &&
                    !feather_fall_timer &&
                    Char.charid == charid_0_kid &&
                    !is_kid_shadow) {
                goto loc_hard_land;
            }
loc_safe_land:
            if (!(key_states[SDL_SCANCODE_LSHIFT] & keystate_held) &&
                    !(key_states[SDL_SCANCODE_RSHIFT] & keystate_held) &&
                    !(joy_button_states[joyinput_x] & keystate_held)) {
                ctrl1_shift2 = control_1_ignore; // prevent restoring control_shift2 to control_m1_held in rest_ctrl_1()
            }
            // custom logic - dead guard should not trigger land sequence unless in feather fall mode
            if (Char.charid == charid_2_guard && guardhp_curr == 0 && feather_fall_timer == 0) {
                set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
                play_sound(sound_17_soft_land);
                seq_id = seq_159_guard_fell_dead;
                goto loc_fall_seq;
            }
            if (Char.charid >= charid_2_guard || Char.sword == sword_2_drawn) {
                Char.sword = sword_2_drawn;
                // custom logic - guard should land on the other side of the tile when kid is near
                if ((Char.charid == charid_2_guard || Char.charid == charid_4_skeleton) &&
                        can_guard_see_kid == 2 &&
                        (tile_is_empty(get_tile_infrontof_char()) || tile_is_empty(get_tile_infrontof2_char())) &&
                        tile_is_floor(get_tile_behind_char())) {
                    Char.x = char_dx_forward(-7);
                }
                // custom logic - guard turns in the direction of kid when he is down below
                if ((Char.charid == charid_2_guard || Char.charid == charid_4_skeleton) &&
                        Kid.room == Char.room &&
                        (Kid.curr_row > Char.curr_row ||
                         (Kid.curr_row == Char.curr_row && Kid.sword == sword_0_sheathed))
                   ) {
                    sbyte orig_direction = Char.direction;
                    Char.direction = Kid.x < Char.x ? dir_FF_left : dir_0_right;
                    if (orig_direction != Char.direction && can_guard_see_kid < 2) {
                        guard_refrac = 6;
                    }
                }
                seq_id = seq_63_guard_active_after_fall; // stand active after landing
            } else {
                seq_id = seq_17_soft_land; // crouch (soft land)
            }
            if (feather_fall_timer == 0) {
                // custom logic - do not land at the very edge
                check_on_edge();
                if (Char.charid != charid_4_skeleton && Char.fall_y >= 11) {
                    set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
                    play_sound(sound_17_soft_land);
                }
            }
            if (Char.charid == charid_0_kid) {
                make_guard_notice(true);
                // custom logic - skeleton waits a little bit after kid lands
                if (current_level == 3 &&
                        Char.room == 3 &&
                        Char.curr_row == 1 &&
                        Char.direction != Opp.direction &&
                        Opp.sword == sword_2_drawn) {
                    guard_refrac = 12;
                }
            }
        } else if (Char.fall_y < 33) {
loc_hard_land:
            // fell 2 rows
            if (Char.charid == charid_1_shadow || Char.charid == charid_4_skeleton) {
                goto loc_safe_land;
            }
            // custom logic - guard can fall 2 rows and lose a life
            if (Char.charid == charid_2_guard) {
                if (take_hp(1) || (current_level == 1 && Char.room == 1) /* first room */) {
                    goto loc_dead_land;
                }
                set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
                play_sound(sound_17_soft_land);
                // custom logic - medium land is a soft land sound followed by hurt sound
                delayed_sound_timer = 1;
                delayed_sound = (struct delayed_sound_type) {
                    sound_12_guard_hurt, Char.room, Char.curr_col, Char.curr_row, 0, 0, 0
                };
                Char.sword = sword_2_drawn;
                seq_id = seq_74_hit_by_sword; // stand active after landing
                // kid
            } else if ((get_tile_at_char() == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0) ||
                       !take_hp(1)) {
                // still alive
                set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
                play_sound(sound_17_soft_land); // medium land
                // custom logic - medium land is a soft land sound followed by hurt sound
                if (curr_tile2 != tiles_11_loose || get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 1) {
                    if (!is_kid_shadow) {
                        delayed_sound_timer = 2;
                        delayed_sound = (struct delayed_sound_type) {
                            sound_13_kid_hurt, 0, 0, 0, 0, 0, 0
                        };
                    }
                } else {
                    make_loose_fall(settings->loose_floor_delay - 1);
                    landed_on_loose = 1;
                }
                make_guard_notice(true);
                seq_id = seq_20_medium_land; // medium land (lose 1 HP, crouch)
            } else {
                // dead (this was the last HP)
                goto loc_crushed_land;
            }
        } else {
            // custom logic - kid can survive falling an extra row on level 6
            if (current_level == 6 &&
                    Char.charid == charid_0_kid &&
                    Char.fall_y < 44 &&
                    !take_hp(1)) {
                set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
                // landed on fat guard
                if (Char.room == 17 && dead_guard3[16].room == 17) {
                    stop_sounds();
                    play_sound(sound_8_bumped);
                    seq_id = seq_22_crushed;
                    Char.y -= 3;
                    Char.x += 3;
                    dead_guard3[16].frame2 = 192; // squashed frame
                    delayed_sound_timer = 22;
                    dead_kid_timer = 24;
                } else {
                    play_sound(sound_17_soft_land); // medium land
                    make_guard_notice(true);
                    delayed_sound_timer = 1;
                    seq_id = seq_20_medium_land;
                }
                // medium land is a soft land sound followed by hurt sound
                if (!is_kid_shadow) {
                    delayed_sound = (struct delayed_sound_type) {
                        sound_13_kid_hurt, 0, 0, 0, 0, 0, 0
                    };
                } else {
                    delayed_sound_timer = 0;
                }
                // start music when kid lands
                music_suspended_timer = 0;
            } else {
                // fell 3 or more rows
                goto loc_dead_land;
            }
        }
    } else {
        // dead
loc_dead_land:
        take_hp(ALL_HP);
loc_crushed_land:
        if (feather_fall_timer == 0) {
            set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
            play_sound(sound_0_fell_to_death); // crashing into the floor
            make_guard_notice(true);
        }
        seq_id = seq_22_crushed; // dead (after falling)
    }
loc_fall_seq:
    seqtbl_offset_char(seq_id);
    play_seq();
    // custom logic - a stuck loose tile gets unstuck when kid/guard lands from 2+ rows above
    if (feather_fall_timer == 0 &&
            (Char.charid != charid_0_kid || !is_kid_shadow) &&
            Char.fall_y >= 22 &&
            get_tile_at_char() == tiles_11_loose &&
            get_tile_ctrl_bits(curr_room, curr_tilepos, 3) == 1) {
        curr_room_tiles[curr_tilepos] &= 0x1F;
        make_loose_fall(4);
    }
    Char.fall_x = 0;
    Char.fall_y = 0;
    // custom logic - reset hard jump fall flag
    hard_jump_fall = landed_on_loose ? 2 : 0;
    // custom logic - reset "holding sword" variable when there is no guard nearby
    if (Char.charid == charid_0_kid && Char.alive < 0 && hitp_curr > 0 && can_guard_see_kid < 2) {
        holding_sword = 0;
    }
    // custom logic - set inactive guard fall timer if there is a dynamic guard on another row
    word room_minus_1 = drawn_room - 1;
    sbyte kid_row;
    if (Char.curr_row == Kid.curr_row && Kid.frame >= frame_135_climbing_1 && Kid.frame <= frame_140_climbing_6) {
        kid_row = Kid.curr_row - 1;
    } else {
        kid_row = Kid.curr_row;
    }
    if (guardhp_curr > 0 &&
            Char.charid == charid_2_guard &&
            Char.alive < 0 &&
            Char.room == drawn_room &&
            Char.curr_row >= 0 &&
            Char.curr_row < 3 &&
            Char.curr_row != kid_row &&
            inactive_guard[room_minus_1].room == 0 &&
            dynamic_guard.room == drawn_room &&
            dynamic_guard.row == kid_row &&
            (dynamic_guard_count[room_minus_1] == 0 ||
             (dynamic_guard_count[room_minus_1] == 1 && dynamic_guard.frame > 0))) {
        inactive_guard_fall_timer = 12;
    }
    // custom logic - horse statue makes a neighing sound when prince appears in the same room
    if (seq_id == sound_17_soft_land || seq_id == seq_20_medium_land) {
        horse_neigh(seq_id == seq_20_medium_land);
    }
}

// seg005:01B7
void spiked(void) {
    // custom logic - is guard a purple wierdo?
    bool disappearing_guard = Char.charid == charid_2_guard && is_disappearing_guard();
    // custom logic - is guard a bird?
    bool bird_guard = Char.charid == charid_2_guard && settings->tbl_guard_type[current_level] == guard_type_bird;
    // custom logic - save data to restore spikes if kid resurrects
    if (Char.charid == charid_0_kid && curr_room_modif[curr_tilepos] != 0xFF) {
        dead_kid_tile = (struct stuck_button_type) {
            curr_tilepos, curr_room, curr_room_tiles[curr_tilepos], 0x8F /* fully out before retracting */
        };
    }
    // If someone falls into spikes (except purple weirdo and Jaffar's vision), those spikes become stuck.
    if (Char.charid != charid_2_guard || !disappearing_guard) {
        curr_room_modif[curr_tilepos] = 0xFF;
        // custom logic - spikes may need to be redrawn when falling from the room above to prevent missing spikes
        set_redraw_full(curr_tilepos, 1);
        set_redraw_full(curr_tilepos + 1, 1);
    }
    Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
    // custom logic - place spiked kid higher in the temple when climbing down
    if (is_temple() && Char.direction == dir_FF_left && Char.frame == frame_91_hanging_5) {
        Char.y -= 4;
    }
    // custom logic - recalculate offscreen spike column
    int spike_col = get_tile_col_at_char();
    Char.x = x_bump[spike_col + ONSCREEN_COLUMN_OFFSET] + CHAR_10_PX_OFFSET;
    // custom logic - do not bump disappearing guards
    // birds have a different dx offset
    if (bird_guard) {
        Char.x = char_dx_forward(1 - Char.direction);
    } else if (!disappearing_guard) {
        Char.x = char_dx_forward(8);
    }
    Char.fall_y = 0;
    // custom logic - do not play spiked sound for guard in room 1 on level 1 from here
    if (!is_positional_audio || current_level != 1 || Char.room != 1 || Char.charid != charid_2_guard) {
        if (Char.charid != charid_0_kid) {
            stop_ambient_music();
            music_suspended_timer = 2;
        }
        set_sound_pos(Char.room, spike_col, Char.curr_row);
        play_sound(sound_48_spiked); // something spiked
    }
    guard_notice_timer = 1;
    take_hp(ALL_HP);
    // custom logic - temple has vertical spikes that require a different sequence for kid
    // while birds have a different spiked frame and can use the standard sequence
    if (is_temple()) {
        if (Char.charid == charid_0_kid) {
            droppedout = 0;
            seqtbl_offset_char(seq_132_kid_spiked_hanging); // kid spiked hanging
        } else {
            seqtbl_offset_char(seq_133_bird_spiked_hanging); // bird guard spiked hanging
        }
    } else {
        seqtbl_offset_char(seq_51_spiked); // spiked
    }
    play_seq();
}

// custom logic - show kid room when he starts moving
void goto_kids_room(void) {
    if (debug_cheats_enabled && Char.charid == charid_0_kid && drawn_room != Char.room) {
        draw_kid_hp(hitp_curr, hitp_max);
        draw_guard_hp(guardhp_curr, guardhp_max);
        next_room = Char.room;
    }
}

// custom logic - control shadow separation based on the turn sequence count
#define SHADOW_SEPARATION_COUNT 4
void control_separate_from_shadow(void) {
    sbyte can_separate;
    int hp = 1;
    can_separate = can_kid_separate_from_shadow();
    if (can_separate == 1) {
        ++separate_from_shadow_counter;
        if (separate_from_shadow_counter >= SHADOW_SEPARATION_COUNT) {
            united_with_shadow = 3;
            if (separate_from_shadow_counter == SHADOW_SEPARATION_COUNT * 2) {
                hp = 2;
                if (hitp_curr > hp) {
                    separate_kid_from_shadow();
                }
                separate_from_shadow_counter = 0;
            }
            play_sound(sound_61_shadow_separation);
            if (take_hp(hp)) {
                return;
            }
        }
    }
}

// custom logic - is starting level 6 sequence playing?
bool is_level_6_sequence(void) {
    return current_level == 6 && drawn_room == level.start_room && level_6_sequence_timer;
}

// custom logic - auto-control kid during the drinking entry on level 6
bool control_level_6_intro(void) {
    short curr_seq = 0;
    short seq_index;
    if (level_6_sequence_timer > 0) {
        --level_6_sequence_timer;
        control_forward = control_backward = control_up = control_down = release_arrows();
        if (level_6_sequence_timer > 0) {
            short moves_num = sizeof(settings->level_6_intro_seq) / sizeof(auto_move_type);
            for (seq_index = 0; seq_index < moves_num; ++seq_index) {
                if (settings->level_6_intro_seq[seq_index].time == level_6_sequence_timer) {
                    curr_seq = settings->level_6_intro_seq[seq_index].move;
                    break;
                }
            }
            switch (curr_seq) {
                case -3: // pick up the potion
                    check_get_item();
                    break;
                case -2: // draw the sword
                    if (had_sword) {
                        play_sound(sound_19_draw_sword);
                    }
                    break;
                case -1: // a gate somewhere near by slams shut
                    if (had_sword) {
                        set_sound_pos(Kid.room, Kid.curr_col + 10, Kid.curr_row - 10);
                        play_sound(sound_6_gate_closing_fast);
                    }
                    break;
                case 0: // no-op
                    break;
                default: // play the kid sequence
                    seqtbl_offset_char(curr_seq);
                    break;
            }
        } else {
            level_6_sequence_timer = -1;
        }
        return true;
    } else if (potion_timer > 0 && potion_timer < 63) {
        control_shift = control_forward = control_backward = control_up = control_down = release_arrows();
        if (had_sword) {
            if (potion_timer == 62) {
                Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
                Char.fall_y = 0;
            }
            short moves_num = sizeof(settings->level_6_drink_seq) / sizeof(auto_move_type);
            for (seq_index = 0; seq_index < moves_num; ++seq_index) {
                if (settings->level_6_drink_seq[seq_index].time == potion_timer) {
                    curr_seq = settings->level_6_drink_seq[seq_index].move;
                    break;
                }
            }
            if (curr_seq != 0) {
                seqtbl_offset_char(curr_seq);
            }
        }
        return true;
    }
    return false;
}

// custom logic - controlling the mouse character
void control_mouse(short char_frame) {
    if (control_up < control_0_released && char_frame != frame_188_mouse_stand) {
        control_up = control_forward = control_backward = release_arrows();
        seqtbl_offset_char(seq_114_mouse_stand);
        if (is_positional_audio && prandom(2) == 1) {
            set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
            play_sound(sound_70_mouse_squeak);
            make_guard_notice(false);
        }
    } else if (control_down < control_0_released && char_frame == frame_188_mouse_stand) {
        control_down = release_arrows();
        seqtbl_offset_char(seq_106_mouse);
    } else if (control_forward < control_0_released) {
        control_forward = control_0_released;
        if (Char.fall_x < MOUSE_HIT_SPEED) {
            ++Char.fall_x;
        }
        if (Char.curr_seq < seqtbl_offsets[seq_160_mouse_forward_manual]) {
            seqtbl_offset_char(seq_160_mouse_forward_manual);
        } else {
            return;
        }
    } else if (control_backward < control_0_released) {
        release_arrows();
        if (char_frame == frame_188_mouse_stand) {
            control_backward = control_1_ignore;
        }
        Char.x = char_dx_forward(-7);
        Char.fall_x = 0;
        Char.direction = ~Char.direction;
        return;
    } else if (char_frame != frame_188_mouse_stand) {
        release_arrows();
        Char.fall_x = 0;
        seqtbl_offset_char(seq_106_mouse);
    }
    play_seq();
}

// seg005:0213
void control(void) {
    short char_frame;
    short char_action;
    char_frame = Char.frame;
    // custom logic - auto-control kid during the drinking entry on level 6
    if (Char.charid == charid_0_kid &&
            current_level == 6 &&
            Char.room == level.start_room &&
            control_level_6_intro()) {
        return;
    }
    // custom logic - immobilize kid during the princess vision on level 6
    if (Char.charid == charid_0_kid &&
            current_level == 6 &&
            Char.room == 1 &&
            Char.x <= 85 &&
            dead_guard3[Char.room - 1].room == Char.room &&
            dead_guard3[Char.room - 1].direction != dir_56_none) {
        if (ogg_fore_sound_playing || check_sound_playing()) {
            release_arrows();
        } else {
            Char.x = char_dx_forward(-14);
            show_black_screen(6, 1);
            text_time_remaining = 48;
            text_time_total = 48;
            display_text_bottom("FIND THE WAY TO OTHER SIDE");
        }
        if (Char.sword != sword_2_drawn) {
            seqtbl_offset_char(seq_2_stand);
        }
        return;
    }
    if (Kid.direction == dir_56_none) {
        return;
    }
    if (Char.alive >= 0) {
        // custom logic - shadow disappears when it dies
        if (Char.charid == charid_0_kid && is_kid_shadow) {
            release_arrows();
            return;
        }
        // custom logic - standing death from any frame in room 15, level 5 (large spilled potion)
        bool room_15_level_5 = current_level == 5 && Kid.room == 15 && took_antidote != 2;
        // custom logic - standing death from any frame in room 1, level 12B, top row (nowhere to run)
        bool room_5_level_13 = current_level == 13 && Kid.room == 1 && Kid.curr_row == 0 && leveldoor_open == 0;
        // kid dies at any frame except for collapsing/dying/falling to death if running or jumping
        if (Char.charid == charid_0_kid &&
                (room_15_level_5 || room_5_level_13) &&
                tile_is_floor(get_tile_at_char()) &&
                Char.action == actions_1_run_jump &&
                (char_frame < frame_177_spiked || char_frame > frame_185_dead)) {
            control_forward = control_backward = control_up = control_down = release_arrows();
            Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
            seqtbl_offset_char(seq_71_dying); // special event: dying
            return;
        }
        if (char_frame == frame_15_stand || // stand
                char_frame == frame_166_stand_inactive || // stand
                char_frame == frame_158_stand_with_sword || // stand with sword
                char_frame == frame_171_stand_with_sword || // stand with sword
                char_frame == frame_109_crouch || // custom logic - crouch (Shift-K handling)
                (Char.charid == charid_0_kid && char_frame >= frame_186_mouse_1 && char_frame <= frame_188_mouse_stand) // custom logic - kid died as mouse
           ) {
            seqtbl_offset_char(seq_71_dying); // dying (not stabbed)
        }
    } else {
        char_action = Char.action;
        // custom logic - try to unite with shadow when down key is pressed
        if (Char.charid == charid_0_kid && is_kid_shadow && control_y == control_1_held_down && unite_shadow_with_kid()) {
            return;
            // custom logic - control shadow separation based on the turn sequence count
        } else if (Char.charid == charid_0_kid &&
                   !is_kid_shadow &&
                   char_frame == frame_45_turn &&
                   Char.curr_seq >= seqtbl_offsets[seq_5_turn] &&
                   Char.curr_seq < seqtbl_offsets[seq_6_run_turn]) {
            control_separate_from_shadow();
        }
        // custom logic - continue running after sprinting through a chomper
        // if shift got released and the forward arrow is pressed
        if (control_forward < control_0_released &&
                control_shift > control_m1_held &&
                char_frame >= frame_6_start_run &&
                Char.curr_seq >= seqtbl_offsets[seq_122_chomper_sprint] &&
                Char.curr_seq < seqtbl_offsets[seq_123_bird_bow]) {
            seqtbl_offset_char(seq_84_run);
            play_seq();
        }
        if (char_action == actions_5_bumped || char_action == actions_4_in_freefall) {
            release_arrows();
            // custom logic - kid sitting in a cell
        } else if (char_action == actions_8_sit) {
            if (control_up < control_0_released) {
                control_up = control_1_ignore;
                Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
                seqtbl_offset_char(seq_49_stand_up_from_crouch);
            } else {
                release_arrows();
            }
            // custom logic - controlling the mouse character
        } else if (char_frame >= frame_186_mouse_1 && char_frame <= frame_188_mouse_stand) {
            control_mouse(char_frame);
        } else if (Char.sword == sword_2_drawn) {
            if (control_forward < control_0_released || control_backward < control_0_released) {
                goto_kids_room();
            }
            control_with_sword();
        } else if (Char.charid >= charid_2_guard) {
            control_guard_inactive();
        } else if (char_frame == frame_15_stand || // standing
                   (char_frame >= frame_50_turn && char_frame < frame_53_runturn) // end of turning
                  ) {
            if (control_forward < control_0_released || control_backward < control_0_released) {
                goto_kids_room();
            }
            control_standing();
        } else if (char_frame == frame_48_turn) { // a frame in turning
            goto_kids_room();
            control_turning();
        } else if (char_frame < frame_4_start_run) { // start run
            goto_kids_room();
            control_startrun();
        } else if (char_frame >= frame_67_start_jump_up_1 && char_frame < frame_70_jumphang) { // start jump up
            goto_kids_room();
            control_jumpup();
        } else if (char_frame < frame_15_stand) { // running
            goto_kids_room();
            control_running();
            // custom logic - stop quicker during run turn
        } else if (char_frame >= frame_63_runturn && char_frame <= frame_65_runturn && control_x == control_0_released) {
            release_arrows();
            Char.x = char_dx_forward(-12);
            Char.direction = ~Char.direction;
            seqtbl_offset_char(seq_13_stop_run); // stop run
        } else if (char_frame >= frame_87_hanging_1 && char_frame <= frame_99_hanging_13) { // hanging
            // custom logic - start music on level 7 when kid grabs in the beginning of the level
            if (current_level == 7 && music_suspended_timer > 0) {
                music_suspended_timer = 0;
            }
            control_hanging();
        } else if (char_frame == frame_109_crouch) { // crouching
            goto_kids_room();
            control_crouched();
        }
        // When ducking with down+forward, give time to release the forward control (prevents unintended crouch-hops)
        else if (Char.curr_seq >= seqtbl_offsets[seq_50_crouch] &&
                 Char.curr_seq < seqtbl_offsets[seq_49_stand_up_from_crouch]) { // while stooping
            if (control_forward < control_1_ignore) control_forward = control_0_released;
        }
        if (char_frame >= frame_191_drink && char_frame <= frame_205_drink) {
            release_arrows();
        }
        if (Char.curr_seq >= seqtbl_offsets[seq_92_put_sword_away] &&
                Char.curr_seq < seqtbl_offsets[seq_93_put_sword_away_fast]) {
            release_arrows();
        }
    }
}

// seg005:02EB
void control_crouched(void) {
    if (need_level1_music == 1 && current_level == 1 && Char.room == 1) {
        // Special event: music when crouching
        if (!check_sound_playing() && !ogg_fore_sound_playing) {
            if (need_level1_music == 1) {
                play_sound(sound_25_presentation); // presentation (level 1 start)
                need_level1_music = 2;
                // custom logic - start music on level 1 after guard's death
                if (music_suspended_timer > 0) {
                    music_suspended_timer = 0;
                    start_ambient_music(true);
                }
                goto loc_control_kid;
            } else {
                need_level1_music = 0;
            }
        }
    } else if (current_level == settings->mouse_2_level &&
               Char.room == settings->mouse_2_room &&
               leveldoor_open > 0 &&
               leveldoor_open < settings->mouse_2_crouch_delay) {
        return;
    } else {
        need_level1_music = 0;
loc_control_kid:
        if (control_shift2 < control_0_released && check_get_item()) {
            return;
        }
        if (control_y != control_1_held_down) {
            seqtbl_offset_char(seq_49_stand_up_from_crouch); // stand up from crouch
        } else {
            if (control_forward < control_0_released) {
                // custom logic - release shift2 when down arrow is pressed
                // to only allow picking up objects with Shift
                if (control_down && control_shift >= control_0_released) {
                    control_shift2 = ctrl1_shift2 = control_0_released;
                }
                if (control_x != control_m1_held) {
                    // custom logic - kid sits on the floor on prison level
                    if (current_level == settings->prison_level) {
                        seqtbl_offset_char(seq_126_crouch_in_cell); // sit
                    } else {
                        seqtbl_offset_char(seq_116_crawl);
                    }
                } else {
                    seqtbl_offset_char(seq_79_crouch_hop); // crouch-hop
                }
            }
        }
    }
}

// seg005:0358
void control_standing(void) {
    int tile_behind, tile_infront, wall_at, wall_infront, wall_behind;
    int distance;
    if (control_shift2 < control_0_released && control_shift < control_0_released && check_get_item()) {
        return;
    }
    // custom logic - reset ignored arrows when standing still
    reset_ignored_arrows();
    // custom logic - draw sword with Alt/Insert
    if (Char.charid == charid_0_kid &&
            // Alt/Insert can be used to start a fist fight when the guard is in the range
            (have_sword ||
             (!is_kid_shadow && can_guard_see_kid == 2) ||
             (!is_kid_shadow && droppedout && !is_guarding_gate && Opp.alive < 0 && Opp.charid == charid_2_guard && Opp.sword == sword_2_drawn)) &&
            Char.room == drawn_room &&
            control_alt < control_0_released) {
        keep_sword_drawn = true;
        // turn around if the guard is behind
        if (can_guard_see_kid == 2 && char_opp_dist() < 0) {
            if (have_sword) {
                back_pressed();
                return;
            } else {
                release_arrows();
                Char.direction = ~Char.direction;
            }
        }
        // do not draw sword between walls/obstacles
        wall_at = wall_type2(get_tile_at_char());
        wall_infront = wall_type2(get_tile_infrontof_char());
        tile_infront = curr_tile2;
        // do not treat stuck loose tiles as loose tiles
        if (tile_infront == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 1) {
            tile_infront = tiles_1_floor;
        }
        wall_behind = wall_type2(get_tile_behind_char());
        tile_behind = curr_tile2;
        // do not treat stuck loose tiles as loose tiles
        if (tile_behind == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 1) {
            tile_behind = tiles_1_floor;
        }
        // treat an abyss or a loose tile in front or behind kid as a wall
        if (wall_infront == wall_type_0_none && (!tile_is_floor(tile_infront) || tile_infront == tiles_11_loose)) {
            wall_infront = wall_type_4_wall;
        }
        if (wall_behind == wall_type_0_none && (!tile_is_floor(tile_behind) || tile_behind == tiles_11_loose)) {
            wall_behind = wall_type_4_wall;
        }
        if ((Char.direction == dir_FF_left &&
                (wall_infront == wall_type_1_right || wall_infront == wall_type_4_wall) &&
                (wall_behind == wall_type_2_left || wall_behind == wall_type_4_wall)) ||
                (Char.direction == dir_0_right &&
                 (wall_infront == wall_type_2_left || wall_infront == wall_type_4_wall) &&
                 (wall_behind == wall_type_1_right || wall_behind == wall_type_4_wall)) ||
                (Char.direction == dir_FF_left && wall_at == wall_type_1_right &&
                 (wall_infront == wall_type_1_right || wall_infront == wall_type_4_wall)) ||
                (Char.direction == dir_0_right && wall_at == wall_type_1_right &&
                 (wall_behind == wall_type_1_right || wall_behind == wall_type_4_wall))) {
            return;
        }
        distance = distance_to_edge_weight();
        // ensure kid does not get pushed behind a gate/tapestry when drawing sword when facing right
        if (Char.direction == dir_0_right && distance >= 10 && wall_behind == wall_type_1_right) {
            Char.x = char_dx_forward(distance - 9);
        }
        // pushing with a sword against a wall/gate/tapestry should push kid back
        if ((wall_infront == wall_type_4_wall ||
                (Char.direction == dir_FF_left && wall_infront == wall_type_1_right)) &&
                wall_behind == wall_type_0_none &&
                !tile_is_floor(tile_behind)) {
            Char.fall_x = -1;
        } else if ((Char.direction == dir_0_right && wall_at == wall_type_1_right) &&
                   wall_behind == wall_type_0_none &&
                   !tile_is_floor(tile_behind)) {
            Char.fall_x = -1;
        }
        if ((!tile_is_floor(get_tile_at_kid(Kid.x)) || curr_tile2 == tiles_11_loose) && distance < 7) {
            Char.x = char_dx_forward(distance - 7);
        }
        // custom logic - do not fall off an edge when manually drawing sword
        check_on_edge();
        draw_sword();
        return;
    }
    if (Char.charid != charid_0_kid && control_down < control_0_released && control_forward < control_0_released) {
        draw_sword();
        return;
    } //else
    if (have_sword) {
        if (offguard != 0 && control_shift >= control_0_released) {
            goto loc_control_forward;
        }
        // custom logic - take out sword near an inactive guard
        if (Char.charid == charid_0_kid &&
                Char.sword == sword_0_sheathed &&
                control_down >= control_0_released &&
                (control_up >= control_0_released || control_shift < control_0_released) &&
                can_guard_see_kid == 2 &&
                Opp.direction == Char.direction &&
                can_inactive_guard_see_kid == 2 &&
                inactive_guard[Char.room - 1].room == Char.room &&
                inactive_guard[Char.room - 1].row == Char.curr_row &&
                inactive_guard[Char.room - 1].direction != Char.direction &&
                ABS(inactive_guard[Char.room - 1].x - Char.x) < 90) {
            holding_sword = 1;
            keep_sword_drawn = false;
            draw_sword();
        } else if (can_guard_see_kid == 2 &&
                   (Char.charid != charid_0_kid ||
                    (control_down >= control_0_released && (control_up >= control_0_released || control_shift < control_0_released)))
                  ) {
            // custom logic - do not draw sword on level 10 after kid is united with shadow
            if (current_level == 10 && leveldoor_open >= DEAD_KID_LEVEL_10_TIMER) {
                make_guard_notice(false);
            } else {
                short kid_guard_dist = char_opp_dist();
                if (kid_guard_dist >= -10 && kid_guard_dist < 90) {
                    holding_sword = 1;
                    if ((word)kid_guard_dist < (word) - 6) {
                        // custom logic - do not keep sword drawn
                        if (Char.charid == charid_0_kid) {
                            keep_sword_drawn = false;
                        }
                        draw_sword();
                        return;
                    } else {
                        if (Opp.direction == Char.direction) {
                            back_pressed();
                        }
                        return;
                    }
                }
            }
        } else {
            offguard = 0;
        }
    }
    if (control_shift < control_0_released) {
        if (control_backward < control_0_released) {
            back_pressed();
        } else if (control_up < control_0_released) {
            up_pressed();
        } else if (control_down < control_0_released) {
            down_pressed();
        } else if (control_x < control_0_released && control_forward < control_0_released) {
            safe_step();
        }
// *INDENT-OFF*
    } else loc_control_forward: if (control_forward < control_0_released) {
        if (is_keyboard_mode && control_up < control_0_released) {
            standing_jump();
        } else {
            forward_pressed();
        }
// *INDENT-ON*
    } else if (control_backward < control_0_released) {
        back_pressed();
    } else if (control_up < control_0_released) {
        if (is_keyboard_mode && control_forward < control_0_released) {
            standing_jump();
        } else {
            up_pressed();
        }
    } else if (control_down < control_0_released) {
        down_pressed();
    } else if (control_x < control_0_released) {
        forward_pressed();
    }
}

// seg005:0482
void up_pressed(void) {
    int leveldoor_tilepos = -1;
    if (get_tile_at_char() == tiles_16_level_door_left || curr_tile2 == tiles_23_balcony_left) {
        leveldoor_tilepos = curr_tilepos;
    } else if (curr_tile2 == tiles_17_level_door_right || curr_tile2 == tiles_24_balcony_right) {
        leveldoor_tilepos = curr_tilepos - 1;
    }
    // custom logic - kid can fall off a balcony
    if (leveldoor_tilepos != -1 && (curr_room_tiles[leveldoor_tilepos] & 0x1F) == tiles_23_balcony_left && !is_kid_shadow) {
        if (get_tile_ctrl_bits(curr_room, leveldoor_tilepos, 0) == 0) { // not a portal/window
            goto loc_leveldoor_jump;
        }
        go_up_leveldoor(leveldoor_tilepos % 10, true, true);
        return;
    }
    if (leveldoor_tilepos != -1) {
        // custom logic - in some scenarios kid tries to walk into a "closed" level door
        bool is_open = level.start_room != drawn_room &&
                       (
                           current_level != settings->inverse_level
                           ? curr_room_modif[leveldoor_tilepos] >= KID_HEIGHT_DOOR_MODIFIER // this door must be fully open
                           : curr_room_modif[leveldoor_tilepos] == 0 // on level 12B level doors act inversely
                       ) &&
                       // custom logic - shadow cannot go to the next level
                       !is_kid_shadow;
        if (curr_room_modif[leveldoor_tilepos] == 0 && current_level != settings->inverse_level) {
            goto loc_leveldoor_jump;
        }
        go_up_leveldoor(leveldoor_tilepos % 10, false, is_open);
    } else {
loc_leveldoor_jump:
        if (control_x < control_0_released) {
            standing_jump();
        } else {
            check_jump_up();
        }
    }
}

// seg005:04C7
void down_pressed(void) {
    control_down = control_1_ignore; // disable automatic repeat
    if (!tile_is_floor(get_tile_infrontof_char()) &&
            distance_to_edge_weight() < 3
       ) {
        // custom logic - turning action disables collision checks in seg004.c
        if (Char.action == actions_7_turn) {
            Char.action = actions_1_run_jump;
        }
        Char.x = char_dx_forward(5);
        load_fram_det_col();
    } else {
        if (!tile_is_floor(get_tile_behind_char()) &&
                distance_to_edge_weight() >= 8
           ) {
            through_tile = get_tile_behind_char();
            get_tile_at_char();
            if (can_grab() && control_forward != control_m1_held &&
                    (Char.direction >= dir_0_right ||
                     get_tile_at_char() != tiles_4_gate ||
                     curr_room_modif[curr_tilepos] >> 2 >= 6)
               ) {
                Char.x = char_dx_forward(distance_to_edge_weight() - 9);
                seqtbl_offset_char(seq_68_climb_down); // climb down
            } else {
                crouch();
            }
        } else {
            crouch();
        }
    }
}

// seg005:0574
void go_up_leveldoor(int col, bool is_balcony, bool is_open) {
    // custom logic - do not force kid into the door/teleport if there is a grabable tile right in front
    if (tile_is_floor(get_tile_front_above_char()) && distance_to_edge_weight() < 7) {
        through_tile = (curr_room_tiles[get_tilepos(Char.curr_col, Char.curr_row - 1)] & 0x1F);
        if (can_grab()) {
            grab_up_with_floor_behind();
            return;
        }
    }
    Char.x = x_bump[col + ONSCREEN_COLUMN_OFFSET] + CHAR_10_PX_OFFSET;
    Char.direction = dir_FF_left;
    if (is_balcony) { // custom logic - kid can run off the balcony and die
        if (current_level != 6) { // level 6 balconies are teleports
            delayed_sound_timer = 7;
            delayed_sound = (struct delayed_sound_type) {
                is_positional_audio ? sound_75_falling_far : sound_1_falling, 0, 0, 0, 0, 0, 0
            };
        }
        seqtbl_offset_char(seq_115_teleport);
    } else {
        if (is_open) {
            // custom logic - stop blinking
            blink_timer = 16;
            // custom logic - suspend ambient music
            music_suspended_timer = 24;
            // remember whether the kid exited through the left door or not
            if (current_level == 2 && Kid.room == 23 && Kid.curr_row == 1) {
                level_2_right_door = Kid.curr_col >= 3 ? 1 : 0;
            } else if (level_2_right_door) {
                level_2_right_door = 0;
            }
            seqtbl_offset_char(seq_70_go_up_on_level_door); // go up on level door
        } else {
            // custom logic - bump into a closed level door
            seqtbl_offset_char(seq_128_bump_into_closed_door);
        }
    }
}

// seg005:058F
void control_turning(void) {
    if (control_shift >= control_0_released && control_x < control_0_released && control_y >= control_0_released) {
        seqtbl_offset_char(seq_43_start_run_after_turn); // start run and run (after turning)
    }

    // Added:
    // When using a joystick, the kid may sometimes jump/duck/turn unintendedly after turning around.
    // To prevent this: clear the remembered controls, so that if the stick has already moved to another/neutral position,
    // the kid will not jump, duck, or turn again.
    if (is_joyst_mode) {
        if (control_up < control_0_released && control_y >= control_0_released) {
            control_up = control_0_released;
        }
        if (control_down < control_0_released && control_y <= control_0_released) {
            control_down = control_0_released;
        }
        if (control_backward < control_0_released && control_x == control_0_released) {
            control_backward = control_0_released;
        }
    }
}

// seg005:05AD
void crouch(void) {
    // custom logic - kid sits on the floor on prison level
    if (current_level == settings->prison_level) {
        seqtbl_offset_char(seq_126_crouch_in_cell); // sit
    } else {
        seqtbl_offset_char(seq_50_crouch); // crouch
    }
    // custom logic - release shift2 when down arrow is pressed
    // to only allow picking up objects with Shift
    if (control_down && control_shift >= control_0_released) {
        control_shift2 = ctrl1_shift2 = control_0_released;
    }
    control_down = release_arrows();
}

// seg005:05BE
void back_pressed(void) {
    word seq_id;
    short distance;
    distance = char_opp_dist();
    release_arrows();
    // After turn, Kid will draw sword if ...
    if (have_sword == 0 || // if Kid has sword
            can_guard_see_kid != 2 || // and can see Guard
            distance > 0 || // and Guard was behind him
            distance_to_edge_weight() < 2
       ) {
        control_backward = control_1_ignore; // disable automatic repeat
        // custom logic - do not fall after turning at the very edge
        check_on_edge();
        seq_id = seq_5_turn; // turn
    } else {
        Char.sword = sword_2_drawn;
        // custom logic - disable automatic repeat if the distance is within striking range
        control_backward = control_0_released;
        control_shift2 = control_1_ignore;
        ctrl1_shift2 = control_1_ignore; // prevent restoring control_shift2 to control_m1_held in rest_ctrl_1()
        justturned_engarde = 1;
        offguard = 0;
        seq_id = seq_89_turn_draw_sword; // turn and draw sword
    }
    seqtbl_offset_char(seq_id);
}

// seg005:060F
void forward_pressed(void) {
    short distance;
    distance = get_edge_distance();
    if (control_down < control_0_released) {
        down_pressed();
        control_forward = control_0_released;
        return;
    }
    if (edge_type == edge_type_1_edge && curr_tile2 != tiles_18_chomper && distance < 8) {
        // If char is near a wall, step instead of run.
        if (control_forward < control_0_released) {
            safe_step();
        }
    } else {
        // custom logic - do not fall off an edge when running
        check_on_edge();
        seqtbl_offset_char(seq_1_start_run); // start run and run
    }
}

// seg005:0649
void control_running(void) {
    if (control_x == control_0_released && (Char.frame == frame_7_run || Char.frame == frame_11_run)) {
        control_forward = release_arrows();
        seqtbl_offset_char(seq_13_stop_run); // stop run
    } else if (control_x > control_0_released) {
        control_backward = release_arrows();
        seqtbl_offset_char(seq_6_run_turn); // run-turn
    } else if (control_y < control_0_released && control_up < control_0_released) {
        run_jump();
    } else if (control_down < control_0_released) {
        // custom logic - release shift2 when down arrow is pressed
        // to only allow picking up objects with Shift
        if (control_shift >= control_0_released) {
            control_shift2 = ctrl1_shift2 = control_0_released;
        }
        control_down = control_1_ignore; // disable automatic repeat
        control_forward = control_m1_held;
        seqtbl_offset_char(seq_26_crouch_while_running); // crouch while running
    }
}

// seg005:06A8
void safe_step(void) {
    short distance;
    control_shift2 = control_1_ignore; // disable automatic repeat
    control_forward = control_1_ignore; // disable automatic repeat
    // custom logic - do not safe step on level 6 before the checkpoint
    if (current_level == 6 && falling_exit_checkpoint < 4) {
        seqtbl_offset_char(seq_39_safe_step_11); // unsafe step (off ledge)
        return;
    }
    distance = get_edge_distance();
    // custom logic - step towards vertical spikes in the temple carefully, like in PoP2
    if (is_temple()) {
        short spike_distance = distance_to_edge_weight();
#define is_retracted(modif) (modif == 0 || (modif >= 6 && modif <= 9))
        if (get_tile_infrontof_char() == tiles_2_spike && spike_distance <= 11 && is_retracted(curr_room_modif[curr_tilepos])) {
            distance = spike_distance + 3;
        } else if (Char.direction == dir_FF_left &&
                   get_tile_at_char() == tiles_2_spike &&
                   spike_distance > 11 &&
                   is_retracted(curr_room_modif[curr_tilepos])) {
            distance = spike_distance - 11;
        }
    }
    // custom logic - stepping closer to the kid before uniting
    if (current_level == 10 && Char.charid == charid_1_shadow) {
        short kid_dist = Char.x - Kid.x;
        if (kid_dist > 0 && kid_dist <= 14) {
            distance = kid_dist;
        }
    }
    if (distance) {
        // custom logic - fix logic that can cause kid to step into chomper from a non-standing frame
        // see: https://forum.princed.org/viewtopic.php?p=15847#p15847
        if (edge_chomper == 1 && Char.frame != frame_15_stand) {
            Char.frame = frame_15_stand;
            distance = get_edge_distance();
        }
        Char.repeat = 1;
        seqtbl_offset_char(distance + 28); // 29..42: safe step to edge
    } else if (edge_type != edge_type_1_edge && Char.repeat != 0) {
        Char.repeat = 0;
        seqtbl_offset_char(seq_44_step_on_edge); // step on edge
    } else {
        // custom logic - briefly sprint through active chomper
        if (edge_chomper == 1) {
            seqtbl_offset_char(seq_122_chomper_sprint);
        } else {
            seqtbl_offset_char(seq_39_safe_step_11); // unsafe step (off ledge)
        }
    }
}

// custom logic - is the tile an non-empty potion?
bool is_nonempty_potion(byte tile) {
    return tile == tiles_10_potion && curr_room_modif[curr_tilepos] < 0x38;
}

// seg005:06F0
int check_get_item(void) {
    // custom logic - do not start drinking on empty/loose tiles
    if (!tile_is_floor(get_tile_at_char()) ||
            (curr_tile2 == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0)) {
        return 0;
    }
    // custom logic - pick potion without moving across tiles
    int curr_wall_type;
    byte is_potion_infront = 0;
    short char_x = Char.x;
    short distance = distance_to_edge_weight();
    bool is_flaming_sword;
    if (is_nonempty_potion(get_tile_infrontof_char()) && distance <= 13) {
        is_potion_infront = 1;
    } else if (get_tile_at_char() == tiles_22_sword) {
        if (!tile_is_floor(get_tile_behind_char()) ||
                wall_type2(curr_tile2) == (Char.direction == dir_0_right ? wall_type_1_right : wall_type_2_left)) {
            return 0;
        }
        short sword_distance = distance_to_edge_weight();
        if (sword_distance >= 7) {
            Char.x = char_dx_forward(-(15 - sword_distance));
        }
    }
    load_fram_det_col();
    // custom logic - do not pick items through walls/chompers/mirrors
    if (is_potion_infront) {
        curr_wall_type = wall_type(curr_room_tiles[get_tilepos(Char.curr_col, Char.curr_row)] & 0x1F);
        if (Char.direction == dir_FF_left && (curr_wall_type == wall_type_2_left || curr_wall_type == wall_type_3_chomper)) {
            return 0;
        } else if (Char.direction == dir_0_right && curr_wall_type == wall_type_1_right) {
            return 0;
        }
    }
    if (is_nonempty_potion(get_tile_at_char())) {
        Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
        load_fram_det_col();
        get_item(0);
        return 1;
    } else if (is_potion_infront) {
        get_tile_infrontof_char();
        get_item(1);
        return 1;
    } else if (get_tile_infrontof_char() == tiles_22_sword) {
pickup_sword:
        // custom logic - swords can only be picked up in 1 direction
        is_flaming_sword = current_level == settings->flame_level &&
                           curr_room == settings->flaming_sword_room &&
                           get_tilepos(tile_col, tile_row) == settings->flaming_sword_tilepos;
        if ((is_flaming_sword && Char.direction != dir_FF_left) ||
                (!is_flaming_sword && Char.direction != dir_0_right)) {
            Char.x = char_x;
            return 0;
        }
        get_item(2);
        return 1;
    } else if (get_tile_infrontof2_char() == tiles_22_sword && distance < 7) {
        Char.x = char_dx_forward(distance + 1);
        goto pickup_sword;
    }
    return 0;
}

// seg005:073E
void get_item(byte align) {
    word potion_type;
    // custom logic - align kid to the object
    if (align) {
        static short align_offset[] = { 0, -2, -10 };
        short distance = get_edge_distance();
        if (edge_type != edge_type_2_floor) {
            short delta_x = distance;
            if (Char.direction >= dir_0_right) {
                delta_x += align_offset[align];
            }
            if (delta_x > 11) {
                Char.x = char_dx_forward(delta_x - 11);
                delta_x = 11;
            }
            if (delta_x >= 6) {
                if (Char.frame == frame_109_crouch) {
                    seqtbl_offset_char(seq_79_crouch_hop);
                } else {
                    seqtbl_offset_char(delta_x + 128); // 134..139: step then crouch
                }
                return;
            } else if (delta_x < 0) {
                Char.x = char_dx_forward(delta_x);
            }
        }
    }
    if (Char.frame != frame_109_crouch) { // crouching
        crouch();
    } else if (curr_tile2 == tiles_22_sword) {
        do_pickup(-1);
        // custom logic - do not pick up sword on level 13 unless kid turned into shadow
        if (current_level != settings->flame_level ||
                Char.room != settings->flaming_sword_room ||
                (is_kid_shadow && !is_shadow_flame)) {
            seqtbl_offset_char(seq_91_get_sword); // get sword
        }
    } else { // potion
        potion_size = curr_room_modif[curr_tilepos] & 0xF8;
        potion_type = potion_size >> 3;
        // custom logic - do not put down posion potions or health potions on level 6
        if (current_level != 6 || potion_type != potion_1_health) {
            potion_room = curr_room;
            potion_pos = curr_tilepos;
        }
        // do not drink empty potions
        if (potion_type > potion_6_open) {
            return;
        }
        // set the replacement pot size
        if (potion_type == potion_0_custom && settings->tbl_empty_potion_size[current_level]) {
            potion_size |= 0x01;
        } else if ((potion_type >= potion_2_life && potion_type < potion_5_hurt) || potion_type == potion_8_empty_small) {
            potion_size |= 0x01;
        } else if (potion_type == potion_6_open) {
            // opener potion has a special size that always keeps it full
            potion_size |= 0x02;
        }
        // pick up potion
        do_pickup(potion_type);
        // custom logic - drinking mouse potion has a different animation
        if (current_level == settings->mouse_2_level &&
                curr_room == settings->mouse_2_room &&
                tile_col == settings->mouse_2_potion_column &&
                tile_row == settings->mouse_2_potion_row) {
            Char.direction = dir_FF_left;
            seqtbl_offset_char(seq_120_drink_mouse_potion);
        } else {
            seqtbl_offset_char(seq_78_drink);
        }
    }
}

// seg005:07FF
void control_startrun(void) {
    if (control_y < control_0_released && control_x < control_0_released) {
        standing_jump();
    }
}

// seg005:0812
void control_jumpup(void) {
    if (control_x < control_0_released || control_forward < control_0_released) {
        standing_jump();
    }
}

// seg005:0825
void standing_jump(void) {
    control_up = control_forward = control_1_ignore; // disable automatic repeat
    // custom logic - do not fall at the very edge
    check_on_edge();
    seqtbl_offset_char(seq_3_standing_jump); // standing jump
}

// seg005:0836
void check_jump_up(void) {
    control_up = release_arrows();
    through_tile = get_tile_above_char();
    get_tile_front_above_char();
    if (can_grab()) {
        grab_up_with_floor_behind();
    } else {
        through_tile = get_tile_behind_above_char();
        get_tile_above_char();
        if (can_grab()) {
            jump_up_or_grab();
        } else {
            jump_up();
        }
    }
}

// seg005:087B
void jump_up_or_grab(void) {
    short distance;
    distance = distance_to_edge_weight();
    if (distance < 6) {
        jump_up();
    } else if (wall_type2(get_tile_at_char()) == wall_type_1_right && Char.direction == dir_FF_left) {
        // custom logic - do not grab up through a closed gate or tapestry
        jump_up();
    } else if (!tile_is_floor(get_tile_behind_char())) {
        // There is no floor behind char.
        if (distance < 7) {
            // custom logic - step back when the distance is 7 pixels
            seqtbl_offset_char(seq_148_step_back_jump_grab);
        } else {
            grab_up_no_floor_behind();
        }
    } else {
        // There is floor behind char, go back a bit.
        if (distance < 11) {
            // custom logic - do a step back animation instead of teleporting kid back
            // when the distance to the edge is between 6 and 10 pixels, which means
            // kid need to step back between 8 and 4 pixels accordingly.
            // seq 143 = step back 4 pixels, jump and grab
            // ...
            // seq 147 = step back 8 pixels, jump and grab
            seqtbl_offset_char(142 + (11 - distance));
        } else {
            Char.x = char_dx_forward(distance - 14);
            load_fram_det_col();
            grab_up_with_floor_behind();
        }
    }
}

// seg005:08C7
void grab_up_no_floor_behind(void) {
    get_tile_above_char();
    Char.x = char_dx_forward(distance_to_edge_weight() - 10);
    seqtbl_offset_char(seq_16_jump_up_and_grab); // jump up and grab (no floor behind)
}

// custom logic - check if kid can grab 2 rows above
bool can_grab_two_rows_above(void) {
    byte tile_above;
    if (tile_is_floor(curr_tile2)) {
        return false;
    }
    if (curr_tile2 == tiles_20_wall) {
        if (curr_tilepos < 10) {
            return false;
        }
        // a non-floor tile in the same room above a wall can be jumped through
        tile_above = (curr_room_tiles[curr_tilepos - 10] & 0x1F);
        return tile_above != tiles_20_wall && !tile_is_floor(tile_above);
    }
    return true;
}

// seg005:08E6
void jump_up(void) {
    int char_col;
    bool is_feather_fall;
    bool is_top_floor;
    short distance;
    word delta_x;
    control_up = release_arrows();
    // custom logic - do not fall off an edge when jumping up
    check_on_edge();
    distance = get_edge_distance();
    if (distance < 4 && edge_type == edge_type_1_edge) {
        Char.x = char_dx_forward(distance - 3);
        // custom logic - move prince away from gate when jumping up
    } else if (Char.direction == dir_0_right &&
               get_tile_behind_char() == tiles_4_gate &&
               can_bump_into_gate() &&
               distance >= 10) {
        Char.x = char_dx_forward(distance - 9);
    }
    // custom logic - prince should be able to grab 2 tiles above from an edge of a tile
    is_feather_fall = feather_fall_timer > 3;
    if (is_feather_fall && !tile_is_floor(get_tile_above_char()) && curr_tile2 != tiles_20_wall) {
        delta_x = Char.direction == dir_FF_left ? 1 : 3;
    } else {
        delta_x = 0;
    }
    char_col = get_tile_div_mod(back_delta_x(delta_x) + dx_weight() - 6);
    get_tile(Char.room, char_col, Char.curr_row - 1);
    // custom logic - super high jump
    if (can_grab_two_rows_above()) {
        if (is_feather_fall) { // super high jump can only happen in feather mode
            if (curr_room == 0 && Char.curr_row == 0) { // there is no room above
                seqtbl_offset_char(seq_14_jump_up_into_ceiling);
            } else {
                get_tile(Char.room, char_col, Char.curr_row - 2); // the target top tile
                is_top_floor = tile_is_floor(curr_tile2) || curr_tile2 == tiles_20_wall;
                if (is_top_floor && curr_tile2 == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0) {
                    is_top_floor = false; // a regular loose floor above should not be treated as a floor
                }
                // kid should jump slightly higher if the top tile is not a floor
                super_jump_timer = is_top_floor ? 22 : 24;
                super_jump_room = curr_room;
                super_jump_col = tile_col;
                super_jump_row = tile_row;
                Char.action = actions_1_run_jump;
                seqtbl_offset_char(seq_48_super_high_jump); // jump up 2 rows with nothing above
            }
        } else {
            seqtbl_offset_char(seq_28_jump_up_with_nothing_above); // jump up with nothing above
        }
    } else {
        seqtbl_offset_char(seq_14_jump_up_into_ceiling); // jump up with wall or floor above
    }
}

// seg005:0968
void control_hanging(void) {
    if (Char.alive < 0) {
        if (grab_timer == 0 && control_y < control_0_released) {
            can_climb_up();
        } else if (control_shift < control_0_released || (super_jump_fall && control_y < control_0_released)) {
            // hanging against a wall or a doortop
            if (Char.action != actions_6_hang_straight &&
                    (get_tile_at_char() == tiles_20_wall ||
                     (Char.direction == dir_FF_left && ( // facing left
                          curr_tile2 == tiles_4_gate || // custom logic - gates should be treated as doortops
                          curr_tile2 == tiles_7_doortop_with_floor ||
                          curr_tile2 == tiles_12_doortop
                      )))
               ) {
                // custom logic - kid falls after a few seconds unless he turned into shadow or the feather mode is on
                seqtbl_offset_char(seq_121_hang_straight_drop);
            } else {
                if (!tile_is_floor(get_tile_above_char())) {
                    hang_fall();
                }
            }
        } else {
            hang_fall();
        }
    } else {
        hang_fall();
    }
}

// seg005:09DF
void can_climb_up(void) {
    short seq_id;
    seq_id = seq_10_climb_up; // climb up
    control_up = control_shift2 = release_arrows();
    // Kid can end up hanging 11 or 12 pixels from the edge.
    // This logic aligns him at 11 pixels to avoid frame artifacts.
    // It also keeps the edge distance consistent after climbing up (no need for Trick 25).
    if (distance_to_edge(Char.x) == 12) {
        Char.x = char_dx_forward(1);
    }
    // custom logic - reset jump flags
    super_jump_fall = 0;
    hard_jump_fall = 0;
    get_tile_above_char();
    // custom logic - shadow can climb up chompers
    if (((curr_tile2 == tiles_13_mirror || (curr_tile2 == tiles_18_chomper && !is_kid_shadow)) &&
            Char.direction == dir_0_right) ||
            (curr_tile2 == tiles_4_gate && Char.direction != dir_0_right &&
             // custom logic - gates on level 12B act inversely
             (current_level != settings->inverse_level ? (curr_room_modif[curr_tilepos] >> 2 < 6) : (curr_room_modif[curr_tilepos] >> 2 >= 6)))
       ) {
        seq_id = seq_73_climb_up_to_closed_gate; // climb up to closed gate and down
    } else if (droppedout && (Opp.alive >= 0 || Opp.curr_row >= Char.curr_row)) {
        // Guard should not follow kid after he climbs up
        droppedout = 0;
    }
    seqtbl_offset_char(seq_id);
    // custom logic - activate the spike when kid is climbing up
    if (!is_temple() && curr_tile2 == tiles_2_spike) {
        start_anim_spike(curr_room, curr_tilepos);
    }
}

// seg005:0A46
void hang_fall(void) {
    control_down = release_arrows();
    // custom logic - reset jump and safe step flags
    super_jump_fall = 0;
    hard_jump_fall = 0;
    Char.repeat = 1;
    if (!tile_is_floor(get_tile_behind_char()) && !tile_is_floor(get_tile_at_char())) {
        // custom logic - kid falls slower when he is tilted towards the wall
        if (Char.frame >= frame_87_hanging_1 && Char.frame <= frame_89_hanging_3) {
            seqtbl_offset_char(seq_11_release_ledge_and_land); // release ledge when swinging forward while hanging down
        } else {
            seqtbl_offset_char(seq_23_release_ledge_and_fall); // release ledge and fall
        }
    } else {
        if (get_tile_at_char() == tiles_20_wall ||
                (Char.direction < dir_0_right && ( // looking left
                     curr_tile2 == tiles_4_gate || // custom logic - gates should be treated as doortops
                     curr_tile2 == tiles_7_doortop_with_floor ||
                     curr_tile2 == tiles_12_doortop
                 ))
           ) {
            Char.x = char_dx_forward(-7);
        }
        seqtbl_offset_char(seq_11_release_ledge_and_land); // end of climb down
    }
}

// seg005:0AA8
void grab_up_with_floor_behind(void) {
    short distance;
    distance = distance_to_edge_weight();
    // custom logic - make the shaking loose tile fall faster when jumping up
    if (get_tile_at_char() == tiles_11_loose &&
            get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0 &&
            curr_room_modif[curr_tilepos] > 0 &&
            feather_fall_timer == 0 &&
            !is_kid_shadow) {
        ++curr_room_modif[curr_tilepos];
    }
    // custom logic - do not try to grab through an active spike in the temple
    if (is_temple() &&
            Char.direction == dir_FF_left &&
            distance >= 10 &&
            curr_tile2 == tiles_2_spike &&
            curr_room_modif[curr_tilepos] == 0) {
        jump_up();
        return;
    }
    // The global variable edge_type (which we need!) gets set as a side effect of get_edge_distance()
    get_edge_distance();
    if (distance < 4 && edge_type != edge_type_1_edge) {
        Char.x = char_dx_forward(distance);
        seqtbl_offset_char(seq_8_jump_up_and_grab_straight); // jump up and grab when there is no need to move kid forward
    } else {
        // custom logic - do a step forward animation instead of teleporting kid back
        // when the distance to the edge is between 8 and 13 pixels, which means
        // kid need to step forward between 4 and 9 pixels accordingly.
        if (distance >= 8) {
            // seq 149 = step forward 9 pixels, jump and grab
            // ...
            // seq 154 = step forward 4 pixels, jump and grab
            seqtbl_offset_char(148 + (14 - distance));
        } else {
            Char.x = char_dx_forward(distance - 4);
            seqtbl_offset_char(seq_24_jump_up_and_grab_forward); // move forward, jump up and grab
        }
    }
}

// seg005:0AF7
void run_jump(void) {
    short tile_count;
    short xpos;
    short col;
    short front_dir;
    short distance_delta;
    if (Char.frame >= frame_7_run) {
        // custom logic - do not align kid on level 6 before the checkpoint
        if (current_level != 6 || falling_exit_checkpoint >= 4) {
            // custom logic - makes a difficult 2-hop running jump easier.
            // Kid must be standing between 1 full tile to 1 full tile + 13 pixels
            // from an abyss that is 1 tile wide.
            xpos = char_dx_forward(10);
            col = get_tile_div_mod_m7(xpos);
            front_dir = dir_front[Char.direction + 1];
            if (!tile_is_floor(get_tile(Char.room, col, Char.curr_row)) && // start the jump before an abyss
                    tile_is_floor(get_tile(Char.room, col + front_dir, Char.curr_row)) &&
                    (wall_type2(curr_tile2) == wall_type_0_none || curr_tile2 == tiles_18_chomper) &&
                    wall_type2(get_tile(Char.room, col + front_dir * 2, Char.curr_row)) < wall_type_4_wall) {
                distance_delta = distance_to_edge(xpos);
                if (distance_delta >= 1 && distance_delta <= 8) {
                    Char.x = char_dx_forward(distance_delta - 1);
                    goto loc_run_jump;
                }
            }
            // Align Kid to edge of floor.
            xpos = char_dx_forward(4);
            col = get_tile_div_mod_m7(xpos);
            for (tile_count = 0; tile_count < 2; ++tile_count) {
                col += dir_front[Char.direction + 1];
                get_tile(Char.room, col, Char.curr_row);
                // custom logic - do not treat fake walls as floors
                if (is_fake_wall()) {
                    curr_tile2 = tiles_0_empty;
                }
                if (curr_tile2 == tiles_2_spike || (!tile_is_floor(curr_tile2) && curr_tile2 != tiles_20_wall)) {
                    distance_delta = distance_to_edge(xpos) + 14 * tile_count - 14;
                    if ((word)distance_delta < (word) - 8 || distance_delta >= 2) {
                        if (distance_delta < 128) return;
                        distance_delta = -3;
                    }
                    Char.x = char_dx_forward(distance_delta + 4);
                    break;
                }
            }
        }
loc_run_jump:
        control_up = release_arrows(); // disable automatic repeat
        seqtbl_offset_char(seq_4_run_jump); // run-jump
    } else {
        // custom logic - allow jumps from 2 full tiles from a standing run
        if (Char.frame == frame_5_start_run &&
                get_tile_at_char() != tiles_18_chomper &&
                tile_is_floor(get_tile_infrontof_char()) &&
                wall_type2(curr_tile2) == wall_type_0_none &&
                (!tile_is_floor(get_tile_infrontof2_char()) && curr_tile2 != tiles_20_wall)) {
            distance_delta = get_edge_distance();
            if (!tile_is_floor(get_tile_behind_char()) && curr_tile2 != tiles_20_wall) {
                ++distance_delta;
            }
            Char.x = char_dx_forward(distance_delta - 14);
        }
    }
}

// sseg005:0BB5
void back_with_sword(void) {
    short frame;
    byte tile_behind;
    short distance;
    frame = Char.frame;
    if (frame == frame_158_stand_with_sword || frame == frame_170_stand_with_sword || frame == frame_171_stand_with_sword) {
        if (Char.charid != charid_0_kid) {
            control_backward = control_1_ignore;
            seqtbl_offset_char(seq_57_back_with_sword);
        } else {
            // custom logic - kid repeatedly retreats when the back arrow is pressed
            control_backward = control_forward = control_0_released;
            tile_behind = get_tile_behind_char();
            distance = distance_to_edge(Char.x);
            // do not back into an abyss, a closer tile or a loose tile
            if (!tile_is_floor(tile_behind) ||
                    (tile_behind == tiles_6_closer && distance < 7) ||
                    (tile_behind == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0)) {
                if (distance >= (tile_behind == tiles_6_closer ? 2 : 4)) {
                    control_backward = control_1_ignore;
                }
                // do not back into a chomper
            } else if ((Char.direction == dir_FF_left && tile_behind == tiles_18_chomper) ||
                       (Char.direction == dir_0_right && get_tile_at_char() == tiles_18_chomper)) {
                if (distance > 1 && distance < 7) {
                    control_backward = control_1_ignore;
                }
            }
            seqtbl_offset_char(seq_162_kid_back_with_sword);
        }
    }
}

// seg005:0BE3
void forward_with_sword(void) {
    short frame;
    short dist;
    frame = Char.frame;
    if (frame == frame_158_stand_with_sword || frame == frame_170_stand_with_sword || frame == frame_171_stand_with_sword) {
        if (Char.charid != charid_0_kid) {
            control_forward = control_1_ignore;
            seqtbl_offset_char(seq_56_guard_forward_with_sword);
        } else {
            // custom logic - kid repeatedly advances when the forward arrow is pressed
            control_forward = control_backward = control_0_released;
            dist = can_guard_see_kid == 2 ? char_opp_dist() : 0;
            if (justturned_engarde || justblocked) {
                justturned_engarde = 0;
            } else if (!have_sword) { // custom logic - advance in a fist fight
                seqtbl_offset_char(seq_164_kid_forward_fist_fight);
            } else if (dist >= 12 && dist < 35) { // custom logic - advance faster near a guard
                Char.action = actions_1_run_jump;
                seqtbl_offset_char(seq_86_forward_with_sword_fast);
            } else {
                seqtbl_offset_char(seq_163_kid_forward_with_sword);
            }
        }
    }
}

// seg005:0C1D
void draw_sword(void) {
    if (Char.charid == charid_0_kid && Char.sword == sword_2_drawn) {
        return;
    }
    word seq_id;
    seq_id = seq_55_draw_sword; // draw sword
    release_arrows();
    control_shift2 = ctrl1_shift2 = (have_sword ? control_1_ignore : control_0_released);
    if (Char.charid == charid_0_kid) {
        // custom logic - fist fight uses a different sequence
        if (!have_sword) {
            // do not fall back off an edge entering the fist fight mode
            short dist = distance_to_edge_weight();
            if (!tile_is_floor(get_tile_behind_char()) && curr_tile2 != tiles_20_wall && dist >= 10) {
                Char.x = char_dx_forward(dist - 9);
            }
            seq_id = seq_155_start_fist_fight;
            make_guard_notice(false);
        }
        offguard = 0;
    } else if (Char.charid != charid_1_shadow) {
        seq_id = seq_90_guard_stand_active; // stand active
    }
    // custom logic - fixes kid getting pushed through gates/doortops after getting hit by guard
    if (Char.charid == charid_0_kid &&
            can_guard_see_kid >= 2 &&
            guard_kid_dist() < 42 &&
            (Char.frame == frame_15_stand || Char.frame >= frame_230_sheathe)) {
        int distance = distance_to_edge_weight();
        if ((get_tile_at_char() == tiles_7_doortop_with_floor ||
                (curr_tile2 == tiles_4_gate && can_bump_into_gate())) &&
                (Char.direction == dir_0_right /* turn */ || distance <= 7)) {
            if (Char.direction == dir_0_right) {
                distance = -distance;
            }
            // move character forward only when there is floor
            if (tile_is_floor(get_tile(Char.room,
                                       get_tile_div_mod_m7(Char.x + dir_front[Char.direction + 1] * distance),
                                       Char.curr_row))) {
                Char.x = char_dx_forward(distance);
            }
        }
    }
    Char.sword = sword_2_drawn;
    justturned_engarde = 0;
    seqtbl_offset_char(seq_id);
}

// seg005:0C67
void control_with_sword(void) {
    short distance;
    short edge_distance;
    if (Char.action < actions_2_hang_climb) {
        // custom logic - check that guard can actually see kid before turning on a loose tile
        if ((get_tile_at_char() == tiles_11_loose && can_guard_see_kid) ||
                can_guard_see_kid == 2) {
            distance = char_opp_dist();
            if ((word)distance < (word)90) {
                swordfight();
                return;
            } else if (distance < 0) {
                if ((word)distance < (word) - 4 ||
                        should_turn_around > 3) { // custom logic - a counter that prevents kid and guard moonwalking
                    should_turn_around = 0;
                    // custom logic - do not turn around until the opponent has landed
                    if (Opp.action == actions_3_in_midair || Opp.action == actions_4_in_freefall) {
                        return;
                    }
                    // custom logic - kid is already turning
                    if (Char.charid == charid_0_kid && (Char.frame >= frame_214_turn_engarde_1 && Char.frame <= frame_216_turn_engarde_3)) {
                        return;
                    }
                    // custom logic - do not turn while kid is kicking
                    if (Opp.frame == frame_213_kick) {
                        return;
                    }
                    edge_distance = distance_to_edge_weight();
                    // custom logic - do not turn when facing a closed gate or a door top
                    if (Char.direction == dir_0_right && Opp.sword == sword_0_sheathed && (
                                (get_tile_at_char() == tiles_4_gate && can_bump_into_gate()) ||
                                curr_tile2 == tiles_7_doortop_with_floor) &&
                            edge_distance >= 11
                       ) {
                        Char.x = char_dx_forward(-11);
                    } else if (Char.direction == dir_0_right && Opp.sword == sword_0_sheathed && (
                                   (get_tile_infrontof_char() == tiles_4_gate && can_bump_into_gate()) ||
                                   curr_tile2 == tiles_7_doortop_with_floor) &&
                               edge_distance <= 2
                              ) {
                        Char.x = char_dx_forward(-11);
                    }
                    // custom logic - prevents kid from turning into a chomper if another guard comes from behind
                    if (Char.charid == charid_0_kid &&
                            Char.direction == dir_FF_left &&
                            dynamic_guard_timer > 0 &&
                            get_tile_at_kid(Char.x - 3) == tiles_18_chomper) {
                        Char.curr_col = tile_col + 1;
                        Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_10_PX_OFFSET;
                    }
                    // turn with sword (after switching places)
                    if (Char.charid == charid_0_kid && have_sword) {
                        if (Char.frame >= frame_207_draw_1 && Char.frame <= frame_210_draw_4) {
                            // custom logic - turn around quickly while drawing sword
                            seqtbl_offset_char(seq_60_turn_with_sword);
                        } else if (Opp.frame != frame_179_collapse_1) {
                            // custom logic - prevents kid from falling off a cliff when turning with sword
                            int delta_weight_x = (cur_frame.flags & FRAME_WEIGHT_X);
                            if ((!tile_is_floor(get_tile_infrontof_char()) && curr_tile2 != tiles_20_wall) || curr_tile2 == tiles_11_loose) {
                                Char.x = char_dx_forward(MIN(-14 - delta_weight_x + edge_distance, 0));
                            } else if ((!tile_is_floor(get_tile_infrontof2_char()) && curr_tile2 != tiles_20_wall) || curr_tile2 == tiles_11_loose) {
                                Char.x = char_dx_forward(MIN(-delta_weight_x + edge_distance, 0));
                            }
                            // custom logic - use an animation for kid to turn around
                            seqtbl_offset_char(seq_129_kid_turn_engarde);
                        }
                    } else if (Opp.frame != frame_179_collapse_1) {
                        if (!have_sword) { // custom logic - kid without a sword stays closer to the guard
                            release_arrows();
                            Char.x = char_dx_forward(-2);
                        }
                        seqtbl_offset_char(seq_60_turn_with_sword);
                    }
                    return;
                } else {
                    swordfight();
                    return;
                }
            }
        }
        if (Char.charid == charid_0_kid && Char.alive < 0) {
            holding_sword = 0;
        }
        if (Char.charid < charid_2_guard &&
                Char.room == drawn_room &&
                (!keep_sword_drawn || !have_sword) && // custom logic - do not put away sword that was taken out manually
                (Opp.alive >= 0 || Opp.sword == sword_0_sheathed || (can_guard_see_kid == 0 && exit_room_timer == 0 && !droppedout)) &&
                can_guard_see_kid != 2 &&
                (is_dynamic_guard_fight != 1 || !have_sword) && // custom logic - put away sword if we are not in dynamic fight mode
                can_inactive_guard_see_kid != 2 // custom logic - put away sword if there is no inactive guard near
           ) {
            // frame 171: stand with sword
            if (Char.frame == frame_171_stand_with_sword) {
                determine_col();
                if (can_guard_see_kid <= 1 &&
                        (get_tile(Char.room, Char.curr_col, Char.curr_row) == tiles_18_chomper ||
                         get_tile(Char.room, Char.curr_col + 1, Char.curr_row) == tiles_18_chomper)) {
                    swordfight();
                } else {
                    Char.sword = sword_0_sheathed;
                    // custom logic - kid stands straight after killing guard with punches
                    if (!have_sword) {
                        guard_punch_count = 0;
                        seqtbl_offset_char(seq_156_stop_fist_fight);
                    } else {
                        seqtbl_offset_char(seq_92_put_sword_away); // put sword away
                    }
                }
            }
        } else {
            swordfight();
        }
    }
}

// custom logic - returns whether kid is between active and inactive guards
bool is_kid_between_guards(void) {
    if (can_guard_see_kid < 2 && can_inactive_guard_see_kid < 2) {
        return true;
    }
    if (can_guard_see_kid < 2 || can_inactive_guard_see_kid < 2) {
        return false;
    }
    if (Guard.direction != ~inactive_guard[Guard.room - 1].direction) {
        return false;
    }
    if (kid_guard_dist() < -14 && inactive_guard_kid_dist() >= 0) {
        return true;
    }
    if (kid_inactive_guard_dist() < -14 && kid_guard_dist() >= 0) {
        return true;
    }
    return false;
}

// seg005:0CDB
void swordfight(void) {
    short frame;
    short seq_id;
    short charid;
    bool standing;
    frame = Char.frame;
    charid = Char.charid;
    standing = (frame == frame_158_stand_with_sword || frame == frame_170_stand_with_sword || frame == frame_171_stand_with_sword);
    // frame 161: parry
    if (frame == frame_161_parry && control_shift2 >= control_0_released) {
        // custom logic - makes it more difficult to throw kid off the edge during a block
        short tile_behind = get_tile(
                                Char.room, get_tile_div_mod_m7(dx_weight() + dir_behind[Char.direction + 1] * 7) + dir_behind[Char.direction + 1],
                                Char.curr_row);
        short distance = distance_to_edge_weight();
        if (Char.charid == charid_0_kid) {
            if (tile_is_empty(tile_behind) && distance >= 7 && distance < 13) {
                Char.x = char_dx_forward(distance_to_edge_weight() - 11);
                seqtbl_offset_char(seq_161_step_back_with_sword); // one step back with sword
            } else {
                seqtbl_offset_char(seq_162_kid_back_with_sword);
            }
        } else {
            seqtbl_offset_char(seq_57_back_with_sword); // back with sword (when parrying)
        }
        return;
    } else if (control_shift2 < control_0_released) {
        if (charid == charid_0_kid) {
            kid_sword_strike = 15;
        }
        // custom logic - Jaffar does not engage unless kid is shadow or is trying to strike
        if (current_level == settings->flame_level &&
                charid == charid_2_guard &&
                Opp.sword == sword_2_drawn &&
                have_sword &&
                !is_kid_shadow) {
            return;
        }
        sword_strike();
        if (control_shift2 == control_1_ignore) return;
    }
    if (control_down < control_0_released) {
        if (standing) {
            control_down = control_1_ignore; // disable automatic repeat
            // custom logic - skeletons do not sheathe swords
            if (charid != charid_4_skeleton) {
                Char.sword = sword_0_sheathed;
            }
            if (charid == charid_0_kid) {
                offguard = 1;
                guard_refrac = 0;
                holding_sword = 0;
                control_forward = control_down = release_arrows();
                keep_sword_drawn = false; // custom logic - do not keep sword drawn
                // custom logic - kid stands exits fist fighting mode
                if (!have_sword) {
                    if (Opp.alive >= 0) {
                        guard_punch_count = 0;
                    }
                    seq_id = seq_156_stop_fist_fight; // stop fighting
                } else {
                    seq_id = seq_93_put_sword_away_fast; // put sword away fast
                }
            } else if (charid == charid_1_shadow) {
                seq_id = seq_92_put_sword_away; // put sword away
            } else {
                seq_id = seq_87_guard_become_inactive; // stand inactive (when Kid leaves sight)
            }
            seqtbl_offset_char(seq_id);
        }
    } else if (control_up < control_0_released && have_sword) { // custom logic - do not parry in a fist fight
        // custom logic - Jaffar does not engage unless kid is shadow or is trying to strike
        if (current_level == settings->flame_level &&
                charid == charid_2_guard &&
                have_sword &&
                !is_kid_shadow) {
            return;
        }
        parry();
    } else if (control_forward < control_0_released) {
        forward_with_sword();
        check_on_floor();
    } else if (control_backward < control_0_released) {
        // custom logic - turn with sword with Shift-back arrow when drawn manually
        // or when kid is between 2 guards
        justturned_engarde = 0;
        if (control_alt == control_m1_held &&
                charid == charid_0_kid &&
                standing &&
                ((keep_sword_drawn && can_guard_see_kid < 2) || is_kid_between_guards())
           ) {
            control_backward = control_0_released;
            // custom logic - prevents kid from falling off a cliff when turning with sword
            int delta_weight_x = (cur_frame.flags & FRAME_WEIGHT_X);
            short distance = distance_to_edge_weight();
            if ((!tile_is_floor(get_tile_infrontof_char()) && curr_tile2 != tiles_20_wall) || curr_tile2 == tiles_11_loose) {
                Char.x = char_dx_forward(MIN(-10 - delta_weight_x + distance, 0));
            } else if ((!tile_is_floor(get_tile_infrontof2_char()) && curr_tile2 != tiles_20_wall) || curr_tile2 == tiles_11_loose) {
                Char.x = char_dx_forward(MIN(4 - delta_weight_x + distance, 0));
            }
            // custom logic - step away from the guard if kid is too close
            if (can_guard_see_kid >= 2) { // between 2 guards
                short guard_distance = guard_kid_dist();
                if (guard_distance < 14) {
                    Char.x = char_dx_forward(guard_distance - 14);
                }
            }
            // custom logic - kid should use the standard turn engarde sequence without the sword
            if (have_sword) {
                seq_id = seq_130_kid_turn_with_sword;
            } else {
                Char.x = char_dx_forward(-7);
                seq_id = seq_60_turn_with_sword;
            }
            seqtbl_offset_char(seq_id);
            return;
        }
        back_with_sword();
        check_on_floor();
    }
}

// seg005:0DB0
void sword_strike(void) {
    short frame;
    short seq_id;
    frame = Char.frame;
    if (frame == frame_157_walk_with_sword || // walk with sword
            frame == frame_158_stand_with_sword || // stand with sword
            frame == frame_170_stand_with_sword || // stand with sword
            frame == frame_171_stand_with_sword || // stand with sword
            frame == frame_165_walk_with_sword // walk with sword
       ) {
        if (Char.charid == charid_0_kid) {
            // custom - do not fall back off an edge after a punch
            if (!have_sword &&
                    !tile_is_floor(get_tile_behind_char()) &&
                    curr_tile2 != tiles_20_wall &&
                    distance_to_edge_weight() == 13) {
                Char.x = char_dx_forward(1);
            }
            if (!have_sword && (key_states[SDL_SCANCODE_LCTRL] || key_states[SDL_SCANCODE_RCTRL] || control_shoulder == control_m1_held)) {
                seq_id = seq_165_kid_kick_fist_fight; // kick (Kid)
            } else {
                seq_id = seq_75_strike; // strike with sword (Kid)
            }
        } else {
            seq_id = seq_58_guard_strike; // strike with sword (Guard)
        }
    } else if (frame == frame_150_parry || frame == frame_161_parry) { // parry
        seq_id = seq_66_strike_after_parry; // strike with sword after parrying
    } else {
        return;
    }
    control_shift2 = control_1_ignore; // disable automatic repeat
    seqtbl_offset_char(seq_id);
}

// seg005:0E0F
void parry(void) {
    short opp_frame;
    short char_frame;
    short is_strike;
    short seq_id;
    short char_charid;
    short parry_limit;
    bool no_parry;
    bool is_parry;
    char_frame = Char.frame;
    opp_frame = Opp.frame;
    char_charid = Char.charid;
    seq_id = seq_62_parry; // defend (parry) with sword
    is_strike = 0;
    is_parry = char_frame == frame_158_stand_with_sword || // stand with sword
               char_frame == frame_170_stand_with_sword || // stand with sword
               char_frame == frame_171_stand_with_sword || // stand with sword
               char_frame == frame_168_back || // back?
               char_frame == frame_165_walk_with_sword;
    // custom logic - Jaffar has improved parry skills
    if (char_charid != charid_0_kid && guard_skill == 10) {
        is_parry = is_parry ||
                   (char_frame >= frame_151_strike_1 && char_frame <= frame_160_fighting);
    }
    if (is_parry) {
        if (char_opp_dist() >= 32 && char_charid != charid_0_kid) {
            back_with_sword();
            return;
        } else if (char_charid == charid_0_kid) {
            if (opp_frame == frame_168_back) return;
            if (opp_frame != frame_151_strike_1 &&
                    opp_frame != frame_152_strike_2 &&
                    opp_frame != frame_162_block_to_strike
               ) {
                if (opp_frame == frame_153_strike_3) { // strike
                    is_strike = 1;
                } else if (char_charid != charid_0_kid) {
                    back_with_sword();
                    return;
                }
            }
        } else {
            // custom logic - Jaffar has improved parry skills
            parry_limit = guardhp_curr > 1 ? 3 : 4;
            if (guard_skill == 10 &&
                    (jaffar_parry_count % parry_limit) != (parry_limit - 1) &&
                    (settings->jaffarparryprob > prandom(255) || guardhp_curr < 3)) {
                no_parry = opp_frame != frame_151_strike_1 &&
                           opp_frame != frame_152_strike_2 &&
                           opp_frame != frame_153_strike_3;
                ++jaffar_parry_count;
            } else {
                jaffar_parry_count = 0;
                no_parry = opp_frame != frame_152_strike_2;
            }
            if (no_parry) {
                return;
            }
        }
    } else {
        if (char_frame != frame_167_blocked) return;
        seq_id = seq_61_parry_after_strike; // parry after striking with sword
    }
    control_up = control_1_ignore; // disable automatic repeat
    seqtbl_offset_char(seq_id);
    if (is_strike) {
        play_seq();
    }
}
