/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  D?vid Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"
#include <time.h>
#include <errno.h>

#if defined WIN32 || _WIN32 || WIN64 || _WIN64
#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN // reduces the number of includes in windows.h
#endif // WIN32_LEAN_AND_MEAN
#endif

#ifdef WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <wchar.h>
#else
#include "dirent.h"
#endif

bool found_exe_dir = false;

// Most functions in this file are different from those in the original game.

void sdlperror(const char* header) {
    const char* error = SDL_GetError();
    printf("%s: %s\n", header, error);
}

// custom logic - display help information for the help command line parameter
void display_help(void) {
#ifdef WIN32_LEAN_AND_MEAN
    MessageBox(0, "See DOCS\\README.TXT", "Help", 0);
#else
    printf ("See DOCS/README.TXT\n");
#endif
}

// custom logic - create the provided folder, if missing
void create_folder_if_missing(const char* folder) {
    struct stat st = {0};

    if (stat(folder, &st) == -1) {
#ifdef WIN32_LEAN_AND_MEAN
        mkdir(folder);
#else
        mkdir(folder, 0700);
#endif
    }
}

void find_exe_dir(void) {
    if (found_exe_dir) return;
    snprintf(exe_dir, sizeof(exe_dir), "%s", g_argv[0]);
    char* last_slash = NULL;
    char* pos = exe_dir;
    for (char c = *pos; c != '\0'; c = *(++pos)) {
        if (c == '/' || c == '\\') {
            last_slash = pos;
        }
    }
    if (last_slash != NULL) {
        *last_slash = '\0';
    }
    found_exe_dir = true;
}

bool file_exists(const char* filename) {
    return (access(filename, F_OK) != -1);
}

const char* locate_file_(const char* filename, char* path_buffer, int buffer_size) {
    if (file_exists(filename)) {
        return filename;
    } else {
        // If failed, it may be that SDLPoP is being run from the wrong different working directory.
        // We can try to rescue the situation by loading from the directory of the executable.
        find_exe_dir();
        snprintf(path_buffer, buffer_size, "%s/%s", exe_dir, filename);
        return (const char*) path_buffer;
    }
}

#ifdef WIN32_LEAN_AND_MEAN
// These macros are from the SDL2 source. (src/core/windows/SDL_windows.h)
// The pointers returned by these macros must be freed with SDL_free().
#define WIN_StringToUTF8(S) SDL_iconv_string("UTF-8", "UTF-16LE", (char *)(S), (SDL_wcslen(S)+1)*sizeof(WCHAR))
#define WIN_UTF8ToString(S) (WCHAR *)SDL_iconv_string("UTF-16LE", "UTF-8", (char *)(S), SDL_strlen(S)+1)

// This hack is needed because SDL uses UTF-8 everywhere (even in argv!), but fopen on Windows uses whatever code page is currently set.
FILE* fopen_UTF8(const char* filename_UTF8, const char* mode_UTF8) {
    WCHAR* filename_UTF16 = WIN_UTF8ToString(filename_UTF8);
    WCHAR* mode_UTF16 = WIN_UTF8ToString(mode_UTF8);
    FILE* result = _wfopen(filename_UTF16, mode_UTF16);
    SDL_free(mode_UTF16);
    SDL_free(filename_UTF16);
    return result;
}

int chdir_UTF8(const char* path_UTF8) {
    WCHAR* path_UTF16 = WIN_UTF8ToString(path_UTF8);
    int result = _wchdir(path_UTF16);
    SDL_free(path_UTF16);
    return result;
}

int access_UTF8(const char* filename_UTF8, int mode) {
    WCHAR* filename_UTF16 = WIN_UTF8ToString(filename_UTF8);
    int result = _waccess(filename_UTF16, mode);
    SDL_free(filename_UTF16);
    return result;
}

int stat_UTF8(const char *filename_UTF8, struct stat *_Stat) {
    WCHAR* filename_UTF16 = WIN_UTF8ToString(filename_UTF8);
#ifdef _MSC_VER
    int result = _wstat(filename_UTF16, _Stat);
#else
    int result = wstat(filename_UTF16, _Stat);
#endif
    SDL_free(filename_UTF16);
    return result;
}
#endif

// OS abstraction for listing directory contents
// Directory listing using dirent.h is available using MinGW on Windows, but not using MSVC (need to use Win32 API).
// - Under GNU/Linux, etc (or if compiling with MinGW on Windows), we can use dirent.h
// - Under Windows, we'd like to directly call the Win32 API. (Note: MSVC does not include dirent.h)
// NOTE: If we are using MinGW, we'll opt to use the Win32 API as well: dirent.h would just wrap Win32 anyway!

dat_type* dat_chain_ptr = NULL;

char last_text_input;

// seg009:000D
int read_key(void) {
    int key = last_key_scancode;
    last_key_scancode = 0;
    return key;
}

// seg009:019A
void clear_kbd_buf(void) {
    last_key_scancode = 0;
    last_text_input = 0;
}

// seg009:040A
word prandom(word max) {
    if (!seed_was_init) {
        // init from current time
        random_seed = (dword) time(NULL);
        seed_was_init = 1;
    }
    random_seed = random_seed * 214013 + 2531011;
    return (random_seed >> 16) % (max + 1);
}

// seg009:0467
int round_xpos_to_byte(int xpos) {
    return xpos;
}

// seg009:0C7A
void quit(int exit_code) {
    restore_stuff();
    exit(exit_code);
}

// seg009:0C90
void restore_stuff(void) {
    free_music();
    SDL_Quit();
#ifdef WIN32_LEAN_AND_MEAN
    // custom logic - restores num lock's initial state if it got changed when starting the game
    if (was_numlock_active) {
        set_numlock_state(true);
    }
#endif
}

// seg009:0E33
int key_test_quit(void) {
    word key;
    key = read_key();
    if (key == (SDL_SCANCODE_Q | WITH_CTRL)) { // ctrl-q
        if (is_menu_shown) menu_was_closed();
        quit(0);
    }
    return key;
}

// seg009:0E54
const char* check_param(const char *param) {
    short arg_index;
    size_t param_len;
    param_len = strlen(param);
    for (arg_index = 1; arg_index < g_argc; ++arg_index) {
        char* curr_arg = g_argv[arg_index];
        // custom logic - do not compare literal arguments of different lengths
        if (strchr(curr_arg, ':') == NULL && param_len != strlen(curr_arg)) {
            continue;
        }
        if (strncasecmp(curr_arg, param, param_len) == 0) {
            return g_argv[arg_index];
        }
    }
    return NULL;
}

// seg009:0EDF
int pop_wait(int timer_index, int time) {
    start_timer(timer_index, time);
    return do_wait(timer_index);
}

static FILE* open_dat_from_root_or_data_dir(const char* filename) {
    FILE* fp = NULL;
    fp = fopen(filename, "rb");

    // if failed, try if the DAT file can be opened in the data/ directory, instead of the main folder
    if (fp == NULL) {
        char data_path[POP_MAX_PATH];
        snprintf(data_path, sizeof(data_path), "DATA/%s", filename);

        // verify that this is a regular file and not a directory (otherwise, don't open)
        struct stat path_stat;
        stat(data_path, &path_stat);
        if (S_ISREG(path_stat.st_mode)) {
            fp = fopen(data_path, "rb");
        }
    }
    return fp;
}

// seg009:0F58
dat_type* open_dat(const char *filename) {
    FILE* fp = NULL;
    fp = open_dat_from_root_or_data_dir(filename);
    dat_header_type dat_header;
    dat_table_type* dat_table = NULL;

    dat_type* pointer = (dat_type*) calloc(1, sizeof(dat_type));
    snprintf(pointer->filename, sizeof(pointer->filename), "%s", filename);
    pointer->next_dat = dat_chain_ptr;
    dat_chain_ptr = pointer;

    if (fp != NULL) {
        if (fread(&dat_header, 6, 1, fp) != 1)
            goto failed;
        dat_table = (dat_table_type*) malloc(dat_header.table_size);
        if (dat_table == NULL ||
                fseek(fp, dat_header.table_offset, SEEK_SET) ||
                fread(dat_table, dat_header.table_size, 1, fp) != 1)
            goto failed;
        pointer->handle = fp;
        pointer->dat_table = dat_table;
    }
out:
    return pointer;
failed:
    perror(filename);
    if (fp)
        fclose(fp);
    if (dat_table)
        free(dat_table);
    goto out;
}

// seg009:9CAC
void set_loaded_palette(dat_pal_type *palette_ptr) {
    int dest_row, dest_index, source_row;
    for (dest_row = dest_index = source_row = 0; dest_row < 16; ++dest_row, dest_index += 0x10) {
        if (palette_ptr->row_bits & (1 << dest_row)) {
            set_pal_arr(dest_index, 16, palette_ptr->vga + source_row * 0x10);
            ++source_row;
        }
    }
}

// data:3356
word chtab_palette_bits = 1;

// seg009:104E
chtab_type* load_sprites_from_file(int resource, int palette_bits) {
    int i;
    int n_images = 0;
    chtab_type* chtab = NULL;
    // custom logic - kid in cheats mode has a purple turban (resource 399)
    int kid_pal_offset = resource == 400 ? kid_cheat_pal_offset : 0;
    dat_shpl_type* shpl = (dat_shpl_type*) load_from_opendats_alloc(resource + kid_pal_offset, "pal", NULL, NULL);
    if (shpl == NULL) {
        if (resource != 1000) {
            printf("Cannot load sprites from resource %d.\n", resource);
        }
        return NULL;
    }
    dat_pal_type* pal_ptr = &shpl->palette;
    if (palette_bits != 0) {
        chtab_palette_bits |= palette_bits;
    }
    pal_ptr->row_bits = palette_bits;
    n_images = shpl->n_images;
    size_t alloc_size = sizeof(chtab_type) + sizeof(void *) * n_images;
    chtab = (chtab_type*) malloc(alloc_size);
    memset(chtab, 0, alloc_size);
    chtab->n_images = n_images;
    for (i = 1; i <= n_images; i++) {
        SDL_Surface* image = load_image(resource + i, pal_ptr);
        chtab->images[i - 1] = image;
    }
    set_loaded_palette(pal_ptr);
    return chtab;
}

// seg009:11A8
void free_chtab(chtab_type *chtab_ptr) {
    image_type *curr_image;
    word id;
    word n_images;
    if (chtab_ptr->has_palette_bits) {
        chtab_palette_bits &= ~ chtab_ptr->chtab_palette_bits;
    }
    n_images = chtab_ptr->n_images;
    for (id = 0; id < n_images; ++id) {
        curr_image = chtab_ptr->images[id];
        if (curr_image) {
            SDL_FreeSurface(curr_image);
        }
    }
    free(chtab_ptr);
}

// seg009:8CE6
void decompress_rle_lr(byte *destination, const byte *source, int dest_length) {
    const byte* src_pos = source;
    byte* dest_pos = destination;
    short rem_length = dest_length;
    while (rem_length) {
        sbyte count = *(src_pos++);
        if (count >= 0) { // copy
            ++count;
            do {
                *(dest_pos++) = *(src_pos++);
                --rem_length;
            } while (--count && rem_length);
        } else { // repeat
            byte al = *(src_pos++);
            count = -count;
            do {
                *(dest_pos++) = al;
                --rem_length;
            } while (--count && rem_length);
        }
    }
}

// seg009:8D1C
void decompress_rle_ud(byte *destination, const byte *source, int dest_length, int width, int height) {
    short rem_height = height;
    const byte* src_pos = source;
    byte* dest_pos = destination;
    short rem_length = dest_length;
    --dest_length;
    --width;
    while (rem_length) {
        sbyte count = *(src_pos++);
        if (count >= 0) { // copy
            ++count;
            do {
                *(dest_pos++) = *(src_pos++);
                dest_pos += width;
                if (--rem_height == 0) {
                    dest_pos -= dest_length;
                    rem_height = height;
                }
                --rem_length;
            } while (--count && rem_length);
        } else { // repeat
            byte al = *(src_pos++);
            count = -count;
            do {
                *(dest_pos++) = al;
                dest_pos += width;
                if (--rem_height == 0) {
                    dest_pos -= dest_length;
                    rem_height = height;
                }
                --rem_length;
            } while (--count && rem_length);
        }
    }
}

// seg009:90FA
byte *decompress_lzg_lr(byte *dest, const byte *source, int dest_length) {
    byte* window = (byte*) malloc(0x400);
    if (window == NULL) return NULL;
    memset(window, 0, 0x400);
    byte* window_pos = window + 0x400 - 0x42; // bx
    short remaining = dest_length; // cx
    byte* window_end = window + 0x400; // dx
    const byte* source_pos = source;
    byte* dest_pos = dest;
    word mask = 0;
    do {
        mask >>= 1;
        if ((mask & 0xFF00) == 0) {
            mask = *(source_pos++) | 0xFF00;
        }
        if (mask & 1) {
            *(window_pos++) = *(dest_pos++) = *(source_pos++);
            if (window_pos >= window_end) window_pos = window;
            --remaining;
        } else {
            word copy_info = *(source_pos++);
            copy_info = (copy_info << 8) | *(source_pos++);
            byte* copy_source = window + (copy_info & 0x3FF);
            byte copy_length = (copy_info >> 10) + 3;
            do {
                *(window_pos++) = *(dest_pos++) = *(copy_source++);
                if (copy_source >= window_end) copy_source = window;
                if (window_pos >= window_end) window_pos = window;
            } while (--remaining && --copy_length);
        }
    } while (remaining);
//	end:
    free(window);
    return dest;
}

// seg009:91AD
byte *decompress_lzg_ud(byte *dest, const byte *source, int dest_length, int stride, int height) {
    byte* window = (byte*) malloc(0x400);
    if (window == NULL) return NULL;
    memset(window, 0, 0x400);
    byte* window_pos = window + 0x400 - 0x42; // bx
    short remaining = height; // cx
    byte* window_end = window + 0x400; // dx
    const byte* source_pos = source;
    byte* dest_pos = dest;
    word mask = 0;
    short length = dest_length - 1;
    do {
        mask >>= 1;
        if ((mask & 0xFF00) == 0) {
            mask = *(source_pos++) | 0xFF00;
        }
        if (mask & 1) {
            *(window_pos++) = *dest_pos = *(source_pos++);
            dest_pos += stride;
            if (--remaining == 0) {
                dest_pos -= length;
                remaining = height;
            }
            if (window_pos >= window_end) window_pos = window;
            --dest_length;
        } else {
            word copy_info = *(source_pos++);
            copy_info = (copy_info << 8) | *(source_pos++);
            byte* copy_source = window + (copy_info & 0x3FF);
            byte copy_length = (copy_info >> 10) + 3;
            do {
                *(window_pos++) = *dest_pos = *(copy_source++);
                dest_pos += stride;
                if (--remaining == 0) {
                    dest_pos -= length;
                    remaining = height;
                }
                if (copy_source >= window_end) copy_source = window;
                if (window_pos >= window_end) window_pos = window;
            } while (--dest_length && --copy_length);
        }
    } while (dest_length);
    free(window);
    return dest;
}

// seg009:938E
void decompr_img(byte *dest, const image_data_type *source, int decomp_size, int cmeth, int stride) {
    switch (cmeth) {
        case 0: // RAW left-to-right
            memcpy(dest, &source->data, decomp_size);
            break;
        case 1: // RLE left-to-right
            decompress_rle_lr(dest, source->data, decomp_size);
            break;
        case 2: // RLE up-to-down
            decompress_rle_ud(dest, source->data, decomp_size, stride, source->height);
            break;
        case 3: // LZG left-to-right
            decompress_lzg_lr(dest, source->data, decomp_size);
            break;
        case 4: // LZG up-to-down
            decompress_lzg_ud(dest, source->data, decomp_size, stride, source->height);
            break;
    }
}

int calc_stride(image_data_type* image_data) {
    int width = image_data->width;
    int flags = image_data->flags;
    int depth = ((flags >> 12) & 7) + 1;
    return (depth * width + 7) / 8;
}

byte* conv_to_8bpp(byte* in_data, int width, int height, int stride, int depth) {
    byte* out_data = (byte*) malloc(width * height);
    int y, x_pixel, x_byte, pixel_in_byte;
    int pixels_per_byte = 8 / depth;
    int mask = (1 << depth) - 1;
    for (y = 0; y < height; ++y) {
        byte* in_pos = in_data + y * stride;
        byte* out_pos = out_data + y * width;
        for (x_pixel = x_byte = 0; x_byte < stride; ++x_byte) {
            byte v = *in_pos;
            int shift = 8;
            for (pixel_in_byte = 0; pixel_in_byte < pixels_per_byte && x_pixel < width; ++pixel_in_byte, ++x_pixel) {
                shift -= depth;
                *out_pos = (v >> shift) & mask;
                ++out_pos;
            }
            ++in_pos;
        }
    }
    return out_data;
}

image_type* decode_image(image_data_type* image_data, dat_pal_type* palette) {
    int height = image_data->height;
    if (height == 0) return NULL;
    int width = image_data->width;
    int flags = image_data->flags;
    int depth = ((flags >> 12) & 7) + 1;
    int cmeth = (flags >> 8) & 0x0F;
    int stride = calc_stride(image_data);
    int dest_size = stride * height;
    byte* dest = (byte*) malloc(dest_size);
    memset(dest, 0, dest_size);
    decompr_img(dest, image_data, dest_size, cmeth, stride);
    byte* image_8bpp = conv_to_8bpp(dest, width, height, stride, depth);
    free(dest);
    dest = NULL;
    image_type* image = SDL_CreateRGBSurface(0, width, height, 8, 0, 0, 0, 0);
    if (image == NULL) {
        sdlperror("SDL_CreateRGBSurface");
        quit(1);
    }
    if (SDL_LockSurface(image) != 0) {
        sdlperror("SDL_LockSurface");
    }
    int y;
    for (y = 0; y < height; ++y) {
        // fill image with data
        memcpy((byte*)image->pixels + y * image->pitch, image_8bpp + y * width, width);
    }
    SDL_UnlockSurface(image);

    free(image_8bpp);
    image_8bpp = NULL;
    SDL_Color colors[16];
    int i;
    for (i = 0; i < 16; ++i) {
        colors[i].r = palette->vga[i].r << 2;
        colors[i].g = palette->vga[i].g << 2;
        colors[i].b = palette->vga[i].b << 2;
        colors[i].a = SDL_ALPHA_OPAQUE;   // SDL2's SDL_Color has a fourth alpha component
    }
    // Force 0th color to be black for non-transparent blitters. (hitpoints, shadow)
    // This is needed to remove the colored rectangles around hitpoints and the shadow, when using Brain's SNES graphics for example.
    colors[0].r = 0;
    colors[0].g = 0;
    colors[0].b = 0;
    colors[0].a = SDL_ALPHA_TRANSPARENT;
    SDL_SetPaletteColors(image->format->palette, colors, 0, 16); // SDL_SetColors = deprecated
    return image;
}

// seg009:121A
image_type* load_image(int resource_id, dat_pal_type* palette) {
    data_location result;
    int size;
    void* image_data = load_from_opendats_alloc(resource_id, "png", &result, &size);
    image_type* image = NULL;
    switch (result) {
        case data_none:
            return NULL;
            break;
        case data_DAT: { // DAT
            // custom logic - change blood color for dead princess
            if (resource_id == 931) {
                palette->vga[12].r = 50;
                palette->vga[12].g = 0;
                palette->vga[12].b = 0;
            }
            image = decode_image((image_data_type*) image_data, palette);
        }
        break;
        case data_directory: { // directory
            SDL_RWops* rw = SDL_RWFromConstMem(image_data, size);
            if (rw == NULL) {
                sdlperror("SDL_RWFromConstMem");
                return NULL;
            }
            image = IMG_Load_RW(rw, 0);
            if (image == NULL) {
                printf("IMG_Load_RW: %s\n", IMG_GetError());
            }
            if (SDL_RWclose(rw) != 0) {
                sdlperror("SDL_RWclose");
            }
        }
        break;
    }
    if (image_data != NULL) free(image_data);
    return image;
}

// seg009:157E
int set_joy_mode(void) {
    if (SDL_NumJoysticks() < 1) {
        is_joyst_mode = 0;
    } else {
        if (SDL_IsGameController(0)) {
            sdl_controller_ = SDL_GameControllerOpen(0);
            if (sdl_controller_ == NULL) {
                is_joyst_mode = 0;
            } else {
                is_joyst_mode = 1;
            }
        }
        // We have a joystick connected, but it's NOT compatible with the SDL_GameController
        // interface, so we resort to the classic SDL_Joystick interface instead
        else {
            sdl_joystick_ = SDL_JoystickOpen(0);
            is_joyst_mode = 1;
            using_sdl_joystick_interface = 1;
        }
    }
    if (settings->controller_rumble && is_joyst_mode) {
        sdl_haptic = SDL_HapticOpen(0);
        SDL_HapticRumbleInit(sdl_haptic); // initialize the device for simple rumble
    } else {
        sdl_haptic = NULL;
    }

    is_keyboard_mode = !is_joyst_mode;
    return is_joyst_mode;
}

// seg009:178B
surface_type* make_offscreen_buffer(const rect_type *rect) {
    // Bit order matches onscreen buffer, good for fading.
    return SDL_CreateRGBSurface(0, rect->right, rect->bottom, 24, 0xFF, 0xFF << 8, 0xFF << 16, 0); //RGB888 (little endian)
}

// seg009:17BD
void free_surface(surface_type *surface) {
    SDL_FreeSurface(surface);
}

// seg009:17EA
void free_peel(peel_type *peel_ptr) {
    SDL_FreeSurface(peel_ptr->peel);
    free(peel_ptr);
}

// seg009:182F
void set_hc_pal(void) {
    set_pal_arr(0, 16, settings->vga_palette);
}

// seg009:2446
void flip_not_ega(byte *memory, int height, int stride) {
    byte* row_buffer = (byte*) malloc(stride);
    byte* top_ptr;
    byte* bottom_ptr;
    bottom_ptr = top_ptr = memory;
    bottom_ptr += (height - 1) * stride;
    short rem_rows = height >> 1;
    do {
        memcpy(row_buffer, top_ptr, stride);
        memcpy(top_ptr, bottom_ptr, stride);
        memcpy(bottom_ptr, row_buffer, stride);
        top_ptr += stride;
        bottom_ptr -= stride;
        --rem_rows;
    } while (rem_rows);
    free(row_buffer);
}

// seg009:19B1
void flip_screen(surface_type *surface) {
    if (SDL_LockSurface(surface) != 0) {
        sdlperror("SDL_LockSurface");
        quit(1);
    }
    flip_not_ega((byte*) surface->pixels, surface->h, surface->pitch);
    SDL_UnlockSurface(surface);
}

static void load_font_character_offsets(rawfont_type* data) {
    int n_chars = data->last_char - data->first_char + 1;
    byte* pos = (byte*) &data->offsets[n_chars];
    for (int index = 0; index < n_chars; ++index) {
        data->offsets[index] = (word) (pos - (byte*) data);
        image_data_type* image_data = (image_data_type*) pos;
        int image_bytes = image_data->height * calc_stride(image_data);
        pos = (byte*) &image_data->data + image_bytes;
    }
}

font_type load_font_from_data(/*const*/ rawfont_type* data) {
    font_type font;
    font.first_char = data->first_char;
    font.last_char = data->last_char;
    font.height_above_baseline = data->height_above_baseline;
    font.height_below_baseline = data->height_below_baseline;
    font.space_between_lines = data->space_between_lines;
    font.space_between_chars = data->space_between_chars;
    int n_chars = font.last_char - font.first_char + 1;
    // Allow loading a font even if the offsets for each character image were not supplied in the raw data.
    if (data->offsets[0] == 0) {
        load_font_character_offsets(data);
    }
    chtab_type* chtab = malloc(sizeof(chtab_type) + sizeof(image_type*) * n_chars);
    int chr, index;
    // Make a dummy palette for decode_image().
    dat_pal_type dat_pal;
    memset(&dat_pal, 0, sizeof(dat_pal));
    dat_pal.vga[1].r = dat_pal.vga[1].g = dat_pal.vga[1].b = 0x3F; // white
    for (index = 0, chr = data->first_char; chr <= data->last_char; ++index, ++chr) {
        /*const*/ image_data_type* image_data = (/*const*/ image_data_type*)((/*const*/ byte*)data + data->offsets[index]);
        //image_data->flags=0;
        if (image_data->height == 0) image_data->height = 1; // HACK: decode_image() returns NULL if height==0.
        image_type* image;
        chtab->images[index] = image = decode_image(image_data, &dat_pal);
        if (SDL_SetColorKey(image, SDL_TRUE, 0) != 0) {
            sdlperror("SDL_SetColorKey");
            quit(1);
        }
    }
    font.chtab = chtab;
    return font;
}

// custom logic - load fonts from files
rawfont_type* load_font_data(const char* filename) {
    FILE* fp = NULL;
    fp = fopen(filename, "rb");
    if (fp == NULL) {
        goto font_error_loc;
    }
    struct stat info;
    if (fstat(fileno(fp), &info) == -1) {
        fclose(fp);
        goto font_error_loc;
    }
    size_t file_size = (size_t) MAX(0, info.st_size);
    void* area = malloc(file_size);
    if (fread(area, sizeof(byte), file_size, fp) != file_size) {
        fclose(fp);
        goto font_error_loc;
    }
    fclose(fp);
    return (rawfont_type*) area;
font_error_loc:
    printf("Cannot open font file: %s", filename);
    return NULL;
}

// custom logic - load fonts from files
void load_font(void) {
    rawfont_type* hc_font_data = NULL;
    rawfont_type* hc_small_font_data = NULL;
    hc_font_data = load_font_data(locate_file("DATA/FONT1.DAT"));
    hc_small_font_data = load_font_data(locate_file("DATA/FONT2.DAT"));
    if (hc_font_data != NULL) {
        hc_font = load_font_from_data(hc_font_data);
    }
    if (hc_small_font_data != NULL) {
        hc_small_font = load_font_from_data(hc_small_font_data);
    }
}

// seg009:35C5
int get_char_width(byte character) {
    font_type* font = textstate.ptr_font;
    int width = 0;
    if (character <= font->last_char && character >= font->first_char) {
        image_type* image = font->chtab->images[character - font->first_char];
        if (image != NULL) {
            width += image->w; //char_ptrs[character - font->first_char]->width;
            if (width) width += font->space_between_chars;
        }
    }
    return width;
}

// seg009:3E99
int find_linebreak(const char *text, int length, int break_width, int x_align) {
    short curr_line_width; // in pixels
    short last_break_pos; // in characters
    int curr_char_pos = 0;
    last_break_pos = 0;
    curr_line_width = 0;
    const char* text_pos = text;
    while (curr_char_pos < length) {
        curr_line_width += get_char_width(*text_pos);
        if (curr_line_width <= break_width) {
            ++curr_char_pos;
            char curr_char = *(text_pos++);
            if (curr_char == '\n') {
                return curr_char_pos;
            }
            if (curr_char == '-' ||
                    (x_align <= 0 && (curr_char == ' ' || *text_pos == ' ')) ||
                    (*text_pos == ' ' && curr_char == ' ')
               ) {
                // May break here.
                last_break_pos = curr_char_pos;
            }
        } else {
            if (last_break_pos == 0) {
                // If the first word is wider than break_width then break it.
                return curr_char_pos;
            } else {
                // Otherwise break at the last space.
                return last_break_pos;
            }
        }
    }
    return curr_char_pos;
}

// seg009:403F
int get_line_width(const char *text, int length) {
    int width = 0;
    const char* text_pos = text;
    while (--length >= 0) {
        width += get_char_width(*(text_pos++));
    }
    return width;
}

// seg009:3706
int draw_text_character(byte character) {
    font_type* font = textstate.ptr_font;
    int width = 0;
    if (character <= font->last_char && character >= font->first_char) {
        image_type* image = font->chtab->images[character - font->first_char]; //char_ptrs[character - font->first_char];
        if (image != NULL) {
            method_3_blit_mono(image, textstate.current_x, textstate.current_y - font->height_above_baseline, textstate.textblit, textstate.textcolor);
            width = font->space_between_chars + image->w;
        }
    }
    textstate.current_x += width;
    return width;
}

// seg009:377F
int draw_text_line(const char *text, int length) {
    int width = 0;
    const char* text_pos = text;
    while (--length >= 0) {
        width += draw_text_character(*(text_pos++));
    }
    return width;
}

// seg009:3755
int draw_cstring(const char *string) {
    int width = 0;
    const char* text_pos = string;
    while (*text_pos) {
        width += draw_text_character(*(text_pos++));
    }
    return width;
}

// seg009:3F01
const rect_type* draw_text(const rect_type *rect_ptr, int x_align, int y_align, const char *text, int length) {
    short rect_top;
    short rect_height;
    short rect_width;
    short num_lines;
    short font_line_distance;
    set_clip_rect(rect_ptr);
    rect_width = rect_ptr->right - rect_ptr->left;
    rect_top = rect_ptr->top;
    rect_height = rect_ptr->bottom - rect_ptr->top;
    num_lines = 0;
    int rem_length = length;
    const char* line_start = text;
#define MAX_LINES 100
    const char* line_starts[MAX_LINES];
    int line_lengths[MAX_LINES];
    do {
        if (num_lines >= MAX_LINES) break;
        int line_length = find_linebreak(line_start, rem_length, rect_width, x_align);
        if (line_length == 0) break;
        line_starts[num_lines] = line_start;
        line_lengths[num_lines] = line_length;
        ++num_lines;
        line_start += line_length;
        rem_length -= line_length;
    } while(rem_length);
    font_type* font = textstate.ptr_font;
    font_line_distance = font->height_above_baseline + font->height_below_baseline + font->space_between_lines;
    int text_height = font_line_distance * num_lines - font->space_between_lines;
    int text_top = rect_top;
    if (y_align >= 0) {
        if (y_align <= 0) {
            // middle
            // The +1 is for simulating SHR + ADC/SBB.
            text_top += (rect_height + 1) / 2 - (text_height + 1) / 2;
        } else {
            // bottom
            text_top += rect_height - text_height;
        }
    }
    textstate.current_y = text_top + font->height_above_baseline;
    int i;
    for (i = 0; i < num_lines; ++i) {
        const char* line_pos = line_starts[i];
        int line_length = line_lengths[i];
        if (x_align < 0 &&
                *line_pos == ' ' &&
                i != 0 &&
                *(line_pos - 1) != '\n'
           ) {
            // Skip over space if it's not at the beginning of a line.
            ++line_pos;
            --line_length;
            if (line_length != 0 &&
                    *line_pos == ' ' &&
                    *(line_pos - 2) == '.'
               ) {
                // Skip over second space after point.
                ++line_pos;
                --line_length;
            }
        }
        int line_width = get_line_width(line_pos, line_length);
        int text_left = rect_ptr->left;
        if (x_align >= 0) {
            if (x_align <= 0) {
                // center
                text_left += rect_width / 2 - line_width / 2;
                if (rect_ptr == &rect_bottom_text && length < 28) {
                    --text_left;
                }
            } else {
                // right
                text_left += rect_width - line_width;
            }
        }
        textstate.current_x = text_left;
        draw_text_line(line_pos, line_length);
        textstate.current_y += font_line_distance;
    }
    reset_clip_rect();
    return rect_ptr;
}

// seg009:3E4F
void show_text(const rect_type *rect_ptr, int x_align, int y_align, const char *text) {
    draw_text(rect_ptr, x_align, y_align, text, (int) strlen(text));
}

// seg009:04FF
void show_text_with_color(const rect_type *rect_ptr, int x_align, int y_align, const char *text, int color) {
    short saved_textcolor;
    saved_textcolor = textstate.textcolor;
    textstate.textcolor = color;
    show_text(rect_ptr, x_align, y_align, text);
    textstate.textcolor = saved_textcolor;
}

// seg009:3A91
void set_curr_pos(int xpos, int ypos) {
    textstate.current_x = xpos;
    textstate.current_y = ypos;
}

// custom logic - about dialog has a different size and font
void init_dialogs(void) {
    copyprot_dialog = make_dialog_info(&dialog_settings, &dialog_rect_1, NULL);
    copyprot_dialog->peel = read_peel_from_screen(&copyprot_dialog->peel_rect);
    about_dialog = make_dialog_info(&dialog_settings, &dialog_rect_2, NULL);
    about_dialog->peel = read_peel_from_screen(&about_dialog->peel_rect);
}

// seg009:0838
int show_message(char *text, byte which) {
    word key;
    word joy_key;
    rect_type rect;
    dialog_type* target_dialog;
    //font_type* saved_font_ptr;
    //surface_type* old_target;
    //old_target = current_target_surface;
    //current_target_surface = onscreen_surface_;
    // In the disassembly there is some messing with the current_target_surface and font (?)
    // However, this does not seem to be strictly necessary
    target_dialog = which ? about_dialog : copyprot_dialog;
    method_1_blit_rect(offscreen_surface, onscreen_surface_, &target_dialog->peel_rect, &target_dialog->peel_rect, 0);
    draw_dialog_frame(target_dialog);
    //saved_font_ptr = textstate.ptr_font;
    //saved_font_ptr = current_target_surface->ptr_font;
    //current_target_surface->ptr_font = ptr_font;
    shrink2_rect(&rect, &target_dialog->text_rect, 2, 1);
    show_text_with_color(&rect, text_align_center, text_align_center, text, color_15_brightwhite);
    //textstate.ptr_font = saved_font_ptr;
    //current_target_surface->ptr_font = saved_font_ptr;
    clear_kbd_buf();
    mouse_clicked = 0;
    mouse_button_clicked_right = 0;
    do {
        idle();
        joy_key = joy_button_states[joyinput_a] & keystate_held ||
                  joy_button_states[joyinput_y] & keystate_held ||
                  joy_button_states[joyinput_x] & keystate_held ||
                  joy_button_states[joyinput_b] & keystate_held ||
                  joy_button_states[joyinput_left_shoulder] & keystate_held ||
                  joy_button_states[joyinput_right_shoulder] & keystate_held ||
                  joy_axis[SDL_CONTROLLER_AXIS_TRIGGERLEFT] > settings->controller_threshold ||
                  joy_axis[SDL_CONTROLLER_AXIS_TRIGGERRIGHT] > settings->controller_threshold;
        key = key_test_quit(); // Press any key to continue...
    } while (key == 0 && joy_key == 0 && mouse_clicked == 0 && mouse_button_clicked_right == 0);
    //restore_dialog_peel_2(copyprot_dialog->peel);
    //current_target_surface = old_target;
    need_full_redraw = 1; // lazy: instead of neatly restoring only the relevant part, just redraw the whole screen
    return key;
}

// seg009:08FB
dialog_type* make_dialog_info(dialog_settings_type *settings, rect_type *text_rect, peel_type *dialog_peel) {
    dialog_type* dialog_info;
    dialog_info = malloc(sizeof(dialog_type));
    dialog_info->settings = settings;
    dialog_info->has_peel = 0;
    dialog_info->peel = dialog_peel;
    if (text_rect != NULL)
        dialog_info->text_rect = *text_rect;
    calc_dialog_peel_rect(dialog_info);
    if (text_rect != NULL) { // does not seem to be quite right; see seg009:0948 (?)
        read_dialog_peel(dialog_info);
    }
    return dialog_info;
}

// seg009:0BE7
void calc_dialog_peel_rect(dialog_type*dialog) {
    dialog_settings_type* settings;
    settings = dialog->settings;
    dialog->peel_rect.left = dialog->text_rect.left - settings->left_border;
    dialog->peel_rect.top = dialog->text_rect.top - settings->top_border;
    dialog->peel_rect.right = dialog->text_rect.right + settings->right_border + settings->shadow_right;
    dialog->peel_rect.bottom = dialog->text_rect.bottom + settings->bottom_border + settings->shadow_bottom;
}

// seg009:0BB0
void read_dialog_peel(dialog_type *dialog) {
    if (dialog->has_peel) {
        if (dialog->peel == NULL) {
            dialog->peel = read_peel_from_screen(&dialog->peel_rect);
        }
        dialog->has_peel = 1;
        draw_dialog_frame(dialog);
    }
}

// seg009:09DE
void draw_dialog_frame(dialog_type *dialog) {
    dialog->settings->method_2_frame(dialog);
}

// A pointer to this function is the first field of dialog_settings (data:2944)
// Perhaps used when replacing a dialog's text with another text (?)
// seg009:096F
void add_dialog_rect(dialog_type *dialog) {
    draw_rect(&dialog->text_rect, color_0_black);
}

// seg009:09F0
void dialog_method_2_frame(dialog_type *dialog) {
    rect_type rect;
    short shadow_right = dialog->settings->shadow_right;
    short shadow_bottom = dialog->settings->shadow_bottom;
    short bottom_border = dialog->settings->bottom_border;
    short outer_border = dialog->settings->outer_border;
    short peel_top = dialog->peel_rect.top;
    short peel_left = dialog->peel_rect.left;
    short peel_bottom = dialog->peel_rect.bottom;
    short peel_right = dialog->peel_rect.right;
    short text_top = dialog->text_rect.top;
    short text_left = dialog->text_rect.left;
    short text_bottom = dialog->text_rect.bottom;
    short text_right = dialog->text_rect.right;
    // Draw outer border
    rect = (rect_type) {
        peel_top, peel_left, peel_bottom - shadow_bottom, peel_right - shadow_right
    };
    draw_rect(&rect, color_0_black);
    // Draw shadow (right)
    rect = (rect_type) {
        text_top, peel_right - shadow_right, peel_bottom, peel_right
    };
    draw_rect(&rect, get_text_color(color_8_darkgray /*dialog's shadow*/, 0));
    // Draw shadow (bottom)
    rect = (rect_type) {
        peel_bottom - shadow_bottom, text_left, peel_bottom, peel_right
    };
    draw_rect(&rect, get_text_color(color_8_darkgray /*dialog's shadow*/, 0));
    // Draw inner border (left)
    rect = (rect_type) {
        peel_top + outer_border, peel_left + outer_border, text_bottom, text_left
    };
    draw_rect(&rect, color_15_brightwhite);
    // Draw inner border (top)
    rect = (rect_type) {
        peel_top + outer_border, text_left, text_top, text_right + dialog->settings->right_border - outer_border
    };
    draw_rect(&rect, color_15_brightwhite);
    // Draw inner border (right)
    rect.top = text_top;
    rect.left =  text_right;
    rect.bottom = text_bottom + bottom_border - outer_border;           // (rect.right stays the same)
    draw_rect(&rect, color_15_brightwhite);
    // Draw inner border (bottom)
    rect = (rect_type) {
        text_bottom, peel_left + outer_border, text_bottom + bottom_border - outer_border, text_right
    };
    draw_rect(&rect, color_15_brightwhite);
}

// seg009:0C44
void show_dialog(const char *text, byte which) {
    char string[256];
    if (which > 1) {
        show_message((char*) text, which - 2);
        return;
    }
    snprintf(string, sizeof(string), "%s\n\nPress any key to go back", text);
    show_message(string, which);
}

// seg009:0791
int get_text_center_y(const rect_type *rect) {
    const font_type *font;
    short empty_height; // height of empty space above+below the line of text
    font = &hc_font;//current_target_surface->ptr_font;
    empty_height = rect->bottom - font->height_above_baseline - font->height_below_baseline - rect->top;
    return ((empty_height - empty_height % 2) >> 1) + font->height_above_baseline + empty_height % 2 + rect->top;
}

// seg009:3E77
int get_cstring_width(const char *text) {
    int width = 0;
    const char* text_pos = text;
    char curr_char;
    while (0 != (curr_char = *(text_pos++))) {
        width += get_char_width(curr_char);
    }
    return width;
}

// seg009:0767
void draw_text_cursor(int xpos, int ypos, int color) {
    set_curr_pos(xpos, ypos);
    textstate.textcolor = color;
    draw_text_character('_');
    textstate.textcolor = 15;
}

// seg009:053C
int input_str(const rect_type *rect, char *buffer, int max_length, const char *initial, int has_initial, int x_offset, int color, int bgcolor) {
    short length;
    word key;
    short cursor_visible;
    short current_xpos;
    short ypos;
    short init_length;
    length = 0;
    cursor_visible = 0;
    draw_rect(rect, bgcolor);
    init_length = strlen(initial);
    if (has_initial) {
        strcpy(buffer, initial);
        length = init_length;
    }
    current_xpos = rect->left + x_offset;
    ypos = get_text_center_y(rect);
    set_curr_pos(current_xpos, ypos);
    textstate.textcolor = color;
    draw_cstring(initial);
    current_xpos += get_cstring_width(initial) + (init_length != 0) * x_offset;
    do {
        key = 0;
        do {
            if (cursor_visible) {
                draw_text_cursor(current_xpos, ypos, color);
            } else {
                draw_text_cursor(current_xpos, ypos, bgcolor);
            }
            cursor_visible = !cursor_visible;
            start_timer(timer_0, 6);
            if (key) {
                if (cursor_visible) {
                    draw_text_cursor(current_xpos, ypos, color);
                    cursor_visible = !cursor_visible;
                }
                if (key == SDL_SCANCODE_RETURN || key == SDL_SCANCODE_KP_ENTER) { // enter
                    buffer[length] = 0;
                    return length;
                } else break;
            }
            while (!has_timer_stopped(timer_0) && (key = key_test_quit()) == 0) idle();
        } while (1);
        // Only use the printable ASCII chars (UTF-8 encoding)
        char entered_char = last_text_input <= 0x7E ? last_text_input : 0;
        clear_kbd_buf();

        if (key == SDL_SCANCODE_ESCAPE) { // esc
            draw_rect(rect, bgcolor);
            buffer[0] = 0;
            return -1;
        }
        if (length != 0 && (key == SDL_SCANCODE_BACKSPACE ||
                            key == SDL_SCANCODE_DELETE)) { // backspace, delete
            --length;
            draw_text_cursor(current_xpos, ypos, bgcolor);
            current_xpos -= get_char_width(buffer[length]);
            set_curr_pos(current_xpos, ypos);
            textstate.textcolor = bgcolor;
            draw_text_character(buffer[length]);
            draw_text_cursor(current_xpos, ypos, color);
        } else if (entered_char >= 0x20 && entered_char <= 0x7E && length < max_length) {
            // Would the new character make the cursor go past the right side of the rect?
            if (get_char_width('_') + get_char_width(entered_char) + current_xpos < rect->right) {
                draw_text_cursor(current_xpos, ypos, bgcolor);
                set_curr_pos(current_xpos, ypos);
                textstate.textcolor = color;
                current_xpos += draw_text_character(buffer[length++] = entered_char);
            }
        }
    } while(1);
}

// seg009:37E8
void draw_rect(const rect_type *rect, int color) {
    method_5_rect(rect, blitters_0_no_transp, color);
}

// seg009:39CE
rect_type* shrink2_rect(rect_type *target_rect, const rect_type *source_rect, int delta_x, int delta_y) {
    target_rect->top    = source_rect->top    + delta_y;
    target_rect->left   = source_rect->left   + delta_x;
    target_rect->bottom = source_rect->bottom - delta_y;
    target_rect->right  = source_rect->right  - delta_x;
    return target_rect;
}

// seg009:3BBA
void restore_peel(peel_type* peel_ptr) {
    method_6_blit_img_to_scr(peel_ptr->peel, peel_ptr->rect.left, peel_ptr->rect.top, /*0x10*/0);
    free_peel(peel_ptr);
}

// seg009:3BE9
peel_type* read_peel_from_screen(const rect_type *rect) {
    peel_type* result;
    result = calloc(1, sizeof(peel_type));
    result->rect = *rect;
    SDL_Surface* peel_surface = SDL_CreateRGBSurface(0, rect->right - rect->left, rect->bottom - rect->top,
                                24, 0xFF, 0xFF << 8, 0xFF << 16, 0);
    if (peel_surface == NULL) {
        sdlperror("SDL_CreateRGBSurface");
        quit(1);
    }
    result->peel = peel_surface;
    rect_type target_rect = {0, 0, rect->right - rect->left, rect->bottom - rect->top};
    method_1_blit_rect(result->peel, current_target_surface, &target_rect, rect, 0);
    return result;
}

// seg009:3D95
int intersect_rect(rect_type *output, const rect_type *input1, const rect_type *input2) {
    short left = MAX(input1->left, input2->left);
    short right = MIN(input1->right, input2->right);
    if (left < right) {
        output->left = left;
        output->right = right;
        short top = MAX(input1->top, input2->top);
        short bottom = MIN(input1->bottom, input2->bottom);
        if (top < bottom) {
            output->top = top;
            output->bottom = bottom;
            return 1;
        }
    }
    memset(output, 0, sizeof(rect_type));
    return 0;
}

// seg009:4063
rect_type * union_rect(rect_type *output, const rect_type *input1, const rect_type *input2) {
    short top = MIN(input1->top, input2->top);
    short left = MIN(input1->left, input2->left);
    short bottom = MAX(input1->bottom, input2->bottom);
    short right = MAX(input1->right, input2->right);
    output->top = top;
    output->left = left;
    output->bottom = bottom;
    output->right = right;
    return output;
}

enum userevents {
    userevent_SOUND,
    userevent_TIMER,
};

short digi_playing = 0;
short midi_playing = 0;
short ogg_playing = 0;

int prev_music_suspended = -1;

// The current buffer, holds the resampled sound data.
byte* digi_buffer = NULL;
// The current size of the buffer (may grow if needed).
size_t digi_buffer_size;
// The current position in digi_buffer.
byte* digi_remaining_pos = NULL;
// The remaining length.
size_t digi_remaining_length = 0;

// The properties of the audio device.
SDL_AudioSpec* digi_audiospec = NULL;
// The desired samplerate. Everything will be resampled to this.
const int digi_samplerate = 44100;

void stop_digi(void) {
    if (!digi_playing) return;
    SDL_LockAudio();
    digi_playing = 0;
    if (prev_music_suspended > -1) {
        music_suspended = prev_music_suspended;
    }
    digi_remaining_length = 0;
    digi_remaining_pos = NULL;
    SDL_UnlockAudio();
}

// Decoder for the currently playing OGG sound. (This also holds the playback position.)
stb_vorbis* ogg_decoder;

void stop_ogg(void) {
    SDL_PauseAudio(1);
    if (!ogg_playing) return;
    ogg_playing = 0;
    if (prev_music_suspended > -1 && current_sound != sound_36_out_of_time) {
        music_suspended = prev_music_suspended;
    }
    SDL_LockAudio();
    ogg_decoder = NULL;
    SDL_UnlockAudio();
}

// seg009:7214
void stop_sounds(void) {
    // custom logic - do not interrupt out of time sound on level 10
    if (current_level == 10 && level_restart_timer > 0) {
        return;
    }
    stop_digi();
    stop_midi();
    stop_ogg();
}

// custom logic - stop all sounds including the delayed ones
void stop_all_sounds(void) {
    stop_digi();
    stop_midi();
    stop_ogg();

    if (feather_fall_timer || ogg_fore_sound_playing) {
        ogg_fore_sound_playing = false;
        Mix_HaltMusic();
    }
    if (delayed_sound_timer > 0 || delayed_sound.sound_id >= 0) {
        delayed_sound_timer = 0;
        delayed_sound = (struct delayed_sound_type) {
            -1, 0, 0, 0, 0, 0, 0
            };
    }
}

// Background: https://auditoryneuroscience.com/spatial-hearing/binaural-cues
int digi_channel_offset[2]; // for interaural time difference
float digi_channel_gain[2]; // for interaural level difference, and for sounds becoming softer in the distance
float max_interaural_time_difference_ms = 0.4f;

void reset_sound_direction(void) {
    memset(digi_channel_offset, 0, sizeof(digi_channel_offset));
    digi_channel_gain[0] = 1.0f;
    digi_channel_gain[1] = 1.0f;
}

size_t set_sound_direction(float sound_dir_x) {
    digi_channel_gain[0] = current_sound_distance_loss;
    digi_channel_gain[1] = current_sound_distance_loss;
    if (sound_dir_x == 0.0f) return 0;
    float samples_per_ms = (float) digi_audiospec->freq * 0.001f;
    float max_delay_samples = max_interaural_time_difference_ms * samples_per_ms;
    int which_channel_to_offset;
    int which_channel_to_make_softer;

    float abs_sound_dir_x;
    if (sound_dir_x >= 0.0f) {
        // Sound comes from the right. -> give the right channel a head start, make the left channel softer
        abs_sound_dir_x = sound_dir_x;
        which_channel_to_offset = 1;
        which_channel_to_make_softer = 0;
    } else {
        // Sound comes from the left. -> give the left channel a head start, make the right channel softer
        abs_sound_dir_x = -sound_dir_x;
        which_channel_to_offset = 0;
        which_channel_to_make_softer = 1;
    }
    int delay_samples = (int) (abs_sound_dir_x * max_delay_samples);
    size_t delay_length = (delay_samples * digi_audiospec->channels * sizeof(short));
    digi_channel_offset[which_channel_to_offset] = delay_length;
    digi_channel_gain[which_channel_to_make_softer] *= (1.0f - (abs_sound_dir_x * 0.2f)); // could tweak this
    return delay_length;
}

void digi_callback(void *userdata, Uint8 *stream, int len) {
    (void) userdata;
    // Don't go over the end of either the input or the output buffer.
    size_t copy_len = MIN((size_t) len, digi_remaining_length);
    if (is_sound_on) {
        // Copy the next part of the input of the output.
        if (enable_positional_audio) {
            size_t sample_len = 2 * sizeof(short);
            int offset_left = digi_channel_offset[0];
            int offset_right = digi_channel_offset[1];
            byte* buffer_pos = digi_remaining_pos;
            short* stream_pos = (short*) stream;
            for (size_t bytes_left_to_copy = copy_len; bytes_left_to_copy > 0; bytes_left_to_copy -= sample_len) {
                short left = *(short*)(buffer_pos + offset_left);
                short right = *(short*)(buffer_pos + offset_right + sizeof(short));
                left = MIN(MAX((int) (left * digi_channel_gain[0]), INT16_MIN), INT16_MAX);
                right = MIN(MAX((int) (right * digi_channel_gain[1]), INT16_MIN), INT16_MAX);
                *stream_pos++ = left;
                *stream_pos++ = right;
                buffer_pos += sample_len;
            }
        } else {
            memcpy(stream, digi_remaining_pos, copy_len);
        }
        // In case the sound does not fill the buffer: fill the rest of the buffer with silence.
        memset(stream + copy_len, digi_audiospec->silence, len - copy_len);
    } else {
        // If sound is off: Mute the sound but keep track of where we are.
        memset(stream, digi_audiospec->silence, len);
    }
    // If the sound ended, push an event.
    if (digi_playing && digi_remaining_length == 0) {
        SDL_Event event;
        memset(&event, 0, sizeof(event));
        event.type = SDL_USEREVENT;
        event.user.code = userevent_SOUND;
        digi_playing = 0;
        if (prev_music_suspended > -1) {
            music_suspended = prev_music_suspended;
        }
        SDL_PushEvent(&event);
    }
    // Advance the pointer.
    digi_remaining_length -= copy_len;
    digi_remaining_pos += copy_len;
}

void ogg_callback(void *userdata, Uint8 *stream, int len) {
    (void) userdata;

    int output_channels = digi_audiospec->channels;
    int bytes_per_sample = sizeof(short) * output_channels;
    int samples_requested = len / bytes_per_sample;

    int samples_filled;
    if (is_sound_on) {
        samples_filled = stb_vorbis_get_samples_short_interleaved(ogg_decoder, output_channels,
                         (short*) stream, len / sizeof(short));
        if (samples_filled < samples_requested) {
            // In case the sound does not fill the buffer: fill the rest of the buffer with silence.
            int bytes_filled = samples_filled * bytes_per_sample;
            int remaining_bytes = (samples_requested - samples_filled) * bytes_per_sample;
            memset(stream + bytes_filled, digi_audiospec->silence, remaining_bytes);
        }
    } else {
        // If sound is off: Mute the sound, but keep track of where we are.
        memset(stream, digi_audiospec->silence, len);
        // Let the decoder run normally (to advance the position), but discard the result.
        byte* discarded_samples = alloca(len);
        samples_filled = stb_vorbis_get_samples_short_interleaved(ogg_decoder, output_channels,
                         (short*) discarded_samples, len / sizeof(short));
    }
    // Push an event if the sound has ended.
    if (samples_filled == 0) {
        SDL_Event event;
        memset(&event, 0, sizeof(event));
        event.type = SDL_USEREVENT;
        event.user.code = userevent_SOUND;
        ogg_playing = 0;
        if (prev_music_suspended > -1) {
            music_suspended = prev_music_suspended;
        }
        SDL_PushEvent(&event);
    }
}

void audio_callback(void* userdata, Uint8* stream, int len) {
    memset(stream, digi_audiospec->silence, len);
    if (digi_playing) {
        digi_callback(userdata, stream, len);
    }
    // Note: music sounds and digi sounds are allowed to play simultaneously (will be blended together)
    // I.e., digi sounds and music will not cut each other short.
    if (midi_playing) {
        midi_callback(userdata, stream, len);
    } else if (ogg_playing) {
        ogg_callback(userdata, stream, len);
    }
}

int digi_unavailable = 0;
void init_digi(void) {
    if (digi_unavailable) return;
    if (digi_audiospec != NULL) return;
    // Open the audio device. Called once.

    SDL_AudioFormat desired_audioformat;
    SDL_version version;
    SDL_GetVersion(&version);

#if SDL_VERSION_ATLEAST(2,0,4)
    desired_audioformat = AUDIO_S16SYS;
#else
    desired_audioformat = AUDIO_U8;
    printf("Your SDL.dll is older than 2.0.4. Using 8-bit audio format to work around resampling bug.");
#endif

    SDL_AudioSpec *desired;
    desired = (SDL_AudioSpec *)malloc(sizeof(SDL_AudioSpec));
    memset(desired, 0, sizeof(SDL_AudioSpec));
    desired->freq = digi_samplerate; //buffer->digi.sample_rate;
    desired->format = desired_audioformat;
    desired->channels = 2;
    desired->samples = 1024;
    desired->callback = audio_callback;
    desired->userdata = NULL;
    if (SDL_OpenAudio(desired, NULL) != 0) {
        sdlperror("SDL_OpenAudio");
        digi_unavailable = 1;
        return;
    }
    digi_audiospec = desired;
    digi_buffer_size = 1024 * 1024;
    digi_buffer = malloc(digi_buffer_size); // persistent
}

const int sound_channel = 0;

void load_sound_names(void) {
    // custom logic - assumes sound/foreground music files are named SOUND??.OGG where ? is a digit
    sound_names = (char**) calloc(sizeof(char*) * (MAX_FORE_SOUND_ID + 1), 1);
    char name[8];
    int index;
    for (index = 24; index <= MAX_FORE_SOUND_ID; index++) {
        snprintf(name, sizeof(name), "SOUND%d", index);
        sound_names[index] = strdup(name);
    }
}

char* sound_name(int index) {
    if (sound_names != NULL && index >= 0 && index <= MAX_FORE_SOUND_ID) {
        return sound_names[index];
    } else {
        return NULL;
    }
}

sound_buffer_type* convert_digi_sound(sound_buffer_type* digi_buffer);

sound_buffer_type* load_sound(int index) {
    sound_buffer_type* result = NULL;
    init_digi();
    if (!digi_unavailable && result == NULL && index >= 0 && index <= MAX_FORE_SOUND_ID) {
        if (enable_positional_audio && sound_names != NULL && sound_name(index) != NULL) {
            do {
                char filename[POP_MAX_PATH];
                snprintf(filename, sizeof(filename), "%s.OGG", sound_name(index));
                size_t file_size;
                void* file_contents = get_enhanced_sound_data(filename, &file_size);
                if (file_contents == NULL) {
                    break;
                }
                // Decoding the entire file immediately would make the loading time much longer.
                // However, we can also create the decoder now, and only use it when we are actually playing the file.
                // (In the audio callback, we'll decode chunks of samples to the output stream, as needed).
                stb_vorbis* decoder = stb_vorbis_open_memory(file_contents, (int) file_size, NULL, NULL);
                if (decoder == NULL) {
                    free(file_contents);
                    break;
                }
                result = malloc(sizeof(sound_buffer_type));
                result->type = sound_ogg;
                result->ogg.total_length = stb_vorbis_stream_length_in_samples(decoder) * sizeof(short);
                result->ogg.file_contents = file_contents; // Remember in case we want to free the sound later.
                result->ogg.decoder = decoder;
            } while(0); // do once (breakable block)
        }
    }
    if (result == NULL) {
        result = (sound_buffer_type*) load_from_opendats_alloc(index + 10000, "bin", NULL, NULL);
    }
    if (result != NULL && (result->type & 7) == sound_digi) {
        sound_buffer_type* converted = convert_digi_sound(result);
        free(result);
        result = converted;
    }
    if (result == NULL) {
        fprintf(stderr, "Failed to load sound %d '%s'\n", index, sound_name(index));
    }
    return result;
}

void play_ogg_sound(sound_buffer_type *buffer) {
    init_digi();
    if (digi_unavailable) return;

    stop_sounds();

    // Need to rewind the music, or else the decoder might continue where it left off, the last time this sound played.
    stb_vorbis_seek_start(buffer->ogg.decoder);

    SDL_LockAudio();
    ogg_decoder = buffer->ogg.decoder;
    SDL_UnlockAudio();
    SDL_PauseAudio(0);

    // custom logic - stop music on certain sounds
    if ((current_sound == sound_25_presentation ||
            current_sound == sound_26_embrace ||
            current_sound == sound_27_cutscene_2_4_6_12 ||
            current_sound == sound_29_meet_Jaffar ||
            current_sound == sound_30_big_potion ||
            current_sound == sound_32_mouse_music ||
            current_sound == sound_35_cutscene_8_9 ||
            current_sound == sound_36_out_of_time ||
            current_sound == sound_37_victory ||
            current_sound == sound_39_low_weight ||
            current_sound == sound_40_cutscene_12_short_time ||
            current_sound == sound_41_end_level_music ||
            current_sound == sound_43_victory_Jaffar ||
            (current_sound >= sound_50_story_2_princess && current_sound <= sound_56_ending_music)) &&
            (music_mode >= 0 || is_heartbeat_music())
       ) {
        prev_music_suspended = !ogg_fore_sound_playing ? music_suspended : 0;
        music_suspended = 1;
        stop_ambient_music();
    }

    ogg_playing = 1;
}

int wave_version = -1;

typedef struct waveinfo_type {
    int sample_rate, sample_size, sample_count;
    byte* samples;
} waveinfo_type;

bool determine_wave_version(sound_buffer_type *buffer, waveinfo_type* waveinfo) {
    int version = wave_version;
    if (version == -1) {
        // Determine the version of the wave data.
        version = 0;
        if (buffer->digi.sample_size == 8) version += 1;
        if (buffer->digi_new.sample_size == 8) version += 2;
        if (version == 1 || version == 2) wave_version = version;
    }

    switch (version) {
        case 1: // 1.0 and 1.1
            waveinfo->sample_rate = buffer->digi.sample_rate;
            waveinfo->sample_size = buffer->digi.sample_size;
            waveinfo->sample_count = buffer->digi.sample_count;
            waveinfo->samples = buffer->digi.samples;
            return true;
        case 2: // 1.3 and 1.4 (and PoP2)
            waveinfo->sample_rate = buffer->digi_new.sample_rate;
            waveinfo->sample_size = buffer->digi_new.sample_size;
            waveinfo->sample_count = buffer->digi_new.sample_count;
            waveinfo->samples = buffer->digi_new.samples;
            return true;
        default: // case 0, ambiguous or unknown
            printf("Warning: Can't determine wave version.\n");
            return false;
    }
}

sound_buffer_type* convert_digi_sound(sound_buffer_type* digi_buffer) {
    init_digi();
    if (digi_unavailable) return NULL;
    waveinfo_type waveinfo;
    if (false == determine_wave_version(digi_buffer, &waveinfo)) return NULL;

    float freq_ratio = (float)waveinfo.sample_rate /  (float)digi_audiospec->freq;

    int source_length = waveinfo.sample_count;
    int expanded_frames = source_length * digi_audiospec->freq / waveinfo.sample_rate;
    int expanded_length = expanded_frames * 2 * sizeof(short);
    sound_buffer_type* converted_buffer = malloc(sizeof(sound_buffer_type) + expanded_length);

    converted_buffer->type = sound_digi_converted;
    converted_buffer->converted.length = expanded_length;

    byte* source = waveinfo.samples;
    short* dest = converted_buffer->converted.samples;

    for (int i = 0; i < expanded_frames; ++i) {
        float src_frame_float = i * freq_ratio;
        int src_frame_0 = (int) src_frame_float; // truncation

        int sample_0 = (source[src_frame_0] | (source[src_frame_0] << 8)) - 32768;
        short interpolated_sample;
        if (src_frame_0 >= waveinfo.sample_count - 1) {
            interpolated_sample = (short)sample_0;
        } else {
            int src_frame_1 = src_frame_0 + 1;
            float alpha = src_frame_float - src_frame_0;
            int sample_1 = (source[src_frame_1] | (source[src_frame_1] << 8)) - 32768;
            interpolated_sample = (short)((1.0f - alpha) * sample_0 + alpha * sample_1);
        }
        for (int channel = 0; channel < digi_audiospec->channels; ++channel) {
            *dest++ = interpolated_sample;
        }
    }

    return converted_buffer;
}

// seg009:74F0
void play_digi_sound(sound_buffer_type *buffer) {
    init_digi();
    if (digi_unavailable) return;
    stop_digi();
    if ((buffer->type & 7) != sound_digi_converted) {
        printf("Tried to play unconverted digi sound.\n");
        return;
    }
    SDL_LockAudio();
    size_t delay_length = 0; // for interaural time difference
    size_t total_length = buffer->converted.length;
    reset_sound_direction();
    if (current_sound_is_positional) {
        delay_length = set_sound_direction(current_sound_dir_x);
        total_length += 2 * delay_length; // pad with silence at start + end.
    }
    // custom logic - sound of potion being put down
    if (current_sound == sound_17_soft_land &&
            (get_tile_at_kid(Kid.x) == tiles_10_potion ||
             get_tile_at_kid(Kid.x + 14 * dir_front[Kid.direction + 1]) == tiles_10_potion) &&
            curr_room_modif[curr_tilepos] >= 0x38 // empty
       ) {
        delayed_sound_timer = 1;
        delayed_sound = (struct delayed_sound_type) {
            -2, 0, 0, 0, 0, 0, 0
            };
    }
    // Grow the buffer if needed.
    if (total_length > digi_buffer_size) {
        byte* new_digi_buffer = realloc(digi_buffer, total_length);
        if (new_digi_buffer == NULL) {
            printf("digi_buffer: realloc failed!");
            quit(1);
        }
        digi_buffer = new_digi_buffer;
        digi_buffer_size = total_length;
    }
    if (delay_length > 0) {
        // pad with silence at start + end.
        memset(digi_buffer, 0, delay_length);
        memset(digi_buffer + delay_length + buffer->converted.length, 0, delay_length);
    }
    memcpy(digi_buffer + delay_length, buffer->converted.samples, buffer->converted.length);
    digi_playing = 1;
    digi_remaining_length = buffer->converted.length + delay_length;
    digi_remaining_pos = digi_buffer;
    SDL_UnlockAudio();
    SDL_PauseAudio(0);
    // custom logic - stop music on certain sounds
    if (current_sound == sound_15_leveldoor_sliding &&
            (music_mode >= 0 || is_heartbeat_music())) {
        prev_music_suspended = !ogg_fore_sound_playing ? music_suspended : 0;
        music_suspended = 1;
        stop_ambient_music();
    }
}

// seg009:7220
void play_sound_from_buffer(sound_buffer_type *buffer) {
    if (buffer == NULL) {
        return;
    }
    switch (buffer->type & 7) {
        case sound_speaker:
            // custom logic - do not play PC speaker sounds. All of them are available in WAV, MIDI or OGG format.
            break;
        case sound_digi_converted:
        case sound_digi:
            play_digi_sound(buffer);
            break;
        case sound_midi:
            play_midi_sound(buffer);
            break;
        case sound_ogg:
            play_ogg_sound(buffer);
            break;
        default:
            printf("Tried to play unimplemented sound type %d.\n", buffer->type);
            quit(1);
            break;
    }
}

// seg009:7273
void turn_sound_on_off(byte new_state) {
    is_sound_on = new_state;
    // custom logic - handle ambient music
    if (is_sound_on) {
        start_ambient_music(true);
    } else {
        Mix_HaltMusic();
    }
}

// seg009:7299
int check_sound_playing(void) {
    return digi_playing || midi_playing || ogg_playing;
}

void apply_aspect_ratio(int width) {
    int height;
    // custom logic - native resolution
    if (settings->native_ratio || width < 320) {
        return;
    }
    // custom logic - calculating native resolution at HD or lower
    width = MIN(width, 1920);
    if (settings->aspect_ratio) {
        height = (width / 4) * 3;
    } else {
        height = (width / 16) * 10;
    }
    // Allow us to use a consistent set of screen co-ordinates, even if the screen size changes.
    SDL_RenderSetLogicalSize(renderer_, width, height); // 4:3 vs 16:10
}

void init_overlay(void) {
    static bool initialized = false;
    if (!initialized) {
        overlay_surface = SDL_CreateRGBSurface(0, 320, 200, 32, 0xFF, 0xFF << 8, 0xFF << 16, 0xFFu << 24);
        merged_surface = SDL_CreateRGBSurface(0, 320, 200, 24, 0xFF, 0xFF << 8, 0xFF << 16, 0);
        initialized = true;
    }
}

void init_scaling(void) {
    if (renderer_ == NULL) return;
    if (texture_sharp == NULL) {
        texture_sharp = SDL_CreateTexture(renderer_, SDL_PIXELFORMAT_RGB24, SDL_TEXTUREACCESS_STREAMING, 320, 200);
    }
    target_texture = texture_sharp;
    if (target_texture == NULL) {
        sdlperror("SDL_CreateTexture");
        quit(1);
    }
}

// custom logic - sets a renderer hint
void set_renderer_hint(void) {
    if (strcmp("directx12", settings->renderer) == 0) {
        SDL_SetHint(SDL_HINT_RENDER_DRIVER, "direct3d12");
    } else if (strcmp("directx", settings->renderer) == 0) {
        SDL_SetHint(SDL_HINT_RENDER_DRIVER, "direct3d");
    } else if (strcmp("metal", settings->renderer) == 0) {
        SDL_SetHint(SDL_HINT_RENDER_DRIVER, "metal");
    } else if (strcmp("opengles", settings->renderer) == 0) {
        SDL_SetHint(SDL_HINT_RENDER_DRIVER, "opengles");
    } else if (strcmp("opengl", settings->renderer) == 0) {
        SDL_SetHint(SDL_HINT_RENDER_DRIVER, "opengl");
    }
}

// custom logic - initializes SDL Mixer
void set_music(void) {
    Mix_AllocateChannels(0); // music only
    Mix_Init(MIX_INIT_OGG);
    if (Mix_OpenAudio(44100, MIX_DEFAULT_FORMAT, 2, 2048) < 0) {
        printf("Could not open audio for music");
    }
}

// seg009:38ED
void set_gr_mode(void) {
#ifdef SDL_HINT_WINDOWS_DISABLE_THREAD_NAMING
    SDL_SetHint(SDL_HINT_WINDOWS_DISABLE_THREAD_NAMING, "1");
#endif
    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_NOPARACHUTE |
                 SDL_INIT_GAMECONTROLLER | SDL_INIT_HAPTIC ) != 0) {
        sdlperror("SDL_Init");
        quit(1);
    }

    Uint32 flags = 0;
    if (!settings->run_in_window) {
        flags |= SDL_WINDOW_FULLSCREEN_DESKTOP;
    }
    flags |= SDL_WINDOW_INPUT_FOCUS;
    flags |= SDL_WINDOW_RESIZABLE;
    flags |= SDL_WINDOW_ALLOW_HIGHDPI; // for Retina displays

#ifdef WIN32_LEAN_AND_MEAN
    // Tell Windows that the application is DPI aware, to prevent unwanted bitmap stretching.
    // SetProcessDPIAware() is only available on Windows Vista and later, so we need to load it dynamically.
    // User32.dll has a 64-bit version in System32. They cannot change the name for compatibility reasons.
    typedef BOOL (WINAPI * dpiaware)(void);
    HMODULE user32dll = LoadLibraryA("User32.dll");
    if (user32dll) {
        dpiaware SetProcessDPIAware = (dpiaware)(void *)GetProcAddress(user32dll, "SetProcessDPIAware");
        if (SetProcessDPIAware) {
            SetProcessDPIAware();
        }
        FreeLibrary(user32dll);
    }
#endif

    window_ = SDL_CreateWindow(WINDOW_TITLE,
                               SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED,
                               pop_window_width, pop_window_height, flags);

    // custom logic - double the window size a high resolution monitor
    SDL_DisplayMode displayMode;
    if (SDL_GetCurrentDisplayMode(SDL_GetWindowDisplayIndex(window_), &displayMode) != 0) {
        sdlperror("SDL_GetCurrentDisplayMode");
        quit(1);
    }
    // custom logic - window should be around 75% of the display size at standard resolution values (640-HD)
    if (displayMode.w >= 2560) {
        pop_window_width = 1920;
    } else if (displayMode.w >= 1920) {
        pop_window_width = 1440;
    } else if (displayMode.w >= 1440) {
        pop_window_width = 1024;
    } else if (displayMode.w >= 1024) {
        pop_window_width = 800;
    } else {
        pop_window_width = 640;
    }
    // Should use different default window dimensions when using 4:3 aspect ratio
    if (settings->aspect_ratio) {
        pop_window_height = pop_window_width / 4 * 3;
    } else {
        pop_window_height = pop_window_width / 16 * 10;
    }
    SDL_SetWindowSize(window_, pop_window_width, pop_window_height);
    SDL_SetWindowPosition(window_, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED);
    Uint32 acceleration;
    if (strcmp("software", settings->renderer) == 0) {
        SDL_SetHint(SDL_HINT_RENDER_VSYNC, "0");
        acceleration = SDL_RENDERER_SOFTWARE;
    } else {
        SDL_SetHint(SDL_HINT_RENDER_VSYNC, settings->enable_vsync ? "1" : "0");
        acceleration = SDL_RENDERER_ACCELERATED;
        set_renderer_hint();
    }
    renderer_ = SDL_CreateRenderer(window_, -1, acceleration | SDL_RENDERER_TARGETTEXTURE);
    SDL_RendererInfo renderer_info;
    if (SDL_GetRendererInfo(renderer_, &renderer_info) == 0) {
        // custom logic - stores the renderer name
        if (renderer_info.name != NULL) {
            strcpy(renderer_name, renderer_info.name);
            to_ucase(renderer_name);
            if ((renderer_info.flags & SDL_RENDERER_PRESENTVSYNC) == SDL_RENDERER_PRESENTVSYNC) {
                if ((renderer_info.flags & SDL_RENDERER_ACCELERATED) == SDL_RENDERER_ACCELERATED) {
                    strcat(renderer_name, "/AC");
                }
                strcat(renderer_name, "/VS");
            } else if ((renderer_info.flags & SDL_RENDERER_ACCELERATED) == SDL_RENDERER_ACCELERATED) {
                strcat(renderer_name, "/ACCEL");
            }
        }
    }

    dat_type* dathandle;
    dathandle = open_dat("PRINCE15.DAT");
    image_type* icon = load_custom_graphics_image(image_4005_icon);
    lighting_mask = load_custom_graphics_image(image_4006_lighting_mask);
    close_dat(dathandle);
    if (icon != NULL) {
        SDL_SetWindowIcon(window_, icon);
    }

    apply_aspect_ratio(displayMode.w);

    // give video and input focus to the window
    SDL_RaiseWindow(window_);

    /* Migration to SDL2: everything is still blitted to onscreen_surface_, however:
     * SDL2 renders textures to the screen instead of surfaces; so, every screen
     * update causes the onscreen_surface_ to be copied into the target_texture, which is
     * subsequently displayed.
     * The function handling the screen updates is update_screen()
     * */
    onscreen_surface_ = SDL_CreateRGBSurface(0, 320, 200, 24, 0xFF, 0xFF << 8, 0xFF << 16, 0);
    if (onscreen_surface_ == NULL) {
        sdlperror("SDL_CreateRGBSurface");
        quit(1);
    }
    init_overlay();
    init_scaling();
    if (!settings->run_in_window) {
        SDL_ShowCursor(SDL_DISABLE);
    }
    load_font();
}

SDL_Surface* get_final_surface(void) {
    if (!is_overlay_displayed) {
        return onscreen_surface_;
    } else {
        return merged_surface;
    }
}

void draw_overlay(void) {
    int overlay = 0;
    int expected_numeric_chars;
    int extra_numeric_chars;
    int line_width;
    int ticks_per_sec;
    char timer_text[32];
    rect_type timer_box_rect;
    rect_type timer_text_rect;
    is_overlay_displayed = false;
    if (!is_ending_sequence && is_timer_displayed && start_level > 0) overlay = 1; // Timer overlay
    else if (!is_cutscene && is_feather_timer_displayed && start_level > 0 && feather_fall_timer > 0) overlay = 3; // Feather timer overlay
    else if (!is_cutscene && is_upside_down_timer_displayed && start_level > 0 && upside_down_timer > 0) overlay = 4; // Upside down timer overlay
    // Menu overlay - not drawn here directly, only copied from the overlay surface.
    if (is_paused && is_menu_shown) overlay = 2;
    if (overlay != 0) {
        is_overlay_displayed = true;
        surface_type* saved_target_surface = current_target_surface;
        current_target_surface = overlay_surface;
        rect_type drawn_rect;
        if (overlay == 1) {
            if (rem_min < 0) {
                snprintf(timer_text, sizeof(timer_text), "%02d:%02d:%02d",
                         -(rem_min + 1), (719 - rem_tick) / 12, (719 - rem_tick) % 12);
            } else {
                snprintf(timer_text, sizeof(timer_text), "%02d:%02d:%02d",
                         rem_min - 1, rem_tick / 12, rem_tick % 12);
            }
            expected_numeric_chars = 6;
            extra_numeric_chars = strnlen(timer_text, sizeof(timer_text)) - 8;
            line_width = 5 + (expected_numeric_chars + extra_numeric_chars) * 9;
            timer_box_rect = (struct rect_type) {
                0, 0, 11, 2 + line_width
            };
            timer_text_rect = (struct rect_type) {
                2, 2, 10, 100
            };
            draw_rect_with_alpha(&timer_box_rect, color_0_black, 128);
            show_text(&timer_text_rect, text_align_left, text_align_top, timer_text);
            drawn_rect = timer_box_rect; // Only need to blit this bit to the merged_surface.
        } else if (overlay == 3) { // Feather timer
            ticks_per_sec = get_ticks_per_sec(timer_1);
            snprintf(timer_text, sizeof(timer_text), "%02d:%02d", feather_fall_timer / ticks_per_sec, feather_fall_timer % ticks_per_sec);
            expected_numeric_chars = 6;
            extra_numeric_chars = strnlen(timer_text, sizeof(timer_text)) - 8;
            line_width = 5 + (expected_numeric_chars + extra_numeric_chars) * 9;
            timer_box_rect = (struct rect_type) {
                0, 0, 11, 2 + line_width
            };
            timer_text_rect = (struct rect_type) {
                2, 2, 10, 100
            };
            draw_rect_with_alpha(&timer_box_rect, color_0_black, 128);
            show_text_with_color(&timer_text_rect, text_align_left, text_align_top, timer_text, color_10_brightgreen);
            drawn_rect = timer_box_rect; // Only need to blit this bit to the merged_surface.
        } else if (overlay == 4) { // Upside down timer
            snprintf(timer_text, sizeof(timer_text), "%02d:%02d", upside_down_timer / 12, upside_down_timer % 12);
            expected_numeric_chars = 6;
            extra_numeric_chars = strnlen(timer_text, sizeof(timer_text)) - 8;
            line_width = 5 + (expected_numeric_chars + extra_numeric_chars) * 9;
            timer_box_rect = (struct rect_type) {
                0, 0, 11, 2 + line_width
            };
            timer_text_rect = (struct rect_type) {
                2, 2, 10, 100
            };
            draw_rect_with_alpha(&timer_box_rect, color_0_black, 128);
            show_text_with_color(&timer_text_rect, text_align_left, text_align_top, timer_text, color_2_green);
            drawn_rect = timer_box_rect; // Only need to blit this bit to the merged_surface.
        } else {
            drawn_rect = screen_rect; // We'll blit the whole contents of overlay_surface to the merged_surface.
        }
        SDL_Rect sdl_rect;
        rect_to_sdlrect(&drawn_rect, &sdl_rect);
        SDL_BlitSurface(onscreen_surface_, NULL, merged_surface, NULL);
        SDL_BlitSurface(overlay_surface, &sdl_rect, merged_surface, &sdl_rect);
        current_target_surface = saved_target_surface;
    }
}

void update_screen(void) {
    draw_level_6_sword_animation();
    draw_overlay();
    SDL_Surface* surface = get_final_surface();
    init_scaling();
    SDL_UpdateTexture(target_texture, NULL, surface->pixels, surface->pitch);
    SDL_RenderClear(renderer_);
    SDL_RenderCopy(renderer_, target_texture, NULL, NULL);
    SDL_RenderPresent(renderer_);
}

// seg009:9289
void set_pal_arr(int start, int count, const rgb_type *array) {
    int i;
    for (i = 0; i < count; ++i) {
        if (array) {
            set_pal(start + i, array[i].r, array[i].g, array[i].b);
        } else {
            set_pal(start + i, 0, 0, 0);
        }
    }
}

rgb_type palette[256];

// seg009:92DF
void set_pal(int index, int red, int green, int blue) {
    palette[index].r = red;
    palette[index].g = green;
    palette[index].b = blue;
}

// seg009:9C36
int find_first_pal_row(int which_rows_mask) {
    word which_row = 0;
    word row_mask = 1;
    do {
        if (row_mask & which_rows_mask) {
            return which_row;
        }
        ++which_row;
        row_mask <<= 1;
    } while (which_row < 16);
    return 0;
}

// seg009:9C6C
int get_text_color(int low_half, int high_half_mask) {
    if (high_half_mask != 0) {
        return (find_first_pal_row(high_half_mask) << 4) + low_half;
    } else {
        return low_half;
    }
}

void load_from_opendats_metadata(int resource_id, const char* extension, FILE** out_fp, data_location* result, byte* checksum, int* size, dat_type** out_pointer) {
    FILE* fp = NULL;
    dat_type* pointer;
    *result = data_none;
    (void) extension;
    // Go through all open DAT files.
    // replace big and small pillars and level door starts in dark mode in dungeon 1
    if (current_level <= 15 &&
            settings->tbl_level_type[current_level] == DUNGEON_LEVEL &&
            settings->tbl_level_subtype[current_level] == DUNGEON_OR_PALACE_1_LEVEL &&
            settings->tbl_level_color[current_level] == 2) {
        if ((resource_id >= 292 && resource_id <= 295) || (resource_id >= 1286 && resource_id <= 1291)) {
            resource_id += 100;
        } else if (resource_id == 344) {
            resource_id += 10;
        }
    }
    for (pointer = dat_chain_ptr; fp == NULL && pointer != NULL; pointer = pointer->next_dat) {
        *out_pointer = pointer;
        if (pointer->handle != NULL) {
            fp = pointer->handle;
            dat_table_type* dat_table = pointer->dat_table;
            int i;
            for (i = 0; i < dat_table->res_count; ++i) {
                if (dat_table->entries[i].id == resource_id) {
                    break;
                }
            }
            if (i < dat_table->res_count) {
                // found
                *result = data_DAT;
                *size = dat_table->entries[i].size;
                if (fseek(fp, dat_table->entries[i].offset, SEEK_SET) ||
                        fread(checksum, 1, 1, fp) != 1) {
                    perror(pointer->filename);
                    fp = NULL;
                }
            } else {
                // not found
                fp = NULL;
            }
        }
    }
    *out_fp = fp;
    if (fp == NULL) {
        *result = data_none;
    }
}

// seg009:9F34
void close_dat(dat_type *pointer) {
    dat_type** prev = &dat_chain_ptr;
    dat_type* curr = dat_chain_ptr;
    while (curr != NULL) {
        if (curr == pointer) {
            *prev = curr->next_dat;
            if (curr->handle) fclose(curr->handle);
            if (curr->dat_table) free(curr->dat_table);
            free(curr);
            return;
        }
        curr = curr->next_dat;
        prev = &((*prev)->next_dat);
    }
}

// seg009:9F80
void* load_from_opendats_alloc(int resource, const char* extension, data_location* out_result, int* out_size) {
    dat_type* pointer;
    data_location result;
    byte checksum;
    int size;
    FILE* fp = NULL;
    load_from_opendats_metadata(resource, extension, &fp, &result, &checksum, &size, &pointer);
    if (out_result != NULL) *out_result = result;
    if (out_size != NULL) *out_size = size;
    if (result == data_none) return NULL;
    void* area = malloc(size);
    if (fread(area, size, 1, fp) != 1) {
        fprintf(stderr, "%s: %s, resource %d, size %d, failed: %s\n",
                __func__, pointer->filename, resource,
                size, strerror(errno));
        free(area);
        area = NULL;
    }
    if (result == data_directory) fclose(fp);
    /* XXX: check checksum */
    return area;
}

// seg009:A172
int load_from_opendats_to_area(int resource, void *area, int length, const char* extension) {
    dat_type* pointer;
    data_location result;
    byte checksum;
    int size;
    FILE* fp = NULL;
    load_from_opendats_metadata(resource, extension, &fp, &result, &checksum, &size, &pointer);
    if (result == data_none) return 0;
    if (fread(area, MIN(size, length), 1, fp) != 1) {
        fprintf(stderr, "%s: %s, resource %d, size %d, failed: %s\n",
                __func__, pointer->filename, resource,
                size, strerror(errno));
        memset(area, 0, MIN(size, length));
    }
    if (result == data_directory) fclose(fp);
    /* XXX: check checksum */
    return 0;
}

// SDL-specific implementations

void rect_to_sdlrect(const rect_type* rect, SDL_Rect* sdlrect) {
    sdlrect->x = rect->left;
    sdlrect->y = rect->top;
    sdlrect->w = rect->right - rect->left;
    sdlrect->h = rect->bottom - rect->top;
}

void method_1_blit_rect(surface_type *target_surface, surface_type *source_surface, const rect_type *target_rect, const rect_type *source_rect, int blit) {
    SDL_Rect src_rect;
    rect_to_sdlrect(source_rect, &src_rect);
    SDL_Rect dest_rect;
    rect_to_sdlrect(target_rect, &dest_rect);

    if (blit == blitters_0_no_transp) {
        // Disable transparency.
        if (SDL_SetColorKey(source_surface, 0, 0) != 0) {
            sdlperror("SDL_SetColorKey");
            quit(1);
        }
    } else {
        // Enable transparency.
        if (SDL_SetColorKey(source_surface, SDL_TRUE, 0) != 0) {
            sdlperror("SDL_SetColorKey");
            quit(1);
        }
    }
    if (SDL_BlitSurface(source_surface, &src_rect, target_surface, &dest_rect) != 0) {
        sdlperror("SDL_BlitSurface");
        quit(1);
    }
}

image_type * method_3_blit_mono(image_type *image, int xpos, int ypos, int blitter, byte color) {
    int w = image->w;
    int h = image->h;
    (void) blitter;
    if (SDL_SetColorKey(image, SDL_TRUE, 0) != 0) {
        sdlperror("SDL_SetColorKey");
        quit(1);
    }
    SDL_Surface* colored_image = SDL_ConvertSurfaceFormat(image, SDL_PIXELFORMAT_ARGB8888, 0);

    SDL_SetSurfaceBlendMode(colored_image, SDL_BLENDMODE_NONE);

    if (SDL_LockSurface(colored_image) != 0) {
        sdlperror("SDL_LockSurface");
        quit(1);
    }

    int y, x;
    rgb_type palette_color;
    // custom logic - use PoP2 color for bottom text and loading screen
    if (color == color_20_yellowish_text) {
        palette_color = (rgb_type) {
            0x3F, 0x35, 0x23
        };
    } else {
        palette_color = palette[color];
    }
    uint32_t rgb_color = SDL_MapRGB(colored_image->format, palette_color.r << 2, palette_color.g << 2, palette_color.b << 2) & 0xFFFFFF;
    int stride = colored_image->pitch;
    for (y = 0; y < h; ++y) {
        uint32_t* pixel_ptr = (uint32_t*) ((byte*)colored_image->pixels + stride * y);
        for (x = 0; x < w; ++x) {
            // set RGB but leave alpha
            *pixel_ptr = (*pixel_ptr & 0xFF000000) | rgb_color;
            ++pixel_ptr;
        }
    }
    SDL_UnlockSurface(colored_image);

    SDL_Rect src_rect = {0, 0, image->w, image->h};
    SDL_Rect dest_rect = {xpos, ypos, image->w, image->h};

    SDL_SetSurfaceBlendMode(colored_image, SDL_BLENDMODE_BLEND);
    SDL_SetSurfaceBlendMode(current_target_surface, SDL_BLENDMODE_BLEND);
    SDL_SetSurfaceAlphaMod(colored_image, 255);
    if (SDL_BlitSurface(colored_image, &src_rect, current_target_surface, &dest_rect) != 0) {
        sdlperror("SDL_BlitSurface");
        quit(1);
    }
    SDL_FreeSurface(colored_image);

    return image;
}

#if !SDL_VERSION_ATLEAST(2,0,4)
// Workaround for a bug in SDL2 (before v2.0.4):
// https://bugzilla.libsdl.org/show_bug.cgi?id=2986
// SDL_FillRect onto a 24-bit surface swaps Red and Blue component

bool RGB24_bug_checked = false;
bool RGB24_bug_affected;

bool RGB24_bug_check(void) {
    if (!RGB24_bug_checked) {
        // Check if the bug occurs in this version of SDL.
        SDL_Surface* test_surface = SDL_CreateRGBSurface(0, 1, 1, 24, 0, 0, 0, 0);
        if (NULL == test_surface) sdlperror("SDL_CreateSurface in RGB24_bug_check");
        // Fill with red.
        SDL_FillRect(test_surface, NULL, SDL_MapRGB(test_surface->format, 0xFF, 0, 0));
        if (0 != SDL_LockSurface(test_surface)) sdlperror("SDL_LockSurface in RGB24_bug_check");
        // Read red component of pixel.
        RGB24_bug_affected = (*(Uint32*)test_surface->pixels & test_surface->format->Rmask) == 0;
        SDL_UnlockSurface(test_surface);
        SDL_FreeSurface(test_surface);
        RGB24_bug_checked = true;
    }
    return RGB24_bug_affected;
}
#endif

int safe_SDL_FillRect(SDL_Surface* dst, const SDL_Rect* rect, Uint32 color) {
#if !SDL_VERSION_ATLEAST(2,0,4)
    if (dst->format->BitsPerPixel == 24 && RGB24_bug_check()) {
        // In the buggy version, SDL_FillRect swaps R and B, so we swap it once more.
        color = ((color & 0xFF) << 16) | (color & 0xFF00) | ((color & 0xFF0000) >> 16);
    }
#endif
    return SDL_FillRect(dst, rect, color);
}
// End of workaround.

const rect_type * method_5_rect(const rect_type *rect, int blit, byte color) {
    SDL_Rect dest_rect;
    (void) blit;
    rect_to_sdlrect(rect, &dest_rect);
    rgb_type palette_color = palette[color];
    uint32_t rgb_color = SDL_MapRGBA(current_target_surface->format, palette_color.r << 2, palette_color.g << 2, palette_color.b << 2, 0xFF);
    if (safe_SDL_FillRect(current_target_surface, &dest_rect, rgb_color) != 0) {
        sdlperror("SDL_FillRect");
        quit(1);
    }
    return rect;
}

void draw_rect_with_alpha(const rect_type* rect, byte color, byte alpha) {
    SDL_Rect dest_rect;
    rect_to_sdlrect(rect, &dest_rect);
    rgb_type palette_color = palette[color];
    uint32_t rgb_color = SDL_MapRGBA(overlay_surface->format, palette_color.r << 2, palette_color.g << 2, palette_color.b << 2, alpha);
    if (safe_SDL_FillRect(current_target_surface, &dest_rect, rgb_color) != 0) {
        sdlperror("SDL_FillRect");
        quit(1);
    }
}

void draw_rect_contours(const rect_type* rect, byte color) {
    // TODO: handle 24 bit surfaces? (currently, 32 bit surface is assumed)
    if (current_target_surface->format->BitsPerPixel != 32) {
        printf("draw_rect_contours: not implemented for %d bit surfaces\n", current_target_surface->format->BitsPerPixel);
        return;
    }
    SDL_Rect dest_rect;
    rect_to_sdlrect(rect, &dest_rect);
    rgb_type palette_color = palette[color];
    uint32_t rgb_color = SDL_MapRGBA(overlay_surface->format, palette_color.r << 2, palette_color.g << 2, palette_color.b << 2, 0xFF);
    if (SDL_LockSurface(current_target_surface) != 0) {
        sdlperror("SDL_LockSurface");
        quit(1);
    }
    int bytes_per_pixel = current_target_surface->format->BytesPerPixel;
    int pitch = current_target_surface->pitch;
    byte* pixels = current_target_surface->pixels;
    int xmin = MIN(dest_rect.x,               current_target_surface->w);
    int xmax = MIN(dest_rect.x + dest_rect.w, current_target_surface->w);
    int ymin = MIN(dest_rect.y,               current_target_surface->h);
    int ymax = MIN(dest_rect.y + dest_rect.h, current_target_surface->h);
    byte* row = pixels + ymin * pitch;
    uint32_t* pixel =  (uint32_t*) (row + xmin * bytes_per_pixel);
    for (int x = xmin; x < xmax; ++x) {
        *pixel++ = rgb_color;
    }
    for (int y = ymin + 1; y < ymax - 1; ++y) {
        row += pitch;
        *(uint32_t*)(row + xmin * bytes_per_pixel) = rgb_color;
        *(uint32_t*)(row + (xmax - 1)*bytes_per_pixel) = rgb_color;
    }
    pixel = (uint32_t*) (pixels + (ymax - 1) * pitch + xmin * bytes_per_pixel);
    for (int x = xmin; x < xmax; ++x) {
        *pixel++ = rgb_color;
    }

    SDL_UnlockSurface(current_target_surface);
}

void blit_xor(SDL_Surface* target_surface, SDL_Rect* dest_rect, SDL_Surface* image, SDL_Rect* src_rect) {
    if (dest_rect->w != src_rect->w || dest_rect->h != src_rect->h) {
        printf("blit_xor: dest_rect and src_rect have different sizes\n");
        quit(1);
    }
    SDL_Surface* helper_surface = SDL_CreateRGBSurface(0, dest_rect->w, dest_rect->h, 24, 0xFF, 0xFF << 8, 0xFF << 16, 0);
    if (helper_surface == NULL) {
        sdlperror("SDL_CreateRGBSurface");
        quit(1);
    }
    SDL_Surface* image_24 = SDL_ConvertSurface(image, helper_surface->format, 0);
    //SDL_CreateRGBSurface(0, src_rect->w, src_rect->h, 24, 0xFF, 0xFF<<8, 0xFF<<16, 0);
    if (image_24 == NULL) {
        sdlperror("SDL_CreateRGBSurface");
        quit(1);
    }
    SDL_Rect dest_rect2 = *src_rect;
    // Read what is currently where we want to draw the new image.
    if (SDL_BlitSurface(target_surface, dest_rect, helper_surface, &dest_rect2) != 0) {
        sdlperror("SDL_BlitSurface");
        quit(1);
    }
    if (SDL_LockSurface(image_24) != 0) {
        sdlperror("SDL_LockSurface");
        quit(1);
    }
    if (SDL_LockSurface(helper_surface) != 0) {
        sdlperror("SDL_LockSurface");
        quit(1);
    }
    int size = helper_surface->h * helper_surface->pitch;
    int i;
    byte *p_src = (byte*) image_24->pixels;
    byte *p_dest = (byte*) helper_surface->pixels;

    // Xor the old area with the image.
    for (i = 0; i < size; ++i) {
        *p_dest ^= *p_src;
        ++p_src;
        ++p_dest;
    }
    SDL_UnlockSurface(image_24);
    SDL_UnlockSurface(helper_surface);
    // Put the new area in place of the old one.
    if (SDL_BlitSurface(helper_surface, src_rect, target_surface, dest_rect) != 0) {
        sdlperror("SDL_BlitSurface 2065");
        quit(1);
    }
    SDL_FreeSurface(image_24);
    SDL_FreeSurface(helper_surface);
}

void draw_colored_torch(int color, SDL_Surface* image, int xpos, int ypos) {
    if (SDL_SetColorKey(image, SDL_TRUE, 0) != 0) {
        sdlperror("SDL_SetColorKey");
        quit(1);
    }

    word blitter = blitters_0_no_transp;

    // custom logic - the blitter color value of 0x40 represents a flaming sword
    if (color == 0x40) {
        color = 11; // argon blue
        blitter = blitters_10h_transp; // transparent background
    }

    SDL_Surface* colored_image = SDL_ConvertSurfaceFormat(image, SDL_PIXELFORMAT_ARGB8888, 0);
    SDL_SetSurfaceBlendMode(colored_image, SDL_BLENDMODE_NONE);

    if (SDL_LockSurface(colored_image) != 0) {
        sdlperror("SDL_LockSurface");
        quit(1);
    }

    int w = colored_image->w;
    int h = colored_image->h;
    int y, x;
    int iRed = ((color >> 4) & 3) * 85;
    int iGreen = ((color >> 2) & 3) * 85;
    int iBlue = ((color >> 0) & 3) * 85;
    uint32_t old_color = SDL_MapRGB(colored_image->format, 0xFC, 0x84, 0x00) & 0xFFFFFF; // the orange in the flame
    uint32_t new_color = SDL_MapRGB(colored_image->format, iRed, iGreen, iBlue) & 0xFFFFFF;
    int stride = colored_image->pitch;
    for (y = 0; y < h; ++y) {
        uint32_t* pixel_ptr = (uint32_t*) ((byte*)colored_image->pixels + stride * y);
        for (x = 0; x < w; ++x) {
            if ((*pixel_ptr & 0xFFFFFF) == old_color) {
                // set RGB but leave alpha
                *pixel_ptr = (*pixel_ptr & 0xFF000000) | new_color;
            }
            ++pixel_ptr;
        }
    }
    SDL_UnlockSurface(colored_image);

    method_6_blit_img_to_scr(colored_image, xpos, ypos, blitter);
    SDL_FreeSurface(colored_image);
}

image_type * method_6_blit_img_to_scr(image_type *image, int xpos, int ypos, int blit) {
    if (image == NULL) {
        return NULL;
    }

    if (blit == blitters_9_black) {
        method_3_blit_mono(image, xpos, ypos, blitters_9_black, 0);
        return image;
    }

    SDL_Rect src_rect = {0, 0, image->w, image->h};
    SDL_Rect dest_rect = {xpos, ypos, image->w, image->h};

    if (blit == blitters_3_xor) {
        blit_xor(current_target_surface, &dest_rect, image, &src_rect);
        return image;
    }

    if (blit >= blitters_colored_flame && blit <= blitters_colored_flaming_sword) {
        draw_colored_torch(blit - blitters_colored_flame, image, xpos, ypos);
        return image;
    }

    SDL_SetSurfaceBlendMode(image, SDL_BLENDMODE_NONE);
    SDL_SetSurfaceAlphaMod(image, 255);

    if (blit == blitters_0_no_transp) {
        SDL_SetColorKey(image, SDL_FALSE, 0);
    } else {
        // custom logic - ensure the color key includes the 0 alpha component to support
        // custom PNGs with transparency under SDL Image 2.8.0
        SDL_SetColorKey(image, SDL_TRUE, SDL_MapRGBA(image->format, 0, 0, 0, 0));
    }

    if (SDL_BlitSurface(image, &src_rect, current_target_surface, &dest_rect) != 0) {
        sdlperror("method_6_blit_img_to_scr: SDL_BlitSurface 2247");
    }

    return image;
}

int fps = 60;
float milliseconds_per_tick = (1000.0f / 60.0f);
Uint64 timer_last_counter[NUM_TIMERS];
int wait_time[NUM_TIMERS];

Uint32 timer_callback(Uint32 interval, void *param) {
    SDL_Event event;
    memset(&event, 0, sizeof(event));
    event.type = SDL_USEREVENT;
    event.user.code = userevent_TIMER;
    event.user.data1 = param;
    SDL_PushEvent(&event);
    return interval;
}

void reset_timer(int timer_index) {
    timer_last_counter[timer_index] = SDL_GetPerformanceCounter();
}

// custom logic - gets ticks per second
double get_ticks_per_sec(int timer_index) {
    return (double) fps / wait_time[timer_index];
}

// custom logic - recalculate remaining feather fall timer
void recalculate_feather_fall_timer(double previous_ticks_per_second, double ticks_per_second) {
    if (feather_fall_timer <= MAX(previous_ticks_per_second, ticks_per_second) ||
            previous_ticks_per_second == ticks_per_second) {
        return;
    }
    feather_fall_timer = feather_fall_timer / previous_ticks_per_second * ticks_per_second;
}

void set_timer_length(int timer_index, int length) {
    if (feather_fall_timer == 0 ||
            wait_time[timer_index] < 3 ||
            wait_time[timer_index] > 7) { // no feather fall recalculation needed
        wait_time[timer_index] = length;
        return;
    }
    // custom logic - recalculate remaining feather fall timer
    double previous_ticks_per_second, ticks_per_second;
    previous_ticks_per_second = get_ticks_per_sec(timer_index);
    wait_time[timer_index] = length;
    ticks_per_second = get_ticks_per_sec(timer_index);
    recalculate_feather_fall_timer(previous_ticks_per_second, ticks_per_second);
}

void start_timer(int timer_index, int length) {
    timer_last_counter[timer_index] = SDL_GetPerformanceCounter();
    wait_time[timer_index] = length;
}

void toggle_fullscreen(void) {
    uint32_t flags = SDL_GetWindowFlags(window_);
    if (flags & SDL_WINDOW_FULLSCREEN_DESKTOP) {
        SDL_SetWindowFullscreen(window_, 0);
        // custom logic - in Linux window does not restore correctly without the next 2 lines
        // if the game was started in full screen mode
        SDL_RestoreWindow(window_);
        SDL_SetWindowSize(window_, pop_window_width, pop_window_height);
        SDL_ShowCursor(SDL_ENABLE);
    } else {
        SDL_SetWindowFullscreen(window_, SDL_WINDOW_FULLSCREEN_DESKTOP);
        SDL_ShowCursor(SDL_DISABLE);
    }
}

bool ignore_tab = false;

void process_events(void) {
    // Process all events in the queue.
    // Previously, this procedure would wait for *one* event and process it, then return.
    // Much like the x86 HLT instruction.
    // (We still want to process all events in the queue. For instance, there might be
    // simultaneous SDL2 KEYDOWN and TEXTINPUT events.)
    SDL_Event event;
    while (SDL_PollEvent(&event) == 1) { // while there are still events to be processed
        switch (event.type) {
            case SDL_KEYDOWN: {
                int modifier = event.key.keysym.mod;
                int scancode = event.key.keysym.scancode;
                // Handle these separately, so they won't interrupt things that are usually interrupted by a keypress. (pause, cutscene)
                if (scancode == SDL_SCANCODE_F12) {
                    // custom logic - do not save screenshots during a cutscene or the demo
                    if (!is_cutscene && !is_ending_sequence && (current_level > 0 || play_demo_level) && current_level < 15) {
                        if (modifier & KMOD_SHIFT) {
                            save_level_screenshot((modifier & KMOD_CTRL) != 0);
                        } else {
                            save_screenshot((modifier & KMOD_CTRL) != 0);
                        }
                    }
                } else if (escape_key_suppressed && scancode == SDL_SCANCODE_ESCAPE) {
                    break; // Prevent repeated keystrokes opening/closing the menu as long as the key is held down.
                } else if ((modifier & KMOD_ALT) &&
                           (scancode == SDL_SCANCODE_RETURN || scancode == SDL_SCANCODE_KP_ENTER)) {
                    // Only if the Enter key was pressed down right now.
                    if ((key_states[scancode] & keystate_held) == 0) {
                        // Alt-Enter: toggle fullscreen mode
                        toggle_fullscreen();
                        key_states[scancode] |= keystate_held | keystate_held_new;
                    }
                } else {
                    key_states[scancode] |= keystate_held | keystate_held_new;
                    switch (scancode) {
                        // Keys that are ignored by themselves:
                        case SDL_SCANCODE_LSHIFT:
                        case SDL_SCANCODE_RSHIFT:
                        case SDL_SCANCODE_LCTRL:
                        case SDL_SCANCODE_LGUI:
                        case SDL_SCANCODE_RCTRL:
                        case SDL_SCANCODE_RGUI:
                        case SDL_SCANCODE_CAPSLOCK:
                        case SDL_SCANCODE_SCROLLLOCK:
                        case SDL_SCANCODE_NUMLOCKCLEAR:
                        case SDL_SCANCODE_APPLICATION:
                        case SDL_SCANCODE_PRINTSCREEN:
                        case SDL_SCANCODE_VOLUMEUP:
                        case SDL_SCANCODE_VOLUMEDOWN:
                        case SDL_SCANCODE_PAUSE:
                            break;

                        default:
                            // If Alt is held down from Alt+Tab: ignore it until it's released.
                            if (scancode == SDL_SCANCODE_TAB && ignore_tab) break;

                            last_key_scancode = scancode;
                            if (modifier & KMOD_SHIFT) last_key_scancode |= WITH_SHIFT;
                            if (modifier & KMOD_CTRL ) last_key_scancode |= WITH_CTRL ;
                            if (modifier & KMOD_ALT  ) last_key_scancode |= WITH_ALT  ;
                    }

                    switch (scancode) {
                        // Keys that are used for keyboard control:
                        case SDL_SCANCODE_LSHIFT:
                        case SDL_SCANCODE_RSHIFT:
                        case SDL_SCANCODE_LALT:
                        case SDL_SCANCODE_RALT:
                        case SDL_SCANCODE_LEFT:
                        case SDL_SCANCODE_RIGHT:
                        case SDL_SCANCODE_UP:
                        case SDL_SCANCODE_DOWN:
                        case SDL_SCANCODE_CLEAR:
                        case SDL_SCANCODE_HOME:
                        case SDL_SCANCODE_END:
                        case SDL_SCANCODE_PAGEDOWN:
                        case SDL_SCANCODE_PAGEUP:
                        case SDL_SCANCODE_KP_1:
                        case SDL_SCANCODE_KP_2:
                        case SDL_SCANCODE_KP_3:
                        case SDL_SCANCODE_KP_4:
                        case SDL_SCANCODE_KP_5:
                        case SDL_SCANCODE_KP_6:
                        case SDL_SCANCODE_KP_7:
                        case SDL_SCANCODE_KP_8:
                        case SDL_SCANCODE_KP_9:
                            if (!is_keyboard_mode) {
                                is_keyboard_mode = 1;
                                is_joyst_mode = 0;
                            }
                            break;
                        // custom logic - playing alpha keys in non-cheat mode
                        case SDL_SCANCODE_A:
                        case SDL_SCANCODE_C:
                        case SDL_SCANCODE_D:
                        case SDL_SCANCODE_E:
                        case SDL_SCANCODE_Q:
                        case SDL_SCANCODE_S:
                        case SDL_SCANCODE_W:
                        case SDL_SCANCODE_X:
                        case SDL_SCANCODE_Z:
                            if (!cheats_enabled && !is_keyboard_mode) {
                                is_keyboard_mode = 1;
                                is_joyst_mode = 0;
                            }
                            break;
                    }
                }
                break;
            }
            case SDL_KEYUP:
                // If Alt was held down from Alt+Tab but now it's released: stop ignoring Tab.
                if (event.key.keysym.scancode == SDL_SCANCODE_TAB && ignore_tab) ignore_tab = false;

                key_states[event.key.keysym.scancode] &= ~keystate_held;
                // Prevent repeated keystrokes opening/closing the menu as long as the key is held down.
                if (event.key.keysym.scancode == SDL_SCANCODE_ESCAPE) {
                    escape_key_suppressed = false;
                }
                break;
            case SDL_CONTROLLERAXISMOTION:
                if (event.caxis.axis < 6) {
                    joy_axis[event.caxis.axis] = event.caxis.value;
                    if (abs(event.caxis.value) > abs(joy_axis_max[event.caxis.axis])) {
                        joy_axis_max[event.caxis.axis] = event.caxis.value;
                    }
                    if (!is_joyst_mode &&
                            (event.caxis.value >= settings->controller_threshold || event.caxis.value <= -(settings->controller_threshold))) {
                        is_joyst_mode = 1;
                        is_keyboard_mode = 0;
                    }
                }
                break;
            case SDL_CONTROLLERDEVICEADDED:
                SDL_GameControllerOpen(event.cdevice.which);
                break;
            case SDL_CONTROLLERDEVICEREMOVED:
                if (sdl_controller_ == SDL_GameControllerFromInstanceID(event.cdevice.which)) {
                    sdl_controller_ = NULL;
                    is_joyst_mode = 0;
                    is_keyboard_mode = 1;
                }
                SDL_GameControllerClose(SDL_GameControllerFromInstanceID(event.cdevice.which));
                break;
            case SDL_CONTROLLERBUTTONDOWN:
                //Make sure sdl_controller_ always points to the active controller
                sdl_controller_ = SDL_GameControllerFromInstanceID(event.cdevice.which);
                if (!is_joyst_mode) {
                    is_joyst_mode = 1;
                    is_keyboard_mode = 0;
                }
                switch (event.cbutton.button) {
                    case SDL_CONTROLLER_BUTTON_DPAD_LEFT:
                        joy_button_states[joyinput_dpad_left] |= keystate_held | keystate_held_new;
                        break; // left
                    case SDL_CONTROLLER_BUTTON_DPAD_RIGHT:
                        joy_button_states[joyinput_dpad_right] |= keystate_held | keystate_held_new;
                        break; // right
                    case SDL_CONTROLLER_BUTTON_DPAD_UP:
                        joy_button_states[joyinput_dpad_up] |= keystate_held | keystate_held_new;
                        break; // up
                    case SDL_CONTROLLER_BUTTON_DPAD_DOWN:
                        joy_button_states[joyinput_dpad_down] |= keystate_held | keystate_held_new;
                        break; // down

                    case SDL_CONTROLLER_BUTTON_A:
                        joy_button_states[joyinput_a] |= keystate_held | keystate_held_new;
                        break; /*** A (down) ***/
                    case SDL_CONTROLLER_BUTTON_Y:
                        joy_button_states[joyinput_y] |= keystate_held | keystate_held_new;
                        break; /*** Y (up) ***/
                    case SDL_CONTROLLER_BUTTON_X:
                        joy_button_states[joyinput_x] |= keystate_held | keystate_held_new;
                        break; /*** X (shift) ***/
                    case SDL_CONTROLLER_BUTTON_B:
                        joy_button_states[joyinput_b] |= keystate_held | keystate_held_new;
                        break; /*** B (alt) ***/

                    // custom logic - shoulder buttons
                    case SDL_CONTROLLER_BUTTON_LEFTSHOULDER:
                        joy_button_states[joyinput_left_shoulder] |= keystate_held | keystate_held_new;
                        break; /*** left shoulder ***/
                    case SDL_CONTROLLER_BUTTON_RIGHTSHOULDER:
                        joy_button_states[joyinput_right_shoulder] |= keystate_held | keystate_held_new;
                        break; /*** right shoulder ***/

                    case SDL_CONTROLLER_BUTTON_START:
                        last_key_scancode = SDL_SCANCODE_ESCAPE;  /*** back (pause game) ***/
                        break;

                    case SDL_CONTROLLER_BUTTON_BACK:
                        last_key_scancode = SDL_SCANCODE_F6; /*** save game ***/
                        break;

                    default:
                        break;
                }
                break;
            case SDL_CONTROLLERBUTTONUP:
                switch (event.cbutton.button) {
                    case SDL_CONTROLLER_BUTTON_DPAD_LEFT:
                        joy_button_states[joyinput_dpad_left] &= ~keystate_held;
                        break; // left
                    case SDL_CONTROLLER_BUTTON_DPAD_RIGHT:
                        joy_button_states[joyinput_dpad_right] &= ~keystate_held;
                        break; // right
                    case SDL_CONTROLLER_BUTTON_DPAD_UP:
                        joy_button_states[joyinput_dpad_up] &= ~keystate_held;
                        break; // up
                    case SDL_CONTROLLER_BUTTON_DPAD_DOWN:
                        joy_button_states[joyinput_dpad_down] &= ~keystate_held;
                        break; // down

                    // custom logic - shoulder buttons (reload the last saved game when pressed at the same time)
                    case SDL_CONTROLLER_BUTTON_LEFTSHOULDER:
                        joy_button_states[joyinput_left_shoulder] &= ~keystate_held;
                        break; /*** left shoulder ***/
                    case SDL_CONTROLLER_BUTTON_RIGHTSHOULDER:
                        joy_button_states[joyinput_right_shoulder] &= ~keystate_held;
                        break; /*** right shoulder ***/

                    case SDL_CONTROLLER_BUTTON_A:
                        joy_button_states[joyinput_a] &= ~keystate_held;
                        break; /*** A (down) ***/
                    case SDL_CONTROLLER_BUTTON_Y:
                        joy_button_states[joyinput_y] &= ~keystate_held;
                        break; /*** Y (up) ***/
                    case SDL_CONTROLLER_BUTTON_X:
                        joy_button_states[joyinput_x] &= ~keystate_held;
                        break; /*** X (shift) ***/
                    case SDL_CONTROLLER_BUTTON_B:
                        joy_button_states[joyinput_b] &= ~keystate_held;
                        break; /*** B (insert) ***/

                    default:
                        break;
                }
                break;
            case SDL_JOYBUTTONDOWN:
            case SDL_JOYBUTTONUP:
            case SDL_JOYAXISMOTION:
                // Only handle the event if the joystick is incompatible with the SDL_GameController interface.
                // (Otherwise it will interfere with the normal action of the SDL_GameController API.)
                if (!using_sdl_joystick_interface) {
                    break;
                }
                if (event.type == SDL_JOYAXISMOTION) {
                    int axis = -1;
                    if (event.jaxis.axis == SDL_JOYSTICK_X_AXIS) {
                        axis = SDL_CONTROLLER_AXIS_LEFTX;
                    } else if (event.jaxis.axis == SDL_JOYSTICK_Y_AXIS) {
                        axis = SDL_CONTROLLER_AXIS_LEFTY;
                    }
                    if (axis == -1) {
                        break;
                    }
                    joy_axis[axis] = event.jaxis.value;
                    if (abs(event.jaxis.value) > abs(joy_axis_max[axis])) {
                        joy_axis_max[axis] = event.jaxis.value;
                    }
                    // Disregard SDL_JOYAXISMOTION events within joystick 'dead zone'
                    int joy_x = joy_axis[SDL_CONTROLLER_AXIS_LEFTX];
                    int joy_y = joy_axis[SDL_CONTROLLER_AXIS_LEFTY];
                    if ((dword)(joy_x * joy_x) + (dword)(joy_y * joy_y) < (dword)(settings->controller_threshold * settings->controller_threshold)) {
                        break;
                    }
                }
                if (!is_joyst_mode) {
                    is_joyst_mode = 1;
                    is_keyboard_mode = 0;
                }
                if (event.type == SDL_JOYBUTTONDOWN) {
                    if (event.jbutton.button == SDL_JOYSTICK_BUTTON_Y) {
                        joy_button_states[joyinput_y] |= keystate_held | keystate_held_new;
                    } else if (event.jbutton.button == SDL_JOYSTICK_BUTTON_X) {
                        joy_button_states[joyinput_x] |= keystate_held | keystate_held_new;
                    } else if (event.jbutton.button == SDL_JOYSTICK_BUTTON_A) {
                        joy_button_states[joyinput_a] |= keystate_held | keystate_held_new;
                    } else if (event.jbutton.button == SDL_JOYSTICK_BUTTON_B) {
                        joy_button_states[joyinput_b] |= keystate_held | keystate_held_new;
                    }
                } else if (event.type == SDL_JOYBUTTONUP) {
                    if (event.jbutton.button == SDL_JOYSTICK_BUTTON_Y) {
                        joy_button_states[joyinput_y] &= ~keystate_held;
                    } else if (event.jbutton.button == SDL_JOYSTICK_BUTTON_X) {
                        joy_button_states[joyinput_x] &= ~keystate_held;
                    } else if (event.jbutton.button == SDL_JOYSTICK_BUTTON_A) {
                        joy_button_states[joyinput_a] &= ~keystate_held;
                    } else if (event.jbutton.button == SDL_JOYSTICK_BUTTON_B) {
                        joy_button_states[joyinput_b] &= ~keystate_held;
                    }
                }
                break;

            case SDL_TEXTINPUT:
                last_text_input = event.text.text[0]; // UTF-8 formatted char text input
                // Make the +/- keys work on the main keyboard, on any keyboard layout.
                // We check SDL_TEXTINPUT instead of SDL_KEYDOWN.
                // If '+' is on Shift+something then we can't detect it in SDL_KEYDOWN,
                // because event.key.keysym.sym only tells us what character would the key type without shift.
                switch (last_text_input) {
                    case '-':
                        last_key_scancode = SDL_SCANCODE_KP_MINUS;
                        break;
                    case '+':
                        last_key_scancode = SDL_SCANCODE_KP_PLUS;
                        break;
                }
                break;
            case SDL_WINDOWEVENT:
                // In case the user switches away while holding a key: do as if all keys were released.
                // (DOSBox does the same.)

                /* // not implemented in SDL2 for now
                 *
                			if ((event.active.state & SDL_APPINPUTFOCUS) && event.active.gain == 0) {
                				memset(key_states, 0, sizeof(key_states));
                			}
                			// Note: event.active.state can contain multiple flags or'ed.
                			// If the game is in full screen, and I switch away (alt-tab) and back, most of the screen will be black, until it is redrawn.
                			if ((event.active.state & SDL_APPACTIVE) && event.active.gain == 1) {
                				update_screen();
                			}
                */
                switch (event.window.event) {
                    case SDL_WINDOWEVENT_EXPOSED:
                        update_screen();
                        break;
                    case SDL_WINDOWEVENT_FOCUS_GAINED:
                        // Fix for this bug: When playing back a recording, Alt+Tabbing back to SDLPoP stops the replay if Alt is released before Tab.
                    {
                        // If Alt is held down from Alt+Tab: ignore it until it's released.
                        const Uint8 *state = SDL_GetKeyboardState(NULL);
                        if (state[SDL_SCANCODE_TAB]) ignore_tab = true;
                    }
                    break;
                }
                break;
            case SDL_MOUSEBUTTONDOWN:
                switch(event.button.button) {
                    case SDL_BUTTON_LEFT:
                        mouse_clicked = true;
                        break;
                    case SDL_BUTTON_RIGHT:
                    case SDL_BUTTON_X1: // 'Back' button (on mice that have these extra buttons).
                        // custom logic - the menu is opened with secondary instead of primary mouse button
                        if (!is_menu_shown) {
                            last_key_scancode = SDL_SCANCODE_ESCAPE;
                        } else {
                            mouse_button_clicked_right = true;
                        }
                        break;
                    default:
                        break;
                }

                break;
            case SDL_MOUSEWHEEL:
                if (is_menu_shown) {
                    menu_control_scroll_y = -event.wheel.y;
                }
                break;
            case SDL_QUIT:
                if (is_menu_shown) {
                    menu_was_closed();
                }
                quit(0);
                break;
        }
    }
}

void idle(void) {
    process_events();
    update_screen();
}

void do_simple_wait(int timer_index) {
    update_screen();
    while (! has_timer_stopped(timer_index)) {
        SDL_Delay(1);
        process_events();
    }
}

word word_1D63A = 1;
int do_wait(int timer_index) {
    update_screen();
    while (! has_timer_stopped(timer_index)) {
        SDL_Delay(1);
        process_events();
        int key = do_paused();
        if (key != 0 && (word_1D63A != 0 || key == 0x1B)) return 1;
    }
    return 0;
}

// seg009:78E9
void init_timer(int frequency) {
    perf_frequency = SDL_GetPerformanceFrequency();
    fps = frequency;
    milliseconds_per_tick = 1000.0f / (float)fps;
    perf_counters_per_tick = perf_frequency / fps;
    milliseconds_per_counter = 1000.0f / perf_frequency;
}

// seg009:35F6
void set_clip_rect(const rect_type *rect) {
    SDL_Rect clip_rect;
    rect_to_sdlrect(rect, &clip_rect);
    SDL_SetClipRect(current_target_surface, &clip_rect);
}

// seg009:365C
void reset_clip_rect(void) {
    SDL_SetClipRect(current_target_surface, NULL);
}

// seg009:1983
void set_bg_attr(int vga_pal_index, int hc_pal_index) {
    if (vga_pal_index == 0) {
        // Make the black pixels transparent.
        if (SDL_SetColorKey(offscreen_surface, SDL_TRUE, 0) != 0) {	// SDL_SRCCOLORKEY old
            sdlperror("SDL_SetColorKey");
            quit(1);
        }
        SDL_Rect rect = {0, 0, 0, 0};
        rect.w = offscreen_surface->w;
        rect.h = offscreen_surface->h;
        rgb_type palette_color = palette[hc_pal_index];
        uint32_t rgb_color = SDL_MapRGB(onscreen_surface_->format, palette_color.r << 2, palette_color.g << 2, palette_color.b << 2) /*& 0xFFFFFF*/;
        // First clear the screen with the color of the flash.
        if (safe_SDL_FillRect(onscreen_surface_, &rect, rgb_color) != 0) {
            sdlperror("SDL_FillRect");
            quit(1);
        }
        if (upside_down_timer > 0) {
            flip_screen(offscreen_surface);
        }
        // Then draw the offscreen image onto it.
        if (SDL_BlitSurface(offscreen_surface, &rect, onscreen_surface_, &rect) != 0) {
            sdlperror("SDL_BlitSurface");
            quit(1);
        }
        if (hc_pal_index == 0) update_lighting(&rect_top);
        if (upside_down_timer > 0) {
            flip_screen(offscreen_surface);
        }
        if (SDL_SetColorKey(offscreen_surface, 0, 0) != 0) {
            sdlperror("SDL_SetColorKey");
            quit(1);
        }
    }
}

// seg009:07EB
rect_type* offset4_rect_add(rect_type *dest, const rect_type *source, int d_left, int d_top, int d_right, int d_bottom) {
    *dest = *source;
    dest->left += d_left;
    dest->top += d_top;
    dest->right += d_right;
    dest->bottom += d_bottom;
    return dest;
}

// seg009:3AA5
rect_type* offset2_rect(rect_type *dest, const rect_type *source, int delta_x, int delta_y) {
    dest->top    = source->top    + delta_y;
    dest->left   = source->left   + delta_x;
    dest->bottom = source->bottom + delta_y;
    dest->right  = source->right  + delta_x;
    return dest;
}

// seg009:19EF
void fade_in_2(int which_rows) {
    palette_fade_type *palette_buffer;
    palette_buffer = make_pal_buffer_fadein(which_rows, 2);
    while (fade_in_frame(palette_buffer) == 0) {
        process_events(); // modified
        do_paused();
    }
    pal_restore_free_fadein(palette_buffer);
}

// seg009:1A51
palette_fade_type* make_pal_buffer_fadein(int which_rows, int wait_time) {
    palette_fade_type *palette_buffer;
    word curr_row;
    word curr_row_mask;
    palette_buffer = (palette_fade_type*) malloc(sizeof(palette_fade_type));
    palette_buffer->which_rows = which_rows;
    palette_buffer->wait_time = wait_time;
    palette_buffer->fade_pos = 0x40;
    palette_buffer->proc_restore_free = &pal_restore_free_fadein;
    palette_buffer->proc_fade_frame = &fade_in_frame;
    read_palette_256(palette_buffer->original_pal);
    memcpy(palette_buffer->faded_pal, palette_buffer->original_pal, sizeof(palette_buffer->faded_pal));
    for (curr_row = 0, curr_row_mask = 1; curr_row < 0x10; ++curr_row, curr_row_mask <<= 1) {
        if (which_rows & curr_row_mask) {
            memset(palette_buffer->faded_pal + (curr_row << 4), 0, sizeof(rgb_type[0x10]));
            set_pal_arr(curr_row << 4, 0x10, NULL);
        }
    }
    return palette_buffer;
}

// seg009:1B64
void pal_restore_free_fadein(palette_fade_type *palette_buffer) {
    set_pal_256(palette_buffer->original_pal);
    free(palette_buffer);
    // for RGB
    method_1_blit_rect(onscreen_surface_, offscreen_surface, &screen_rect, &screen_rect, 0);
}

// seg009:1B88
int fade_in_frame(palette_fade_type *palette_buffer) {
    rgb_type* faded_pal_ptr;
    word start;
    word column;
    rgb_type* original_pal_ptr;
    word current_row_mask;
    start_timer(timer_1, palette_buffer->wait_time); // too slow?

    --palette_buffer->fade_pos;
    for (start = 0, current_row_mask = 1; start < 0x100; start += 0x10, current_row_mask <<= 1) {
        if (palette_buffer->which_rows & current_row_mask) {
            original_pal_ptr = palette_buffer->original_pal + start;
            faded_pal_ptr = palette_buffer->faded_pal + start;
            for (column = 0; column < 0x10; ++column) {
                if (original_pal_ptr[column].r > palette_buffer->fade_pos) {
                    ++faded_pal_ptr[column].r;
                }
                if (original_pal_ptr[column].g > palette_buffer->fade_pos) {
                    ++faded_pal_ptr[column].g;
                }
                if (original_pal_ptr[column].b > palette_buffer->fade_pos) {
                    ++faded_pal_ptr[column].b;
                }
            }
        }
    }
    column = 0;
    for (start = 0, current_row_mask = 1; start < 0x100; start += 0x10, current_row_mask <<= 1) {
        if (palette_buffer->which_rows & current_row_mask) {
            set_pal_arr(start, 0x10, palette_buffer->faded_pal + start);
        }
    }

    int h = offscreen_surface->h;
    if (SDL_LockSurface(onscreen_surface_) != 0) {
        sdlperror("SDL_LockSurface");
        quit(1);
    }
    if (SDL_LockSurface(offscreen_surface) != 0) {
        sdlperror("SDL_LockSurface");
        quit(1);
    }
    int y, x;
    int on_stride = onscreen_surface_->pitch;
    int off_stride = offscreen_surface->pitch;
    int fade_pos = palette_buffer->fade_pos;
    for (y = 0; y < h; ++y) {
        byte* on_pixel_ptr = (byte*)onscreen_surface_->pixels + on_stride * y;
        byte* off_pixel_ptr = (byte*)offscreen_surface->pixels + off_stride * y;
        for (x = 0; x < on_stride; ++x) {
            //if (*off_pixel_ptr > palette_buffer->fade_pos) *pixel_ptr += 4;
            int v = *off_pixel_ptr - fade_pos * 4;
            if (v < 0) v = 0;
            *on_pixel_ptr = v;
            ++on_pixel_ptr;
            ++off_pixel_ptr;
        }
    }
    SDL_UnlockSurface(onscreen_surface_);
    SDL_UnlockSurface(offscreen_surface);

    /**/do_simple_wait(1); // can interrupt fading of cutscene
    return palette_buffer->fade_pos == 0;
}

// seg009:1CC9
void fade_out_2(int rows) {
    palette_fade_type *palette_buffer;
    palette_buffer = make_pal_buffer_fadeout(rows, 2);
    while (fade_out_frame(palette_buffer) == 0) {
        process_events(); // modified
        do_paused();
    }
    pal_restore_free_fadeout(palette_buffer);
}

// seg009:1D28
palette_fade_type* make_pal_buffer_fadeout(int which_rows, int wait_time) {
    palette_fade_type *palette_buffer;
    palette_buffer = (palette_fade_type*) malloc(sizeof(palette_fade_type));
    palette_buffer->which_rows = which_rows;
    palette_buffer->wait_time = wait_time;
    palette_buffer->fade_pos = 00; // modified
    palette_buffer->proc_restore_free = &pal_restore_free_fadeout;
    palette_buffer->proc_fade_frame = &fade_out_frame;
    read_palette_256(palette_buffer->original_pal);
    memcpy(palette_buffer->faded_pal, palette_buffer->original_pal, sizeof(palette_buffer->faded_pal));
    // for RGB
    method_1_blit_rect(onscreen_surface_, offscreen_surface, &screen_rect, &screen_rect, 0);
    return palette_buffer;
}

// seg009:1DAF
void pal_restore_free_fadeout(palette_fade_type *palette_buffer) {
    surface_type* surface;
    surface = current_target_surface;
    current_target_surface = onscreen_surface_;
    draw_rect(&screen_rect, color_0_black);
    current_target_surface = surface;
    set_pal_256(palette_buffer->original_pal);
    free(palette_buffer);
    // for RGB
    method_5_rect(&screen_rect, 0, 0);
}

// seg009:1DF7
int fade_out_frame(palette_fade_type *palette_buffer) {
    rgb_type* faded_pal_ptr;
    word start;
    word result;
    word column;
    word current_row_mask;
    byte* curr_color_ptr;
    result = 1;
    ++palette_buffer->fade_pos; // modified
    /**/start_timer(timer_1, palette_buffer->wait_time); // too slow?
    for (start = 0, current_row_mask = 1; start < 0x100; start += 0x10, current_row_mask <<= 1) {
        if (palette_buffer->which_rows & current_row_mask) {
            faded_pal_ptr = palette_buffer->faded_pal + start;
            for (column = 0; column < 0x10; ++column) {
                curr_color_ptr = &faded_pal_ptr[column].r;
                if (*curr_color_ptr != 0) {
                    --*curr_color_ptr;
                    result = 0;
                }
                curr_color_ptr = &faded_pal_ptr[column].g;
                if (*curr_color_ptr != 0) {
                    --*curr_color_ptr;
                    result = 0;
                }
                curr_color_ptr = &faded_pal_ptr[column].b;
                if (*curr_color_ptr != 0) {
                    --*curr_color_ptr;
                    result = 0;
                }
            }
        }
    }
    column = 0;
    for (start = 0, current_row_mask = 1; start < 0x100; start += 0x10, current_row_mask <<= 1) {
        if (palette_buffer->which_rows & current_row_mask) {
            set_pal_arr(start, 0x10, palette_buffer->faded_pal + start);
        }
    }

    int h = offscreen_surface->h;
    if (SDL_LockSurface(onscreen_surface_) != 0) {
        sdlperror("SDL_LockSurface");
        quit(1);
    }
    if (SDL_LockSurface(offscreen_surface) != 0) {
        sdlperror("SDL_LockSurface");
        quit(1);
    }
    int y, x;
    int on_stride = onscreen_surface_->pitch;
    int off_stride = offscreen_surface->pitch;
    int fade_pos = palette_buffer->fade_pos;
    for (y = 0; y < h; ++y) {
        byte* on_pixel_ptr = (byte*)onscreen_surface_->pixels + on_stride * y;
        byte* off_pixel_ptr = (byte*)offscreen_surface->pixels + off_stride * y;
        for (x = 0; x < on_stride; ++x) {
            //if (*pixel_ptr >= 4) *pixel_ptr -= 4;
            int v = *off_pixel_ptr - fade_pos * 4;
            if (v < 0) v = 0;
            *on_pixel_ptr = v;
            ++on_pixel_ptr;
            ++off_pixel_ptr;
        }
    }
    SDL_UnlockSurface(onscreen_surface_);
    SDL_UnlockSurface(offscreen_surface);

    do_simple_wait(timer_1); // can interrupt fading of cutscene
    return result;
}

// seg009:1F28
void read_palette_256(rgb_type *target) {
    int i;
    for (i = 0; i < 256; ++i) {
        target[i] = palette[i];
    }
}

// seg009:1F5E
void set_pal_256(rgb_type *source) {
    int i;
    for (i = 0; i < 256; ++i) {
        palette[i] = source[i];
    }
}

#define DEAD_IMAGE_ID 79

void set_chtab_palette(chtab_type* chtab, byte* colors, int n_colors) {
    if (chtab != NULL) {
        SDL_Color* scolors = (SDL_Color*) malloc(n_colors * sizeof(SDL_Color));
        int i;
        for (i = 0; i < n_colors; ++i) {
            scolors[i].r = *colors << 2;
            ++colors;
            scolors[i].g = *colors << 2;
            ++colors;
            scolors[i].b = *colors << 2;
            ++colors;
            scolors[i].a = SDL_ALPHA_OPAQUE; // the SDL2 SDL_Color struct has an alpha component
        }

        // Color 0 of the palette data is not used, it is replaced by the background color.
        // Needed for correct alternate colors (v1.3) of level 8.
        scolors[0].r = scolors[0].g = scolors[0].b = 0;
        scolors[0].a = SDL_ALPHA_TRANSPARENT;

        // custom logic - is this the shadow palette?
        bool is_shadow_palette = chtab == chtab_addrs[id_chtab_2_kid] && is_kid_shadow && is_shadow_flame;

        for (i = 0; i < chtab->n_images; ++i) {
            // custom logic - do not change color of dead kid on level 13
            if (is_shadow_palette && i == DEAD_IMAGE_ID && dead_kid_timer != 7) {
                continue;
            }

            image_type* current_image = chtab->images[i];

            if (current_image != NULL) {
                int n_colors_to_be_set = n_colors;
                SDL_Palette* current_palette = current_image->format->palette;

                // one of the guard images (i=25) is only a single transparent pixel
                // this caused SDL_SetPaletteColors to fail, I think because that palette contains only 2 colors
                if (current_palette != NULL) {
                    if (current_palette->ncolors < n_colors_to_be_set) {
                        n_colors_to_be_set = current_palette->ncolors;
                    }
                    if (SDL_SetPaletteColors(current_palette, scolors, 0, n_colors_to_be_set) != 0) {
                        sdlperror("SDL_SetPaletteColors");
                        quit(1);
                    }
                }
            }
        }
        free(scolors);
    }
}

int has_timer_stopped(int timer_index) {
    Uint64 current_counter = SDL_GetPerformanceCounter();
    int ticks_elapsed = (int)((current_counter / perf_counters_per_tick) - (timer_last_counter[timer_index] / perf_counters_per_tick));
    int overshoot = ticks_elapsed - wait_time[timer_index];
    if (overshoot >= 0) {
        if (overshoot > 0 && overshoot <= 3) {
            current_counter -= overshoot * perf_counters_per_tick;
        }
        timer_last_counter[timer_index] = current_counter;
        return true;
    } else {
        return false;
    }
}

// custom logic - calculates size of a const char* array terminated by NULL
int str_arr_size(const char** arr) {
    int size = 0;
    while (arr[size]) {
        ++size;
    }
    return size;
}

// custom logic - converts a character array to upper case
void to_ucase(char* str) {
    while (*str) {
        *str = toupper(*str);
        str++;
    }
}

// custom logic - displays "press button to continue"
void display_press_button(void) {
    if (is_joyst_mode) {
        display_text_bottom("PRESS BUTTON TO CONTINUE");
    } else {
        display_text_bottom("PRESS KEY TO CONTINUE");
    }
}

#ifdef WIN32_LEAN_AND_MEAN
// custom logic - turns off num lock in Windows because Shift is not detected when Num Lock is on
void set_numlock_state(bool active) {
    byte keyState[256];
    GetKeyboardState((LPBYTE) &keyState);
    if ((keyState[VK_NUMLOCK] & 1) != active) {
        keybd_event(VK_NUMLOCK, 0x45, KEYEVENTF_EXTENDEDKEY, 0);
        keybd_event(VK_NUMLOCK, 0x45, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
        was_numlock_active = 1;
    }
}
#endif
