/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#ifndef TYPES_H
#define TYPES_H

#define STB_VORBIS_HEADER_ONLY
#include "stb_vorbis.c"

#ifdef __has_include
#if __has_include(<SDL2/SDL.h>)
#include "SDL2/SDL.h"
#include "SDL2/SDL_image.h"
#include "SDL2/SDL_mixer.h"
#else
#include "SDL.h"
#include "SDL_image.h"
#include "SDL_mixer.h"
#endif
#else
#include "SDL.h"
#include "SDL_image.h"
#include "SDL_mixer.h"
#endif

#if SDL_BYTEORDER != SDL_LIL_ENDIAN
#error This program is not (yet) prepared for big endian CPUs, please contact the author.
#endif

// In Mac OS the Command key should be used instead of the Control key
#if defined __APPLE__ || __MACH__
#define KMOD_CTRL KMOD_GUI
#endif

// This macro is from SDL_types.h / SDL_stdinc.h .
// It used to be #undefined at the end of that file, but since some time in 2006 it's kept available.
// And SDL's definition changed in SDL 2.0.6, which caused a warning at this redefinition.
// So we should just use the macro from SDL and not define our own.
/* Make sure the types really have the right sizes */
/*
#define SDL_COMPILE_TIME_ASSERT(name, x)               \
       typedef int SDL_dummy_ ## name[(x) * 2 - 1]
*/

typedef Uint8 byte;
typedef Sint8 sbyte;
typedef Uint16 word;
typedef Uint32 dword;

typedef struct rect_type {
    short top, left, bottom, right;
} rect_type;

typedef struct piece {
    byte base_id;
    byte floor_left;
    sbyte base_y;
    byte right_id;
    byte floor_right;
    sbyte right_y;
    byte stripe_id;
    byte topright_id;
    byte bottom_id;
    byte fore_id;
    byte fore_x;
    sbyte fore_y;
} piece;
typedef struct tile_and_mod {
    byte tiletype;
    byte modifier;
} tile_and_mod;

typedef int (*add_table_type)(short chtab_id, int id, sbyte xh, sbyte xl, int ybottom, int blit, byte peel);

typedef struct back_table_type {
    sbyte xh;
    sbyte xl;
    short y;
    byte chtab_id;
    byte id;
    int blit;
} back_table_type;

typedef struct midtable_type {
    sbyte xh;
    sbyte xl;
    short y;
    byte chtab_id;
    byte id;
    byte peel;
    rect_type clip;
    int blit;
    byte color;
} midtable_type;

typedef struct wipetable_type {
    short left;
    short bottom;
    sbyte height;
    short width;
    sbyte color;
    sbyte layer;
} wipetable_type;

enum soundflags { sfDigi = 1, sfMidi = 2, soundflags_4 = 4, sfLoop = 0x80 };

// custom logic - guard colors (only apply to regular guards)
enum guard_colors {
    guard_light_blue = 1,
    guard_dark_red = 2,
    guard_dark_orange = 3,
    guard_green = 4,
    guard_dark_blue = 5,
    guard_purple = 6,
    guard_yellow = 7,
    guard_light_orange = 8,
    guard_prison_fashion = 9,
};

// custom logic - guard types
enum guard_types {
    guard_type_guard = 0,
    guard_type_fat = 1,
    guard_type_skeleton = 2,
    guard_type_Jaffar = 3,
    guard_type_kid = 4, // experimental
    guard_type_Jaffar_vision = 5,
    guard_type_purple = 6,
    guard_type_bird = 7,
};

enum tiles {
    tiles_0_empty = 0,
    tiles_1_floor = 1,
    tiles_2_spike = 2,
    tiles_3_pillar = 3,
    tiles_4_gate = 4,
    tiles_5_stuck = 5,
    tiles_6_closer = 6, // a.k.a. drop button
    tiles_7_doortop_with_floor = 7, // a.k.a. tapestry
    tiles_8_bigpillar_bottom = 8,
    tiles_9_bigpillar_top = 9,
    tiles_10_potion = 10,
    tiles_11_loose = 11,
    tiles_12_doortop = 12, // a.k.a. tapestry top
    tiles_13_mirror = 13,
    tiles_14_debris = 14, // a.k.a. broken floor
    tiles_15_opener = 15, // a.k.a. raise button
    tiles_16_level_door_left = 16, // a.k.a. exit door
    tiles_17_level_door_right = 17,
    tiles_18_chomper = 18,
    tiles_19_torch = 19,
    tiles_20_wall = 20,
    tiles_21_skeleton = 21,
    tiles_22_sword = 22,
    tiles_23_balcony_left = 23,
    tiles_24_balcony_right = 24,
    tiles_25_lattice_pillar = 25,
    tiles_26_lattice_down = 26, // a.k.a. lattice support
    tiles_27_lattice_small = 27,
    tiles_28_lattice_left = 28,
    tiles_29_lattice_right = 29,
    tiles_30_torch_with_debris = 30,
    tiles_26_torch_empty = 26, // custom logic - same value as lattice support, only works in dungeons
};

enum potions {
    potion_0_custom = 0,
    potion_1_health = 1,
    potion_2_life = 2,
    potion_3_feather = 3,
    potion_4_invert = 4,
    potion_5_hurt = 5,
    potion_6_open = 6,
    potion_7_empty_small = 7,
    potion_8_empty_small = 8,
    potion_9_spilled_left_small = 9,
    potion_10_spilled_right_small = 10,
    potion_11_spilled_left_large = 11,
    potion_12_spilled_right_large = 12,
};

enum chtabs {
    id_chtab_0_sword = 0,
    id_chtab_1_flameswordpotion = 1,
    id_chtab_2_kid = 2,
    id_chtab_3_princessinstory = 3,
    id_chtab_4_jaffarinstory_princessincutscenes = 4,
    id_chtab_5_guard = 5,
    id_chtab_6_environment = 6,
    id_chtab_7_environmentwall = 7,
    id_chtab_8_princessroom = 8,
    id_chtab_9_princessbed = 9,
    // custom logic - fat guard can co-exist with others therefore he needs a chtab
    id_chtab_10_fat_guard = 10,
    // custom logic - Jaffar's vision has a sword without a partial hand
    id_chtab_11_sword_no_hand = 11,
    // custom logic - skeleton sword
    id_chtab_12_skel_sword = 12,
    // custom logic - second skeleton sword
    id_chtab_13_alt_skel_sword = 13,
    // custom logic - flame's sword
    id_chtab_14_flaming_sword = 14,
    // custom logic - PoP2 style sword on the floor
    id_chtab_15_pop2_sword = 15,
};

enum blitters {
    blitters_0_no_transp = 0,
    blitters_3_xor = 3, // used for the shadow
    blitters_9_black = 9,
    blitters_Fh_white = 0x0F,
    blitters_10h_transp = 0x10,
    /* 0x40..0x4F will draw a monochrome image with color 0..15 */
    blitters_40h_mono = 0x40,
    blitters_46h_mono_6 = 0x46, // used for palace wall patterns
    blitters_4Ch_mono_12 = 0x4C, // used for chomper blood
    blitters_colored_flame = 0x100,
    blitters_colored_flame_last = 0x13F,
    // custom logic - drawn with color 11 and transparent background
    blitters_colored_flaming_sword = 0x140,
};

enum full_image_id {
    TITLE_PRESENTS = 0,
    TITLE_GAME,
    TITLE_POP,
    TITLE_HACKED_BY,
    TITLE_TUTORIAL,
    HOF_POP,
    STORY_FRAME,
    STORY_ABSENCE,
    STORY_MARRY,
    STORY_HAIL,
    STORY_CREDITS,
    MAX_FULL_IMAGES
};

// custom logic - room link direction
enum room_dir {
    room_dir_left = 0,
    room_dir_right = 1,
    room_dir_up = 2,
    room_dir_down = 3,
};

#pragma pack(push,1)
typedef struct link_type {
    byte left, right, up, down;
} link_type;

typedef struct level_type {
    byte fg[720];
    byte bg[720];
    byte doorlinks1[256];
    byte doorlinks2[256];
    link_type roomlinks[24];
    byte used_rooms;
    byte opener_col[24];
    byte opener_row[24];
    byte fill_1[15];
    byte start_room;
    byte start_pos;
    sbyte start_dir;
    byte fill_2[4];
    byte guards_tile[24];
    byte guards_dir[24];
    byte guards_x[24];
    byte guards_seq_lo[24];
    byte guards_skill[24];
    byte guards_seq_hi[24];
    byte guards_color[24];
    byte fill_3[18];
} level_type;
SDL_COMPILE_TIME_ASSERT(level_size, sizeof(level_type) == 2305);
#pragma pack(pop)

typedef SDL_Surface surface_type;
typedef SDL_Surface image_type;
typedef struct peel_type {
    SDL_Surface* peel;
    rect_type rect;
} peel_type;

typedef struct chtab_type {
    word n_images;
    word chtab_palette_bits;
    word has_palette_bits;
    // This is a variable-size array, with n_images elements.
    image_type* images[];
} chtab_type;

typedef struct full_image_type {
    int id;
    chtab_type** chtab;
    enum blitters blitter;
    int xpos, ypos;
} full_image_type;

#pragma pack(push,1)
typedef struct rgb_type {
    byte r, g, b;
} rgb_type;
typedef struct dat_pal_type {
    word row_bits;
    byte n_colors;
    rgb_type vga[16];
    byte cga[16];
    byte ega[32];
} dat_pal_type;
typedef struct dat_shpl_type {
    byte n_images;
    dat_pal_type palette;
} dat_shpl_type;
SDL_COMPILE_TIME_ASSERT(dat_shpl_size, sizeof(dat_shpl_type) == 100);
#pragma pack(pop)

typedef struct char_type {
    byte frame;
    byte x;
    byte y;
    sbyte direction;
    sbyte curr_col;
    sbyte curr_row;
    byte action;
    sbyte fall_x;
    sbyte fall_y;
    byte room;
    byte repeat;
    byte charid;
    byte sword;
    sbyte alive;
    word curr_seq;
    byte color;
} char_type;

// custom logic - used for dead guards, kid separated from shadow and inactive guards
typedef struct char_type2 {
    byte room;
    sbyte direction;
    byte x;
    sbyte row;
    byte frame;
    byte color;
    byte frame2;
    short x_left_coll;
    short x_right_coll;
    byte alt_frame;
} char_type2;

// custom logic - used for guards falling out of the room
typedef struct falling_guard_type {
    byte room;
    sbyte direction;
    byte x;
    byte y;
    sbyte fall_y;
    word skill;
    word hitp;
    byte color;
    byte start_fall;
} falling_guard_type;

enum charids {
    charid_0_kid      = 0,
    charid_1_shadow   = 1,
    charid_2_guard    = 2,
    charid_3          = 3,
    charid_4_skeleton = 4,
    charid_5_princess = 5,
    charid_6_vizier   = 6,
    charid_24_mouse   = 0x18
};

enum sword_status {
    sword_0_sheathed = 0,
    sword_2_drawn = 2
};

typedef struct auto_move_type {
    short time, move;
} auto_move_type;

enum obj_types {
    objtype_0_cutscene_char = 0,
    objtype_0_kid = 0,
    objtype_1_shadow = 1,
    objtype_2_guard = 2,
    objtype_3_sword = 3,
    objtype_4_mirror_image = 4,
    objtype_5_hurt_splash = 5,
    objtype_128_loose_floor = 0x80
};

typedef struct objtable_type {
    sbyte xh;
    sbyte xl;
    short y;
    byte chtab_id;
    byte id;
    sbyte direction;
    byte obj_type;
    rect_type clip;
    byte tilepos;
    byte color;
} objtable_type;

typedef struct frame_type {
    byte image;
    byte sword;
    sbyte dx;
    sbyte dy;
    byte flags;
} frame_type;

enum frame_flags {
    FRAME_WEIGHT_X = 0x1F,
    FRAME_THIN = 0x20,
    FRAME_NEEDS_FLOOR = 0x40,
    FRAME_EVEN_ODD_PIXEL = 0x80,
};

typedef struct trob_type {
    byte tilepos;
    byte room;
    sbyte type;
} trob_type;

typedef struct mob_type {
    byte xh;
    byte y;
    byte room;
    sbyte speed;
    byte type;
    byte row;
} mob_type;

// custom logic - stuck button type
typedef struct stuck_button_type {
    byte tilepos;
    byte room;
    sbyte type;
    byte modifier;
} stuck_button_type;

enum directions {
    dir_0_right = 0x00,
    dir_56_none = 0x56,
    dir_FF_left = -1
};

enum actions {
    actions_0_stand            = 0,
    actions_1_run_jump         = 1,
    actions_2_hang_climb       = 2,
    actions_3_in_midair        = 3,
    actions_4_in_freefall      = 4,
    actions_5_bumped           = 5,
    actions_6_hang_straight    = 6,
    actions_7_turn             = 7,
    // custom logic - sit in cell
    actions_8_sit              = 8,
    actions_99_hurt            = 99,
};

enum edge_types {
    edge_type_0_object_or_loose = 0,
    edge_type_1_edge            = 1,
    edge_type_2_floor           = 2,
};

enum wall_types {
    wall_type_0_none    = 0,
    wall_type_1_right   = 1,
    wall_type_2_left    = 2,
    wall_type_3_chomper = 3, // wall on the left
    wall_type_4_wall    = 4,
};

typedef struct sword_table_type {
    byte id;
    sbyte x;
    sbyte y;
} sword_table_type;

// custom logic - custom image type id and coordinates
typedef struct custom_image_type {
    word id;
    word x;
    word y;
} custom_image_type;

#pragma pack(push,1)
typedef struct dat_header_type {
    Uint32 table_offset;
    Uint16 table_size;
} dat_header_type;
SDL_COMPILE_TIME_ASSERT(dat_header_size, sizeof(dat_header_type) == 6);

typedef struct dat_res_type {
    Uint16 id;
    Uint32 offset;
    Uint16 size;
} dat_res_type;
SDL_COMPILE_TIME_ASSERT(dat_res_size, sizeof(dat_res_type) == 8);

typedef struct dat_table_type {
    Uint16 res_count;
    dat_res_type entries[];
} dat_table_type;
SDL_COMPILE_TIME_ASSERT(dat_table_size, sizeof(dat_table_type) == 2);

typedef struct image_data_type {
    Uint16 height;
    Uint16 width;
    Uint16 flags;
    byte data[];
} image_data_type;
SDL_COMPILE_TIME_ASSERT(image_data_size, sizeof(image_data_type) == 6);
#pragma pack(pop)

typedef struct dat_type {
    struct dat_type* next_dat;
    FILE* handle;
    char filename[POP_MAX_PATH];
    dat_table_type* dat_table;
    // handle and dat_table are NULL if the DAT is a directory.
} dat_type;

typedef void (*cutscene_ptr_type)(void);

typedef struct palette_fade_type {
    word which_rows;
    word wait_time;
    word fade_pos;
    rgb_type original_pal[256];
    rgb_type faded_pal[256];
    int (*proc_fade_frame)(struct palette_fade_type *palette_buffer);
    void (*proc_restore_free)(struct palette_fade_type *palette_buffer);
} palette_fade_type;

#ifndef O_BINARY
#define O_BINARY 0
#endif

typedef struct font_type {
    byte first_char;
    byte last_char;
    short height_above_baseline;
    short height_below_baseline;
    short space_between_lines;
    short space_between_chars;
    chtab_type* chtab;
} font_type;

typedef struct textstate_type {
    short current_x;
    short current_y;
    short textblit;
    short textcolor;
    font_type* ptr_font;
} textstate_type;

#pragma pack(push,1)
typedef struct rawfont_type {
    byte first_char;
    byte last_char;
    short height_above_baseline;
    short height_below_baseline;
    short space_between_lines;
    short space_between_chars;
    word offsets[];
} rawfont_type;
SDL_COMPILE_TIME_ASSERT(rawfont_type, sizeof(rawfont_type) == 10);
#pragma pack(pop)

typedef enum data_location {
    data_none = 0,
    data_DAT = 1,
    data_directory = 2
} data_location;

enum sound_type {
    sound_speaker = 0, // not used
    sound_digi = 1,
    sound_midi = 2,
    sound_chunk = 3,
    sound_music = 4,
    sound_ogg = 5,
    sound_digi_converted = 6,
};

#pragma pack(push,1)

typedef struct note_type {
    word frequency; // 0x00 or 0x01 = rest, 0x12 = end
    byte length;
} note_type;
SDL_COMPILE_TIME_ASSERT(note_type, sizeof(note_type) == 3);

typedef struct speaker_type { // IBM
    word tempo;
    note_type notes[];
} speaker_type;
SDL_COMPILE_TIME_ASSERT(speaker_type, sizeof(speaker_type) == 2);

typedef struct digi_type { // wave in 1.0 and 1.1
    word sample_rate;
    word sample_count;
    word unknown;
    byte sample_size; // =8
    byte samples[];
} digi_type;
SDL_COMPILE_TIME_ASSERT(digi_type, sizeof(digi_type) == 7);

typedef struct digi_new_type { // wave in 1.3 and 1.4 (and PoP2)
    word sample_rate;
    byte sample_size; // =8
    word sample_count;
    word unknown;
    word unknown2;
    byte samples[];
} digi_new_type;
SDL_COMPILE_TIME_ASSERT(digi_new_type, sizeof(digi_new_type) == 9);

typedef struct midi_type {
    char chunk_type[4];
    dword chunk_length;
    union {
        struct {
            word format;
            word num_tracks;
            word delta;
            byte tracks[0];
        } header;
        byte data[0];
    };
} midi_type;

typedef struct ogg_type {
    //byte sample_size; // =16
    int total_length;
    byte* file_contents;
    stb_vorbis* decoder;
} ogg_type;

typedef struct converted_audio_type {
    int length;
    short samples[];
} converted_audio_type;

typedef struct sound_buffer_type {
    byte type;
    union {
        speaker_type speaker;
        digi_type digi;
        digi_new_type digi_new;
        midi_type midi;
        ogg_type ogg;
        converted_audio_type converted;
    };
} sound_buffer_type;


typedef struct midi_raw_chunk_type {
    char chunk_type[4];
    dword chunk_length;
    union {
        struct {
            word format;
            word num_tracks;
            word time_division;
            byte tracks[0];
        } header;
        byte data[0];
    };
} midi_raw_chunk_type;

#pragma pack(pop)

typedef struct midi_event_type {
    dword delta_time;
    byte event_type;
    union {
        struct {
            byte channel;
            byte param1;
            byte param2;
        } channel;
        struct {
            dword length;
            byte* data;
        } sysex;
        struct {
            byte type;
            dword length;
            byte* data;
        } meta;
    };

} midi_event_type;

typedef struct midi_track_type {
    dword size;
    int num_events;
    midi_event_type* events;
    int event_index;
    int64_t next_pause_tick;
} midi_track_type;

typedef struct parsed_midi_type {
    int num_tracks;
    midi_track_type* tracks;
    dword ticks_per_beat;
} parsed_midi_type;

#pragma pack(push, 1)

typedef struct operator_type {
    byte mul;
    byte ksl_tl;
    byte a_d;
    byte s_r;
    byte waveform;
} operator_type;
SDL_COMPILE_TIME_ASSERT(operator_type, sizeof(operator_type) == 5);

typedef struct instrument_type {
    byte blocknum_low;
    byte blocknum_high;
    byte FB_conn;
    operator_type operators[2];
    byte percussion;
    byte unknown[2];
} instrument_type;
SDL_COMPILE_TIME_ASSERT(instrument_type, sizeof(instrument_type) == 16);

#pragma pack(pop)

struct dialog_type; // (declaration only)
typedef struct dialog_settings_type {
    void (* method_1) (struct dialog_type *dialog);
    void (* method_2_frame) (struct dialog_type *dialog);
    short top_border;
    short left_border;
    short bottom_border;
    short right_border;
    short shadow_bottom;
    short shadow_right;
    short outer_border;
} dialog_settings_type;

typedef struct dialog_type {
    dialog_settings_type* settings;
    rect_type text_rect;
    rect_type peel_rect;
    word has_peel;
    peel_type* peel;
} dialog_type;

typedef struct room_offset_type {
    sbyte dx;
    sbyte dy;
    byte distance;
    bool visited; // For Dijkstra's algorithm
} room_offset_type;

// custom logic - delayed sound structure
typedef struct delayed_sound_type {
    int sound_id, room, col, row;
    short button_type;
    byte play_victory_sound;
    byte repeat;
} delayed_sound_type;

enum soundids {
    sound_0_fell_to_death = 0,
    sound_1_falling = 1,
    sound_2_tile_crashing = 2,
    sound_3_button_pressed = 3,
    sound_4_gate_closing = 4,
    sound_5_gate_opening = 5,
    sound_6_gate_closing_fast = 6,
    sound_7_gate_stop = 7,
    sound_8_bumped = 8,
    sound_9_grab = 9,
    sound_10_sword_vs_sword = 10,
    sound_11_sword_moving = 11,
    sound_12_guard_hurt = 12,
    sound_13_kid_hurt = 13,
    sound_14_leveldoor_closing = 14,
    sound_15_leveldoor_sliding = 15,
    sound_16_fat_fart = 16,
    sound_17_soft_land = 17,
    sound_18_drink = 18,
    sound_19_draw_sword = 19,
    sound_20_loose_shake_1 = 20,
    sound_21_loose_shake_2 = 21,
    sound_22_loose_shake_3 = 22,
    sound_23_footstep = 23,
    sound_24_death_regular = 24,
    sound_25_presentation = 25,
    sound_26_embrace = 26,
    sound_27_cutscene_2_4_6_12 = 27,
    sound_28_death_in_fight = 28,
    sound_29_meet_Jaffar = 29,
    sound_30_big_potion = 30,
    sound_31_flame_hurts_Jaffar = 31,
    sound_32_mouse_music = 32,
    sound_33_small_potion = 33,
    sound_34_hit_by_sword = 34,
    sound_35_cutscene_8_9 = 35,
    sound_36_out_of_time = 36,
    sound_37_victory = 37,
    sound_38_blink = 38,
    sound_39_low_weight = 39,
    sound_40_cutscene_12_short_time = 40,
    sound_41_end_level_music = 41,
    sound_42_guard_falling = 42,
    sound_43_victory_Jaffar = 43,
    sound_44_skel_alive = 44,
    sound_45_jump_through_mirror = 45,
    sound_46_chomped = 46,
    sound_47_chomper = 47,
    sound_48_spiked = 48,
    sound_49_spikes_out = 49,
    sound_50_story_2_princess = 50,
    sound_51_wall_sliding = 51,
    sound_52_story_4_Jaffar_leaves = 52,
    sound_53_story_3_Jaffar_comes = 53,
    sound_54_intro_music = 54,
    sound_55_story_1_absence = 55,
    sound_56_ending_music = 56,
    sound_57_sword_striked = 57,
    sound_58_say_whew = 58,
    sound_59_guard_snoring = 59,
    sound_60_spikes_in = 60,
    sound_61_shadow_separation = 61,
    // custom logic - additional sounds in positional audio mode
    sound_62_princess_door_opening = 62,
    sound_63_temple_victory = 63,
    sound_64_death_temple = 64,
    sound_65_camera_shutter = 65,
    sound_66_death_in_fight_temple = 66,
    sound_67_Jaffar_spell = 67,
    sound_68_come_to_princess = 68,
    sound_69_upside_down = 69,
    sound_70_mouse_squeak = 70,
    sound_71_horse_neigh = 71,
    sound_72_guards_kill_kid = 72,
    sound_73_shadow_level_music = 73,
    sound_74_bridge_collapse_music = 74,
    sound_75_falling_far = 75,
    sound_76_skel_presentation = 76,
    sound_77_get_sword = 77,
    sound_78_separated_from_shadow = 78,
    sound_79_united_with_shadow = 79,
    sound_80_leveldoor_opening = 80,
    sound_81_fat_guard_falling = 81,
};

#define NUM_TIMERS 3
enum timerids {
    timer_0 = 0,
    timer_1 = 1,
    timer_2 = 2,
};

enum colorids {
    color_0_black = 0,
    color_1_blue = 1,
    color_2_green = 2,
    color_3_cyan = 3,
    color_4_red = 4,
    color_5_magenta = 5,
    color_6_brown = 6,
    color_7_lightgray = 7,
    color_8_darkgray = 8,
    color_9_brightblue = 9,
    color_10_brightgreen = 10,
    color_11_brightcyan = 11,
    color_12_brightred = 12,
    color_13_brightmagenta = 13,
    color_14_brightyellow = 14,
    color_15_brightwhite = 15,
    // custom logic - use PoP2 color for bottom text and loading screen
    color_20_yellowish_text = 20,
};

#define COUNT(array) (sizeof(array)/sizeof(array[0]))

// These are or'ed with SDL_SCANCODE_* constants in last_key_scancode.
enum key_modifiers {
    WITH_SHIFT = 0x8000,
    WITH_CTRL  = 0x4000,
    WITH_ALT   = 0x2000,
};

enum control_states {
    control_0_released = 0,
    control_1_ignore = 1,
    control_m1_held = -1,
    control_m2_set = -2, // This is for the control_alt to take out sword but do not make prince turn with arrows
    control_m1_held_left = -1,
    control_1_held_right = 1,
    control_m1_held_up = -1,
    control_1_held_down = 1,
};

enum keystates {
    keystate_held = 0x01,
    keystate_held_new = 0x02,
};

enum {
    joyinput_dpad_left,
    joyinput_dpad_right,
    joyinput_dpad_up,
    joyinput_dpad_down,
    joyinput_a,
    joyinput_b,
    joyinput_x,
    joyinput_y,
    joyinput_left_shoulder,
    joyinput_right_shoulder,

    JOYINPUT_NUM
};

// custom logic - custom graphic resource ids
enum custom_resources {
    image_0_palace = 0,
    image_1_the_end = 1,
    image_2_dead_kid = 2,
    image_3_turned_kid = 3,
    image_4_fire_kid = 4,
    image_5_flaming_sword = 5,
    image_6_dragon_wall = 6,
    image_7_jaffar_vision = 7,
    image_8_jaffar_vision_2 = 8,
    image_9_sword = 9,
    image_10_prince_captured = 10,
    image_11_birdman = 11,
    image_12_horse_statue = 12,
    image_13_serpent_symbol = 13,
    image_14_birdman_statue = 14,
    image_15_gate_frame_piece = 15,
    image_4005_icon = 4005,
    image_4006_lighting_mask = 4006,
};

enum text_align {
    text_align_left = -1,
    text_align_top = -1,
    text_align_center = 0,
    text_align_right = 1,
    text_align_bottom = 1,
};

// custom logic - dialog types
enum dialog_types {
    default_dialog = 0,
    small_text_dialog = 1,
    instructions_dialog = 3,
};

#define MAX_OPTION_VALUE_NAME_LENGTH 20
typedef struct key_value_type {
    char key[MAX_OPTION_VALUE_NAME_LENGTH];
    int value;
} key_value_type;

typedef struct names_list_type {
    byte type; // 0 = names list, 1 = key/value pair list
    union {
        struct {
            const char (* data)[][MAX_OPTION_VALUE_NAME_LENGTH];
            word count;
        } names;
        struct {
            key_value_type* data;
            word count;
        } kv_pairs;
    };
} names_list_type;

// Macros for declaring and initializing a names_list_type (for names lists and key/value pair lists).
#define NAMES_LIST(listname, ...) const char listname[][MAX_OPTION_VALUE_NAME_LENGTH] = __VA_ARGS__; \
    names_list_type listname##_list = {.type=0, .names = {&listname, COUNT(listname)}}

#define KEY_VALUE_LIST(listname, ...) const key_value_type listname[] = __VA_ARGS__; \
    names_list_type listname##_list = {.type=1, .kv_pairs= {(key_value_type*)&listname, COUNT(listname)}}

#define NUM_GUARD_SKILLS 12

typedef struct game_settings_type {
    word start_minutes_left;
    word start_ticks_left;
    word start_hitp;
    word max_hitp_allowed;
    rgb_type vga_palette[16];
    byte level_6_seq_potion_column;
    byte level_6_seq_potion_row;
    byte level_6_seq_sword_left_x;
    byte level_6_seq_sword_top_y;
    word mirror_level;
    byte mirror_room;
    byte mirror_column;
    byte mirror_row;
    word falling_exit_level;
    byte falling_exit_room;
    word falling_entry_level;
    byte falling_entry_room;
    word mouse_level;
    byte mouse_room;
    byte mouse_start_x;
    word mouse_2_level;
    byte mouse_2_room;
    byte mouse_2_room_left_first_empty_col;
    byte mouse_2_potion_column;
    byte mouse_2_potion_row;
    byte mouse_2_start_x;
    byte mouse_2_crouch_delay;
    word loose_tiles_level;
    byte loose_tiles_room_1;
    byte loose_tiles_room_2;
    byte loose_tiles_first_tile;
    byte loose_tiles_last_tile;
    byte loose_floor_delay;
    byte loose_shake_count;
    byte jaffar_victory_flash_time;
    word win_level;
    byte win_room;
    word flame_level;
    byte flaming_sword_room;
    byte flaming_sword_tilepos;
    byte falling_bridge_delay;
    word inverse_level;
    byte moving_wall_delay;
    byte moving_wall_level;
    byte moving_wall_room;
    byte moving_wall_row;
    byte moving_wall_col_left;
    byte moving_wall_col_right;
    byte prison_level;
    word fat_guard_level;
    word fat_guard_skill;

    byte default_game_speed; // default timer length (non-fight mode, fight mode is always faster by 1), smaller is faster
    byte one_hp_game_speed;  // timer length with 1 hp
    byte fast_game_speed;    // timer length on level 6
    byte slow_game_speed;    // timer length on level 6B

    byte tbl_level_type[16];
    byte tbl_level_subtype[16];
    word tbl_level_color[16];
    short tbl_guard_type[16];
    byte tbl_guard_hp[16];
    byte tbl_cutscenes_by_index[16];
    byte tbl_entry_pose[16];
    sbyte tbl_seamless_exit[16];

    sbyte tbl_empty_potion_color[16];
    byte tbl_empty_potion_size[16];

    const char* tbl_guard_dat[8];
    const char* tbl_level_dat[16]; // custom level graphic files
    byte tbl_level_wda[16]; // wda flag for palace levels

    // demo skills
    word demo_guard_skill;
    word demo_kid_skill;

    // guard skills
    word strikeprob   [NUM_GUARD_SKILLS];
    word restrikeprob [NUM_GUARD_SKILLS];
    word blockprob    [NUM_GUARD_SKILLS];
    word impblockprob [NUM_GUARD_SKILLS];
    word advprob      [NUM_GUARD_SKILLS];
    word refractimer  [NUM_GUARD_SKILLS];
    word extrastrength[NUM_GUARD_SKILLS];
    word refractimer2 [NUM_GUARD_SKILLS];
    word jaffarparryprob;

    // shadow initialization
    byte init_shad_6b[9];
    byte init_shad_7[9];
    byte init_shad_8[9];
    byte init_shad_9[9];
    byte init_shad_9_2[9];
    byte init_shad_10[9];

    // auto moves
    auto_move_type demo_moves[27];
    auto_move_type demo_moves_2[23];
    auto_move_type level_6_intro_seq[6];
    auto_move_type level_6_drink_seq[3];
    auto_move_type shad_level_6b[8];
    auto_move_type shad_level_7[6];
    auto_move_type shad_level_8[4];
    auto_move_type shad_level_9[10];
    auto_move_type shad_level_9_2[4];

    // dead guards
    char_type2 dead_guards[16];

    // inactive guards
    char_type2 inactive_guards[16];

    // dynamic guards
    char_type2 dynamic_guards[16][24];

    // torch colors
    int torch_level_colors[16];
    int torch_level_colors_dimmed[16];

    // dark levels
    int dark_levels[16];

    // level names
    const char* level_names[16];

    // sound priority table
    byte sound_prio_table[MAX_FORE_SOUND_ID + 2];

    // quotes at the end of the demo
    const char* tbl_quotes[4];

    // INI driven parameters
    bool aspect_ratio;
    bool controller_rumble;
    bool native_ratio;
    bool positional_audio;
    bool run_in_window;
    bool enable_vsync;
    char renderer[50];
    int controller_threshold;
} game_settings_type;

typedef struct directory_listing_type directory_listing_type;

#endif
